unit DBFIndex;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  OCI, o8Env, o8Session, db, o8Filter, DBCommon;

type

  TDBFIndexType = (itNotDefined, itNTX, itNDX, itMDX, itCDX, itIDX);

  TOnSubIndex = procedure(Sender: TObject; var ItemKey: String; RecordNum: DWORD) of object;
  TOnSubNtx = procedure(Sender: TObject; var ItemKey: String; RecordNum: DWORD; var Accept: boolean) of object;
  TOnEvaluteKey = procedure(Sender: TObject; out Key: String) of object;
  TOnEvaluteFor = procedure(Sender: TObject; out ForValue: boolean) of object;
  TOnCompareKeys = procedure(Sender: TObject; CurrentKey, ItemKey: PChar; MaxLen: Cardinal; out c: Integer) of object;

  TIndex = class;

  TIndexes = class(TOwnedCollection)
  private

    FIndexType: TDBFIndexType;
    FActiveObject: TIndex;
    function GetCollectionOwner: TPersistent;
    function GetItem(Index: Integer): TIndex;
    procedure SetItem(Index: Integer; const Value: TIndex);

  public

    constructor Create(AOwner: TPersistent; ItemClass: TCollectionItemClass);
    procedure AssignValues(Value: TIndexes);
    function FindIndex(const Value: string): TIndex;
    function IsEqual(Value: TIndexes): Boolean;
    function CreateIndex(IndexName: String): TIndex;
    procedure CloseAll;
    property Owner: TPersistent read GetCollectionOwner;
    property Items[Index: Integer]: TIndex read GetItem write SetItem; default;
    property IndexType: TDBFIndexType read FIndexType write FIndexType;
    property ActiveObject: TIndex read FActiveObject;

  end;

  TIndex = class(TCollectionItem)
  private

    b_R: boolean;

    procedure SetActive(const Value: boolean);

  protected

    FName: String;
    FIndexes: TIndexes;
    FActive: boolean;

    FOnSubIndex: TOnSubIndex;
    FOnEvaluteFor: TOnEvaluteFor;
    FOnEvaluteKey: TOnEvaluteKey;
    FOnCompareKeys: TOnCompareKeys;

    function GetIsRanged: boolean; virtual;
    procedure AssignIndex(oInd: TIndex);
    function GetDisplayName: string; override;
    function InternalFirst: TGetResult; virtual;
    function InternalNext: TGetResult; virtual;
    function InternalPrior: TGetResult; virtual;
    function InternalLast: TGetResult; virtual;
    function GetCurrentKey: String; virtual;
    function GetCurrentRec: DWORD; virtual;

  public

    FOldEditKey: String;
    FOldEditRec: Longint;

    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    function IsEqual(Value: TIndex): Boolean; virtual;
    function Open: boolean; virtual;
    procedure Close; virtual;
    function IsOpen: boolean; virtual;
    function SetToRecord: boolean; overload; virtual;
    function SetToRecord(Key: String; Rec: Longint): boolean; overload; virtual;
    function SetToRecord(Rec: Longint): boolean; overload; virtual;
    function Seek(Key: String; SoftSeek: boolean = false): boolean; virtual;
    function SeekFirst(Key: String; SoftSeek: boolean = false): boolean; virtual;
    function SeekFields(const KeyFields: string; const KeyValues: Variant): Integer; virtual;
    function SubIndex(LowKey, HiKey: String): boolean; virtual;
    function FillFirstBufRecords(DBFHandler: Integer; FBuffer: pChar; FRecordsPerBuf: Integer; FRecordSize: Integer; FBufInd: pLongint; data_offset: Word): longint; virtual;
    function FillLastBufRecords(DBFHandler: Integer; FBuffer: pChar; FRecordsPerBuf: Integer; FRecordSize: Integer; FBufInd: pLongint; data_offset: Word): longint; virtual;
    function EvaluteKeyExpr: String; virtual;
    function SuiteFieldList(fl: String; out m: Integer): Integer; virtual;
    function EvaluteForExpr: boolean; virtual;
    function GetRecordByIndex(GetMode: TGetMode; var cRec: Longint): TGetResult; virtual;
    function GetFirstByIndex(var cRec: Longint): TGetResult; virtual;
    function GetLastByIndex(var cRec: Longint): TGetResult; virtual;
    procedure First; virtual;
    procedure Next; virtual;
    procedure Prior; virtual;
    procedure Last; virtual;
    function LastKey(out LastKey: String; out LastRec: LongInt): boolean; virtual;
    function NextBuffer(DBFHandler: Integer; FBuffer: pChar; FRecordsPerBuf: Integer; FRecordSize: Integer; FBufInd: pLongint; data_offset: Word): Longint; virtual;
    function PriorBuffer(DBFHandler: Integer; FBuffer: pChar; FRecordsPerBuf: Integer; FRecordSize: Integer; FBufInd: pLongint; data_offset: Word): Longint; virtual;
    procedure CreateIndex; virtual;
    procedure Reindex; virtual;
    procedure SetRangeFields(FieldList: String; FieldValues: array of const); overload; virtual;
    procedure SetRangeFields(FieldList: String; FieldValues: variant); overload; virtual;
    function InRange(Key: String): boolean; virtual;

    function FLock: boolean; virtual;
    function FUnLock: boolean; virtual;

    procedure StartUpdate(UnLock: boolean = true); virtual;
    procedure Flush; virtual;

    procedure DeleteKey(sKey: String; nRec: Longint); virtual;
    procedure AddKey(sKey: String; nRec: Longint); virtual;

    property IsRanged: boolean read GetIsRanged;
    property CurrentKey: String read GetCurrentKey;
    property CurrentRec: DWORD read GetCurrentRec;

    property OnSubIndex: TOnSubIndex read FOnSubIndex write FOnSubIndex;
    property OnEvaluteKey: TOnEvaluteKey read FOnEvaluteKey write FOnEvaluteKey;
    property OnEvaluteFor: TOnEvaluteFor read FOnEvaluteFor write FOnEvaluteFor;
    property OnCompareKeys: TOnCompareKeys read FOnCompareKeys write FOnCompareKeys;

  published

    property Name: String read FName write FName;
    property Active: boolean read FActive write SetActive;

  end;

implementation

uses
   DBFDataSet, DBFNTX, DBFCDX;

{ TIndexes }

procedure TIndexes.AssignValues(Value: TIndexes);
var
  I: Integer;
  P: TIndex;
begin
  for I := 0 to Value.Count - 1 do
  begin
    P := FindIndex(Value[I].Name);
    if P <> nil then
      P.Assign(Value[I]);
  end;
end;

procedure TIndexes.CloseAll;
var
  I: Integer;

  function FindOpened(var Ind: Integer): boolean;
  var
    i: Integer;
  begin
    Result := false;
    for i := 0 to Count - 1 do
      if Items[i].IsOpen then begin
        Ind := i;
        Result := true;
        Exit;
      end;
  end;

begin
  while FindOpened(I) do Items[I].Close;
end;

constructor TIndexes.Create(AOwner: TPersistent;
  ItemClass: TCollectionItemClass);
begin
  inherited Create(AOwner, ItemClass);
  if ItemClass.ClassName = 'TNTXIndex' then
    FIndexType := itNTX
  else if ItemClass.ClassName = 'TNDXIndex' then
    FIndexType := itNDX
  else if ItemClass.ClassName = 'TMDXIndex' then
    FIndexType := itMDX
  else if ItemClass.ClassName = 'TCDXIndex' then
    FIndexType := itCDX
  else
    FIndexType := itNotDefined;
  FActiveObject := nil;
end;

function TIndexes.CreateIndex(IndexName: String): TIndex;
begin
  case FIndexType of
    itNotDefined: raise Exception.Create('TIndex: IndexType not defined.');
    itNTX: Result := Add as TNTXIndex;
    //itNDX: Result := Add as TNDXIndex;
    //itMDX: Result := Add as TMDXIndex;
    itCDX: Result := Add as TCDXIndex;
  else
    Result := Add as TIndex;
  end;
  Result.Name := IndexName;
end;

function TIndexes.FindIndex(const Value: string): TIndex;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
  begin
    Result := TIndex(inherited Items[I]);
    if AnsiCompareText(Result.Name, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TIndexes.GetCollectionOwner: TPersistent;
begin
  Result := GetOwner;
end;

function TIndexes.GetItem(Index: Integer): TIndex;
begin
  Result := TIndex(inherited Items[Index]);
end;

function TIndexes.IsEqual(Value: TIndexes): Boolean;
var
  I: Integer;
begin
  Result := (Count = Value.Count);
  if Result then
    for I := 0 to Count - 1 do
    begin
      Result := TIndex(Items[I]).IsEqual(TIndex(Value.Items[I]));
      if not Result then Break;
    end
end;

procedure TIndexes.SetItem(Index: Integer; const Value: TIndex);
begin
  inherited SetItem(Index, TCollectionItem(Value));
end;

{ TIndex }

procedure TIndex.AddKey(sKey: String; nRec: Integer);
begin
  //
end;

procedure TIndex.Assign(Source: TPersistent);
begin
  if Source is TIndex then
    AssignIndex(TIndex(Source))
  else
    inherited Assign(Source);
end;

procedure TIndex.AssignIndex(oInd: TIndex);
begin
  if oInd <> nil then
  begin
    Name := oInd.Name;
  end;
end;

procedure TIndex.Close;
begin
  //
end;

constructor TIndex.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FIndexes := TIndexes(Collection);
  b_R := false;
end;

procedure TIndex.CreateIndex;
begin
  //
end;

procedure TIndex.DeleteKey(sKey: String; nRec: Integer);
begin
  //
end;

destructor TIndex.Destroy;
begin
  inherited Destroy;
end;

function TIndex.EvaluteForExpr: boolean;
begin
  Result := false;
end;

function TIndex.EvaluteKeyExpr: String;
begin
  Result := '';
end;

function TIndex.FillFirstBufRecords(DBFHandler: Integer; FBuffer: pChar;
  FRecordsPerBuf, FRecordSize: Integer;
  FBufInd: pLongint; data_offset: Word): longint;
begin
  Result := 0;
end;

function TIndex.FillLastBufRecords(DBFHandler: Integer; FBuffer: pChar;
  FRecordsPerBuf, FRecordSize: Integer; FBufInd: pLongint;
  data_offset: Word): longint;
begin
  Result := 0;
end;

procedure TIndex.First;
begin
  //
end;

function TIndex.FLock: boolean;
begin
  Result := false;
end;

function TIndex.FUnLock: boolean;
begin
  Result := false;
end;

function TIndex.GetCurrentKey: String;
begin
  Result := '';
end;

function TIndex.GetCurrentRec: DWORD;
begin
  Result := 0;
end;

function TIndex.GetDisplayName: string;
begin
  if Name <> '' then
    Result := Name
  else
    Result := inherited GetDisplayName;
end;

function TIndex.GetFirstByIndex(var cRec: Integer): TGetResult;
begin
  Result := grError;
end;

function TIndex.GetLastByIndex(var cRec: Integer): TGetResult;
begin
  Result := grError;
end;

function TIndex.GetRecordByIndex(GetMode: TGetMode;
  var cRec: Integer): TGetResult;
begin
  Result := grError;
end;

function TIndex.InternalFirst: TGetResult;
begin
  Result := grError;
end;

function TIndex.InternalLast: TGetResult;
begin
  Result := grError;
end;

function TIndex.InternalNext: TGetResult;
begin
  Result := grError;
end;

function TIndex.InternalPrior: TGetResult;
begin
  Result := grError;
end;

function TIndex.IsEqual(Value: TIndex): Boolean;
begin
  Result := false;
end;

function TIndex.IsOpen: boolean;
begin
  Result := false;
end;

procedure TIndex.Last;
begin
  //
end;

function TIndex.LastKey(out LastKey: String; out LastRec: Integer): boolean;
begin
  LastKey := '';
  LastRec := -1;
  Result := false;
end;

procedure TIndex.Next;
begin
  //
end;

function TIndex.NextBuffer(DBFHandler: Integer; FBuffer: pChar; FRecordsPerBuf: Integer; FRecordSize: Integer; FBufInd: pLongint; data_offset: Word): Longint;
begin
  Result := 0;
end;

function TIndex.Open: boolean;
begin
  Result := false;
end;

procedure TIndex.Prior;
begin
  //
end;

function TIndex.PriorBuffer(DBFHandler: Integer; FBuffer: pChar; FRecordsPerBuf: Integer; FRecordSize: Integer; FBufInd: pLongint; data_offset: Word): Longint;
begin
  Result := 0;
end;

function TIndex.Seek(Key: String; SoftSeek: boolean): boolean;
begin
  Result := false;
end;

function TIndex.SeekFields(const KeyFields: string;
  const KeyValues: Variant): Integer;
begin
  Result := 0;
end;

function TIndex.SeekFirst(Key: String; SoftSeek: boolean): boolean;
begin
  Result := false;
end;

procedure TIndex.SetActive(const Value: boolean);
var
  i: Integer;
  oW: TDBFNTX;
begin
  if FActive <> Value then begin
    oW := TDBFNTX(FIndexes.Owner);
    if Value then begin
      try
        b_R := true;
        for i := 0 to FIndexes.Count - 1 do
          FIndexes.Items[i].Active := false;
      finally
        b_R := false;
      end;
      FIndexes.FActiveObject := self;
      FActive := true;
      if oW.Active then oW.RecNo := oW.RecNo;
    end else begin
      if ( FIndexes.FActiveObject <> nil ) and ( FIndexes.FActiveObject = self ) then begin
        FIndexes.FActiveObject := nil;
        if not b_R then if oW.Active then oW.RecNo := oW.RecNo;
      end;
    end;
    FActive := Value;
  end;
end;

function TIndex.SetToRecord: boolean;
begin
  Result := false;
end;

function TIndex.SetToRecord(Key: String; Rec: Integer): boolean;
begin
  Result := false;
end;

procedure TIndex.SetRangeFields(FieldList: String;
  FieldValues: array of const);
begin
  //
end;

procedure TIndex.SetRangeFields(FieldList: String; FieldValues: variant);
begin
  //
end;

function TIndex.SetToRecord(Rec: Integer): boolean;
begin
  Result := false;
end;

function TIndex.SubIndex(LowKey, HiKey: String): boolean;
begin
  Result := false;
end;

function TIndex.SuiteFieldList(fl: String; out m: Integer): Integer;
begin
  m := 0;
  Result := 0;
end;

function TIndex.GetIsRanged: boolean;
begin
  Result := false;
end;

function TIndex.InRange(Key: String): boolean;
begin
  Result := false;
end;

procedure TIndex.Flush;
begin
  //
end;

procedure TIndex.Reindex;
begin
  //
end;

procedure TIndex.StartUpdate(UnLock: boolean = true);
begin
//
end;

end.
