{Copyright:      Vlad Karpov  mailto:KarpovVV@protek.ru
 Author:         Vlad Karpov
 Remarks:        freeware, but this Copyright must be included

 Description:    TDSet2Ora - component for migration DataSet to ORACLE

 Version:        1.0.0

 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}
unit DSet2Ora;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, o8Stmt, o8Env, o8Session, o8StmtOrd, OCI;

const
    cBatchCount: Integer = 1000;

type

  TAfterBatch = procedure(Sender: TObject; BatchedRecord: Integer) of object;
  TAfterCommit = procedure(Sender: TObject; CommitedRecord: Integer) of object;

  TDSet2Ora = class(TComponent)
  private

    FDataSet: TDataSet;
    FSession: To8Session;
    FTableName: String;
    FBatchCount: Integer;
    FCreateTable: Boolean;
    FCommitCount: Integer;
    FAfterBatch: TAfterBatch;
    FAfterCommit: TAfterCommit;

  protected
    { Protected declarations }

  public
    { Public declarations }

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure BatchMove;

  published
    { Published declarations }

    property DataSet: TDataSet read FDataSet write FDataSet;
    property oraSession: To8Session read FSession write FSession;
    property TableName: String read FTableName write FTableName;
    property BatchCount: Integer read FBatchCount write FBatchCount;
    property CommitCount: Integer read FCommitCount write FCommitCount;
    property CreateTable: Boolean read FCreateTable write FCreateTable;

    property AfterCommit: TAfterCommit read FAfterCommit write FAfterCommit;
    property AfterBatch: TAfterBatch read FAfterBatch write FAfterBatch;

  end;

procedure Register;

implementation

//******************************************************************************
procedure Register;
begin
  RegisterComponents('VK NLO', [TDSet2Ora]);
end;

{ TDSet2Ora }

procedure TDSet2Ora.BatchMove;
var
  i, j, k, pNum, l, m, n: Integer;
  qq: array of integer;
  oSQL, oSQL1: To8StmtOrd;
  ss, ss0, ss1, ss2, ss3, s1, s2, s3: String;
  oField: TField;
  oPar: TOraParam;
  Val: Variant;
  HasLob: Boolean;
  BlobStream: TStream;
  DestStream: TOraLobStream;
begin
  HasLob := false;
  if (DataSet <> nil) and (DataSet.Active) then
    if (oraSession <> nil) and (oraSession.Active) then
    begin
      if TableName = '' then TableName := DataSet.Name;
      TableName := AnsiUpperCase(TableName);
      oSQL := To8StmtOrd.Create(self);
      oSQL.oraSession := oraSession;
      oSQL1 := To8StmtOrd.Create(self);
      oSQL1.oraSession := oraSession;
      SetLength(qq, DataSet.FieldCount);
      try
        oSQL.SQL.Clear;
        oSQL1.SQL.Clear;
        ss0 := '';
        ss1 := '';
        ss2 := '';
        ss3 := '';
        oSQL.SQL.Add(Format('CREATE TABLE %s (', [TableName]));
        l := 0;
        for i := 0 to DataSet.Fields.Count - 1 do
        begin
          ss := '';
          s1 := '';
          s2 := '';
          s3 := '';
          oField := DataSet.Fields[i];
          qq[i] := l;
          case oField.DataType of
            ftString, ftFixedChar, ftWideString:
            begin
              ss := Format('%s  VARCHAR2(%d)', [oField.FieldName , oField.Size]);
              s1 := oField.FieldName;
              s2 := ':p' + oField.FieldName;
              Inc(l);
            end;
            ftSmallint, ftLargeint, ftInteger, ftAutoInc, ftWord, ftBoolean:
            begin
              ss := Format('%s  INTEGER', [oField.FieldName]);
              s1 := oField.FieldName;
              s2 := ':p' + oField.FieldName;
              Inc(l);
            end;
            ftFloat, ftCurrency, ftBCD:
            begin
              ss := Format('%s  NUMBER', [oField.FieldName]);
              s1 := oField.FieldName;
              s2 := ':p' + oField.FieldName;
              Inc(l);
            end;
            ftDate, ftTime, ftDateTime:
            begin
              ss := Format('%s  DATE', [oField.FieldName]);
              s1 := oField.FieldName;
              s2 := ':p' + oField.FieldName;
              Inc(l);
            end;
            ftBytes, ftVarBytes, ftTypedBinary:
            begin
              ss := Format('%s  ROW(%d)', [oField.FieldName , oField.Size]);
              s1 := oField.FieldName;
              s2 := ':p' + oField.FieldName;
              Inc(l);
            end;
            ftBlob, ftGraphic, ftOraBlob, ftParadoxOle, ftDBaseOle:
            begin
              ss := Format('%s  BLOB', [oField.FieldName]);
              s1 := oField.FieldName;
              s2 := 'EMPTY_BLOB()'; //':p' + oField.FieldName;
              s3 := oField.FieldName;
              HasLob := true;
              qq[i] := -1;
            end;
            ftMemo, ftFmtMemo, ftOraClob:
            begin
              ss := Format('%s  CLOB', [oField.FieldName]);
              s1 := oField.FieldName;
              s2 := 'EMPTY_CLOB()'; //':p' + oField.FieldName;
              s3 := oField.FieldName;
              HasLob := true;
              qq[i] := -1;
            end;
            ftCursor, ftDataSet,
            ftADT, ftArray, ftReference, ftVariant,
            ftInterface, ftIDispatch, ftGuid: qq[i] := -1;
          end;

          if ss0 <> '' then ss := ', ' + ss;
          if ss1 <> '' then s1 := ', ' + s1;
          if ss2 <> '' then s2 := ', ' + s2;
          if (ss3 <> '') and (qq[i] = -1) then s3 := ', ' + s3;

          oSQL.SQL.Add(ss);

          ss0 := ss0 + ss;
          ss1 := ss1 + s1;
          ss2 := ss2 + s2;
          ss3 := ss3 + s3;

        end;
        oSQL.SQL.Add(')');
        if FCreateTable then oSQL.ExecSql;
        oSQL.SQL.Clear;
        oSQL.SQL.Add('insert into ' + TableName + ' (' + ss1 + ') values (' + ss2 + ')');
        if HasLob then
        begin

          oSQL.SQL.Add('   returning RowId into :pRowId');

          oSQL1.SQL.Clear;
          oSQL1.SQL.Add('select ' + ss3 + ' from ' + TableName + ' where RowId = :pRowId for update');
          oSQL1.Params[0].DataType := ptRowID;

        end;
        for i := 0 to DataSet.Fields.Count - 1 do
        begin
          oField := DataSet.Fields[i];
          pNum := qq[i];
          if pNum <> -1 then
          begin
            oPar := oSQL.Params[pNum];
            case oField.DataType of
              ftString, ftFixedChar, ftWideString:
              begin
                oPar.DataType := ptString;
                oPar.MaxLengthStringBuffer := oField.Size + 10;
              end;
              ftSmallint, ftLargeint, ftInteger, ftAutoInc, ftWord, ftBoolean:
              begin
                oPar.DataType := ptInteger;
                oPar.MaxLengthStringBuffer := 10;
              end;
              ftFloat, ftCurrency, ftBCD:
              begin
                oPar.DataType := ptFloat;
                oPar.MaxLengthStringBuffer := 10;
              end;
              ftDate, ftTime, ftDateTime:
              begin
                oPar.DataType := ptDateTime;
                oPar.MaxLengthStringBuffer := 10;
              end;
              ftBytes, ftVarBytes, ftTypedBinary:
              begin
                oPar.DataType := ptUnknown;
                oPar.MaxLengthStringBuffer := oField.Size + 10;
              end;
              {ftBlob, ftGraphic, ftOraBlob, ftParadoxOle, ftDBaseOle:
              begin
                oPar.DataType := ptBLob;
              end;
              ftMemo, ftFmtMemo, ftOraClob:
              begin
                oPar.DataType := ptCLob;
              end;
              ftCursor, ftDataSet:
              ftADT, ftArray, ftReference, ftVariant:
              ftInterface, ftIDispatch, ftGuid:}
            end;
          end;
        end;
        if HasLob then
        begin
          BatchCount := 1;
          oSQL.ArrayOfPar := false;
          oPar := oSQL.ParamByName['pRowId'];
          oPar.DataType := ptRowID;
          oPar.MaxLengthStringBuffer := 10;
        end else begin
          oSQL.Portion := BatchCount;
          oSQL.ArrayOfPar := true;    //After set portion !!!!!
        end;
        j := 0;
        k := 0;
        m := 0;
        n := 0;
        DataSet.DisableControls;
        DataSet.First;
        while not DataSet.Eof do
        begin
          //
          for i := 0 to DataSet.Fields.Count - 1 do
          begin
            oField := DataSet.Fields[i];
            pNum := qq[i];
            if pNum <> -1 then
            begin
              if HasLob then
              begin
                oPar := oSQL.Params[pNum];
                Val := oField.Value;
                if Val <> NULL then
                  oPar.ParamValue :=  Val
                else
                  oPar.IsNull := true;
              end else begin
                case oField.DataType of
                ftBlob, ftGraphic, ftOraBlob, ftParadoxOle, ftDBaseOle: begin end;
                ftMemo, ftFmtMemo, ftOraClob: begin end;
                ftCursor, ftDataSet: begin end;
                ftADT, ftArray, ftReference, ftVariant: begin end;
                ftInterface, ftIDispatch, ftGuid: begin end;
                else
                  begin
                    Val := oField.Value;
                    if Val <> NULL then
                      oSQL.ArrayParValue[j, pNum] := Val
                    else
                      oSQL.ArrayParNull[j, pNum] := true;
                  end
                end;
              end;
            end;
          end;
          //
          Inc(j);
          if Cardinal(j) = Cardinal(BatchCount) then
          begin
            oSQL.ExecSql;
            if HasLob then
            begin
              oSQL1.Active := false;
              oSQL1.Params[0].RowId := oSQL.ParamByName['pRowId'].RowId;
              oSQL1.Active := true;
              if not oSQL1.Eof then
              begin
                for i := 0 to DataSet.Fields.Count - 1 do
                begin
                  oField := DataSet.Fields[i];
                  case oField.DataType of
                  ftBlob, ftGraphic, ftOraBlob, ftParadoxOle, ftDBaseOle, ftMemo, ftFmtMemo, ftOraClob:
                    begin
                      BlobStream := DataSet.CreateBlobStream(oField, bmRead);
                      DestStream := oSQL1.OraFieldByName[oField.FieldName].LobStream;
                      try
                        DestStream.CopyFrom(BlobStream, BlobStream.Size);
                      finally
                        BlobStream.Free;
                        DestStream.Free;
                      end;
                    end;
                  end;
                end;
              end;
              oSQL1.Active := false;
            end;
            k := k + j;
            if Assigned(AfterBatch) then AfterBatch(self, k);
            j := 0;
            Application.ProcessMessages;
          end;
          Inc(m);
          if Cardinal(m) = Cardinal(CommitCount) then
          begin
            oSQL.oraSession.TransCommit;
            n := n + m;
            if Assigned(AfterCommit) then AfterCommit(self, n);
            m := 0;
            Application.ProcessMessages;
          end;
          DataSet.Next;
        end;
        DataSet.First;
        DataSet.EnableControls;
        if j <> 0 then
        begin
          oSQL.ExecSql;
          if HasLob then
          begin
            oSQL1.Active := false;
            oSQL1.Params[0].RowId := oSQL.ParamByName['pRowId'].RowId;
            oSQL1.Active := true;
            if not oSQL1.Eof then
            begin
              for i := 0 to DataSet.Fields.Count - 1 do
              begin
                oField := DataSet.Fields[i];
                case oField.DataType of
                ftBlob, ftGraphic, ftOraBlob, ftParadoxOle, ftDBaseOle, ftMemo, ftFmtMemo, ftOraClob:
                  begin
                    BlobStream := DataSet.CreateBlobStream(oField, bmRead);
                    DestStream := oSQL1.OraFieldByName[oField.FieldName].LobStream;
                    try
                      DestStream.CopyFrom(BlobStream, BlobStream.Size);
                    finally
                      BlobStream.Free;
                      DestStream.Free;
                    end;
                  end;
                end;
              end;
            end;
            oSQL1.Active := false;
          end;
          k := k + j;
          if Assigned(AfterBatch) then AfterBatch(self, k);
          Application.ProcessMessages;
        end;
        if m <> 0 then
        begin
          oSQL.oraSession.TransCommit;
          n := n + m;
          if Assigned(AfterCommit) then AfterCommit(self, n);
          Application.ProcessMessages;
        end;
      finally
        oSQL.Destroy;
        oSQL1.Destroy;
      end;
    end;
end;

constructor TDSet2Ora.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FDataSet := nil;
  FSession := nil;
  FBatchCount := cBatchCount;
  FCommitCount := cBatchCount;
  FCreateTable := true;
end;

destructor TDSet2Ora.Destroy;
begin
  inherited Destroy;
end;

end.
