unit MemMgr;

interface

uses
  contnrs, Dialogs;

type

  {TOneAlloc}
  TOneAlloc = class
  private

    FMemory: Pointer;
    FCaller: TObject;
    FCaption: String;
    FSize: Cardinal;

  public

    constructor Create; overload;
    constructor Create(Caller: TObject; Caption: String; Size: Cardinal); overload;
    constructor Create(Caller: TObject; Size: Cardinal); overload;
    destructor Destroy; override;

    procedure GetMem(Size: Cardinal);
    procedure ReallocMem(NewSize: Cardinal);
    procedure FreeMem;

    property Memory: Pointer read FMemory;
    property Caller: TObject read FCaller write FCaller;
    property Caption: String read FCaption write FCaption;
    property Size: Cardinal read FSize;

  end;

  {TMemMgr}
  TMemMgr = class(TObjectList)
  private
  public
    constructor Create;
    destructor Destroy; override;

    function FindIndex(p: Pointer; out Ind: Integer): boolean;
    function FindCaption(Capt: String; out Ind: Integer): boolean;

    procedure FreeForCaption(Capt: String);

    function GetMem(Caller: TObject; size: Integer): Pointer; overload;
    function ReallocMem(p: Pointer; size: Integer): Pointer; overload;
    procedure FreeMem(p: Pointer); overload;

    function GetMem(Capt: String; size: Integer): Pointer; overload;

    function GetSize(p: Pointer): Integer;

  end;

var
  oMem: TMemMgr;

implementation

{ TMemMgr }

constructor TMemMgr.Create;
begin
  inherited Create;
end;

destructor TMemMgr.Destroy;
begin
  inherited Destroy;
end;

function TMemMgr.FindCaption(Capt: String; out Ind: Integer): boolean;
var
  i: Integer;
begin
  Result := false;
  Ind := -1;
  for i := 0 to Count - 1 do
    if TOneAlloc(Items[i]).Caption = Capt then begin
      Result := true;
      Ind := i;
      Exit;
    end;
end;

function TMemMgr.FindIndex(p: Pointer; out Ind: Integer): boolean;
var
  B: TOneAlloc;
  beg, Mid: Integer;
begin
  Ind := Count;
  if ( Ind > 0 ) then begin
    beg := 0;
    B := TOneAlloc(Items[beg]);
    if ( Integer(p) > Integer(B.FMemory) ) then begin
      repeat
        Mid := (Ind + beg) div 2;
        B := TOneAlloc(Items[Mid]);
        if ( Integer(p) > Integer(B.FMemory) ) then
           beg := Mid
        else
           Ind := Mid;
      until ( ((Ind - beg) div 2) = 0 );
    end else
      Ind := beg;
    if Ind < Count then begin
      B := TOneAlloc(Items[Ind]);
      Result := (Integer(p) = Integer(B.FMemory));
    end else
      Result := false;
  end else
    Result := false;
end;

procedure TMemMgr.FreeForCaption(Capt: String);
var
  i: Integer;
begin
  while FindCaption(Capt, i) do Delete(i);
end;

procedure TMemMgr.FreeMem(p: Pointer);
var
  i: Integer;
begin
  if (p <> nil) and FindIndex(p, i) then Delete(i);
end;

function TMemMgr.GetMem(Caller: TObject; size: Integer): Pointer;
var
  Obj: TOneAlloc;
  i: Integer;
begin
  Obj := TOneAlloc.Create(Caller, size);
  FindIndex(Obj.FMemory, i);
  Insert(i, Obj);
  Result := Obj.FMemory;
end;

function TMemMgr.GetMem(Capt: String; size: Integer): Pointer;
var
  Obj: TOneAlloc;
  i: Integer;
begin
  Obj := TOneAlloc.Create(nil, Capt, size);
  FindIndex(Obj.FMemory, i);
  Insert(i, Obj);
  Result := Obj.FMemory;
end;

function TMemMgr.GetSize(p: Pointer): Integer;
var
  Obj: TOneAlloc;
  i: Integer;
begin
  if p <> nil then begin
    if FindIndex(p, i) then begin
      Obj := TOneAlloc(Items[i]);
      Result := Obj.FSize;
    end else
      Result := 0;
  end else
    Result := 0;
end;

function TMemMgr.ReallocMem(p: Pointer; size: Integer): Pointer;
var
  Obj: TOneAlloc;
  i: Integer;
  Old: Pointer;
begin
  if p <> nil then begin
    if FindIndex(p, i) then begin
      Obj := TOneAlloc(Items[i]);
      Old := Obj.FMemory;
      Obj.ReallocMem(size);
      Result := Obj.FMemory;
      if Integer(Old) <> Integer(Obj.FMemory) then begin
        OwnsObjects := false;
        try
          Delete(i);
          FindIndex(Obj.FMemory, i);
          Insert(i, Obj);
        finally
          OwnsObjects := true;
        end;
      end;
    end else
      Result := nil;
  end else
    Result := self.GetMem(self, size);
end;

{ TOneAlloc }

constructor TOneAlloc.Create;
begin
  FMemory := nil;
  FCaller := nil;
  FSize := 0;
  FCaption := '';
end;

constructor TOneAlloc.Create(Caller: TObject; Caption: String; Size: Cardinal);
begin
  Create;
  self.GetMem(Size);
  FCaller := Caller;
  FCaption := Caption;
end;

constructor TOneAlloc.Create(Caller: TObject; Size: Cardinal);
begin
  Create;
  self.GetMem(Size);
  FCaller := Caller;
  if FCaller <> nil then
    FCaption := FCaller.ClassName;
end;

destructor TOneAlloc.Destroy;
begin
  self.FreeMem;
  inherited Destroy;
end;

procedure TOneAlloc.FreeMem;
begin
  if FMemory <> nil then begin
    System.FreeMem(FMemory);
    FMemory := nil;
    FSize := 0;
  end;
end;

procedure TOneAlloc.GetMem(Size: Cardinal);
begin
  if FMemory = nil then begin
    System.GetMem(FMemory, Size);
    FSize := Size;
  end else begin
    System.ReallocMem(FMemory, Size);
    FSize := Size;
  end;
end;

procedure TOneAlloc.ReallocMem(NewSize: Cardinal);
begin
  System.ReallocMem(FMemory, NewSize);
  FSize := NewSize;
end;

initialization

    oMem := TMemMgr.Create;

finalization

    oMem.Free;

end.
