{Copyright:      Vlad Karpov  mailto:KarpovVV@protek.ru
 Author:         Vlad Karpov
 Remarks:        freeware, but this Copyright must be included

 Description:    OCI function defenitions from OCI.DLL

 Version:        1.0.0

 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}
unit OCI;

interface

Uses Windows, SysUtils, MemMgr;

const

{---------------------------------------------------------------------------
                     PUBLIC TYPES AND CONSTANTS
  ---------------------------------------------------------------------------}

{-----------------------------Modes ----------------------------------------}
   OCI_DEFAULT = 0; { the default value for parameters and attributes }
   OCI_THREADED =1;      { the application is in threaded environment }
   OCI_OBJECT  = 2;        { the application is in object environment }
   OCI_NON_BLOCKING =4 ;             { non blocking mode of operation }
   OCI_ENV_NO_MUTEX =8; { the environment handle will not be protected
                 by a mutex internally }

{-----------------------------Handle Types----------------------------------}
                                           { handle types range from 1 - 49 }
   OCI_HTYPE_FIRST =1  ;                    { start value of handle type }
   OCI_HTYPE_ENV =1    ;                            { environment handle }
   OCI_HTYPE_ERROR =2  ;                                  { error handle }
   OCI_HTYPE_SVCCTX =3 ;                                { service handle }
   OCI_HTYPE_STMT =4   ;                              { statement handle }
   OCI_HTYPE_BIND =5   ;                                   { bind handle }
   OCI_HTYPE_DEFINE =6 ;                                 { define handle }
   OCI_HTYPE_DESCRIBE =7 ;                             { describe handle }
   OCI_HTYPE_SERVER =8   ;                               { server handle }
   OCI_HTYPE_SESSION =9  ;                       { authentication handle }
   OCI_HTYPE_TRANS =10   ;                          { transaction handle }
   OCI_HTYPE_COMPLEXOBJECT =11;        { complex object retrieval handle }
   OCI_HTYPE_SECURITY =12     ;                        { security handle }
   OCI_HTYPE_LAST =12         ;            { last value of a handle type }
{---------------------------------------------------------------------------}


{-------------------------Descriptor Types----------------------------------}
                                    { descriptor values range from 50 - 255 }
   OCI_DTYPE_FIRST =50         ;        { start value of descriptor type }
   OCI_DTYPE_LOB =50           ;                          { lob  locator }
   OCI_DTYPE_SNAP =51          ;                   { snapshot descriptor }
   OCI_DTYPE_RSET =52          ;                 { result set descriptor }
   OCI_DTYPE_PARAM =53 ; { a parameter descriptor obtained from ocigparm }
   OCI_DTYPE_ROWID  =54         ;                     { rowid descriptor }
   OCI_DTYPE_COMPLEXOBJECTCOMP  =55 ;
                                      { complex object retrieval descriptor }
   OCI_DTYPE_FILE =56                ;                { File Lob locator }
   OCI_DTYPE_LAST =56                ; { last value of a descriptor type }
{---------------------------------------------------------------------------}

{---------------------------------------------------------------------------}

{-------------------------Object Ptr Types----------------------------------}
   OCI_OTYPE_NAME =1                  ;    { object name }
   OCI_OTYPE_REF  =2                  ;     { REF to TDO }
   OCI_OTYPE_PTR  =3                  ;     { PTR to TDO }
{---------------------------------------------------------------------------}

{-----------------------------Attribute Types-------------------------------}
   OCI_ATTR_FNCODE  =1                 ;
   OCI_ATTR_OBJECT   =2                ;
   OCI_ATTR_NONBLOCKING_MODE  =3       ;
   OCI_ATTR_SQLCODE  =4                ;
   OCI_ATTR_ENV  =5                    ;
   OCI_ATTR_SERVER =6                  ;
   OCI_ATTR_SESSION =7                 ;
   OCI_ATTR_TRANS   =8                 ;
   OCI_ATTR_ROW_COUNT =  9             ;
   OCI_ATTR_SQLFNCODE =10              ;
   OCI_ATTR_PREFETCH_ROWS = 11         ;
   OCI_ATTR_NESTED_PREFETCH_ROWS= 12   ;
   OCI_ATTR_PREFETCH_MEMORY =13        ;
   OCI_ATTR_NESTED_PREFETCH_MEMORY =14 ;
   OCI_ATTR_CHAR_COUNT  =15            ;
            { this specifies the bind and define size in characters }
   OCI_ATTR_PDSCL  = 16                ;
   OCI_ATTR_PDFMT  = 17                ;
   OCI_ATTR_PARAM_COUNT =18            ;
   OCI_ATTR_ROWID   =19                ;
   OCI_ATTR_CHARSET = 20               ;
   OCI_ATTR_NCHAR   =21                ;
   OCI_ATTR_USERNAME =22               ;            { username attribute }
   OCI_ATTR_PASSWORD =23               ;            { password attribute }
   OCI_ATTR_STMT_TYPE =  24            ;                { statement type }
   OCI_ATTR_INTERNAL_NAME  = 25        ;
   OCI_ATTR_EXTERNAL_NAME  = 26        ;
   OCI_ATTR_XID    = 27                ;
   OCI_ATTR_TRANS_LOCK =28             ;
   OCI_ATTR_TRANS_NAME  =    29        ;
   OCI_ATTR_HEAPALLOC =30              ;  { memory allocated on the heap }
   OCI_ATTR_CHARSET_ID =31             ;              { Character Set ID }
   OCI_ATTR_CHARSET_FORM =32           ;            { Character Set Form }
   OCI_ATTR_MAXDATA_SIZE =33       ;{ Maximumsize of data on the server  }
   OCI_ATTR_CACHE_OPT_SIZE =34     ;         { object cache optimal size }
   OCI_ATTR_CACHE_MAX_SIZE =35   ;{ object cache maximum size percentage }
   OCI_ATTR_PINOPTION= 36         ;    { object cache default pin option }
   OCI_ATTR_ALLOC_DURATION =37     ;
                            { object cache default allocation duration }
   OCI_ATTR_PIN_DURATION =38       ; { object cache default pin duration }
   OCI_ATTR_FDO          =39       ;{ Format Descriptor object attribute }
   OCI_ATTR_POSTPROCESSING_CALLBACK =40;
                                    { Callback to process outbind data }
   OCI_ATTR_POSTPROCESSING_CONTEXT =41  ;
                            { Callback context to process outbind data }
   OCI_ATTR_ROWS_RETURNED =42           ;
             { Number of rows returned in current iter - for Bind handles }
   OCI_ATTR_FOCBK        =43             ; { Failover Callback attribute }
   OCI_ATTR_IN_V8_MODE   =44 ;{ is the server/service context in V8 mode }
   OCI_ATTR_LOBEMPTY     =45  ;
   OCI_ATTR_SESSLANG     =46  ;                { session language handle }

{-------------------------Parameter Attribute Types-------------------------}

   OCI_ATTR_UNK              =101 ;                  { unknown attribute }
   OCI_ATTR_NUM_COLS         =102 ;                  { number of columns }
   OCI_ATTR_LIST_COLUMNS     =103 ;       { parameter of the column list }
   OCI_ATTR_RDBA             =104 ;          { DBA of the segment header }
   OCI_ATTR_CLUSTERED        =105 ;     { whether the table is clustered }
   OCI_ATTR_PARTITIONED      =106 ;   { whether the table is partitioned }
   OCI_ATTR_INDEX_ONLY       =107 ;    { whether the table is index only }
   OCI_ATTR_LIST_ARGUMENTS   =108 ;     { parameter of the argument list }
   OCI_ATTR_LIST_SUBPROGRAMS =109 ;   { parameter of the subprogram list }
   OCI_ATTR_REF_TDO          =110 ;         { REF to the type descriptor }
   OCI_ATTR_LINK             =111 ;             { the database link name }
   OCI_ATTR_MIN              =112 ;                      { minimum value }
   OCI_ATTR_MAX              =113 ;                      { maximum value }
   OCI_ATTR_INCR             =114 ;                    { increment value }
   OCI_ATTR_CACHE            =115 ;  { number of sequence numbers cached }
   OCI_ATTR_ORDER            =116 ;    { whether the sequence is ordered }
   OCI_ATTR_HW_MARK          =117 ;                    { high-water mark }
   OCI_ATTR_TYPE_SCHEMA      =118 ;                 { type's schema name }
   OCI_ATTR_TIMESTAMP        =119 ;            { timestamp of the object }
   OCI_ATTR_NUM_ATTRS        =120 ;               { number of sttributes }
   OCI_ATTR_NUM_PARAMS       =121 ;               { number of parameters }
   OCI_ATTR_OBJID            =122 ;      { object id for a table or view }
   OCI_ATTR_PTYPE            =123 ;          { type of info described by }
   OCI_ATTR_PARAM            =124 ;               { parameter descriptor }
   OCI_ATTR_OVERLOAD_ID      =125 ;    { overload ID for funcs and procs }
   OCI_ATTR_TABLESPACE       =126 ;                   { table name space }

{-------------------------Credential Types----------------------------------}
   OCI_CRED_RDBMS    =1           ;         { database username/password }
   OCI_CRED_EXT      =2           ;    { externally provided credentials }
{---------------------------------------------------------------------------}

{------------------------Error Return Values--------------------------------}
   OCI_SUCCESS =0                ;      { maps to SQL_SUCCESS of SAG CLI }
   OCI_SUCCESS_WITH_INFO =1      ;       { maps to SQL_SUCCESS_WITH_INFO }
   OCI_NO_DATA =100              ;                 { maps to SQL_NO_DATA }
   OCI_ERROR =-1                 ;                   { maps to SQL_ERROR }
   OCI_INVALID_HANDLE =-2        ;          { maps to SQL_INVALID_HANDLE }
   OCI_NEED_DATA =99             ;               { maps to SQL_NEED_DATA }
   OCI_STILL_EXECUTING =-3123    ;               { OCI would block error }
   OCI_CONTINUE =-24200    ;{ Continue with the body of the OCI function }
{---------------------------------------------------------------------------}

{------------------------Parsing Syntax Types-------------------------------}
   OCI_V7_SYNTAX =2   ;                                    { V7 language }
   OCI_V8_SYNTAX =3   ;                                    { V8 language }
   OCI_NTV_SYNTAX =1  ;  { Use what so ever is the native lang of server }
                     { these values must match the values defined in kpul.h }
{---------------------------------------------------------------------------}

{------------------------Scrollable Cursor Options--------------------------}
   OCI_FETCH_NEXT =2        ;
   OCI_FETCH_FIRST =4       ;
   OCI_FETCH_LAST =8        ;
   OCI_FETCH_PRIOR =16      ;
   OCI_FETCH_ABSOLUTE =32   ;
   OCI_FETCH_RELATIVE =64   ;
{---------------------------------------------------------------------------}

{------------------------Bind and Define Options----------------------------}
   OCI_SB2_IND_PTR   =1     ;
   OCI_DATA_AT_EXEC  =2     ;
   OCI_DYNAMIC_FETCH =2     ;
   OCI_PIECEWISE     =4     ;
{---------------------------------------------------------------------------}

{------------------------Execution Modes------------------------------------}
   OCI_BATCH_MODE        =1  ;
   OCI_EXACT_FETCH       =2  ;
   OCI_KEEP_FETCH_STATE  =4  ;
   OCI_SCROLLABLE_CURSOR =8  ;
   OCI_DESCRIBE_ONLY     =16 ;
   OCI_COMMIT_ON_SUCCESS =32 ;
{---------------------------------------------------------------------------}

{------------------------Authentication Modes-------------------------------}
   OCI_MIGRATE     =1        ;            { migratable auth context }
   OCI_SYSDBA      =2        ;           { for SYSDBA authorization }
   OCI_SYSOPER     =4        ;          { for SYSOPER authorization }
   OCI_PRELIM_AUTH =8        ;      { for preliminary authorization }
{---------------------------------------------------------------------------}

{------------------------Piece Information----------------------------------}
   OCI_PARAM_IN =1           ;
   OCI_PARAM_OUT =2          ;
{---------------------------------------------------------------------------}

{------------------------ Transaction Start Flags --------------------------}
   OCI_TRANS_OLD           =0      ;
   OCI_TRANS_NEW           =1      ;
   OCI_TRANS_JOIN          =2      ;
   OCI_TRANS_RESUME        =4      ;
   OCI_TRANS_STARTMASK     =255    ;

   OCI_TRANS_READONLY      =256    ;
   OCI_TRANS_READWRITE     =512    ;
   OCI_TRANS_SERIALIZABLE  =1024   ;
   OCI_TRANS_ISOLMASK      =65280  ;

   OCI_TRANS_LOOSE         =65536  ;
   OCI_TRANS_TIGHT         =131072 ;
   OCI_TRANS_TYPEMASK      =983040 ;

   OCI_TRANS_NOMIGRATE     =1048576  ;

{---------------------------------------------------------------------------}

{------------------------ Transaction End Flags ----------------------------}
   OCI_TRANS_TWOPHASE              = 16777216 ;
{---------------------------------------------------------------------------}

{--------------------------- FILE open modes -------------------------------}
  OCI_FILE_READONLY                = 1;   { readonly mode open for FILE types }
{---------------------------------------------------------------------------}

{----------------------- LOB Buffering Flush Flags -------------------------}
  OCI_LOB_BUFFER_FREE     = 1;
  OCI_LOB_BUFFER_NOFREE   = 2;
{---------------------------------------------------------------------------}


{-----------------------Object Types----------------------------------------}
   OCI_OTYPE_UNK           =0            ;
   OCI_OTYPE_TABLE         =1            ;
   OCI_OTYPE_VIEW          =2            ;
   OCI_OTYPE_SYN           =3            ;
   OCI_OTYPE_PROC          =4            ;
   OCI_OTYPE_FUNC          =5            ;
   OCI_OTYPE_PKG           =6            ;
   OCI_OTYPE_STMT          =7            ;
{---------------------------------------------------------------------------}

{-----------------------Describe Handle Parameter Attributes ---------------}
{ Attributes common to Columns and Stored Procs }
   OCI_ATTR_DATA_SIZE      =1            ;
   OCI_ATTR_DATA_TYPE      =2            ;
   OCI_ATTR_DISP_SIZE      =3            ;
   OCI_ATTR_NAME           =4            ;
   OCI_ATTR_PRECISION      =5            ;
   OCI_ATTR_SCALE          =6            ;
   OCI_ATTR_IS_NULL        =7            ;
   OCI_ATTR_TYPE_NAME      =8            ;
   OCI_ATTR_SCHEMA_NAME    =9            ;
   OCI_ATTR_SUB_NAME       =10           ;
   OCI_ATTR_POSITION       =11           ;

{ complex object retrieval parameter attributes }
   OCI_ATTR_COMPLEXOBJECTCOMP_TYPE             =50;
   OCI_ATTR_COMPLEXOBJECTCOMP_TYPE_LEVEL       =51;
   OCI_ATTR_COMPLEXOBJECT_LEVEL                =52;
   OCI_ATTR_COMPLEXOBJECT_COLL_OUTOFLINE       =53;

{Only Columns }
   OCI_ATTR_DISP_NAME      =100                    ;

{Only Stored Procs }
   OCI_ATTR_OVERLOAD       =210                     ;
   OCI_ATTR_LEVEL          =211                      ;
   OCI_ATTR_HAS_DEFAULT    =212                     ;
   OCI_ATTR_IOMODE         =213                     ;
   OCI_ATTR_RADIX          =214                     ;
   OCI_ATTR_NUM_ARGS       =215                     ;

{ only user-defined Type's }
   OCI_ATTR_TYPECODE                  =216          ;
   OCI_ATTR_COLLECTION_TYPECODE       =217          ;
   OCI_ATTR_VERSION                   =218          ;
   OCI_ATTR_IS_INCOMPLETE_TYPE        =219          ;
   OCI_ATTR_IS_SYSTEM_TYPE            =220          ;
   OCI_ATTR_IS_PREDEFINED_TYPE        =221          ;
   OCI_ATTR_IS_TRANSIENT_TYPE         =222          ;
   OCI_ATTR_IS_SYSTEM_GENERATED_TYPE  =223          ;
   OCI_ATTR_HAS_NESTED_TABLE          =224          ;
   OCI_ATTR_HAS_LOB                   =225          ;
   OCI_ATTR_HAS_FILE                  =226          ;
   OCI_ATTR_COLLECTION_ELEMENT        =227          ;
   OCI_ATTR_NUM_TYPE_ATTRS            =228          ;
   OCI_ATTR_LIST_TYPE_ATTRS           =229          ;
   OCI_ATTR_NUM_TYPE_METHODS          =230          ;
   OCI_ATTR_LIST_TYPE_METHODS         =231          ;
   OCI_ATTR_MAP_METHOD                =232          ;
   OCI_ATTR_ORDER_METHOD              =233          ;

{ only collection element }
   OCI_ATTR_NUM_ELEMS                 =234          ;

{ only type methods }
   OCI_ATTR_ENCAPSULATION             =235          ;
   OCI_ATTR_IS_SELFISH                =236          ;
   OCI_ATTR_IS_VIRTUAL                =237          ;
   OCI_ATTR_IS_INLINE                 =238          ;
   OCI_ATTR_IS_CONSTANT               =239          ;
   OCI_ATTR_HAS_RESULT                =240          ;
   OCI_ATTR_IS_CONSTRUCTOR            =241          ;
   OCI_ATTR_IS_DESTRUCTOR             =242          ;
   OCI_ATTR_IS_OPERATOR               =243         ;
   OCI_ATTR_IS_MAP                    =244         ;
   OCI_ATTR_IS_ORDER                  =245         ;
   OCI_ATTR_IS_RNDS                   =246         ;
   OCI_ATTR_IS_RNPS                   =247         ;
   OCI_ATTR_IS_WNDS                   =248         ;
   OCI_ATTR_IS_WNPS                   =249         ;

{---------------------------------------------------------------------------}

{---------------------------ocicpw Modes------------------------------------}
   OCI_AUTH         =8        ; { Change the password but donot login }


{------------------------Other Constants------------------------------------}
   OCI_MAX_FNS   =100          ;           { max number of OCI Functions }
   OCI_SQLSTATE_SIZE =5        ;
   OCI_ERROR_MAXMSG_SIZE   =1024 ;
   OCI_LOBMAXSIZE = 4294967295;  {MINUB4MAXVAL;}
   OCI_ROWID_LEN             =23 ;
{---------------------------------------------------------------------------}

{------------------------ Fail Over Events ---------------------------------}
   OCI_FO_END          =1        ;
   OCI_FO_ABORT        =2        ;
   OCI_FO_REAUTH       =4        ;
   OCI_FO_BEGIN        =8        ;
{---------------------------------------------------------------------------}

{------------------------- Fail Over Types ---------------------------------}
   OCI_FO_NONE           =1       ;
   OCI_FO_SESSION        =2       ;
   OCI_FO_SELECT         =4       ;
   OCI_FO_TXNAL          =8       ;
{---------------------------------------------------------------------------}

{-----------------------Function Codes--------------------------------------}
   OCI_FNCODE_INITIALIZE     =1    ;
   OCI_FNCODE_HANDLEALLOC  =2      ;
   OCI_FNCODE_HANDLEFREE  =3       ;
   OCI_FNCODE_DESCRIPTORALLOC = 4  ;
   OCI_FNCODE_DESCRIPTORFREE = 5   ;
   OCI_FNCODE_ENVINIT   =6         ;
   OCI_FNCODE_SERVERATTACH  = 7    ;
   OCI_FNCODE_SERVERDETACH  = 8    ;
   { unused         9 }
   OCI_FNCODE_SESSIONBEGIN  =10    ;
   OCI_FNCODE_SESSIONEND   =11     ;
   OCI_FNCODE_PASSWORDCHANGE  = 12 ;
   OCI_FNCODE_STMTPREPARE   =13    ;
   OCI_FNCODE_STMTBINDBYPOS  =14   ;
   OCI_FNCODE_STMTBINDBYNAME = 15  ;
   { unused        16 }
   OCI_FNCODE_BINDDYNAMIC   =17    ;
   OCI_FNCODE_BINDOBJECT  =18      ;
   OCI_FNCODE_BINDARRAYOFSTRUCT  = 20;
   OCI_FNCODE_STMTEXECUTE  =21        ;
   { unused 22, 23 }
   OCI_FNCODE_DEFINE  =24             ;
   OCI_FNCODE_DEFINEOBJECT  =25       ;
   OCI_FNCODE_DEFINEDYNAMIC  = 26     ;
   OCI_FNCODE_DEFINEARRAYOFSTRUCT = 27;
   OCI_FNCODE_STMTFETCH  = 28         ;
   OCI_FNCODE_STMTGETBIND  = 29       ;
   { 30, 31 unused }
   OCI_FNCODE_DESCRIBEANY = 32        ;
   OCI_FNCODE_TRANSSTART  =33         ;
   OCI_FNCODE_TRANSDETACH = 34        ;
   OCI_FNCODE_TRANSCOMMIT = 35        ;
   { 36 unused }
   OCI_FNCODE_ERRORGET   =37          ;
   OCI_FNCODE_LOBOPENFILE = 38        ;
   OCI_FNCODE_LOBCLOSEFILE = 39       ;
   OCI_FNCODE_LOBCREATEFILE = 40      ;
   OCI_FNCODE_LOBDELFILE  =41         ;
   OCI_FNCODE_LOBCOPY  =42            ;
   OCI_FNCODE_LOBAPPEND = 43          ;
   OCI_FNCODE_LOBERASE  =44           ;
   OCI_FNCODE_LOBLENGTH = 45          ;
   OCI_FNCODE_LOBTRIM  =46            ;
   OCI_FNCODE_LOBREAD  =47            ;
   OCI_FNCODE_LOBWRITE = 48           ;
   { 49 unused }
   OCI_FNCODE_SVCCTXBREAK =50         ;
   OCI_FNCODE_SERVERVERSION = 51      ;
   { unused 52, 53 }
   OCI_FNCODE_ATTRGET =54             ;
   OCI_FNCODE_ATTRSET =55             ;
   OCI_FNCODE_PARAMSET =56            ;
   OCI_FNCODE_PARAMGET =57            ;
   OCI_FNCODE_STMTGETPIECEINFO  = 58  ;
   OCI_FNCODE_LDATOSVCCTX =59         ;
   OCI_FNCODE_RESULTSETTOSTMT =60     ;
   OCI_FNCODE_STMTSETPIECEINFO  = 61  ;
   OCI_FNCODE_TRANSFORGET =62         ;
   OCI_FNCODE_TRANSPREPARE =63        ;
   OCI_FNCODE_TRANSROLLBACK  =64      ;

   OCI_FNCODE_DEFINEBYPOS =65         ;
   OCI_FNCODE_BINDBYPOS =66           ;
   OCI_FNCODE_BINDBYNAME =67          ;

   OCI_FNCODE_LOBASSIGN = 68          ;
   OCI_FNCODE_LOBISEQUAL = 69         ;
   OCI_FNCODE_LOBISINIT = 70          ;
   OCI_FNCODE_LOBLOCATORSIZE = 71     ;
   OCI_FNCODE_LOBCHARSETID  =72       ;
   OCI_FNCODE_LOBCHARSETFORM = 73     ;
   OCI_FNCODE_LOBFILESETNAME = 74     ;
   OCI_FNCODE_LOBFILEGETNAME = 75     ;

   OCI_FNCODE_LOGON =76               ;
   OCI_FNCODE_LOGOFF =77              ;
{-------------------------------------------------------------------------------}
   OCI_ONE_PIECE =0;
   OCI_FIRST_PIECE =1;
   OCI_NEXT_PIECE =2;
   OCI_LAST_PIECE =3;
{-------------------------------------------------------------------------------}

   OCI_STMT_SELECT  =1 ;                             { select statement }
   OCI_STMT_UPDATE  =2 ;                             { update statement }
   OCI_STMT_DELETE  =3 ;                             { delete statement }
   OCI_STMT_INSERT  =4 ;                             { Insert Statement }
   OCI_STMT_CREATE  =5 ;                             { create statement }
   OCI_STMT_DROP    =6 ;                               { drop statement }
   OCI_STMT_ALTER   =7 ;                              { alter statement }
   OCI_STMT_BEGIN   =8 ;                  { begin ... (pl/sql statement)}
   OCI_STMT_DECLARE =9 ;               { declare .. (pl/sql statement ) }
{---------------------------------------------------------------------------}

{--------------------------- OCI Parameter Types ---------------------------}
   OCI_PTYPE_UNK       =0 ;                              { unknown   }
   OCI_PTYPE_TABLE     =1 ;                              { table     }
   OCI_PTYPE_VIEW      =2 ;                              { view      }
   OCI_PTYPE_PROC      =3 ;                              { procedure }
   OCI_PTYPE_FUNC      =4 ;                              { function  }
   OCI_PTYPE_PKG       =5 ;                              { package   }
   OCI_PTYPE_TYPE      =6 ;                      { user-defined type }
   OCI_PTYPE_SYN       =7 ;                              { synonym   }
   OCI_PTYPE_SEQ       =8 ;                              { sequence  }
   OCI_PTYPE_COL       =9 ;                              { column    }
   OCI_PTYPE_ARG       =10;                              { argument  }
   OCI_PTYPE_LIST      =11;                              { list      }
   OCI_PTYPE_TYPE_ATTR =    12 ;         { user-defined type's attribute }
   OCI_PTYPE_TYPE_COLL =    13 ;             { collection type's element }
   OCI_PTYPE_TYPE_METHOD =  14 ;            { user-defined type's method }
   OCI_PTYPE_TYPE_ARG    =  15 ;   { user-defined type method's argument }
   OCI_PTYPE_TYPE_RESULT =  16 ;     { user-defined type method's result }
{---------------------------------------------------------------------------}

{----------------------------- OCI List Types ------------------------------}
   OCI_LTYPE_UNK          = 0  ;                             { unknown   }
   OCI_LTYPE_COLUMN       = 1  ;                           { column list }
   OCI_LTYPE_ARG_PROC     = 2  ;               { procedure argument list }
   OCI_LTYPE_ARG_FUNC     = 3  ;                { function argument list }
   OCI_LTYPE_SUBPRG       = 4  ;                       { subprogram list }
   OCI_LTYPE_TYPE_ATTR    = 5  ;                        { type attribute }
   OCI_LTYPE_TYPE_METHOD  = 6  ;                           { type method }
   OCI_LTYPE_TYPE_ARG_PROC =7  ;  { type method w/o result argument list }
   OCI_LTYPE_TYPE_ARG_FUNC =8  ;    { type method w/result argument list }
{---------------------------------------------------------------------------}

   EWORDMAXVAL   = 32767;
   EWORDMINVAL   =     0;
   UWORDMAXVAL   = 65535;
   UWORDMINVAL   =     0;
   SWORDMAXVAL   = 32767;
   SWORDMINVAL   = -32767;
   MINEWORDMAXVAL  =32767;
   MAXEWORDMINVAL  =0;
   MINUWORDMAXVAL  =65535;
   MAXUWORDMINVAL  =0;
   MINSWORDMAXVAL  =32767;
   MAXSWORDMINVAL  =-32767;

   UB1MINVAL =0;
   SB1MAXVAL =127;
   SB1MINVAL =-127;
   MINEB1MAXVAL =127;
   MAXEB1MINVAL =0;
   MINUB1MAXVAL =255;
   MAXUB1MINVAL =0;
   MINSB1MAXVAL =127;
   MAXSB1MINVAL =-127;

   EB2MAXVAL    =32767;
   EB2MINVAL    =0;
   UB2MAXVAL    =65535;
   UB2MINVAL    =0;
   SB2MAXVAL    =32767;
   SB2MINVAL    =-32767;
   MINEB2MAXVAL =32767;
   MAXEB2MINVAL =0;
   MINUB2MAXVAL =65535;
   MAXUB2MINVAL =0;
   MINSB2MAXVAL =32767;
   MAXSB2MINVAL =-32767;

   EB4MAXVAL    =  2147483647;
   EB4MINVAL    =    0;
   UB4MAXVAL    = 4294967295;
   UB4MINVAL    =    0;
   SB4MAXVAL    =  2147483647;
   SB4MINVAL    =  -2147483647;
   MINEB4MAXVAL = 2147483647;
   MAXEB4MINVAL =      0;
   MINUB4MAXVAL = 4294967295;
   MAXUB4MINVAL =      0;
   MINSB4MAXVAL = 2147483647;
   MAXSB4MINVAL =-2147483647;


   CDA_SIZE = 64;

   OCI_EV_DEF =0   ;               { default single-threaded environment }
   OCI_EV_TSF =1   ;                           { thread-safe environment }

{ OCI Logon Modes for olog call }
   OCI_LM_DEF =0    ;                                    { default login }
   OCI_LM_NBL =1    ;                               { non-blocking logon }

{ OCI_*_PIECE defines the piece types that are returned or set }
   //OCI_ONE_PIECE   =0 ;                { there or this is the only piece }
   //OCI_FIRST_PIECE =1 ;                       { the first of many pieces }
   //OCI_NEXT_PIECE  =2 ;                        { the next of many pieces }
   //OCI_LAST_PIECE  =3 ;                  { the last piece of this column }

{ input data types }
   SQLT_CHR  =1        ;                { (ORANET TYPE) character string }
   SQLT_NUM  =2        ;                  { (ORANET TYPE) oracle numeric }
   SQLT_INT  =3        ;                         { (ORANET TYPE) integer }
   SQLT_FLT  =4        ;           { (ORANET TYPE) Floating point number }
   SQLT_STR  =5        ;                        { zero terminated string }
   SQLT_VNU  =6        ;                { NUM with preceding length byte }
   SQLT_PDN  =7        ;          { (ORANET TYPE) Packed Decimal Numeric }
   SQLT_LNG  =8        ;                                          { long }
   SQLT_VCS  =9        ;                     { Variable character string }
   SQLT_NON  =10       ;               { Null/empty PCC Descriptor entry }
   SQLT_RID  =11       ;                                         { rowid }
   SQLT_DAT  =12       ;                         { date in oracle format }
   SQLT_VBI  =15       ;                          { binary in VCS format }
   SQLT_BIN  =23       ;                           { binary data(DTYBIN) }
   SQLT_LBI  =24       ;                                   { long binary }
   SQLT_UIN  =68       ;                              { unsigned integer }
   SQLT_SLS  =91       ;                 { Display sign leading separate }
   SQLT_LVC  =94       ;                           { Longer longs (char) }
   SQLT_LVB  =95       ;                           { Longer long binary }
   SQLT_AFC  =96       ;                               { Ansi fixed char }
   SQLT_AVC  =97       ;                                 { Ansi Var char }
   SQLT_CUR  =102      ;                                  { cursor  type }
   SQLT_RDD  =104      ;                              { rowid descriptor }
   SQLT_LAB  =105      ;                                    { label type }
   SQLT_OSL  =106      ;                                  { oslabel type }

   SQLT_NTY  =108      ;                             { named object type }
   SQLT_REF  =110      ;                                      { ref type }
   SQLT_CLOB =112      ;                                 { character lob }
   SQLT_BLOB =113      ;                                    { binary lob }
   SQLT_BFILE =114     ;                               { binary file lob }
   SQLT_CFILE =115     ;                            { character file lob }
   SQLT_RSET =116      ;                               { result set type }
   SQLT_NCO  =122      ;{ named collection type (varray or nested table) }
   SQLT_VST  =155      ;                                { OCIString type }
   SQLT_ODT  =156      ;                                  { OCIDate type }

{ cxcheng: this has been added for backward compatibility -it needs to be here because ocidfn.h can get included ahead of sqldef.h }
   SQLT_FILE =SQLT_BFILE ;                             { binary file lob }
   SQLT_CFIL =SQLT_CFILE ;
   SQLT_BFIL =SQLT_BFILE ;

{ CHAR/NCHAR/VARCHAR2/NVARCHAR2/CLOB/NCLOB char set "form" information }
   SQLCS_IMPLICIT =1  ;   { for CHAR, VARCHAR2, CLOB w/o a specified set }
   SQLCS_NCHAR    =2  ;                { for NCHAR, NCHAR VARYING, NCLOB }
   SQLCS_EXPLICIT =3  ; { for CHAR, etc, with "CHARACTER SET ..." syntax }
   SQLCS_FLEXIBLE =4  ;               { for PL/SQL "flexible" parameters }
   SQLCS_LIT_NULL =5  ;    { for typecheck of NULL and empty_clob() lits }

   HOSTDATAAREA = 512;

{-------------------------- OBJECT Duration --------------------------------}

  OCI_DURATION_BEGIN = 10;
{beginning sequence of duration}
  OCI_DURATION_NULL = OCI_DURATION_BEGIN - 1;
{null duration}
  OCI_DURATION_DEFAULT = OCI_DURATION_BEGIN - 2; { default }
  OCI_DURATION_NEXT = OCI_DURATION_BEGIN-3;
{next special duration}
  OCI_DURATION_SESSION = OCI_DURATION_BEGIN;
{the end of user session}
  OCI_DURATION_TRANS = OCI_DURATION_BEGIN + 1;
{the end of user transaction}

type

  OCITypeGetOpt = (OCI_TYPEGET_HEADER, OCI_TYPEGET_ALL);

   eword = Longint;           { use where sign not important }
   uword = Longword;      { use where unsigned important }
   sword = Longint;           { use where   signed important }
   peword = ^eword;
   puword = ^uword;
   psword = ^sword;
   ppeword = ^peword;
   ppuword = ^puword;
   ppsword = ^psword;

   eb1 = Shortint;          { use where sign not important }
   ub1 = Byte;          { use where unsigned important }
   sb1 = Shortint;          { use where   signed important }
   peb1 = ^eb1;
   pub1 = ^ub1;
   psb1 = ^sb1;
   ppeb1 = ^peb1;
   ppub1 = ^pub1;
   ppsb1 = ^psb1;

   eb2 = Smallint;           { use where sign not important }
   ub2 = Word;           { use where unsigned important }
   sb2 = Smallint;           { use where   signed important }
   peb2 = ^eb2;
   pub2 = ^ub2;
   psb2 = ^sb2;
   ppeb2 = ^peb2;
   ppub2 = ^pub2;
   ppsb2 = ^psb2;

   eb4 = Longint;          { use where sign not important }
   ub4 = Longword;          {* use where unsigned important}
   sb4 = Longint;          {* use where   signed important }
   peb4 = ^eb4;
   pub4 = ^ub4;
   psb4 = ^sb4;
   ppeb4 = ^peb4;
   ppub4 = ^pub4;
   ppsb4 = ^psb4;

   ubig_ora = Longword;         { use where unsigned important }
   sbig_ora = Longint;         { use where   signed important }
   pubig_ora = ^ubig_ora;
   psbig_ora = ^sbig_ora;
   ppubig_ora = ^pubig_ora;
   ppsbig_ora = ^psbig_ora;

   ub8 = Int64;
   sb8 = Int64;
   pub8 = ^ub8;
   psb8 = ^sb8;
   ppub8 = ^pub8;
   ppsb8 = ^psb8;

   size_t = Longword;
   psize_t = ^size_t;
   ppsize_t = ^psize_t;

   bool = Longword;
   pbool = ^bool;

{---------------------LDA CDA DEF-------------------------------------------}

{$A-}

   rd = packed record
      rcs4: ub4;
      rcs5: ub2;
      rcs6: ub1;
   end;

   rid = packed record                         { rowid structure }
      rd1: rd;
      rcs7: ub4;
      rcs8: ub2;
   end;

   cda_head = packed record
      v2_rc: sb2;                              { V2 return code }
      ft: ub2;                                 { SQL function type }
      rpc: ub4;                                { rows processed count }
      peo: ub2;                                { parse error offset }
      fc: ub1;                                 { OCI function code }
      rcs1: ub1;                               { filler area }
      rc: ub2;                                 { V7 return code }
      wrn: ub1;                                { warning flags }
      rcs2: ub1;                               { reserved }
      rcs3: sword;                             { reserved }
      rid1: rid;
      ose: sword;                              { OSD dependent error }
      chk: ub1;
      rcsp: Pointer;                           { pointer to reserved area }
   end;

   cda_def = packed record
      v2_rc: sb2;                              { V2 return code }
      ft: ub2;                                 { SQL function type }
      rpc: ub4;                                { rows processed count }
      peo: ub2;                                { parse error offset }
      fc: ub1;                                 { OCI function code }
      rcs1: ub1;                               { filler area }
      rc: ub2;                                 { V7 return code }
      wrn: ub1;                                { warning flags }
      rcs2: ub1;                               { reserved }
      rcs3: sword;                             { reserved }
      rid1: rid;
      ose: sword;                              { OSD dependent error }
      chk: ub1;
      rcsp: Pointer;                           { pointer to reserved area }
      rcs9: packed array [1..(CDA_SIZE - sizeof(cda_head))] of ub1;   { filler }
   end;

{$A+}

   {the logon data area (LDA) is the same shape as the CDA}
   lda_def = cda_def;
   pCda_Def = ^Cda_Def;
   ppCda_Def = ^pCda_Def;
   pLda_Def = ^Lda_Def;
   ppLda_Def = ^pLda_Def;

{---------------------------------------------------------------------------}
{-----------------------Handle Definitions----------------------------------}
   pPointer = ^Pointer;
   pPChar = ^PChar;
   { OCI environment handle }
   pOCIEnv = Pointer;
   ppOCIEnv = ^pOCIEnv;
   { OCI error handle }
   pOCIError = Pointer;
   ppOCIError = ^pOCIError;
   { OCI service handle }
   pOCISvcCtx = Pointer;
   ppOCISvcCtx = ^pOCISvcCtx;
   { OCI statement handle }
   pOCIStmt = Pointer;
   ppOCIStmt = ^pOCIStmt;
   { OCI bind handle }
   pOCIBind = Pointer;
   ppOCIBind = ^pOCIBind;
   { OCI Define handle }
   pOCIDefine = Pointer;
   ppOCIDefine = ^pOCIDefine;
   { OCI Describe handle }
   pOCIDescribe = Pointer;
   ppOCIDescribe = ^pOCIDescribe;
   { OCI Server handle }
   pOCIServer = Pointer;
   ppOCIServer = ^pOCIServer;
   { OCI Authentication handle }
   pOCISession = Pointer;
   ppOCISession = ^pOCISession;
   { OCI COR handle }
   pOCIComplexObject = Pointer;
   ppOCIComplexObject = ^pOCIComplexObject;
   { OCI Transaction handle }
   pOCITrans = Pointer;
   ppOCITrans = ^pOCITrans;
   { OCI Security handle }
   pOCISecurity = Pointer;
   ppOCISecurity = ^pOCISecurity;
{-----------------------Descriptor Definitions------------------------------}
   { OCI snapshot descriptor }
   pOCISnapshot = Pointer;
   ppOCISnapshot = ^pOCISnapshot;
   { OCI Result Set Descriptor }
   pOCIResult = Pointer;
   ppOCIResult = ^pOCIResult;
   { OCI Lob Locator descriptor }
   pOCILobLocator = Pointer;
   ppOCILobLocator = ^pOCILobLocator;
   { OCI PARameter descriptor }
   pOCIParam = Pointer;
   ppOCIParam = ^pOCIParam;
   { OCI COR descriptor }
   pOCIComplexObjectComp = Pointer;
   ppOCIComplexObjectComp = ^pOCIComplexObjectComp;
   { OCI ROWID descriptor }
   pOCIRowid  = Pointer;
   ppOCIRowid = ^pOCIRowid;

   { Lob typedefs for Pro*C }
   { OCI Character LOB Locator }
   //pOCILobLocator = Pointer;
   //ppOCIClobLocator = ^pOCIClobLocator;
   { OCI Binary LOB Locator }
   pOCIBlobLocator = pOCILobLocator;
   ppOCIBlobLocator = ppOCILobLocator;
   { OCI Binary LOB File Locator }
   pOCIBFileLocator = pOCILobLocator;
   ppOCIBFileLocator = ppOCILobLocator;

   pOCIDescriptor = Pointer;
   ppOCIDescriptor = ^pOCIDescriptor;

{----------------------------- TYPE DESCRIPTION ----------------------------}
   pOCIType = Pointer;
   ppOCIType = ^pOCIType;
{------------------------- TYPE ELEMENT DESCRIPTION ------------------------}
   pOCITypeElem = Pointer;
   ppOCITypeElem = ^pOCITypeElem;
{--------------------------- METHOD DESCRIPTION ---------------------------}
   pOCITypeMethod = Pointer;
   ppOCITypeMethod = ^pOCITypeMethod;
{--------------------------- TYPE ACCESS ITERATOR --------------------------}
   pOCITypeIter = Pointer;
   ppOCITypeIter = ^pOCITypeIter;
{---------------------------------------------------------------------------}

   Tmalocfp = function(ctxp: Pointer; size: size_t): Pointer; cdecl;
   Tralocfp = function(ctxp: Pointer; memptr: Pointer; newsize: size_t): Pointer; cdecl;
   Tmfreefp = procedure(ctxp: Pointer; memptr: Pointer); cdecl;


{-----------------------Dynamic Callback Function Pointers------------------}

   OCICallbackInBind = function( ictxp:Pointer;
                                 bindp: pOCIBind;
                                 iter: ub4;
                                 index: ub4;
                                 bufpp: pPointer;
                                 alenp: pub4;
                                 piecep: pub1;
                                 indp: pPointer): sb4; cdecl;

   OCICallbackOutBind = function(   octxp: Pointer;
                                    bindp: pOCIBind;
                                    iter: ub4;
                                    index: ub4;
                                    bufpp: pPointer;
                                    alenp: ppub4;
                                    piecer: pub1;
                                    indp: pPointer;
                                    rcodep: ppub2): sb4; cdecl;

   OCICallbackDefine = function( octxp: Pointer;
                                 defnp: pOCIDefine;
                                 iter: ub4;
                                 bufpp: pPointer;
                                 alenp: ppub4;
                                 piecep: pub1;
                                 indp: pPointer;
                                 rcodep: ppub2): sb4; cdecl;

   OCICallbackLobRead = function(   ctxp: Pointer;
                                    bufp: Pointer;
                                    len: ub4;
                                    piece: ub1): sb4; cdecl;

   OCICallbackLobWrite = function(  ctxp: Pointer;
                                    bufp: Pointer;
                                    lenp: pub4;
                                    piece: pub1): sb4; cdecl;

{--------------------------Failover Callback Structure ---------------------}
   OCICallbackFailover = function(  svcctx: Pointer;
                                    envctx: Pointer;
                                    fo_ctx: Pointer;
                                    fo_type: ub4;
                                    fo_event: ub4): sb4; cdecl;

   OCIFocbkStruct = record
      callback_function: OCICallbackFailover;
      fo_ctx: Pointer;
   end;
{---------------------------------------------------------------------------}
   Tcbfp = function( ctxp: Pointer;
                     bufp: Pointer;
                     len: ub4;
                     piece: ub1): sb4; cdecl;
{---------------------------------------------------------------------------}

{****************************************************************************
                         ACTUAL PROTOTYPE DECLARATIONS
******************************************************************************}
   TOCIInitialize = function(
      mode: ub4;
      ctxp: Pointer;
      malocfp: Tmalocfp;
      ralocfp: Tralocfp;
      mfreefp: Tmfreefp): sword; cdecl;

   TOCIHandleAlloc = function(
      parenth: Pointer;
      hndlpp: pPointer;
      type_h: ub4;
      xtramem_sz: size_t;
      usrmempp: Pointer): sword; cdecl;

   TOCIHandleFree = function(
      hndlp: Pointer;
      type_h: ub4): sword; cdecl;

   TOCIDescriptorAlloc = function(
      parenth: Pointer;
      descpp: pPointer;
      type_h: ub4;
      xtramem_sz: size_t;
      usrmempp: pPointer): sword; cdecl;

   TOCIDescriptorFree = function(
      descp: Pointer;
      type_h: ub4): sword; cdecl;

   TOCIEnvInit = function(
      envp: ppOCIEnv;
      mode: ub4;
      xtramem_sz: size_t;
      usrmempp: Pointer): sword; cdecl;

   TOCIServerAttach = function(
      srvhp: pOCIServer;
      errhp: pOCIError;
      dblink: PChar;
      dblink_len: sb4;
      mode: ub4): sword; cdecl;

   TOCIServerDetach = function(
      srvhp: pOCIServer;
      errhp: pOCIError;
      mode: ub4): sword; cdecl;

   TOCISessionBegin = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      usrhp: pOCISession;
      credt: ub4;
      mode: ub4): sword; cdecl;

   TOCISessionEnd = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      usrhp: pOCISession;
      mode: ub4): sword; cdecl;

   TOCILogon = function(
      envhp: pOCIEnv;
      errhp: pOCIError;
      svchp: ppOCISvcCtx;
      username: PChar;
      uname_len: ub4;
      password: PChar;
      passwd_len: ub4;
      dbname: PChar;
      dbname_len: ub4): sword; cdecl;

   TOCILogoff = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError): sword; cdecl;

   TOCIPasswordChange = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      user_name: PChar;
      usernm_len: ub4;
      opasswd: PChar;
      opasswd_len: ub4;
      npasswd: PChar;
      npasswd_len: ub4;
      mode: ub4): sword; cdecl;

   TOCIStmtPrepare = function(
      stmtp: pOCIStmt;
      errhp: pOCIError;
      stmt: PChar;
      stmt_len: ub4;
      language: ub4;
      mode: ub4): sword; cdecl;

   TOCIBindByPos = function(
      stmtp: pOCIStmt;
      bindp: ppOCIBind;
      errhp: pOCIError;
		  position: ub4;
      valuep: Pointer;
      value_sz: sb4;
      dty: ub2;
      indp: Pointer;
      alenp: pub2;
      rcodep: pub2;
		  maxarr_len: ub4;
      curelep: pub4;
      mode: ub4): sword; cdecl;

   TOCIBindByName = function(
      stmtp: pOCIStmt;
      bindp: ppOCIBind;
      errhp: pOCIError;
			placeholder: PChar;
      placeh_len: sb4;
      valuep: Pointer;
      value_sz: sb4;
      dty: ub2;
      indp: Pointer;
      alenp: pub2;
      rcodep: pub2;
      maxarr_len: ub4;
      curelep: pub4;
      mode: ub4): sword; cdecl;

   TOCIBindObject = function(
      bindp: pOCIBind;
      errhp: pOCIError;
      type_h: pOCIType;
			pgvpp: pPointer;
      pvszsp: pub4;
      indpp: pPointer;
			indszp: pub4): sword; cdecl;

   TOCIBindDynamic = function(
      bindp: pOCIBind;
      errhp:pOCIError;
      ictxp: Pointer;
			icbfp: OCICallbackInBind;
      octxp: Pointer;
			ocbfp: OCICallbackOutBind): sword; cdecl;

   TOCIBindArrayOfStruct = function(
      bindp: pOCIBind;
      errhp: pOCIError;
      pvskip: ub4;
      indskip: ub4;
      alskip: ub4;
      rcskip: ub4): sword; cdecl;

   TOCIStmtGetPieceInfo = function(
      stmtp: pOCIStmt;
      errhp: pOCIError;
      hndlpp: pPointer;
      typep: pub4;
      in_outp: pub1;
      iterp: pub4;
      idxp: pub4;
      piecep: pub1): sword; cdecl;

   TOCIStmtSetPieceInfo = function(
      hndlp: Pointer;
      type_h: ub4;
      errhp: pOCIError;
      bufp: Pointer;
      alenp: pub4;
      piece: ub1;
      indp: Pointer;
      rcodep: pub2): sword; cdecl;

   TOCIStmtExecute = function(
      svchp: pOCISvcCtx;
      stmtp: pOCIStmt;
      errhp: pOCIError;
      iters: ub4;
      rowoff: ub4;
      snap_in: pOCISnapshot;
      snap_out: pOCISnapshot;
      mode: ub4): sword; cdecl;

   TOCIDefineByPos = function(
      stmtp: pOCIStmt;
      defnp: ppOCIDefine;
      errhp: pOCIError;
			position: ub4;
      valuep: Pointer;
      value_sz: sb4;
      dty: ub2;
			indp: Pointer;
      rlenp: pub2;
      rcodep: pub2;
      mode: ub4): sword; cdecl;

   TOCIDefineObject = function(
      defnp: pOCIDefine;
      errhp: pOCIError;
      type_h: pOCIType;
      pgvpp: pPointer;
      pvszsp: pub4;
      indpp: pPointer;
      indszp: pub4): sword; cdecl;

   TOCIDefineDynamic = function(
      defnp: pOCIDefine;
      errhp: pOCIError;
      octxp: Pointer;
      ocbfp: OCICallbackDefine): sword; cdecl;

   TOCIDefineArrayOfStruct = function(
      defnp: pOCIDefine;
      errhpP: pOCIError;
      pvskip: ub4;
      indskip: ub4;
      rlskip: ub4;
      rcskip: ub4): sword; cdecl;

   TOCIStmtFetch = function(
      stmtp: pOCIStmt;
      errhp: pOCIError;
      nrows: ub4;
      orientation: ub2;
      mode: ub4): sword; cdecl;

   TOCIStmtGetBindInfo = function(
      stmtp: pOCIStmt;
      errhp: pOCIError;
      size: ub4;
      startloc: ub4;
      found: psb4;
      bvnp: pPChar;
      bvnl: pub1;
      invp: pPChar;
      inpl: pub1;
      dupl: pub1;
      hndl: ppOCIBind): sword; cdecl;

   TOCIDescribeAny = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      objptr: Pointer;
      objnm_len: ub4;
      objptr_typ: ub1;
      info_level: ub1;
			objtyp: ub1;
      dschp: pOCIDescribe): sword; cdecl;

   TOCIParamGet = function(
      hndlp: Pointer;
      htype: ub4;
      errhp: pOCIError;
      parmdpp: pPointer;
      pos: ub4): sword; cdecl;

   TOCIParamSet = function (
      hdlp: Pointer;
      htyp: ub4;
      errhp: pOCIError;
      dscp: Pointer;
      dtyp: ub4;
      pos: ub4): sword; cdecl;

   TOCITransStart = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      timeout: uword;
      flags: ub4): sword; cdecl;

   TOCITransDetach = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      flags: ub4): sword; cdecl;

   TOCITransCommit = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      flags: ub4): sword; cdecl;

   TOCITransRollback = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      flags: ub4): sword; cdecl;

   TOCITransPrepare = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      flags: ub4): sword; cdecl;

   TOCITransForget = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      flags: ub4): sword; cdecl;

   TOCIErrorGet = function(
      hndlp: Pointer;
      recordno: ub4;
      sqlstate: PChar;
      errcodep: psb4;
      bufp: PChar;
      bufsiz: ub4;
      type_h: ub4): sword; cdecl;

   TOCILobAppend = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      dst_locp: pOCILobLocator;
      src_locp: pOCILobLocator): sword; cdecl;

   TOCILobAssign = function(
      envhp: pOCIEnv;
      errhp: pOCIError;
      src_locp: pOCILobLocator;
      dst_locpp: ppOCILobLocator): sword; cdecl;

   TOCILobCharSetForm = function(
      envhp: pOCIEnv;
      errhp: pOCIError;
      locp: pOCILobLocator;
      csfrm: pub1): sword; cdecl;

   TOCILobCharSetId = function(
      envhp: pOCIEnv;
      errhp: pOCIError;
      locp: pOCILobLocator;
      csid: pub2): sword; cdecl;

   TOCILobCopy = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      dst_locp: pOCILobLocator;
      src_locp: pOCILobLocator;
      amount: ub4;
      dst_offset: ub4;
      src_offset: ub4): sword; cdecl;

   TOCILobDisableBuffering = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      locp: pOCILobLocator): sword; cdecl;

   TOCILobEnableBuffering = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      locp: pOCILobLocator): sword; cdecl;

   TOCILobErase = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      locp: pOCILobLocator;
      amount: ub4;
      offset: ub4): sword; cdecl;

   TOCILobFileClose = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      filep: pOCILobLocator): sword; cdecl;

   TOCILobFileCloseAll = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError): sword; cdecl;

   TOCILobFileExists = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      filep: pOCILobLocator;
			flag: pbool): sword; cdecl;

   TOCILobFileGetName = function(
      envhp: pOCIEnv;
      errhp: pOCIError;
      filep: pOCILobLocator;
      dir_alias: PChar;
      d_length: pub2;
      filename: PChar;
      f_length: pub2): sword; cdecl;

   TOCILobFileIsOpen = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      filep: pOCILobLocator;
      flag: pbool): sword; cdecl;

   TOCILobFileOpen = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      filep: pOCILobLocator;
      mode: ub1): sword; cdecl;

   TOCILobFileSetName = function(
      envhp: pOCIEnv;
      errhp: pOCIError;
      filepp: ppOCILobLocator;
      dir_alias: PChar;
      d_length: ub2;
      filename: PChar;
      f_length: ub2): sword; cdecl;

   TOCILobFlushBuffer = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      locp: pOCILobLocator;
      flag: ub4): sword; cdecl;

   TOCILobGetLength = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      locp: pOCILobLocator;
      lenp: pub4): sword; cdecl;

   TOCILobIsEqual = function(
      envhp: pOCIEnv;
      x: pOCILobLocator;
      y: pOCILobLocator;
      is_equal: pbool): sword; cdecl;

   TOCILobLoadFromFile = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      dst_locp: pOCILobLocator;
      src_filep: pOCILobLocator;
      amount: ub4;
      dst_offset: ub4;
      src_offset: ub4): sword; cdecl;

   TOCILobLocatorIsInit = function(
      envhp: pOCIEnv;
      errhp: pOCIError;
      locp: pOCILobLocator;
      is_initialized: pbool): sword; cdecl;

   TOCILobRead = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      locp: pOCILobLocator;
      amtp: pub4;
      offset: ub4;
      bufp: Pointer;
      bufl: ub4;
      ctxp: Pointer;
      cbfp: Tcbfp;
      csid: ub2;
      csfrm: ub1): sword; cdecl;

   TOCILobTrim = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      locp: pOCILobLocator;
      newlen: ub4): sword; cdecl;

   TOCILobWrite = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      locp: pOCILobLocator;
      amtp: pub4;
      offset: ub4;
      bufp: Pointer;
      buflen: ub4;
      piece: ub1;
      ctxp: Pointer;
      cbfp: Tcbfp;
      csid: ub2;
      csfrm: ub1): sword; cdecl;

   TOCIBreak = function(
      hndlp: Pointer;
      errhp: pOCIError): sword; cdecl;

   TOCIServerVersion = function(
      hndlp: Pointer;
      errhp: pOCIError;
      bufp: PChar;
      bufsz: ub4;
      hndltype: ub1): sword; cdecl;

   TOCIAttrGet = function(
      trgthndlp: Pointer;
      trghndltyp: ub4;
      attributep: Pointer;
      sizep: pub4;
      attrtype: ub4;
      errhp: pOCIError): sword; cdecl;

   TOCIAttrSet = function(
      trgthndlp: Pointer;
      trghndltyp: ub4;
      attributep: Pointer;
      size: ub4;
      attrtype: ub4;
      errhp: pOCIError): sword; cdecl;

   TOCISvcCtxToLda = function(
      svchp: pOCISvcCtx;
      errhp: pOCIError;
      ldap: pLda_Def): sword; cdecl;

   TOCILdaToSvcCtx = function(
      svchpp: ppOCISvcCtx;
      errhp: pOCIError;
      ldap: pLda_Def): sword; cdecl;

   TOCIResultSetToStmt = function(
      rsetdp: pOCIResult;
      errhp: pOCIError): sword; cdecl;

  TOCITypeByName = function(
      env: pOCIEnv;
      err: pOCIError;
      const svc: pOCISvcCtx;
      const schema_name: pChar;
      s_length: ub4;
      const type_name: pChar;
      t_length: ub4;
      const version_name: pChar;
      v_length: ub4;
      pin_duration: ub2;
      get_option: ub2;
      tdo: ppOCIType    ): sword; cdecl;

  //TOCITerminate = function(mode: ub4): sword; cdecl;

  function InitOCI(OCIDllName: String): Boolean;
  procedure TerminateOCI;

  function  Tmalocf(ctxp: Pointer; size: size_t): Pointer; cdecl;
  function  Tralocf(ctxp: Pointer; memptr: Pointer; newsize: size_t): Pointer; cdecl;
  procedure Tmfreef(ctxp: Pointer; memptr: Pointer); cdecl;

var

  OCIInitialize:                  TOCIInitialize;
  OCIHandleAlloc:                 TOCIHandleAlloc;
  OCIHandleFree:                  TOCIHandleFree;
  OCIDescriptorAlloc:             TOCIDescriptorAlloc;
  OCIDescriptorFree:              TOCIDescriptorFree;
  OCIEnvInit:                     TOCIEnvInit;
  OCIServerAttach:                TOCIServerAttach;
  OCIServerDetach:                TOCIServerDetach;
  OCISessionBegin:                TOCISessionBegin;
  OCISessionEnd:                  TOCISessionEnd;
  OCILogon:                       TOCILogon;
  OCILogoff:                      TOCILogoff;
  OCIPasswordChange:              TOCIPasswordChange;
  OCIStmtPrepare:                 TOCIStmtPrepare;
  OCIBindByPos:                   TOCIBindByPos;
  OCIBindByName:                  TOCIBindByName;
  OCIBindObject:                  TOCIBindObject;
  OCIBindDynamic:                 TOCIBindDynamic;
  OCIBindArrayOfStruct:           TOCIBindArrayOfStruct;
  OCIStmtGetPieceInfo:            TOCIStmtGetPieceInfo;
  OCIStmtSetPieceInfo:            TOCIStmtSetPieceInfo;
  OCIStmtExecute:                 TOCIStmtExecute;
  OCIDefineByPos:                 TOCIDefineByPos;
  OCIDefineObject:                TOCIDefineObject;
  OCIDefineDynamic:               TOCIDefineDynamic;
  OCIDefineArrayOfStruct:         TOCIDefineArrayOfStruct;
  OCIStmtFetch:                   TOCIStmtFetch;
  OCIStmtGetBindInfo:             TOCIStmtGetBindInfo;
  OCIDescribeAny:                 TOCIDescribeAny;
  OCIParamGet:                    TOCIParamGet;
  OCIParamSet:                    TOCIParamSet;
  OCITransStart:                  TOCITransStart;
  OCITransDetach:                 TOCITransDetach;
  OCITransCommit:                 TOCITransCommit;
  OCITransRollback:               TOCITransRollback;
  OCITransPrepare:                TOCITransPrepare;
  OCITransForget:                 TOCITransForget;
  OCIErrorGet:                    TOCIErrorGet;
  OCILobAppend:                   TOCILobAppend;
  OCILobAssign:                   TOCILobAssign;
  OCILobCharSetForm:              TOCILobCharSetForm;
  OCILobCharSetId:                TOCILobCharSetId;
  OCILobCopy:                     TOCILobCopy;
  OCILobDisableBuffering:         TOCILobDisableBuffering;
  OCILobEnableBuffering:          TOCILobEnableBuffering;
  OCILobErase:                    TOCILobErase;
  OCILobFileClose:                TOCILobFileClose;
  OCILobFileCloseAll:             TOCILobFileCloseAll;
  OCILobFileExists:               TOCILobFileExists;
  OCILobFileGetName:              TOCILobFileGetName;
  OCILobFileIsOpen:               TOCILobFileIsOpen;
  OCILobFileOpen:                 TOCILobFileOpen;
  OCILobFileSetName:              TOCILobFileSetName;
  OCILobFlushBuffer:              TOCILobFlushBuffer;
  OCILobGetLength:                TOCILobGetLength;
  OCILobIsEqual:                  TOCILobIsEqual;
  OCILobLoadFromFile:             TOCILobLoadFromFile;
  OCILobLocatorIsInit:            TOCILobLocatorIsInit;
  OCILobRead:                     TOCILobRead;
  OCILobTrim:                     TOCILobTrim;
  OCILobWrite:                    TOCILobWrite;
  OCIBreak:                       TOCIBreak;
  OCIServerVersion:               TOCIServerVersion;
  OCIAttrGet:                     TOCIAttrGet;
  OCIAttrSet:                     TOCIAttrSet;
  OCISvcCtxToLda:                 TOCISvcCtxToLda;
  OCILdaToSvcCtx:                 TOCILdaToSvcCtx;
  OCIResultSetToStmt:             TOCIResultSetToStmt;
  OCITypeByName:                  TOCITypeByName;
  //OCITerminate:                   TOCITerminate;

//+++++++++++
  FOCIInitialized: Boolean;
  FOCILibraryLoaded: Boolean;
  FOCIhDll: Integer;
  FOCIDllName: String;

implementation

{$R OCI.res}

function InitOCI(OCIDllName: String): Boolean;

  function GetOCIProcAddress(const AProcName: String): Pointer;
  begin
    Result := GetProcAddress(FOCIhDll, PChar(AProcName));
    if Result = nil then
      raise Exception.Create(Format('Entry point [%s] not found in OCI DLL', [AProcName]))
  end;

begin
  if not FOCILibraryLoaded then
  begin
    FOCIDllName := OCIDllName;
    FOCIhDll := LoadLibrary(PChar(FOCIDllName));
    if FOCIhDll <> 0 then
    begin
      FOCILibraryLoaded := true;
      @OCIInitialize            := GetOCIProcAddress('OCIInitialize');
      @OCIHandleAlloc           := GetOCIProcAddress('OCIHandleAlloc');
      @OCIHandleFree            := GetOCIProcAddress('OCIHandleFree');
      @OCIDescriptorAlloc       := GetOCIProcAddress('OCIDescriptorAlloc');
      @OCIDescriptorFree        := GetOCIProcAddress('OCIDescriptorFree');
      @OCIEnvInit               := GetOCIProcAddress('OCIEnvInit');
      @OCIServerAttach          := GetOCIProcAddress('OCIServerAttach');
      @OCIServerDetach          := GetOCIProcAddress('OCIServerDetach');
      @OCISessionBegin          := GetOCIProcAddress('OCISessionBegin');
      @OCISessionEnd            := GetOCIProcAddress('OCISessionEnd');
      @OCILogon                 := GetOCIProcAddress('OCILogon');
      @OCILogoff                := GetOCIProcAddress('OCILogoff');
      @OCIPasswordChange        := GetOCIProcAddress('OCIPasswordChange');
      @OCIStmtPrepare           := GetOCIProcAddress('OCIStmtPrepare');
      @OCIBindByPos             := GetOCIProcAddress('OCIBindByPos');
      @OCIBindByName            := GetOCIProcAddress('OCIBindByName');
      @OCIBindObject            := GetOCIProcAddress('OCIBindObject');
      @OCIBindDynamic           := GetOCIProcAddress('OCIBindDynamic');
      @OCIBindArrayOfStruct     := GetOCIProcAddress('OCIBindArrayOfStruct');
      @OCIStmtGetPieceInfo      := GetOCIProcAddress('OCIStmtGetPieceInfo');
      @OCIStmtSetPieceInfo      := GetOCIProcAddress('OCIStmtSetPieceInfo');
      @OCIStmtExecute           := GetOCIProcAddress('OCIStmtExecute');
      @OCIDefineByPos           := GetOCIProcAddress('OCIDefineByPos');
      @OCIDefineObject          := GetOCIProcAddress('OCIDefineObject');
      @OCIDefineDynamic         := GetOCIProcAddress('OCIDefineDynamic');
      @OCIDefineArrayOfStruct   := GetOCIProcAddress('OCIDefineArrayOfStruct');
      @OCIStmtFetch             := GetOCIProcAddress('OCIStmtFetch');
      @OCIStmtGetBindInfo       := GetOCIProcAddress('OCIStmtGetBindInfo');
      @OCIDescribeAny           := GetOCIProcAddress('OCIDescribeAny');
      @OCIParamGet              := GetOCIProcAddress('OCIParamGet');
      @OCIParamSet              := GetOCIProcAddress('OCIParamSet');
      @OCITransStart            := GetOCIProcAddress('OCITransStart');
      @OCITransDetach           := GetOCIProcAddress('OCITransDetach');
      @OCITransCommit           := GetOCIProcAddress('OCITransCommit');
      @OCITransRollback         := GetOCIProcAddress('OCITransRollback');
      @OCITransPrepare          := GetOCIProcAddress('OCITransPrepare');
      @OCITransForget           := GetOCIProcAddress('OCITransForget');
      @OCIErrorGet              := GetOCIProcAddress('OCIErrorGet');
      @OCILobAppend             := GetOCIProcAddress('OCILobAppend');
      @OCILobAssign             := GetOCIProcAddress('OCILobAssign');
      @OCILobCharSetForm        := GetOCIProcAddress('OCILobCharSetForm');
      @OCILobCharSetId          := GetOCIProcAddress('OCILobCharSetId');
      @OCILobCopy               := GetOCIProcAddress('OCILobCopy');
      @OCILobDisableBuffering   := GetOCIProcAddress('OCILobDisableBuffering');
      @OCILobEnableBuffering    := GetOCIProcAddress('OCILobEnableBuffering');
      @OCILobErase              := GetOCIProcAddress('OCILobErase');
      @OCILobFileClose          := GetOCIProcAddress('OCILobFileClose');
      @OCILobFileCloseAll       := GetOCIProcAddress('OCILobFileCloseAll');
      @OCILobFileExists         := GetOCIProcAddress('OCILobFileExists');
      @OCILobFileGetName        := GetOCIProcAddress('OCILobFileGetName');
      @OCILobFileIsOpen         := GetOCIProcAddress('OCILobFileIsOpen');
      @OCILobFileOpen           := GetOCIProcAddress('OCILobFileOpen');
      @OCILobFileSetName        := GetOCIProcAddress('OCILobFileSetName');
      @OCILobFlushBuffer        := GetOCIProcAddress('OCILobFlushBuffer');
      @OCILobGetLength          := GetOCIProcAddress('OCILobGetLength');
      @OCILobIsEqual            := GetOCIProcAddress('OCILobIsEqual');
      @OCILobLoadFromFile       := GetOCIProcAddress('OCILobLoadFromFile');
      @OCILobLocatorIsInit      := GetOCIProcAddress('OCILobLocatorIsInit');
      @OCILobRead               := GetOCIProcAddress('OCILobRead');
      @OCILobTrim               := GetOCIProcAddress('OCILobTrim');
      @OCILobWrite              := GetOCIProcAddress('OCILobWrite');
      @OCIBreak                 := GetOCIProcAddress('OCIBreak');
      @OCIServerVersion         := GetOCIProcAddress('OCIServerVersion');
      @OCIAttrGet               := GetOCIProcAddress('OCIAttrGet');
      @OCIAttrSet               := GetOCIProcAddress('OCIAttrSet');
      @OCISvcCtxToLda           := GetOCIProcAddress('OCISvcCtxToLda');
      @OCILdaToSvcCtx           := GetOCIProcAddress('OCILdaToSvcCtx');
      @OCIResultSetToStmt       := GetOCIProcAddress('OCIResultSetToStmt');
      @OCITypeByName            := GetOCIProcAddress('OCITypeByName');
      //@OCITerminate             := GetOCIProcAddress('OCITerminate');
      if OCIInitialize( ub4(OCI_THREADED or OCI_OBJECT),
                        nil,
                        @Tmalocf,
                        @Tralocf,
                        @Tmfreef)= OCI_SUCCESS then
      //if OCIInitialize( ub4(OCI_THREADED or OCI_OBJECT),
      //                  nil,
      //                  nil,
      //                  nil,
      //                  nil)= OCI_SUCCESS then
      begin
        FOCIInitialized := true;
        Result := true
      end else
        raise Exception.Create('Error when initializing OCI environment');
    end else
      raise Exception.Create(Format('Error when loading OCI dll [%s]', [FOCIDllName]));
  end else
    Result := true;
end;

procedure TerminateOCI;
begin
  if FOCIInitialized then
  begin
    //OCITerminate(OCI_DEFAULT);
    FOCIInitialized := False;
  end;
  if FOCILibraryLoaded then
  begin
    FreeLibrary(FOCIhDll);
    FOCILibraryLoaded := false;
  end;
end;

function Tmalocf(ctxp: Pointer; size: size_t): Pointer; cdecl;
begin
  Result := MemMgr.oMem.GetMem('OCI Env', size);
end;

function Tralocf(ctxp: Pointer; memptr: Pointer; newsize: size_t): Pointer; cdecl;
begin
  Result := MemMgr.oMem.ReallocMem(memptr, newsize);
end;

procedure Tmfreef(ctxp: Pointer; memptr: Pointer); cdecl;
begin
  MemMgr.oMem.FreeMem(memptr);
end;

initialization

    FOCIInitialized := false;
    FOCILibraryLoaded := false;

finalization

    TerminateOCI;

end.
