unit Streams;

interface

uses
  Classes, MemMgr, SysUtils, Windows, Consts;

const
  BLOCK_SIZE = 4096;

type

  {TAllStream}
  TAllStream = class(TStream)
  public
    procedure qq; virtual; abstract;
  end;

  {TMemStream}
  TMemStream = class(TAllStream)
  private
    FMemory: Pointer;
    FSize, FPosition: Longint;
  public
    constructor Create;
    destructor Destroy; override;
    procedure SetSize(NewSize: Longint); override;
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    procedure qq; override;
  end;

  {TMyHandleStream class}
  TMyHandleStream = class(TAllStream)
  private
    FHandle: Integer;
  protected
    procedure SetSize(NewSize: Longint); override;
  public
    constructor Create(AHandle: Integer);
    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;
    procedure qq; override;
    property Handle: Integer read FHandle;
  end;

  {TMyFileStream class}
  TMyFileStream = class(TMyHandleStream)
  public
    constructor Create(const FileName: string; Mode: Word);
    destructor Destroy; override;
    procedure qq; override;
  end;

  {TFStream}
  TFStream = class(TMyFileStream)
  public
    procedure qq; override;
  end;

implementation

{ TMemStream }

constructor TMemStream.Create;
begin
  inherited Create;
  FMemory := nil; //MemMgr.oMem.GetMem(BLOCK_SIZE);
  FSize := 0;
  FPosition := 0;
end;

destructor TMemStream.Destroy;
begin
  if FMemory <> nil then MemMgr.oMem.FreeMem(FMemory);
  FMemory := nil;
  FSize := 0;
  FPosition := 0;
  inherited Destroy;
end;

procedure TMemStream.qq;
begin
  FSize := 0;
  FPosition := 0;
end;

function TMemStream.Read(var Buffer; Count: Integer): Longint;
begin
  if (FPosition >= 0) and (Count >= 0) then
  begin
    Result := FSize - FPosition;
    if Result > 0 then
    begin
      if Result > Count then Result := Count;
      System.Move(Pointer(Longint(FMemory) + FPosition)^, Buffer, Result);
      Inc(FPosition, Result);
      Exit;
    end;
  end;
  Result := 0;
end;

function TMemStream.Seek(Offset: Integer; Origin: Word): Longint;
begin
  case Origin of
    soFromBeginning: FPosition := Offset;
    soFromCurrent: Inc(FPosition, Offset);
    soFromEnd: FPosition := FSize + Offset;
  end;
  Result := FPosition;
end;

procedure TMemStream.SetSize(NewSize: Integer);
var
  OldPosition: Longint;
  LSize, nBlock: Integer;
begin
  OldPosition := FPosition;
  LSize := MemMgr.oMem.GetSize(FMemory);
  if NewSize > LSize then
  begin
    nBlock := (NewSize - LSize) div BLOCK_SIZE + 1;
    FMemory := MemMgr.oMem.ReallocMem(FMemory, nBlock * BLOCK_SIZE);
  end;
  FSize := NewSize;
  if OldPosition > NewSize then Seek(0, soFromEnd);
end;

function TMemStream.Write(const Buffer; Count: Integer): Longint;
var
  Pos: Longint;
  LSize, nBlock: Integer;
begin
  if (FPosition >= 0) and (Count >= 0) then
  begin
    Pos := FPosition + Count;
    if Pos > 0 then
    begin
      if Pos > FSize then
      begin
        LSize := MemMgr.oMem.GetSize(FMemory);
        if Pos > LSize then begin
          nBlock := Pos div BLOCK_SIZE + 1;
          FMemory := MemMgr.oMem.ReallocMem(FMemory, nBlock * BLOCK_SIZE);
        end;
        FSize := Pos;
      end;
      System.Move(Buffer, Pointer(Longint(FMemory) + FPosition)^, Count);
      FPosition := Pos;
      Result := Count;
      Exit;
    end;
  end;
  Result := 0;
end;

{ TFStream }

procedure TFStream.qq;
begin
  Seek(0, soFromBeginning);
  Size := 0;
end;

{ TMyHandleStream }

constructor TMyHandleStream.Create(AHandle: Integer);
begin
  FHandle := AHandle;
end;

procedure TMyHandleStream.qq;
begin
  inherited;
end;

function TMyHandleStream.Read(var Buffer; Count: Integer): Longint;
begin
  Result := FileRead(FHandle, Buffer, Count);
  if Result = -1 then Result := 0;
end;

function TMyHandleStream.Seek(Offset: Integer; Origin: Word): Longint;
begin
  Result := FileSeek(FHandle, Offset, Origin);
end;

procedure TMyHandleStream.SetSize(NewSize: Integer);
begin
  Seek(NewSize, soFromBeginning);
  Win32Check(SetEndOfFile(FHandle));
end;

function TMyHandleStream.Write(const Buffer; Count: Integer): Longint;
begin
  Result := FileWrite(FHandle, Buffer, Count);
  if Result = -1 then Result := 0;
end;

{ TMyFileStream }

constructor TMyFileStream.Create(const FileName: string; Mode: Word);
begin
  if Mode = fmCreate then
  begin
    FHandle := FileCreate(FileName);
    if FHandle < 0 then
      raise EFCreateError.CreateResFmt(@SFCreateError, [FileName]);
  end else
  begin
    FHandle := FileOpen(FileName, Mode);
    if FHandle < 0 then
      raise EFOpenError.CreateResFmt(@SFOpenError, [FileName]);
  end;
end;

destructor TMyFileStream.Destroy;
begin
  if FHandle >= 0 then FileClose(FHandle);
end;

procedure TMyFileStream.qq;
begin
  inherited;
end;

end.
