{Copyright:      Vlad Karpov  mailto:KarpovVV@protek.ru
 Author:         Vlad Karpov
 Remarks:        freeware, but this Copyright must be included

 Description:    To8DataSet - DataSet component for ORACLE SQL

 Version:        1.0.0

 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}
unit o8DataSet;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, o8Stmt, o8Env, o8Session, o8StmtOrd, OCI, Math, o8Filter, DBcommon, MemMgr;

type

  {$A-}
  TRecInfo = packed record
     RecordRowID: ub4;
     UpdateStatus: TUpdateStatus;
     BookmarkFlag: TBookmarkFlag;
  end;
  {$A+}
  pTRecInfo = ^TRecInfo;
  ppTRecInfo = ^pTRecInfo;

  pDouble = ^Double;
  pInteger = ^Integer;

  To8DataSet = class;

  TRecNoStatus = (rnDbase, rnParadox, rnNotSupported);
  TUpdateAction = (uaFail, uaAbort, uaSkip, uaRetry, uaApplied);
  TUpdateRecordTypes = set of (rtModified, rtInserted, rtDeleted, rtUnmodified);
  TUpdateErrorEvent = procedure(DataSet: TDataSet; E: EDatabaseError;
    UpdateKind: TUpdateKind; var UpdateAction: TUpdateAction) of object;
  TUpdateRecordEvent = procedure(DataSet: TDataSet; UpdateKind: TUpdateKind;
    var UpdateAction: TUpdateAction) of object;

  {To8DataSetUpdateObject}
  To8DataSetUpdateObject = class(TComponent)
  protected
    function GetDataSet: To8DataSet; virtual; abstract;
    procedure SetDataSet(ADataSet: To8DataSet); virtual; abstract;
    procedure Apply(UpdateKind: TUpdateKind); virtual; abstract;
    property DataSet: To8DataSet read GetDataSet write SetDataSet;
  end;

  {TVKDL1}
  TVKDL1 = class(TDataLink)
  private
    Fo8DataSet: To8DataSet;
  protected
    procedure ActiveChanged; override;
    procedure RecordChanged(Field: TField); override;
    procedure CheckBrowseMode; override;
  public
    property o8DataSet: To8DataSet read Fo8DataSet write Fo8DataSet;
  end;

  {To8DataSet}
  To8DataSet = class(TDataSet)
  private
    { Private declarations }
    FSTMT: To8Stmt;
    FFilterParser: To8ExprParser;
    FStreamedActive: boolean;
    FRecordBuffer: TList;
    FDeletedRecord: TList;
    FCurrentIndex: sb4;
    FRowID: ub4;
    FTempRecord: pChar;
    FFilterRecord: pChar;
    FUpdateObject: To8DataSetUpdateObject;
    FLocateType: TLocateType;
    FFetchAll: boolean;
    FIsLookup: Boolean;
    FDataLink: TVKDL1;
    procedure ClearRecordBuffers;
    procedure ClearDeletedRecordBuffers;
    procedure SetFSTMT(oStatment: To8Stmt);
    function GetRecordBufferSize: Integer;
    property RecordBufferSize: Integer read GetRecordBufferSize;
    function GetRecordUpdateStatus: TUpdateStatus;
    procedure SetUpdateObject(Value: To8DataSetUpdateObject);
    procedure SetRecordUpdateStatus(const Value: TUpdateStatus);
    function GetActiveRecBuf(var RecBuf: PChar): Boolean;
    function GetRecordByBuffer(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
    function AcceptRecord: Boolean;
    procedure SetParamsFromCursor;
    procedure SetDataSource(const Value: TDataSource);
    procedure ReopenFromSrc;
  protected
    { Protected declarations }
    procedure SetActive(Value: boolean); override;
    procedure Loaded; override;

    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function GetRecordSize: Word; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;

    function GetRecordCount: Integer; override;
    function GetRecNo: Integer; override;

    function GetFldType(rs: pRStruct): TFieldType;

    function LocateRecord(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions; SyncCursor: Boolean): Boolean;
    function FindRecord(Restart, GoForward: Boolean): Boolean; override;

    function GetStateFieldValue(State: TDataSetState; Field: TField): Variant; override;
    function CompareLocateField(const Fields: TList; const KeyValues: Variant; Options: TLocateOptions): Integer;

    procedure SetFiltered(Value: Boolean); override;

    function GetDataSource: TDataSource; override;

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure ApplyUpdates;

    function Locate(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions): Boolean; override;
    function Lookup(const KeyFields: string; const KeyValues: Variant; const ResultFields: string): Variant; override;
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    function GetOldFieldData(Field: TField): Variant;
    function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    function GetAsRowID(Field: TField): pOCIRowid;
    function GetAsResultSet(Field: TField): pOCIResult;

    procedure FetchAll;

    property CurrentIndex: sb4 read FCurrentIndex;
    property RecordUpdateStatus: TUpdateStatus read GetRecordUpdateStatus write SetRecordUpdateStatus;
    property IsLookup: boolean read FIsLookup write FIsLookup;

  published
    { Published declarations }
    property Active;
    property STMT: To8Stmt read FSTMT write SetFSTMT;
    property LocateType: TLocateType read FLocateType write FLocateType;
    property UpdateObject: To8DataSetUpdateObject read FUpdateObject write SetUpdateObject;

    property DataSource: TDataSource read GetDataSource write SetDataSource;

    property Filter;
    property Filtered;
    property FilterOptions;

    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;

  end;

{ To8UpdateSQL }

  To8UpdateSQL = class(To8DataSetUpdateObject)
  private
    FDataSet: To8DataSet;
    FQueries: array[TUpdateKind] of To8StmtOrd;
    FSQLText: array[TUpdateKind] of TStrings;
    function GetQuery(UpdateKind: TUpdateKind): To8StmtOrd;
    function GetSQL(UpdateKind: TUpdateKind): TStrings;
    function GetSQLIndex(Index: Integer): TStrings;
    procedure SetSQL(UpdateKind: TUpdateKind; Value: TStrings);
    procedure SetSQLIndex(Index: Integer; Value: TStrings);
  protected
    function GetDataSet: To8DataSet; override;
    procedure SetDataSet(ADataSet: To8DataSet); override;
    procedure SQLChanged(Sender: TObject);
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure Apply(UpdateKind: TUpdateKind); override;
    procedure ExecSQL(UpdateKind: TUpdateKind);
    procedure SetParams(UpdateKind: TUpdateKind);
    property DataSet;
    property Query[UpdateKind: TUpdateKind]: To8StmtOrd read GetQuery;
    property SQL[UpdateKind: TUpdateKind]: TStrings read GetSQL write SetSQL;
  published
    property ModifySQL: TStrings index 0 read GetSQLIndex write SetSQLIndex;
    property InsertSQL: TStrings index 1 read GetSQLIndex write SetSQLIndex;
    property DeleteSQL: TStrings index 2 read GetSQLIndex write SetSQLIndex;
  end;

procedure Register;

implementation

//******************************************************************************
procedure Register;
begin
  RegisterComponents('VK NLO', [To8DataSet, To8UpdateSQL]);
end;

//******************************************************************************
procedure To8DataSet.SetActive(Value: boolean);
begin
  if csReading in ComponentState then
    FStreamedActive := Value
  else
    if FSTMT <> nil then begin
      if FSTMT.Active then
      begin
        inherited SetActive(Value);
      end else inherited SetActive(false);
    end else
      inherited SetActive(false);
end;

//******************************************************************************
procedure To8DataSet.SetFSTMT(oStatment: To8Stmt);
begin
  if FSTMT <> nil then FSTMT.RemoveDataSet(self);
  Active := false;
  if oStatment <> nil then
  begin
    FSTMT := oStatment;
    FSTMT.AddDataSet(self);
  end else FSTMT := nil;
end;

//******************************************************************************
procedure To8DataSet.Loaded;
begin
  inherited Loaded();
  if FStreamedActive then Active := true;
end;

//******************************************************************************
constructor To8DataSet.Create(AOwner: TComponent);
var
  FieldMap: TFieldMap;
begin
  inherited Create(AOwner);
  FFilterParser := To8ExprParser.Create(self, '', [], [poExtSyntax], '', nil, FieldMap);
  FRecordBuffer := TList.Create;
  FDeletedRecord := TList.Create;
  FStreamedActive := false;
  FLocateType := ltLoop;
  FFetchAll := false;
  FIsLookup := false;
  FDataLink := TVKDL1.Create;
  FDataLink.o8DataSet := self;
end;

//******************************************************************************
destructor To8DataSet.Destroy;
begin
  FDataLink.o8DataSet := nil;
  FDataLink.Destroy;
  FDataLink := nil;
  ClearRecordBuffers;
  FRecordBuffer.Destroy;
  ClearDeletedRecordBuffers;
  FDeletedRecord.Destroy;
  FFilterParser.Destroy;
  if FSTMT <> nil then FSTMT.RemoveDataSet(self);
  inherited Destroy;
end;

//******************************************************************************
function To8DataSet.AllocRecordBuffer: PChar;
begin
  Result := AllocMem(RecordBufferSize);
end;

//******************************************************************************
procedure To8DataSet.FreeRecordBuffer(var Buffer: PChar);
begin
  FreeMem(Buffer);
end;

//******************************************************************************
procedure To8DataSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  ub4(Data^) := pTRecInfo(Buffer + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID;
end;

//******************************************************************************
function To8DataSet.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := pTRecInfo(Buffer + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).BookmarkFlag;
end;

//******************************************************************************
function To8DataSet.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var
  SaveState: TDataSetState;
  Accept: Boolean;
begin
  if not Filtered then
    Result := GetRecordByBuffer(Buffer, GetMode, DoCheck)
  else begin
    if GetMode <> gmCurrent then
    begin
      repeat
        Result := GetRecordByBuffer(FFilterRecord, GetMode, DoCheck);
        SaveState := SetTempState(dsFilter);
        Accept := AcceptRecord;
        RestoreState(SaveState);
      until Accept or (Result <> grOK);
      if Accept then
        Move(FFilterRecord^, Buffer^, RecordBufferSize)
      else
        Move(FTempRecord^, Buffer^, RecordBufferSize);
    end else
      Result := GetRecordByBuffer(Buffer, GetMode, DoCheck);
  end;
end;

function To8DataSet.GetRecordByBuffer(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var
  cc: PChar;
begin
  Result := grOK;
  case GetMode of
    gmCurrent:
      begin
        if FCurrentIndex <> -1 then
        begin
          if FCurrentIndex < FRecordBuffer.Count then
          begin
            cc := pChar(FRecordBuffer.Items[FCurrentIndex]);
            pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).BookmarkFlag := bfCurrent;
            Move(cc^, Buffer^, (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2 + sizeof(TRecInfo));
          end else begin
            InternalInitRecord(FTempRecord);
            Move(FTempRecord^, Buffer^, RecordBufferSize);
            Result := grError;
          end;
        end else begin
          InternalInitRecord(FTempRecord);
          Move(FTempRecord^, Buffer^, RecordBufferSize);
          Result := grError;
        end;
      end;
    gmNext:
      begin
        if FCurrentIndex < FRecordBuffer.Count then
        begin
          Inc(FCurrentIndex);
          if FCurrentIndex = FRecordBuffer.Count then
          begin
            if FCurrentIndex <> 0 then STMT.Next;
            if not STMT.Eof then
            begin
              Inc(FRowID);
              cc := MemMgr.oMem.GetMem(self, RecordBufferSize);
              InternalInitRecord(cc);
              pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).BookmarkFlag := bfCurrent;
              pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID := FRowID;
              pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).UpdateStatus := usUnmodified;
              Move(STMT.RawBufferGet^, cc^, STMT.RecordSize);
              Move(STMT.RawIndGet^   , (cc + STMT.RecordSize)^, STMT.FieldCount*sizeof(sb2));
              Move(STMT.RawRCodeGet^ , (cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2))^, STMT.FieldCount*sizeof(sb2));
              Move(STMT.RawRlenGet^  , (cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*2)^, STMT.FieldCount*sizeof(sb2));
              Move(cc^, Buffer^, RecordBufferSize);
              FRecordBuffer.Add(cc);
            end else begin
              Dec(FCurrentIndex);
              InternalInitRecord(FTempRecord);
              Move(FTempRecord^, Buffer^, RecordBufferSize);
              Result := grEOF;
            end;
          end else begin
            cc := pChar(FRecordBuffer.Items[FCurrentIndex]);
            pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).BookmarkFlag := bfCurrent;
            Move(cc^, Buffer^, RecordBufferSize);
          end;
        end else begin
          FCurrentIndex := FRecordBuffer.Count;
          InternalInitRecord(FTempRecord);
          Move(FTempRecord^, Buffer^, RecordBufferSize);
          Result := grEOF;
        end;
      end;
    gmPrior:
      begin
        Dec(FCurrentIndex);
        if FCurrentIndex >= 0 then
        begin
          cc := pChar(FRecordBuffer.Items[FCurrentIndex]);
          pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).BookmarkFlag := bfCurrent;
          Move(cc^, Buffer^, RecordBufferSize);
        end else begin
          FCurrentIndex := -1;
          InternalInitRecord(FTempRecord);
          Move(FTempRecord^, Buffer^, RecordBufferSize);
          Result := grBOF;
        end;
      end;
  end;
  if Result = grOK then
    GetCalcFields(Buffer);
end;


//******************************************************************************
function To8DataSet.GetRecordSize: Word;
begin
  Result := Word(STMT.RecordSize);
end;

//******************************************************************************
procedure To8DataSet.InternalAddRecord(Buffer: Pointer; Append: Boolean);
var
  cc: pChar;
begin
  CheckActive;
  if Append then
  begin
    Inc(FRowID);
    cc := MemMgr.oMem.GetMem(self, RecordBufferSize);
    Move(ActiveBuffer^, cc^, RecordBufferSize);
    pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID := FRowID;
    pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).UpdateStatus := usInserted;
    FRecordBuffer.Add(cc);
    FCurrentIndex := FRecordBuffer.Count - 1;
  end else begin
    Inc(FRowID);
    cc := MemMgr.oMem.GetMem(self, RecordBufferSize);
    Move(ActiveBuffer^, cc^, RecordBufferSize);
    pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID := FRowID;
    pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).UpdateStatus := usInserted;
    InternalSetToRecord(ActiveBuffer);
    FRecordBuffer.Insert(FCurrentIndex, cc);
  end;
end;

//******************************************************************************
procedure To8DataSet.InternalClose;
begin
  ClearRecordBuffers;
  ClearDeletedRecordBuffers;
  FreeRecordBuffer(FFilterRecord);
  FreeRecordBuffer(FTempRecord);
  BindFields(false);
  if DefaultFields then DestroyFields;
end;

//******************************************************************************
procedure To8DataSet.InternalDelete;
var
  cc: pChar;
begin
  CheckActive;
  if (FCurrentIndex >= 0) and (FCurrentIndex < FRecordBuffer.Count) then
  begin
    cc := MemMgr.oMem.GetMem(self, RecordBufferSize);
    Move(ActiveBuffer^, cc^, RecordBufferSize);
    //pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).UpdateStatus := usDeleted;
    FDeletedRecord.Add(cc);
    MemMgr.oMem.FreeMem(FRecordBuffer.Items[FCurrentIndex]);
    FRecordBuffer.Delete(FCurrentIndex);
    if FCurrentIndex = 0 then
      FCurrentIndex := -1;
  end else begin
    FCurrentIndex := -1;
  end;
end;

//******************************************************************************
procedure To8DataSet.InternalFirst;
begin
  FCurrentIndex := -1;
end;

//******************************************************************************
procedure To8DataSet.InternalGotoBookmark(Bookmark: Pointer);
var
  ff: ub4;
  i: Integer;
begin
  ff := pTRecInfo(Bookmark).RecordRowID;
  for i := 0 to FRecordBuffer.Count - 1 do
    if pTRecInfo(pChar(FRecordBuffer.Items[i]) + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID = ff then
    begin
      FCurrentIndex := i;
      break;
    end;
end;

//******************************************************************************
procedure To8DataSet.InternalHandleException;
begin
  Application.HandleException(self);
end;

//******************************************************************************
procedure To8DataSet.InternalInitFieldDefs;
var
   i: Integer;
   RS: pRStruct;
begin
  FieldDefs.BeginUpdate;
  FieldDefs.Clear;
  for i := 0 to STMT.FieldCount -1 do
  begin
    FieldDefs.AddFieldDef;
    FieldDefs.Items[i].Name := STMT.FieldName[i+1];
    RS := STMT.RawFieldStruct[i];
    FieldDefs.Items[i].DataType := GetFldType(RS);
    case FieldDefs.Items[i].DataType of
      ftString, ftBytes:
        FieldDefs.Items[i].Size := Word(STMT.RawFieldStruct[i].S_LENGTH + 1);
    else
      FieldDefs.Items[i].Size := 0;
    end;
    FieldDefs.Items[i].Precision := STMT.RawFieldStruct[i].S_PREC;
    if RS.S_NULLOK = 1 then
      FieldDefs.Items[i].Required := false
    else
      FieldDefs.Items[i].Required := true;
  end;
  FieldDefs.EndUpdate;
end;

//------------------------------------------------------------------------------
function To8DataSet.GetFldType(rs: pRStruct): TFieldType;
begin
  case rs.S_EXTERNALTYPE of
    SQLT_CHR:   Result := ftString;
    SQLT_AFC:   Result := ftString;
    SQLT_INT:   Result := ftInteger;
    SQLT_FLT:   Result := ftFloat;
    SQLT_RDD:   Result := ftBytes;
    SQLT_DAT:   Result := ftDateTime;
    SQLT_BIN:   Result := ftBytes;
    SQLT_LNG:   Result := ftMemo;
    SQLT_LBI:   Result := ftBlob;
    SQLT_CLOB:  Result := ftMemo;
    SQLT_BLOB:  Result := ftBlob;
    SQLT_CFILE: Result := ftMemo;
    SQLT_BFILE: Result := ftBlob;
    SQLT_RSET:  Result := ftDataSet;
    SQLT_NTY:   Result := ftADT;
  else
    Result := ftUnknown;
  end;
end;

//******************************************************************************
procedure To8DataSet.InternalInitRecord(Buffer: PChar);
begin
  if Buffer <> nil then FillChar(Buffer^, RecordBufferSize, -1);
end;

//******************************************************************************
procedure To8DataSet.InternalLast;
var
  cc: pChar;
begin
  while true do
  begin
    STMT.Next;
    if not STMT.Eof then
    begin
      Inc(FRowID);
      cc := MemMgr.oMem.GetMem(self, RecordBufferSize);
      InternalInitRecord(cc);
      pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).BookmarkFlag := bfCurrent;
      pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID := FRowID;
      pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).UpdateStatus := usUnmodified;
      Move(STMT.RawBufferGet^, cc^, STMT.RecordSize);
      Move(STMT.RawIndGet^   , (cc + STMT.RecordSize)^, STMT.FieldCount*sizeof(sb2));
      Move(STMT.RawRCodeGet^ , (cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2))^, STMT.FieldCount*sizeof(sb2));
      Move(STMT.RawRlenGet^  , (cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*2)^, STMT.FieldCount*sizeof(sb2));
      FRecordBuffer.Add(cc);
    end else
      break;
  end;
  FCurrentIndex := FRecordBuffer.Count;
end;

//******************************************************************************
procedure To8DataSet.InternalOpen;
begin
  ClearRecordBuffers;
  ClearDeletedRecordBuffers;
  FCurrentIndex := -1;
  FRowID := 0;
  ObjectView := true;
  InternalInitFieldDefs();
  if DefaultFields then CreateFields;
  BindFields(true);
  BookmarkSize := sizeof(ub4);
  FTempRecord := AllocRecordBuffer;
  FFilterRecord := AllocRecordBuffer;
  STMT.First;
  if Filtered and (Filter <> '') then
    FFilterParser.SetExprParams(Filter, FilterOptions, [poExtSyntax], '');
end;

//******************************************************************************
procedure To8DataSet.InternalPost;
var
  cc: pChar;
  l: Integer;
begin
  CheckActive;
  if State = dsEdit then
  begin
    Move(ActiveBuffer^, (FRecordBuffer.Items[FCurrentIndex])^, RecordBufferSize);
  end else begin
    l := Integer(pTRecInfo(pChar(ActiveBuffer) + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID);
    if l = -1 then
    begin
      Inc(FRowID);
      cc := MemMgr.oMem.GetMem(self, RecordBufferSize);
      Move(ActiveBuffer^, cc^, RecordBufferSize);
      pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID := FRowID;
      pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).UpdateStatus := usInserted;
      FRecordBuffer.Add(cc);
      FCurrentIndex := FRecordBuffer.Count - 1;
    end else begin
      Inc(FRowID);
      cc := MemMgr.oMem.GetMem(self, RecordBufferSize);
      Move(ActiveBuffer^, cc^, RecordBufferSize);
      pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID := FRowID;
      pTRecInfo(cc + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).UpdateStatus := usInserted;
      InternalSetToRecord(ActiveBuffer);
      FRecordBuffer.Insert(FCurrentIndex, cc);
    end;
  end;
end;

//******************************************************************************
procedure To8DataSet.InternalSetToRecord(Buffer: PChar);
var
  i: Integer;
  l: ub4;
begin
  FCurrentIndex := -1;
  l := pTRecInfo(Buffer + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID;
  for i := 0 to FRecordBuffer.Count - 1 do
    if l = pTRecInfo(pChar(FRecordBuffer.Items[i]) + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID then
    begin
      FCurrentIndex := sb4(i);
      Break;
    end;
end;

//******************************************************************************
function To8DataSet.IsCursorOpen: Boolean;
begin
   if Assigned(STMT) then Result := STMT.Active else Result := false;
end;

//******************************************************************************
procedure To8DataSet.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  pTRecInfo(Buffer + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).BookmarkFlag := Value;
end;

//******************************************************************************
procedure To8DataSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  pTRecInfo(Buffer + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID := ub4(Data^);
end;

//******************************************************************************
procedure To8DataSet.SetFieldData(Field: TField; Buffer: Pointer);
var
    rr: pRStruct;
    pInd:   psb2;
    RLen:   pub2;
    ActiveBuf: PChar;
    pIndOld:   psb2;
    RLenOld:   pub2;
    ActiveBufOld: PChar;
    j: Integer;
    oDate, oDateOld: pORA_Date;
    OraDate: ORA_Date;
    c: PChar;
    c1: PChar;
    dd: double;
    sTS: TTimeStamp;
    US: pTRecInfo;
begin
  case Field.FieldKind of
  fkData:
    begin
      j := Field.FieldNo-1;
      rr := STMT.RawFieldStruct[j];
      GetActiveRecBuf(ActiveBuf);
      if (Buffer <> nil) and (ActiveBuf <> nil) then
      begin
        pInd  := psb2((psb2(ActiveBuf + STMT.RecordSize + j*sizeof(sb2))));
        RLen  := pub2((pub2(ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(ub2)*2 + j*sizeof(ub2))));
        US := pTRecInfo(ActiveBuf + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2);
        ActiveBufOld := ActiveBuf + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3);
        pIndOld  := psb2((psb2(ActiveBufOld + STMT.RecordSize + j*sizeof(sb2))));
        RLenOld  := pub2((pub2(ActiveBufOld + STMT.RecordSize + STMT.FieldCount*sizeof(ub2)*2 + j*sizeof(ub2))));
        Field.Validate(Buffer);
        case Field.DataType of
          ftString, ftInteger, ftFloat, ftBytes, ftVarBytes:
            begin
              c := ActiveBuf + rr.S_OFFSET;
              c1 := ActiveBufOld + rr.S_OFFSET;
              if sb2(RLenOld^) = -1 then
              begin
                Move(c^, c1^, rr.S_LENGTH);
                RLenOld^ := RLen^;
                pIndOld^ := pInd^;
                if US.UpdateStatus = usUnmodified then US.UpdateStatus := usModified;
              end;
              Move(Buffer^, c^, rr.S_LENGTH);
              pInd^ := 0;
              if  Field.DataType in [ftString, ftBytes, ftVarBytes] then
              begin
                RLen^ := Length(pChar(Buffer));
                if RLen^ > rr.S_LENGTH then RLen^ := rr.S_LENGTH;
              end else
                RLen^ := rr.S_LENGTH;
              if  Field.DataType in [ftString, ftBytes, ftVarBytes]  then pChar(Buffer)[RLen^] := Char(0);
              SetModified(true);
            end;
          ftDateTime:
            begin
              oDate := pORA_Date(ActiveBuf + rr.S_OFFSET);
              oDateOld := pORA_Date(ActiveBufOld + rr.S_OFFSET);
              if sb2(RLenOld^) = -1 then
              begin
                Move(oDate^, oDateOld^, rr.S_LENGTH);
                RLenOld^ := RLen^;
                pIndOld^ := pInd^;
                if US.UpdateStatus = usUnmodified then US.UpdateStatus := usModified;
              end;
              dd := double(Buffer^);
              sTS.Date := Trunc(dd/(3600.0*24*1000));
              sTS.Time := Trunc(dd - sTS.Date*((3600.0*24*1000)));
              OraDate := DateTimeToOraDate(TimeStampToDateTime(sTS));
              Move(OraDate, oDate^, rr.S_LENGTH);
              RLen^ := rr.S_LENGTH;
              pInd^ := 0;
              SetModified(true);
            end;
        end;
      end;
    end;
  fkCalculated:
    begin
      GetActiveRecBuf(ActiveBuf);
      if (Buffer <> nil) and (ActiveBuf <> nil) then
      begin
        pInd  := psb2(ActiveBuf + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2 + sizeof(TRecInfo) + ub4(CalcFieldsSize) + Field.Offset);
        //RLen  := pub2(ActiveBuf + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2 + sizeof(TRecInfo) + ub4(CalcFieldsSize)*2 + Field.Offset);
        c := ActiveBuf + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2 + sizeof(TRecInfo) + Field.Offset;
        Move(Buffer^, c^, Field.DataSize);
        pInd^ := 0;
        //if  Field.DataType in [ftString, ftBytes, ftVarBytes] then
        //begin
        //  RLen^ := Length(pChar(Buffer));
        //  if RLen^ > Field.DataSize then RLen^ := Field.DataSize;
        //end else
        //  RLen^ := Field.DataSize;
        //if  Field.DataType in [ftString, ftBytes, ftVarBytes]  then pChar(Buffer)[RLen^] := Char(0);
      end;
    end;
  end;
  if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
    DataEvent(deFieldChange, Longint(Field));
end;

//******************************************************************************
function To8DataSet.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
    rr: pRStruct;
    pInd:   sb2;
    RLen:   ub2;
    ActiveBuf: PChar;
    j: Integer;
    oDate: pORA_Date;
    oDateTime: TDateTime;
    wYear: Integer;
    dd: double;
    sTS: TTimeStamp;
    c: PChar;
    oDS: TDataSet;
    LookupResult: Variant;
begin
    Result := false;
    case Field.FieldKind of
    fkData:
      begin
        j := Field.FieldNo-1;
        rr := STMT.RawFieldStruct[j];
        if GetActiveRecBuf(ActiveBuf) then
        begin
          pInd  := sb2((psb2(ActiveBuf + STMT.RecordSize + j*sizeof(sb2)))^);
          RLen  := ub2((pub2(ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(ub2)*2 + j*sizeof(ub2))^));
        end else begin
          ActiveBuf := nil;
          pInd  := -1;
          RLen  := 0;
        end;
        if Buffer <> nil then
        begin
          if ActiveBuf <> nil then
          begin
            case Field.DataType of
              ftString, ftInteger, ftFloat, ftBytes, ftVarBytes:
                begin
                  case pInd of
                    0:
                      begin
                        c := ActiveBuf + rr.S_OFFSET;
                        if RLen <> 0 then Move(c^, Buffer^, RLen) else FillChar(Buffer^, rr.S_LENGTH, 0);
                        if  Field.DataType in [ftString]  then pChar(Buffer)[RLen] := Char(0);
                        Result := true;
                      end;
                    -1: Result := false;
                  end;
                end;
              ftBlob,ftMemo:
                begin
                   if rr.S_INERNALTYPE in [SQLT_BLOB, SQLT_CLOB, SQLT_CFILE, SQLT_BFILE] then
                   begin
                    if pInd = -1 then
                      Result := false
                    else begin
                      c := ActiveBuf + rr.S_OFFSET;
                      if RLen <> 0 then Move(c^, Buffer^, RLen) else FillChar(Buffer^, rr.S_LENGTH, 0);
                      Result := true;
                    end;
                   end;
                   if rr.S_INERNALTYPE in [SQLT_LNG, SQLT_LBI] then
                   begin
                    if pInd = -1 then
                      Result := false
                    else begin
                      c := ActiveBuf + rr.S_OFFSET;
                      if RLen <> 0 then Move(c^, Buffer^, RLen) else FillChar(Buffer^, rr.S_LENGTH, 0);
                      pChar(Buffer)[RLen] := Char(0);
                      Result := true;
                    end;
                   end;
                end;
              ftCursor:
                begin
                  if pInd = -1 then
                    Result := false
                  else begin
                    c := ActiveBuf + rr.S_OFFSET;
                    if RLen <> 0 then Move(c^, Buffer^, RLen) else FillChar(Buffer^, rr.S_LENGTH, 0);
                    pChar(Buffer)[RLen] := Char(0);
                    Result := true;
                  end;
                end;
              ftDateTime:
                begin
                  case pInd of
                    0:
                      begin
                        oDate := pORA_Date(ActiveBuf + rr.S_OFFSET);
                        try
                          wYear := (oDate.cCentury-100)*100+(oDate.cYear-100);
                          if (wYear > 0) and ( wYear < 9999) and ( oDate.cMonth >= 1) and ( oDate.cMonth <= 12) then
                             oDateTime := EncodeDate(wYear, oDate.cMonth, oDate.cDay) +
                                      EncodeTime(oDate.cHour-1, oDate.cMinute-1, oDate.cSecond-1, 0)
                          else
                             oDateTime := 0;
                          sTS := DateTimeToTimeStamp(oDateTime);
                          dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                          double(Buffer^) := dd;
                          Result := true;
                        except
                          Result := false;
                        end;
                      end;
                    -1: Result := false;
                  end;
                end;
            end;
          end else
            Result := false;
        end else begin
          if pInd = -1 then
            Result := false
          else
            Result := true;
        end;
      end;
    fkCalculated:
      begin
        if GetActiveRecBuf(ActiveBuf) then
        begin
          pInd  := sb2(psb2(ActiveBuf + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2 + sizeof(TRecInfo) + ub4(CalcFieldsSize) + Field.Offset)^);
        end else begin
          pInd  := -1;
        end;
        if Buffer <> nil then
        begin
          if (ActiveBuf <> nil) and not (csDesigning in ComponentState) then
          begin
            case pInd of
              0:
                begin
                  c := ActiveBuf + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2 + sizeof(TRecInfo) + Field.Offset;
                  Move(c^, Buffer^, Field.DataSize);
                  if Field.DataType in [ftString, ftBytes, ftVarBytes]  then pChar(Buffer)[Field.DataSize] := Char(0);
                  Result := true;
                end;
              -1: Result := false;
            end;
          end else begin
            FillChar(Buffer^, Field.DataSize, 0);
            Result := false;
          end;
        end else
          Result := false;
      end;
    fkLookup:
      begin
        Result := false;
        if Buffer <> nil then
        begin
          if GetActiveRecBuf(ActiveBuf) then
          begin
            oDS := Field.LookupDataSet;
            if (oDS <> nil) and (oDS.Active) then
            begin
              LookupResult := oDS.Lookup(Field.LookupKeyFields, FieldValues[Field.KeyFields], Field.LookupResultField);
              if (not VarIsEmpty(LookupResult)) and (not VarIsNull(LookupResult)) then
              begin
                if Field.DataType = ftString then
                begin
                  c := TVarData(LookupResult).VPointer;
                  Move(c^, Buffer^, Length(c) + 1);
                end else if Field.DataType = ftDateTime then
                begin
                  sTS := DateTimeToTimeStamp(LookupResult);
                  dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                  double(Buffer^) := dd;
                end else begin
                  c := @TVarData(LookupResult).VAny;
                  Move(c^, Buffer^, Field.DataSize);
                end;
                Result := true;
              end;
            end;
          end else begin
            FillChar(Buffer^, Field.DataSize, 0);
            Result := false;
          end;
        end;
      end;
    end;
end;

//******************************************************************************
function To8DataSet.GetOldFieldData(Field: TField): Variant;
var
  rr: pRStruct;
  pInd:   sb2;
  RLen:   ub2;
  ActiveBuf: PChar;
  j: Integer;
  oDate: pORA_Date;
  oDateTime: TDateTime;
  wYear: Integer;
  dd: double;
  sTS: TTimeStamp;
  c: PChar;
  s: String;
begin
  Result := NULL;
  case Field.FieldKind of
  fkData:
    begin
      j := Field.FieldNo-1;
      rr := STMT.RawFieldStruct[j];
      if GetActiveRecBuf(ActiveBuf) then
      begin
        ActiveBuf := ActiveBuf + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3);
        pInd  := sb2((psb2(ActiveBuf + STMT.RecordSize + j*sizeof(sb2)))^);
        RLen  := ub2((pub2(ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(ub2)*2 + j*sizeof(ub2)))^);
      end else begin
        ActiveBuf := nil;
        pInd  := -1;
        RLen  := 0;
      end;
      if ActiveBuf <> nil then
      begin
        if sb2(RLen) <> -1 then
        begin
          if pInd = 0 then
          begin
            case Field.DataType of
              ftString:
                begin
                  c := ActiveBuf + rr.S_OFFSET;
                  SetString(s, c, RLen);
                  Result := s;
                end;
              ftInteger: Result := Integer(pInteger(ActiveBuf + rr.S_OFFSET)^);
              ftFloat: Result := Double(pDouble(ActiveBuf + rr.S_OFFSET)^);
              ftBytes,
              ftVarBytes,
              ftBlob,ftMemo:
                begin
                  Result := NULL;
                end;
              ftCursor:
                begin
                  Result := NULL;
                end;
              ftDateTime:
                begin
                  oDate := pORA_Date(ActiveBuf + rr.S_OFFSET);
                  try
                    wYear := (oDate.cCentury-100)*100+(oDate.cYear-100);
                    if (wYear > 0) and ( wYear < 9999) and ( oDate.cMonth >= 1) and ( oDate.cMonth <= 12) then
                       oDateTime := EncodeDate(wYear, oDate.cMonth, oDate.cDay) +
                                EncodeTime(oDate.cHour-1, oDate.cMinute-1, oDate.cSecond-1, 0)
                    else
                       oDateTime := 0;
                    sTS := DateTimeToTimeStamp(oDateTime);
                    dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                    Result := dd;
                  except
                    Result := NULL;
                  end;
                end;
            end;
          end else
            Result := NULL;
        end else
          Result := Unassigned;
      end else
        Result := NULL;
    end;
  fkCalculated:
    begin
      Result := NULL;
    end;
  fkLookup:
    begin
      Result := NULL;
    end;
  fkInternalCalc:
    begin
      Result := NULL;
    end;
  fkAggregate:
    begin
      Result := NULL;
    end;
  end;
end;

//******************************************************************************
function To8DataSet.GetActiveRecBuf(var RecBuf: PChar): Boolean;
begin
  if not FIsLookup then
    case State of
      dsBrowse: if IsEmpty then RecBuf := nil else RecBuf := ActiveBuffer;
      dsEdit, dsInsert: RecBuf := ActiveBuffer;
      dsFilter: RecBuf := FFilterRecord;
      dsNewValue, dsOldValue, dsCurValue: RecBuf := ActiveBuffer;
      dsCalcFields: RecBuf := CalcBuffer;
    else
      RecBuf := nil;
    end
  else
    RecBuf := FRecordBuffer.Items[FCurrentIndex];
  Result := RecBuf <> nil;
end;

//******************************************************************************
function To8DataSet.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
var
  j: Integer;
  rr: pRStruct;
  pInd:   sb2;
  ActiveBuf: PChar;
begin
  j := Field.FieldNo-1;
  rr := STMT.RawFieldStruct[j];
  if GetActiveRecBuf(ActiveBuf) then
  begin
    pInd  := sb2((psb2(ActiveBuf + STMT.RecordSize + j*sizeof(sb2)))^);
  end else begin
    ActiveBuf := nil;
    pInd  := -1;
  end;
  if (ActiveBuf <> nil) and (pInd <> -1) then
    Result := self.STMT.GetBlobStream(Pointer(ActiveBuf + rr.S_OFFSET), rr.S_INERNALTYPE)
  else
    Result := TOraEmptyStream.Create(Pointer(ActiveBuf + rr.S_OFFSET), self.STMT, rr.S_INERNALTYPE);
end;

function To8DataSet.GetAsResultSet(Field: TField): pOCIResult;
var
  j: Integer;
  rr: pRStruct;
  pInd:   sb2;
  ActiveBuf: PChar;
begin
  j := Field.FieldNo-1;
  rr := STMT.RawFieldStruct[j];
  if GetActiveRecBuf(ActiveBuf) then
  begin
    pInd  := sb2((psb2(ActiveBuf + STMT.RecordSize + j*sizeof(sb2)))^);
  end else begin
    ActiveBuf := nil;
    pInd  := -1;
  end;
  if (rr.S_INERNALTYPE = SQLT_RSET) and (ActiveBuf <> nil) and (pInd <> -1) then
    Result := PpOCIResult(ActiveBuf + rr.S_OFFSET)^
  else
    Result := nil;
end;

function To8DataSet.GetAsRowID(Field: TField): pOCIRowid;
var
  j: Integer;
  rr: pRStruct;
  pInd:   sb2;
  ActiveBuf: PChar;
begin
  j := Field.FieldNo-1;
  rr := STMT.RawFieldStruct[j];
  if GetActiveRecBuf(ActiveBuf) then
  begin
    pInd  := sb2((psb2(ActiveBuf + STMT.RecordSize + j*sizeof(sb2)))^);
  end else begin
    ActiveBuf := nil;
    pInd  := -1;
  end;
  if (rr.S_INERNALTYPE = SQLT_RDD) and (ActiveBuf <> nil) and (pInd <> -1) then
    Result := ppOCIRowid(ActiveBuf + rr.S_OFFSET)^
  else
    Result := nil;
end;

function To8DataSet.AcceptRecord: Boolean;
begin
  Result := true;
  if Assigned(OnFilterRecord) then
    OnFilterRecord(self, Result)
  else if Filter <> '' then
    Result := FFilterParser.Execute;
end;

function To8DataSet.Lookup(const KeyFields: string; const KeyValues: Variant; const ResultFields: string): Variant;
begin
  Result := Null;
  FIsLookup := true;
  o8Env.o8CrSection.Acquire;
  try
    if LocateRecord(KeyFields, KeyValues, [], false) then
    begin
      SetTempState(dsCalcFields);
      try
        Result := FieldValues[ResultFields];
      finally
        RestoreState(dsBrowse);
      end;
    end;
  finally
    FIsLookup := false;
    o8Env.o8CrSection.Release;
  end;
end;

procedure To8DataSet.SetFiltered(Value: Boolean);
begin
  if Active then
  begin
    CheckBrowseMode;
    if ((not Filtered) and Value) and (Filter <> '') then
      FFilterParser.SetExprParams(Filter, FilterOptions, [poExtSyntax], '');
    if Filtered <> Value then
      inherited SetFiltered(Value);
    First;
  end else
    inherited SetFiltered(Value);
end;

function To8DataSet.Locate(const KeyFields: string;
  const KeyValues: Variant; Options: TLocateOptions): Boolean;
begin
  DoBeforeScroll;
  FIsLookup := true;
  o8Env.o8CrSection.Acquire;
  try
    Result := LocateRecord(KeyFields, KeyValues, Options, True);
  finally
    FIsLookup := false;
    o8Env.o8CrSection.Release;
  end;
  if Result then
  begin
    Resync([]);
    DoAfterScroll;
  end;
end;

function To8DataSet.LocateRecord(const KeyFields: string;
  const KeyValues: Variant; Options: TLocateOptions;
  SyncCursor: Boolean): Boolean;
var
  Fields: TList;
  i: Integer;
  bm: String;
  LowR, HighR, l: sb4;
begin
  Result := false;
  if not VarIsArray(KeyValues) then
  begin
    if VarIsNull(KeyValues) or VarIsEmpty(KeyValues) then
    begin
      Result := false;
      Exit;
    end;
  end else
    for i := 0 to VarArrayHighBound(KeyValues, 1) do
      if VarIsNull(KeyValues[i]) or VarIsEmpty(KeyValues[i]) then
      begin
        Result := false;
        Exit;
      end;
  if FLocateType = ltLoop then
  begin
    DisableControls;
    if not FFetchAll then FetchAll;
    if State <> dsBrowse then Post;
    bm := Bookmark;
    Result := false;
    Fields := TList.Create;
    try
      GetFieldList(Fields, KeyFields);
      FCurrentIndex := 0;
      while FCurrentIndex < FRecordBuffer.Count do
      begin
        l := CompareLocateField(Fields, KeyValues, Options);
        if l = 0 then
        begin
          Result := true;
          Break;
        end;
        Inc(FCurrentIndex);
      end;
      if not Result then Bookmark := bm;
    finally
      Fields.Free;
      EnableControls;
    end;
  end else if FLocateType = ltOrdered then begin
    DisableControls;
    if not FFetchAll then FetchAll;
    if State <> dsBrowse then Post;
    bm := Bookmark;
    Result := false;
    Fields := TList.Create;
    try
      GetFieldList(Fields, KeyFields);
      //
      LowR := 0;
      HighR := FRecordBuffer.Count - 1;
      while LowR < HighR do
      begin
        FCurrentIndex := LowR + Trunc((HighR - LowR)/2);
        l := CompareLocateField(Fields, KeyValues, Options);
        if l < 0 then
        begin
          if LowR <> FCurrentIndex then
          begin
            LowR := FCurrentIndex;
            continue;
          end else
            Break;
        end;
        if l = 0 then
        begin
          Result := true;
          Break;
        end;
        if l > 0 then
        begin
          if HighR <> FCurrentIndex then
          begin
            HighR := FCurrentIndex;
            continue;
          end else
            Break;
        end;
      end;
      //
      if not Result then Bookmark := bm;
    finally
      Fields.Free;
      EnableControls;
    end;
  end else if FLocateType = ltOrderedDesc then begin
    DisableControls;
    if not FFetchAll then FetchAll;
    if State <> dsBrowse then Post;
    bm := Bookmark;
    Result := false;
    Fields := TList.Create;
    try
      GetFieldList(Fields, KeyFields);
      //
      LowR := 0;
      HighR := FRecordBuffer.Count - 1;
      while LowR < HighR do
      begin
        FCurrentIndex := LowR + Trunc((HighR - LowR)/2);
        l := CompareLocateField(Fields, KeyValues, Options);
        if l < 0 then
        begin
          if HighR <> FCurrentIndex then
          begin
            HighR := FCurrentIndex;
            continue;
          end else
            Break;
        end;
        if l = 0 then
        begin
          Result := true;
          Break;
        end;
        if l > 0 then
        begin
          if  LowR <> FCurrentIndex then
          begin
            LowR := FCurrentIndex;
            continue;
          end else
            Break;
        end;
      end;
      //
      if not Result then Bookmark := bm;
    finally
      Fields.Free;
      EnableControls;
    end;
  end;
end;

function To8DataSet.CompareLocateField(const Fields: TList;
  const KeyValues: Variant; Options: TLocateOptions): Integer;
var
  FieldCount: Integer;
  Field: TField;
  KeyVal: Variant;
  v1, v2: String;
  i1, i2: Integer;
  w1, w2: Word;
  b1, b2: boolean;
  f1, f2: double;
  i: Integer;
begin
  Result := 1;
  FieldCount := Fields.Count;
  if FieldCount = 1 then
  begin
    try
      Field := TField(Fields.First);
    except
      Field := nil;
    end;
    try
      if VarIsArray(KeyValues) then
         KeyVal := KeyValues[0]
      else
         KeyVal := KeyValues;
    except
      KeyVal := NULL;
    end;
    case Field.DataType of
      ftFixedChar, ftWideString, ftString:
      begin
        v1 := Field.AsString;
        if not VarIsNull(KeyVal) then v2 := KeyVal else v2 := '';
        if loCaseInsensitive in Options then
          Result := AnsiCompareText(v1, v2)
        else
          Result := AnsiCompareStr(v1, v2);
      end;
      ftLargeint, ftSmallint, ftInteger:
      begin
        i1 := Field.AsInteger;
        if not VarIsNull(KeyVal) then i2 := KeyVal else i2 := 0;
        Result := i1 - i2;
      end;
      ftWord:
      begin
        w1 := Field.AsInteger;
        if not VarIsNull(KeyVal) then w2 := KeyVal else w2 := 0;
        Result := w1 - w2;
      end;
      ftBoolean:
      begin
        b1 := Field.AsBoolean;
        if not VarIsNull(KeyVal) then b2 := KeyVal else b2 := false;
        if (not b1) and b2 then
          Result := -1;
        if b1 = b2 then
          Result := 0;
        if b1 and (not b2) then
          Result := 1;
      end;
      ftFloat, ftCurrency:
      begin
        f1 := Field.AsFloat;
        if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
        if f1 < f2 then
          Result := -1;
        if f1 = f2 then
          Result := 0;
        if f1 > f2 then
          Result := 1;
      end;
      ftDate, ftTime, ftDateTime:
      begin
        f1 := Field.AsDateTime;
        if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
        if f1 < f2 then
          Result := -1;
        if f1 = f2 then
          Result := 0;
        if f1 > f2 then
          Result := 1;
      end;
    end;
  end else begin
    for i := 0 to FieldCount - 1 do
    begin
      //
      try
        Field := TField(Fields.Items[i]);
      except
        Field := nil;
      end;
      try
        if VarIsArray(KeyValues) then
           KeyVal := KeyValues[i]
        else
           KeyVal := NULL;
      except
        KeyVal := NULL;
      end;
      case Field.DataType of
        ftFixedChar, ftWideString, ftString:
        begin
          v1 := Field.AsString;
          if not VarIsNull(KeyVal) then v2 := KeyVal else v2 := '';
          if loCaseInsensitive in Options then
            Result := AnsiCompareText(v1, v2)
          else
            Result := AnsiCompareStr(v1, v2);
        end;
        ftLargeint, ftSmallint, ftInteger:
        begin
          i1 := Field.AsInteger;
          if not VarIsNull(KeyVal) then i2 := KeyVal else i2 := 0;
          Result := i1 - i2;
        end;
        ftWord:
        begin
          w1 := Field.AsInteger;
          if not VarIsNull(KeyVal) then w2 := KeyVal else w2 := 0;
          Result := w1 - w2;
        end;
        ftBoolean:
        begin
          b1 := Field.AsBoolean;
          if not VarIsNull(KeyVal) then b2 := KeyVal else b2 := false;
          if (not b1) and b2 then
            Result := -1;
          if b1 = b2 then
            Result := 0;
          if b1 and (not b2) then
            Result := 1;
        end;
        ftFloat, ftCurrency:
        begin
          f1 := Field.AsFloat;
          if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
          if f1 < f2 then
            Result := -1;
          if f1 = f2 then
            Result := 0;
          if f1 > f2 then
            Result := 1;
        end;
        ftDate, ftTime, ftDateTime:
        begin
          f1 := Field.AsDateTime;
          if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
          if f1 < f2 then
            Result := -1;
          if f1 = f2 then
            Result := 0;
          if f1 > f2 then
            Result := 1;
        end;
      end;
      //
      if Result <> 0 then
        Exit;
    end;
  end;
end;

function To8DataSet.FindRecord(Restart, GoForward: Boolean): Boolean;
var
  SaveState: TDataSetState;
  Accept: Boolean;
  Ret: TGetResult;
begin
  if (not Filtered) and (Filter <> '') then
      FFilterParser.SetExprParams(Filter, FilterOptions, [poExtSyntax], '');
  CheckBrowseMode;
  DoBeforeScroll;
  SetFound(False);
  UpdateCursorPos;
  CursorPosChanged;
  if GoForward then
  begin
    if Restart then InternalFirst;
    repeat
      Ret := GetRecordByBuffer(FFilterRecord, gmNext, false);
      SaveState := SetTempState(dsFilter);
      Accept := AcceptRecord;
      RestoreState(SaveState);
    until Accept or (Ret <> grOK);
  end else
  begin
    if Restart then InternalLast;
    repeat
      Ret := GetRecordByBuffer(FFilterRecord, gmPrior, false);
      SaveState := SetTempState(dsFilter);
      Accept := AcceptRecord;
      RestoreState(SaveState);
    until Accept or (Ret <> grOK);
  end;
  if Ret = grOK then
  begin
    Resync([rmExact, rmCenter]);
    SetFound(True);
  end else
    InternalSetToRecord(ActiveBuffer);
  Result := Found;
  if Result then DoAfterScroll;
end;


procedure To8DataSet.FetchAll;
begin
  CheckBrowseMode;
  while not eof do Next;
  FFetchAll := true;
  CursorPosChanged;
end;

procedure To8DataSet.ClearRecordBuffers;
var
  i: Integer;
begin
  if FRecordBuffer.Count > 0 then
  begin
    for i := 0 to FRecordBuffer.Count - 1 do
      MemMgr.oMem.FreeMem(pChar(FRecordBuffer.Items[i]));
    FRecordBuffer.Clear;
  end;
end;

function To8DataSet.GetRecNo: Integer;
begin
  Result := Integer(pTRecInfo(ActiveBuffer + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).RecordRowID);
end;

function To8DataSet.GetRecordCount: Integer;
begin
  Result := FRowID;
end;

function To8DataSet.GetRecordBufferSize: Integer;
begin
  Result := (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2 + sizeof(TRecInfo) + ub4(CalcFieldsSize)*2;
end;

procedure To8DataSet.ClearDeletedRecordBuffers;
var
  i: Integer;
begin
  for i := 0 to FDeletedRecord.Count - 1 do
    MemMgr.oMem.FreeMem(pChar(FDeletedRecord.Items[i]));
  FDeletedRecord.Clear;
end;

function To8DataSet.GetRecordUpdateStatus: TUpdateStatus;
begin
  Result := pTRecInfo(ActiveBuffer + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).UpdateStatus;
end;

procedure To8DataSet.SetRecordUpdateStatus(const Value: TUpdateStatus);
begin
  pTRecInfo(ActiveBuffer + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).UpdateStatus := Value;
end;

function To8DataSet.GetStateFieldValue(State: TDataSetState; Field: TField): Variant;
begin
  Result := NULL;
  if State in [dsNewValue, dsCurValue] then
    Result := inherited GetStateFieldValue(State, Field);
  if State = dsOldValue then
    Result := GetOldFieldData(Field);
end;

procedure To8DataSet.ApplyUpdates;
var
  i, j, k: sb4;
  ActiveBuf: pChar;
  RecordBufferCount: Integer;
begin
  if FUpdateObject <> nil then begin
    if State <> dsBrowse then Post;
    DisableControls;
    k := FCurrentIndex;
    FIsLookup := true;
    try

      //
      RecordBufferCount := FRecordBuffer.Count;
      while FDeletedRecord.Count > 0 do
      begin
        j := FRecordBuffer.Add(FDeletedRecord.Items[0]);
        pTRecInfo(pChar(FRecordBuffer.Items[j]) + (STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3)*2).UpdateStatus := usDeleted;
        FDeletedRecord.Delete(0);
      end;
      //

      for i := 0 to FRecordBuffer.Count - 1 do begin

        FCurrentIndex := sb4(i);

        j := STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3;
        ActiveBuf := pChar(FRecordBuffer.Items[FCurrentIndex]) + j;
        case pTRecInfo(ActiveBuf + j).UpdateStatus of
          usModified:
            begin
              FUpdateObject.Apply(ukModify);
              FillChar(ActiveBuf^, j, -1);
              pTRecInfo(ActiveBuf + j).UpdateStatus := usUnmodified;
              GetRecord(ActiveBuffer, gmCurrent, True);
            end;
          usInserted:
            begin
              FUpdateObject.Apply(ukInsert);
              FillChar(ActiveBuf^, j, -1);
              pTRecInfo(ActiveBuf + j).UpdateStatus := usUnmodified;
              GetRecord(ActiveBuffer, gmCurrent, True);
            end;
          usDeleted:
            begin
              FUpdateObject.Apply(ukDelete);
            end;
        end;

      end;
      //
      while FRecordBuffer.Count > RecordBufferCount do
      begin
        MemMgr.oMem.FreeMem(FRecordBuffer.Items[FRecordBuffer.Count - 1]);
        FRecordBuffer.Delete(FRecordBuffer.Count - 1);
      end;
      //
      FCurrentIndex := k;
    finally
      FIsLookup := false;
    end;

    EnableControls;
  end;
end;

procedure To8DataSet.SetUpdateObject(Value: To8DataSetUpdateObject);
begin
  if Value <> FUpdateObject then
  begin
    if Assigned(FUpdateObject) and (FUpdateObject.DataSet = Self) then
      FUpdateObject.DataSet := nil;
    FUpdateObject := Value;
    if Assigned(FUpdateObject) then
    begin
      { If another dataset already references this updateobject, then
        remove the reference }
      if Assigned(FUpdateObject.DataSet) and
        (FUpdateObject.DataSet <> Self) then
        FUpdateObject.DataSet.UpdateObject := nil;
      FUpdateObject.DataSet := Self;
    end;
  end;
end;

{ To8UpdateSQL }

procedure To8UpdateSQL.Apply(UpdateKind: TUpdateKind);
begin
  SetParams(UpdateKind);
  ExecSQL(UpdateKind);
end;

constructor To8UpdateSQL.Create(AOwner: TComponent);
var
  UpdateKind: TUpdateKind;
begin
  inherited Create(AOwner);
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
  begin
    FSQLText[UpdateKind] := TStringList.Create;
    TStringList(FSQLText[UpdateKind]).OnChange := SQLChanged;
  end;
end;

destructor To8UpdateSQL.Destroy;
var
  UpdateKind: TUpdateKind;
begin
  if Assigned(FDataSet) and (FDataSet.UpdateObject = Self) then
    FDataSet.UpdateObject := nil;
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
    FSQLText[UpdateKind].Free;
  inherited Destroy;
end;

procedure To8UpdateSQL.ExecSQL(UpdateKind: TUpdateKind);
begin
  with Query[UpdateKind] do
  begin
    Prepare;
    ExecSql;
  end;
end;

function To8UpdateSQL.GetDataSet: To8DataSet;
begin
  Result := FDataSet;
end;

function To8UpdateSQL.GetQuery(UpdateKind: TUpdateKind): To8StmtOrd;
begin
  if not Assigned(FQueries[UpdateKind]) then
  begin
    FQueries[UpdateKind] := To8StmtOrd.Create(Self);
    FQueries[UpdateKind].SQL.Assign(FSQLText[UpdateKind]);
    if (FDataSet is To8DataSet) then
      FQueries[UpdateKind].oraSession := To8DataSet(FDataSet).STMT.oraSession;
  end;
  Result := FQueries[UpdateKind];
end;

function To8UpdateSQL.GetSQL(UpdateKind: TUpdateKind): TStrings;
begin
  Result := FSQLText[UpdateKind];
end;

function To8UpdateSQL.GetSQLIndex(Index: Integer): TStrings;
begin
  Result := FSQLText[TUpdateKind(Index)];
end;

procedure To8UpdateSQL.SetDataSet(ADataSet: To8DataSet);
begin
  FDataSet := ADataSet;
end;

procedure To8UpdateSQL.SetParams(UpdateKind: TUpdateKind);
var
  I: Integer;
  Old: Boolean;
  Param: TOraParam;
  PName: string;
  Field: TField;
  Value: Variant;
  qr: To8StmtOrd;
begin
  if not Assigned(FDataSet) then Exit;
  qr := Query[UpdateKind];
  for i := 0 to qr.Params.Count - 1 do
  begin
    Param := TOraParam(qr.Params.Items[I]);
    PName := Param.Name;
    Old := CompareText(Copy(PName, 1, 4), 'OLD_') = 0;
    if Old then System.Delete(PName, 1, 4);
    Field := FDataSet.FindField(PName);
    if not Assigned(Field) then Continue;
    if Old then
    begin
      Value := Field.OldValue;
      if VarIsEmpty(Value) then Value := Field.NewValue;
      Param.AssignFieldValue(Field, Value);
    end else begin
      Value := Field.NewValue;
      Param.AssignFieldValue(Field, Value);
    end;
  end;
end;

procedure To8UpdateSQL.SetSQL(UpdateKind: TUpdateKind; Value: TStrings);
begin
  FSQLText[UpdateKind].Assign(Value);
end;

procedure To8UpdateSQL.SetSQLIndex(Index: Integer; Value: TStrings);
begin
  SetSQL(TUpdateKind(Index), Value);
end;

procedure To8UpdateSQL.SQLChanged(Sender: TObject);
var
  UpdateKind: TUpdateKind;
begin
  for UpdateKind := Low(TUpdateKind) to High(TUpdateKind) do
    if Sender = FSQLText[UpdateKind] then
    begin
      if Assigned(FQueries[UpdateKind]) then
      begin
        FQueries[UpdateKind].Params.Clear;
        FQueries[UpdateKind].SQL.Assign(FSQLText[UpdateKind]);
      end;
      Break;
    end;
end;

function To8DataSet.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure To8DataSet.SetDataSource(const Value: TDataSource);
begin
  if IsLinkedTo(Value) then DatabaseError('SCircularDataLink', Self);
  FDataLink.DataSource := Value;
end;

procedure To8DataSet.SetParamsFromCursor;
var
  i: Integer;
  o8: To8StmtOrd;
  DataSet: TDataSet;
begin
  if FDataLink.DataSource <> nil then
  begin
    DataSet := FDataLink.DataSource.DataSet;
    if DataSet <> nil then
    begin
      if FSTMT is To8StmtOrd then begin
        o8 := To8StmtOrd(FSTMT);
        DataSet.FieldDefs.Update;
        for i := 0 to o8.Params.Count - 1 do
          o8.Params[i].ParamValue := DataSet.FieldByName(o8.Params[i].Name).AsVariant;
      end;
    end;
  end;
end;

procedure To8DataSet.ReopenFromSrc;
var
  o8: To8StmtOrd;
begin
  o8 := nil;
  if FSTMT is To8StmtOrd then
    o8 := To8StmtOrd(FSTMT);
  if Active then begin
    SetParamsFromCursor;
    if o8 <> nil then o8.ReExecute;
    ClearRecordBuffers;
    ClearDeletedRecordBuffers;
    FCurrentIndex := -1;
    FRowID := 0;
    Resync([]);
  end else begin
    if o8 <> nil then begin
      o8.Active := false;
      SetParamsFromCursor;
      o8.Active := true;
    end;
    Active := true;
  end;
end;

{ TVKDL1 }

procedure TVKDL1.ActiveChanged;
begin
  o8DataSet.ReopenFromSrc;
end;

procedure TVKDL1.CheckBrowseMode;
begin
  if o8DataSet.Active then o8DataSet.CheckBrowseMode;
end;

procedure TVKDL1.RecordChanged(Field: TField);
begin
  if (Field = nil) and o8DataSet.Active then o8DataSet.ReopenFromSrc;
end;

end.
