{Copyright:      Vlad Karpov  mailto:KarpovVV@protek.ru
 Author:         Vlad Karpov
 Remarks:        freeware, but this Copyright must be included

 Description:    To8Env component for manage environment handler

 Version:        1.0.0

 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}
unit o8Env;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, OCI,
  syncobjs, registry, db, MemMgr;

const

   MaxErrorMessage = OCI_ERROR_MAXMSG_SIZE;

type

  To8Env = class(TComponent)
  private

    { Private declarations }

    envhp: pOCIEnv; { the environment handle }
    AllocHandlerInv: boolean;

    Factive: boolean;
    FStreamedActive: boolean;

    SessionList: TList;
    FOCIDllName: String;
    FOCIOracleHome: String;
    FOCIVersion: String;

    function InitializeOci: boolean;
    procedure GetOCIDllName;

    function GetActive: boolean;
    procedure SetActive(bActive: boolean);

    procedure FreeHandlers;
    function AllocHandlers: boolean;
    procedure SetOCIDllName(const Value: String);
    function GetClientVersion: String;

  protected
    { Protected declarations }

    procedure Loaded; override;

  public

    { Public declarations }

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure AddSession(oOBJ: TObject);
    procedure RemoveSession(oOBJ: TObject);

    class procedure GetError(errhp: pOCIError; bShowError: boolean; out ErrCode: sb4; out ErrMes: String);

    property hEnv: pOCIEnv read envhp;

    property OCIDllName: String read FOCIDllName write SetOCIDllName;
    property OCIClientVersion: String read GetClientVersion;

  published
    { Published declarations }

    property Active: Boolean read GetActive write SetActive;

  end;

function NormalizePath(const APath: String): String;

procedure Register;

var
  o8CrSection: TCriticalSection;

implementation

uses
   o8Session;

//******************************************************************************
procedure Register;
begin
  RegisterComponents('VK NLO', [To8Env]);
end;

//******************************************************************************
function NormalizePath(const APath: String): String;
begin
    if (APath <> '') and (APath[Length(APath)] = '\') then
        Result := Copy(APath, 1, Length(APath) - 1)
    else
        Result := APath;
end;

//******************************************************************************
constructor To8Env.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   SessionList := TList.Create;
   envhp := nil;
   Factive := false;
   FStreamedActive := false;
   AllocHandlerInv := false;
end;

//******************************************************************************
destructor To8Env.Destroy;
var
   oSess: To8Session;
begin
   Active := false;
   while SessionList.Count > 0 do
   begin
      oSess := To8Session(SessionList.Items[0]);
      oSess.Environment := nil;
   end;
   SessionList.Destroy;
   FreeHandlers;
   inherited Destroy;
end;

//******************************************************************************
function To8Env.GetActive: boolean;
begin
   Result := Factive;
end;

//******************************************************************************
procedure To8Env.SetActive(bActive: boolean);
var
  i: Integer;
begin
  if ( csReading in ComponentState ) then FStreamedActive := bActive
  else begin
    if bActive then
    begin
      if InitializeOci then
      begin
        FOCIDllName := OCI.FOCIDllName;
        if AllocHandlers then Factive := true;
      end;
    end else begin
      for i := 0 to SessionList.Count - 1 do To8Session(SessionList.Items[i]).Active := false;
      Factive := false;
      FreeHandlers;
    end;
  end;
end;

//******************************************************************************
procedure To8Env.Loaded;
begin
    inherited Loaded;
    if FStreamedActive then Active := true;
end;

//******************************************************************************
procedure To8Env.AddSession(oOBJ: TObject);
begin
    if Assigned(SessionList) then
    begin
      SessionList.Add(oOBJ);
    end;
end;

//******************************************************************************
procedure To8Env.RemoveSession(oOBJ: TObject);
begin
   if Assigned(SessionList) then
   begin
      SessionList.Remove(oOBJ);
      SessionList.Pack;
   end;
end;

function To8Env.AllocHandlers: boolean;
begin
  FreeHandlers;
  if ( OCIEnvInit(@envhp, OCI_DEFAULT, 0, nil) = OCI_SUCCESS ) then
    AllocHandlerInv := true;
  Result := AllocHandlerInv;
end;

procedure To8Env.FreeHandlers;
begin
  if AllocHandlerInv then
  begin
    OCIHandleFree(envhp, OCI_HTYPE_ENV);
    envhp := nil;
    AllocHandlerInv := false;
  end;
end;

class procedure To8Env.GetError(errhp: pOCIError; bShowError: boolean; out ErrCode: sb4; out ErrMes: String);
var
  buf: array [1..MaxErrorMessage] of char;
  sRet: sword;
begin
  ErrCode := 0;
  fillchar(buf, MaxErrorMessage, 0);
  sRet := OCIErrorGet(  errhp,
                        1,
                        nil,
                        @ErrCode,
                        @buf[1],
                        MaxErrorMessage,
                        OCI_HTYPE_ERROR);
  if sRet in [OCI_SUCCESS, OCI_SUCCESS_WITH_INFO] then
    ErrMes := Trim(buf)
  else begin
    if sRet = OCI_NO_DATA then
    begin
      ErrCode := 2000020;
      ErrMes := '   .';
    end else begin
      ErrCode := 2000010;
      ErrMes := ' ORACLE  .   Error handler.';
    end;
  end;
  if bShowError then
    if ErrCode <> 0 then
      raise Exception.Create(ErrMes);
end;

function To8Env.InitializeOci: boolean;
begin
  if OCIDllName = '' then GetOCIDllName;
  Result := OCI.InitOCI(OCIDllName);
end;

procedure To8Env.GetOCIDllName;
var
  s: String;
  bestInd, i, n: Integer;
  path: String;
  reg: TRegistry;

  function GetRegPathVar(AReg: TRegistry; const AName: String): String;
  begin
    Result := NormalizePath(AReg.ReadString(AName));
  end;

  procedure InitBest;
  var
    sz: DWORD;
  begin
    bestInd := MAXINT;
    FOCIOracleHome := '';
    sz := GetEnvironmentVariable(PChar('PATH'), nil, 0);
    SetString(path, nil, sz);
    GetEnvironmentVariable(PChar('PATH'), PChar(path), sz);
    path := AnsiUpperCase(Copy(path, 1, Length(path) - 1));
  end;

  function TestBestHome(const AKey: String): Boolean;
  var
    i: Integer;
    bin, home: String;
  begin
    Result := reg.OpenKey(AKey, False);
    if Result then begin
      home := GetRegPathVar(reg, 'ORACLE_HOME');
      bin := home + '\bin';
      i := 1;
      while i <= Length(path) do begin
        if AnsiCompareText(bin, NormalizePath(ExtractFieldName(path, i))) = 0 then begin
          if bestInd > i then begin
            bestInd := i;
            FOCIOracleHome := home;
            FOCIDllName := reg.ReadString('ORAOCI');
          end;
          Break;
        end;
      end;
    end;
  end;

begin
  if OCI.FOCILibraryLoaded then
    FOCIDllName := OCI.FOCIDllName
  else begin
    reg := TRegistry.Create;
    with reg do
    try
      RootKey := HKEY_LOCAL_MACHINE;
      InitBest;
      if TestBestHome('\Software\Oracle') then
        if OpenKey('\Software\Oracle\All_Homes', False) then
        begin
          s := ReadString('HOME_COUNTER');
          if s = '' then
            n := 1
          else
            n := StrToInt(s);
          for i := 0 to n - 1 do
            TestBestHome('\Software\Oracle\Home' + IntToStr(i));
        end;
      if FOCIDllName = '' then
      begin
        if FileExists(FOCIOracleHome + '\BIN\OCI.DLL') then
          FOCIDllName := FOCIOracleHome + '\BIN\OCI.DLL';
        // single case, than things differs - 8.0.3
        if not FileExists(FOCIDllName) then
          FOCIDllName := FOCIOracleHome + '\Bin\ORA803.DLL';
      end;
    finally
        Free;
    end;
  end;
end;

procedure To8Env.SetOCIDllName(const Value: String);
begin
  if OCI.FOCILibraryLoaded then
    FOCIDllName := OCI.FOCIDllName
  else
    FOCIDllName := Value;
end;

function To8Env.GetClientVersion: String;
const
  s: String = '\StringFileInfo\040904B0\FileVersion';
  undef: String = '<UNDEFINED>';
var
  hndl, sz: DWORD;
  buf: Pointer;
  len: UINT;
  pStrVer: PChar;
begin
  FOCIVersion := '';
  if Active then
  begin
    pStrVer := nil;
    sz := GetFileVersionInfoSize(PChar(FOCIDllName), hndl);
    if sz > 0 then
    begin
      buf := MemMgr.oMem.GetMem(self, sz);
      try
        if GetFileVersionInfo(PChar(FOCIDllName), hndl, sz, buf) then
          if VerQueryValue(buf, PChar(s), Pointer(pStrVer), Len) then
            FOCIVersion := String(pStrVer);
      finally
        MemMgr.oMem.FreeMem(buf);
      end;
    end;
  end;
  Result := FOCIVersion;
end;

initialization

  o8CrSection := TCriticalSection.Create;

finalization

  o8CrSection.Destroy;

end.
