{Copyright:      Vlad Karpov  mailto:KarpovVV@protek.ru
 Author:         Vlad Karpov
 Remarks:        freeware, but this Copyright must be included

 Description:    Filter in DataSet realised

 Version:        1.0.0

 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}
unit o8Filter;

interface

uses dbcommon, Windows, classes, db, o8Util;

type

  PNLOExprNode = ^TNLOExprNode;
  TNLOExprNode = record
    FNext: PNLOExprNode;
    FKind: TExprNodeKind;
    FPartial: Boolean;
    FOperator: TCANOperator;
    FData: Variant;
    FLeft: PNLOExprNode;
    FRight: PNLOExprNode;
    FDataType: TFieldType;
    FDataSize: Integer;
    FDataLen: Integer;
    FDataPrec: Integer;
    FArgs: TList;
    FScopeKind: TExprScopeKind;
    FField: TField;
  end;

  {To8FilterExpr}
  To8FilterExpr = class
  private
    FDataSet: TDataSet;
    FFieldMap: TFieldMap;
    FOptions: TFilterOptions;
    FParserOptions: TParserOptions;
    FNodes: PNLOExprNode;
    FFieldName: string;
    FDependentFields: TBits;
    function GetFieldByName(Name: string) : TField;
  public
    constructor Create(DataSet: TDataSet; Options: TFilterOptions;
      ParseOptions: TParserOptions; const FieldName: string; DepFields: TBits;
      FieldMap: TFieldMap);
    destructor Destroy; override;
    function NewCompareNode(Field: TField; Operator: TCANOperator;
      const Value: Variant): PNLOExprNode;
    function NewNode(Kind: TExprNodeKind; Operator: TCANOperator;
      const Data: Variant; Left, Right: PNLOExprNode): PNLOExprNode;
    property DataSet: TDataSet write FDataSet;
  end;

  {To8ExprParser}
  To8ExprParser = class(To8FilterExpr)
    FFilter: To8FilterExpr;
    FFieldMap: TFieldMap;
    FText: string;
    FSourcePtr: PChar;
    FTokenPtr: PChar;
    FTokenString: string;
    FStrTrue: string;
    FStrFalse: string;
    FToken: TExprToken;
    FPrevToken: TExprToken;
    FNumericLit: Boolean;
    FParserOptions: TParserOptions;
    FFieldName: string;
    FDataSet: TDataSet;
    FDependentFields: TBits;
    FIndexKeyValue: boolean;
    FFields: TList;
    FKeyValues: Variant;
    FKeyFromValues: boolean;
    FFC: Char;
    procedure NextToken;
    function NextTokenIsLParen : Boolean;
    function ParseExpr: PNLOExprNode;
    function ParseExpr2: PNLOExprNode;
    function ParseExpr3: PNLOExprNode;
    function ParseExpr4: PNLOExprNode;
    function ParseExpr5: PNLOExprNode;
    function ParseExpr6: PNLOExprNode;
    function ParseExpr7: PNLOExprNode;
    function TokenName: string;
    function TokenSymbolIs(const S: string): Boolean;
    function TokenSymbolIsFunc(const S: string) : Boolean;
    procedure GetFuncResultInfo(Node: PNLOExprNode);
    procedure TypeCheckArithOp(Node: PNLOExprNode);
    procedure GetScopeKind(Root, Left, Right : PNLOExprNode);
    function Execute(Root: PNLOExprNode): Variant; overload;
  private
    FLastRoot: PNLOExprNode;
    FValue: Variant;
    FKey: String;
    function GetDataLen: Integer;
    function GetDataPrec: Integer;
  public
    constructor Create(DataSet: TDataSet; const Text: string;
      Options: TFilterOptions; ParserOptions: TParserOptions;
      const FieldName: string; DepFields: TBits; FieldMap: TFieldMap);
    destructor Destroy; override;
    procedure SetExprParams(const Text: string; Options: TFilterOptions;
      ParserOptions: TParserOptions; const FieldName: string);
    procedure SetExprParams1(const Text: string; Options: TFilterOptions;
      ParserOptions: TParserOptions; const FieldName: string);
    function Execute: Variant; overload;
    function EvaluteKey: String; overload;
    function EvaluteKey(const KeyFields: string; const KeyValues: Variant; const CF: Char = #$20): String; overload;
    function SuiteFieldList(fl: String; out m: Integer): Integer;
    property IndexKeyValue: boolean read FIndexKeyValue write FIndexKeyValue;
    property Value: Variant read FValue;
    property Key: String read FKey;
    property Len: Integer read GetDataLen; //FLastRoot.FDataLen;
    property Prec: Integer read GetDataPrec; //FLastRoot.FDataPrec;
  end;

  function LikeOperator(const Var1, Var2: Variant; const CaseInsensitive: Boolean; const ManyChars, OneChar: Char): boolean;

implementation

uses SysUtils, DBConsts, DBFDataSet;

const
  StringFieldTypes = [ftString, ftFixedChar, ftWideString, ftGuid];
  BlobFieldTypes = [ftBlob, ftMemo, ftGraphic, ftFmtMemo, ftParadoxOle, ftDBaseOle,
    ftTypedBinary, ftOraBlob, ftOraClob];

function IsNumeric(DataType: TFieldType): Boolean;
begin
  Result := DataType in [ftSmallint, ftInteger, ftWord, ftFloat, ftCurrency,
    ftBCD, ftAutoInc, ftLargeint];
end;

function IsTemporal(DataType: TFieldType): Boolean;
begin
  Result := DataType in [ftDate, ftTime, ftDateTime];
end;

function LikeOperator(const Var1, Var2: Variant; const CaseInsensitive: Boolean; const ManyChars, OneChar: Char): boolean;
var
  sStr, sPatt: String;
begin
  if VarIsNull(Var1) or VarIsNull(Var2) then
    Result := False
  else begin
    sStr := Var1;
    sPatt := Var2;
    if CaseInsensitive then
    begin
      sStr := AnsiUpperCase(sStr);
      sPatt := AnsiUpperCase(sPatt);
    end;
    Result := wildc(pChar(sPatt), pChar(sStr), Length(sStr), ManyChars, OneChar);
  end;
end;

{To8FilterExpr}
constructor To8FilterExpr.Create(DataSet: TDataSet; Options: TFilterOptions;
  ParseOptions: TParserOptions; const FieldName: string; DepFields: TBits;
  FieldMap: TFieldMap);
begin
  FFieldMap := FieldMap;
  FDataSet := DataSet;
  FOptions := Options;
  FFieldName := FieldName;
  FParserOptions := ParseOptions;
  FDependentFields := DepFields;
end;

destructor To8FilterExpr.Destroy;
var
  Node: PNLOExprNode;
begin
  while FNodes <> nil do
  begin
    Node := FNodes;
    FNodes := Node^.FNext;
    if (Node^.FKind = enFunc) and (Node^.FArgs <> nil) then
      Node^.FArgs.Free;
    Dispose(Node);
  end;
end;

function To8FilterExpr.NewCompareNode(Field: TField; Operator: TCANOperator;
  const Value: Variant): PNLOExprNode;
var
  ConstExpr: PNLOExprNode;
begin
  ConstExpr := NewNode(enConst, coNOTDEFINED, Value, nil, nil);
  ConstExpr^.FDataType := Field.DataType;
  ConstExpr^.FDataSize := Field.Size;
  Result := NewNode(enOperator, Operator, Unassigned,
    NewNode(enField, coNOTDEFINED, Field.FieldName, nil, nil), ConstExpr);
end;

function To8FilterExpr.NewNode(Kind: TExprNodeKind; Operator: TCANOperator;
  const Data: Variant; Left, Right: PNLOExprNode): PNLOExprNode;
var
  Field : TField;
begin
  New(Result);
  with Result^ do
  begin
    FNext := FNodes;
    FKind := Kind;
    FPartial := False;
    FOperator := Operator;
    FData := Data;
    FLeft := Left;
    FRight := Right;
    FDataLen := 0;
    FDataPrec := 0;
    FDataType := ftUnknown;
    FArgs := nil;
  end;
  FNodes := Result;
  if Kind = enField then
  begin
    Field := GetFieldByName(Data);
    if Field = nil then
      DatabaseErrorFmt(SFieldNotFound, [Data]);
    Result^.FDataType := Field.DataType;
    Result^.FDataSize := Field.Size;
    Result^.FField := Field;
  end;
end;

function To8FilterExpr.GetFieldByName(Name: string) : TField;
//var
//  I: Integer;
//  F: TField;
//  FieldInfo: TFieldInfo;
begin
//  Result := nil;
//  if poFieldNameGiven in FParserOptions then
//    Result := FDataSet.FieldByName(UpperCase(FFieldName))
//  else if poUseOrigNames in FParserOptions then begin
//    for I := 0 to FDataset.FieldCount - 1 do
//    begin
//      F := FDataSet.Fields[I];
//      if GetFieldInfo(F.Origin, FieldInfo) and
//         (AnsiCompareStr(Name, FieldInfo.OriginalFieldName) = 0) then
//      begin
//        Result := F;
//        Exit;
//      end;
//    end;
//  end;
//  if Result = nil then
    Result := FDataSet.FieldByName(UpperCase(Name));
//  if (Result <> nil) and (Result.FieldKind = fkCalculated) and (poAggregate in FParserOptions) then
//    DatabaseErrorFmt(SExprNoAggOnCalcs, [Result.FieldName]);
//  if (poFieldDepend in FParserOptions) and (Result <> nil) and
//     (FDependentFields <> nil) then
//    FDependentFields[Result.FieldNo-1] := True;
end;

{To8ExprParser}
constructor To8ExprParser.Create(DataSet: TDataSet; const Text: string;
  Options: TFilterOptions; ParserOptions: TParserOptions; const FieldName: string;
  DepFields: TBits; FieldMap: TFieldMap);
begin
  FFieldMap := FieldMap;
  FStrTrue := STextTrue;
  FStrFalse := STextFalse;
  FDataSet := DataSet;
  FDependentFields := DepFields;
  FIndexKeyValue := false;
  FFilter := To8FilterExpr.Create(DataSet, Options, ParserOptions, FieldName,
    DepFields, FieldMap);
  if Text <> '' then
    SetExprParams(Text, Options, ParserOptions, FieldName);
  FFields := nil;
  FKeyFromValues := false;
  FFC := #$20; //' '
end;

destructor To8ExprParser.Destroy;
begin
  FFilter.Free;
end;

procedure  To8ExprParser.SetExprParams(const Text: string; Options: TFilterOptions;
  ParserOptions: TParserOptions; const FieldName: string);
var
  Root, DefField: PNLOExprNode;
begin
  FParserOptions := ParserOptions;
  if FFilter <> nil then
    FFilter.Free;
  FFilter := To8FilterExpr.Create(FDataSet, Options, ParserOptions, FieldName,
    FDependentFields, FFieldMap);
  FText := Text;
  FSourcePtr := PChar(Text);
  FFieldName := FieldName;
  NextToken;
  Root := nil;
  if FToken <> etEnd then Root := ParseExpr;
  FValue := NULL;
  if Root <> nil then begin
    if FToken <> etEnd then DatabaseError(SExprTermination);

    if (poAggregate in FParserOptions) and (Root^.FScopeKind <> skAgg) then
       DatabaseError(SExprNotAgg);
    if (not (poAggregate in FParserOptions)) and (Root^.FScopeKind = skAgg) then
       DatabaseError(SExprNoAggFilter);
    if poDefaultExpr in ParserOptions then
    begin
      DefField := FFilter.NewNode(enField, coNOTDEFINED, FFieldName, nil, nil);
      if (IsTemporal(DefField^.FDataType) and (Root^.FDataType in StringFieldTypes)) or
         ((DefField^.FDataType = ftBoolean ) and (Root^.FDataType in StringFieldTypes)) then
        Root^.FDataType := DefField^.FDataType;

      if not ((IsTemporal(DefField^.FDataType) and IsTemporal(Root^.FDataType))
         or (IsNumeric(DefField^.FDataType) and IsNumeric(Root^.FDataType))
         or ((DefField^.FDataType in StringFieldTypes) and (Root^.FDataType in StringFieldTypes))
         or ((DefField^.FDataType = ftBoolean) and (Root^.FDataType = ftBoolean))) then
        DatabaseError(SExprTypeMis);
      Root := FFilter.NewNode(enOperator, coASSIGN, Unassigned, Root, DefField);
    end;

    if not (poAggregate in FParserOptions) and not(poDefaultExpr in ParserOptions)
       and (Root^.FDataType <> ftBoolean ) then
       DatabaseError(SExprIncorrect);


    FValue := Execute(Root);

  end;

  FLastRoot := Root;

end;

procedure To8ExprParser.SetExprParams1(const Text: string;
  Options: TFilterOptions; ParserOptions: TParserOptions;
  const FieldName: string);
var
  Root: PNLOExprNode;
begin
  FParserOptions := ParserOptions;
  if FFilter <> nil then
    FFilter.Free;
  FFilter := To8FilterExpr.Create(FDataSet, Options, ParserOptions, FieldName,
    FDependentFields, FFieldMap);
  FText := Text;
  FSourcePtr := PChar(Text);
  FFieldName := FieldName;
  NextToken;
  Root := nil;
  if FToken <> etEnd then Root := ParseExpr;
  if Root <> nil then
    if FToken <> etEnd then DatabaseError(SExprTermination);

  FLastRoot := Root;

end;

function To8ExprParser.NextTokenIsLParen : Boolean;
var
  P : PChar;
begin
  P := FSourcePtr;
  while (P^ <> #0) and (P^ <= ' ') do Inc(P);
  Result := P^ = '(';
end;

procedure To8ExprParser.NextToken;
type
  ASet = Set of Char;
var
  P, TokenStart: PChar;
  L: Integer;
  StrBuf: array[0..255] of Char;

  function IsKatakana(const Chr: Byte): Boolean;
  begin
    Result := (SysLocale.PriLangID = LANG_JAPANESE) and (Chr in [$A1..$DF]);
  end;

  procedure Skip(TheSet: ASet);
  begin
    while TRUE do
    begin
      if P^ in LeadBytes then
        Inc(P, 2)
      else if (P^ in TheSet) or IsKatakana(Byte(P^)) then
        Inc(P)
      else
        Exit;
    end;
  end;

  procedure Litr(ltr: Char);
  begin
    Inc(P);
    L := 0;
    while True do
    begin
      if P^ = #0 then DatabaseError(SExprStringError);
      if P^ = ltr then
      begin
        Inc(P);
        if P^ <> ltr then Break;
      end;
      if L < SizeOf(StrBuf) then
      begin
        StrBuf[L] := P^;
        Inc(L);
      end;
      Inc(P);
    end;
    SetString(FTokenString, StrBuf, L);
    FToken := etLiteral;
    FNumericLit := False;
  end;

begin
  FPrevToken := FToken;
  FTokenString := '';
  P := FSourcePtr;
  while (P^ <> #0) and (P^ <= ' ') do Inc(P);
  if (P^ <> #0) and (P^ = '/') and (P[1] <> #0) and (P[1] = '*')then
  begin
    P := P + 2;
    while (P^ <> #0) and (P^ <> '*') do Inc(P);
    if (P^ = '*') and (P[1] <> #0) and (P[1] =  '/')  then
      P := P + 2
    else
      DatabaseErrorFmt(SExprInvalidChar, [P^]);
  end;
  while (P^ <> #0) and (P^ <= ' ') do Inc(P);
  FTokenPtr := P;
  case P^ of
    'A'..'Z', 'a'..'z', '_', #$81..#$fe:
      begin
        TokenStart := P;
        if not SysLocale.FarEast then
        begin
          Inc(P);
          while P^ in ['A'..'Z', 'a'..'z', '0'..'9', '_', '.', '[', ']'] do Inc(P);
        end
        else
          Skip(['A'..'Z', 'a'..'z', '0'..'9', '_', '.', '[', ']']);
        SetString(FTokenString, TokenStart, P - TokenStart);
        FToken := etSymbol;
        if CompareText(FTokenString, 'LIKE') = 0 then   { do not localize }
          FToken := etLIKE
        else if CompareText(FTokenString, 'IN') = 0 then   { do not localize }
          FToken := etIN
        else if CompareText(FTokenString, 'IS') = 0 then    { do not localize }
        begin
          while (P^ <> #0) and (P^ <= ' ') do Inc(P);
          TokenStart := P;
          Skip(['A'..'Z', 'a'..'z']);
          SetString(FTokenString, TokenStart, P - TokenStart);
          if CompareText(FTokenString, 'NOT')= 0 then  { do not localize }
          begin
            while (P^ <> #0) and (P^ <= ' ') do Inc(P);
            TokenStart := P;
            Skip(['A'..'Z', 'a'..'z']);
            SetString(FTokenString, TokenStart, P - TokenStart);
            if CompareText(FTokenString, 'NULL') = 0 then
              FToken := etISNOTNULL
            else
              DatabaseError(SInvalidKeywordUse);
          end
          else if CompareText (FTokenString, 'NULL') = 0  then  { do not localize }
          begin
            FToken := etISNULL;
          end
          else
            DatabaseError(SInvalidKeywordUse);
        end;
      end;
    '[':
      begin
        Inc(P);
        TokenStart := P;
        P := AnsiStrScan(P, ']');
        if P = nil then DatabaseError(SExprNameError);
        SetString(FTokenString, TokenStart, P - TokenStart);
        FToken := etName;
        Inc(P);
      end;
    '''': Litr('''');
    '"': Litr('"');
    '-', '0'..'9':
      begin
        if (FPrevToken <> etLiteral) and (FPrevToken <> etName) and
           (FPrevToken <> etSymbol)and (FPrevToken <> etRParen) then
          begin
            TokenStart := P;
            Inc(P);
            while (P^ in ['0'..'9', DecimalSeparator, 'e', 'E', '+', '-']) do
              Inc(P);
            if ((P-1)^ = ',') and (DecimalSeparator = ',') and (P^ = ' ') then
              Dec(P);
            SetString(FTokenString, TokenStart, P - TokenStart);
            FToken := etLiteral;
            FNumericLit := True;
          end
        else
         begin
           FToken := etSUB;
           Inc(P);
         end;
      end;
    '(':
      begin
        Inc(P);
        FToken := etLParen;
      end;
    ')':
      begin
        Inc(P);
        FToken := etRParen;
      end;
    '<':
      begin
        Inc(P);
        case P^ of
          '=':
            begin
              Inc(P);
              FToken := etLE;
            end;
          '>':
            begin
              Inc(P);
              FToken := etNE;
            end;
        else
          FToken := etLT;
        end;
      end;
    '=':
      begin
        Inc(P);
        FToken := etEQ;
      end;
    '>':
      begin
        Inc(P);
        if P^ = '=' then
        begin
          Inc(P);
          FToken := etGE;
        end else
          FToken := etGT;
      end;
    '+':
      begin
        Inc(P);
        FToken := etADD;
      end;
    '*':
      begin
        Inc(P);
        FToken := etMUL;
      end;
    '/':
      begin
        Inc(P);
        FToken := etDIV;
      end;
    ',':
      begin
        Inc(P);
        FToken := etComma;
      end;
    #0:
      FToken := etEnd;
    '.':
      begin
        TokenStart := P;
        Skip(['A'..'Z', 'a'..'z', '0'..'9', '_', '.', '[', ']']);
        SetString(FTokenString, TokenStart, P - TokenStart);
        if CompareText(FTokenString, '.T.') = 0 then FTokenString := 'TRUE';
        if CompareText(FTokenString, '.F.') = 0 then FTokenString := 'FALSE';
        if CompareText(FTokenString, '.AND.') = 0 then FTokenString := 'AND';
        if CompareText(FTokenString, '.OR.') = 0 then FTokenString := 'OR';
        if CompareText(FTokenString, '.XOR.') = 0 then FTokenString := 'XOR';
        if CompareText(FTokenString, '.NOT.') = 0 then FTokenString := 'NOT';
        FToken := etSymbol;
      end;
    '!':
      begin
        Inc(P);
        FTokenString := 'NOT';
        FToken := etSymbol;
      end;
  else
    DatabaseErrorFmt(SExprInvalidChar, [P^]);
  end;
  FSourcePtr := P;
end;

function To8ExprParser.ParseExpr: PNLOExprNode;
begin
  Result := ParseExpr2;
  while TokenSymbolIs('OR') do
  begin
    NextToken;
    Result := FFilter.NewNode(enOperator, coOR, Unassigned,
      Result, ParseExpr2);
    GetScopeKind(Result, Result^.FLeft, Result^.FRight);
    Result^.FDataType := ftBoolean;
  end;
end;

function To8ExprParser.ParseExpr2: PNLOExprNode;
begin
  Result := ParseExpr3;
  while TokenSymbolIs('AND') do
  begin
    NextToken;
    Result := FFilter.NewNode(enOperator, coAND, Unassigned,
      Result, ParseExpr3);
    GetScopeKind(Result, Result^.FLeft, Result^.FRight);
    Result^.FDataType := ftBoolean;
  end;
end;

function To8ExprParser.ParseExpr3: PNLOExprNode;
begin
  if TokenSymbolIs('NOT') then
  begin
    NextToken;
    Result := FFilter.NewNode(enOperator, coNOT, Unassigned,
      ParseExpr4, nil);
    Result^.FDataType := ftBoolean;
  end else
    Result := ParseExpr4;
  GetScopeKind(Result, Result^.FLeft, Result^.FRight);
end;


function To8ExprParser.ParseExpr4: PNLOExprNode;
const
  Operators: array[etEQ..etLT] of TCANOperator = (
    coEQ, coNE, coGE, coLE, coGT, coLT);
var
  Operator: TCANOperator;
  Left, Right: PNLOExprNode;
begin
  Result := ParseExpr5;
  if (FToken in [etEQ..etLT]) or (FToken = etLIKE)
     or (FToken = etISNULL) or (FToken = etISNOTNULL)
     or (FToken = etIN) then
  begin
    case FToken of
      etEQ..etLT:
        Operator := Operators[FToken];
      etLIKE:
        Operator := coLIKE;
      etISNULL:
        Operator := coISBLANK;
      etISNOTNULL:
        Operator := coNOTBLANK;
      etIN:
        Operator := coIN;
      else
        Operator := coNOTDEFINED;
    end;
    NextToken;
    Left := Result;
    if Operator = coIN then
    begin
      if FToken <> etLParen then
        DatabaseErrorFmt(SExprNoLParen, [TokenName]);
      NextToken;
      Result := FFilter.NewNode(enOperator, coIN, Unassigned,
                 Left, nil);
      Result.FDataType := ftBoolean;
      if FToken <> etRParen then
      begin
        Result.FArgs := TList.Create;
        repeat
          Right := ParseExpr;
          if IsTemporal(Left.FDataType) then
            Right.FDataType := Left.FDataType;
          Result.FArgs.Add(Right);
          if (FToken <> etComma) and (FToken <> etRParen) then
            DatabaseErrorFmt(SExprNoRParenOrComma, [TokenName]);
          if FToken = etComma then NextToken;
        until (FToken = etRParen) or (FToken = etEnd);
        if FToken <> etRParen then
          DatabaseErrorFmt(SExprNoRParen, [TokenName]);
        NextToken;
      end else
        DatabaseError(SExprEmptyInList);
    end else
    begin
      if (Operator <> coISBLANK) and (Operator <> coNOTBLANK) then
        Right := ParseExpr5
      else
        Right := nil;
      Result := FFilter.NewNode(enOperator, Operator, Unassigned,
        Left, Right);
      if Right <> nil then
      begin
        if (Left^.FKind = enField) and (Right^.FKind = enConst) then
          begin
            Right^.FDataType := Left^.FDataType;
            Right^.FDataSize := Left^.FDataSize;
          end
        else if (Right^.FKind = enField) and (Left^.FKind = enConst) then
          begin
            Left^.FDataType := Right^.FDataType;
            Left^.FDataSize := Right^.FDataSize;
          end;
      end;
      if (Left^.FDataType in BlobFieldTypes) and (Operator = coLIKE) then
      begin
        if Right^.FKind = enConst then Right^.FDataType := ftString;
      end
      else if (Operator <> coISBLANK) and (Operator <> coNOTBLANK)
         and ((Left^.FDataType in (BlobFieldTypes + [ftBytes])) or
         ((Right <> nil) and (Right^.FDataType in (BlobFieldTypes + [ftBytes])))) then
        DatabaseError(SExprTypeMis);
      Result.FDataType := ftBoolean;
      if Right <> nil then
      begin
        if IsTemporal(Left.FDataType) and (Right.FDataType in StringFieldTypes) then
          Right.FDataType := Left.FDataType
        else if IsTemporal(Right.FDataType) and (Left.FDataType in StringFieldTypes) then
          Left.FDataType := Right.FDataType;
      end;
      GetScopeKind(Result, Left, Right);
    end;
  end;
end;

function To8ExprParser.ParseExpr5: PNLOExprNode;
const
  Operators: array[etADD..etDIV] of TCANOperator = (
    coADD, coSUB, coMUL, coDIV);
var
  Operator: TCANOperator;
  Left, Right: PNLOExprNode;
begin
  Result := ParseExpr6;
  while FToken in [etADD, etSUB] do
  begin
    if not (poExtSyntax in FParserOptions) then
      DatabaseError(SExprNoArith);
    Operator := Operators[FToken];
    Left := Result;
    NextToken;
    Right := ParseExpr6;
    Result := FFilter.NewNode(enOperator, Operator, Unassigned, Left, Right);
    TypeCheckArithOp(Result);
    GetScopeKind(Result, Left, Right);
  end;
end;

function To8ExprParser.ParseExpr6: PNLOExprNode;
const
  Operators: array[etADD..etDIV] of TCANOperator = (
    coADD, coSUB, coMUL, coDIV);
var
  Operator: TCANOperator;
  Left, Right: PNLOExprNode;
begin
  Result := ParseExpr7;
  while FToken in [etMUL, etDIV] do
  begin
    if not (poExtSyntax in FParserOptions) then
      DatabaseError(SExprNoArith);
    Operator := Operators[FToken];
    Left := Result;
    NextToken;
    Right := ParseExpr7;
    Result := FFilter.NewNode(enOperator, Operator, Unassigned, Left, Right);
    TypeCheckArithOp(Result);
    GetScopeKind(Result, Left, Right);
  end;
end;


function To8ExprParser.ParseExpr7: PNLOExprNode;
var
  FuncName: string;
begin
  case FToken of
    etSymbol:
      if (poExtSyntax in FParserOptions)
         and  NextTokenIsLParen and TokenSymbolIsFunc(FTokenString) then
        begin
          Funcname := FTokenString;
          NextToken;
          if FToken <> etLParen then
            DatabaseErrorFmt(SExprNoLParen, [TokenName]);
          NextToken;
          if (CompareText(FuncName,'count') = 0) and (FToken = etMUL) then
          begin
            FuncName := 'COUNT(*)';
            NextToken;
          end;
          Result := FFilter.NewNode(enFunc, coNOTDEFINED, FuncName,
                    nil, nil);
          if FToken <> etRParen then
          begin
            Result.FArgs := TList.Create;
            repeat
              Result.FArgs.Add(ParseExpr);
              if (FToken <> etComma) and (FToken <> etRParen) then
                DatabaseErrorFmt(SExprNoRParenOrComma, [TokenName]);
              if FToken = etComma then NextToken;
            until (FToken = etRParen) or (FToken = etEnd);
          end else
            Result.FArgs := nil;

          GetFuncResultInfo(Result);
        end
      else if TokenSymbolIs('NULL') then
        begin
          Result := FFilter.NewNode(enConst, coNOTDEFINED, System.Null, nil, nil);
          Result.FScopeKind := skConst;
        end
      else if TokenSymbolIs(FStrTrue) then
        begin
          Result := FFilter.NewNode(enConst, coNOTDEFINED, True, nil, nil);
          Result.FScopeKind := skConst;
        end
      else if TokenSymbolIs(FStrFalse) then
        begin
          Result := FFilter.NewNode(enConst, coNOTDEFINED, False, nil, nil);
          Result.FScopeKind := skConst;
        end
      else
        begin
          Result := FFilter.NewNode(enField, coNOTDEFINED, FTokenString, nil, nil);
          Result.FScopeKind := skField;
        end;
    etName:
      begin
        Result := FFilter.NewNode(enField, coNOTDEFINED, FTokenString, nil, nil);
        Result.FScopeKind := skField;
      end;
    etLiteral:
      begin
        Result := FFilter.NewNode(enConst, coNOTDEFINED, FTokenString, nil, nil);
        if FNumericLit then Result^.FDataType := ftFloat else
           Result^.FDataType := ftString;
        Result.FScopeKind := skConst;
      end;
    etLParen:
      begin
        NextToken;
        Result := ParseExpr;
        if FToken <> etRParen then DatabaseErrorFmt(SExprNoRParen, [TokenName]);
      end;
  else
    DatabaseErrorFmt(SExprExpected, [TokenName]);
    Result := nil;
  end;
  NextToken;
end;

procedure  To8ExprParser.GetScopeKind(Root, Left, Right : PNLOExprNode);
begin
  if (Left = nil) and (Right = nil) then Exit;
  if Right = nil then
  begin
    Root.FScopeKind := Left.FScopeKind;
    Exit;
  end;
  if ((Left^.FScopeKind = skField) and (Right^.FScopeKind = skAgg))
     or ((Left^.FScopeKind = skAgg) and (Right^.FScopeKind = skField)) then
    DatabaseError(SExprBadScope);
  if (Left^.FScopeKind = skConst) and (Right^.FScopeKind = skConst) then
    Root^.FScopeKind := skConst
  else if (Left^.FScopeKind = skAgg) or (Right^.FScopeKind = skAgg) then
    Root^.FScopeKind := skAgg
  else if (Left^.FScopeKind = skField) or (Right^.FScopeKind = skField) then
    Root^.FScopeKind := skField;
end;

procedure To8ExprParser.GetFuncResultInfo(Node : PNLOExprNode);
begin
  Node^.FDataType := ftString;
  if (CompareText(Node^.FData, 'COUNT(*)') <> 0 )
     and (CompareText(Node^.FData,'GETDATE') <> 0 )
     and ( (Node^.FArgs = nil ) or ( Node^.FArgs.Count = 0) ) then
      DatabaseError(SExprTypeMis);

  if (Node^.FArgs <> nil) and (Node^.FArgs.Count > 0) then
     Node^.FScopeKind := PNLOExprNode(Node^.FArgs.Items[0])^.FScopeKind;
  if (CompareText(Node^.FData , 'SUM') = 0) or
     (CompareText(Node^.FData , 'AVG') = 0) then
  begin
    Node^.FDataType := ftFloat;
    Node^.FScopeKind := skAgg;
  end
  else if (CompareText(Node^.FData , 'MIN') = 0) or
          (CompareText(Node^.FData , 'MAX') = 0) then
  begin
    Node^.FDataType := PNLOExprNode(Node^.FArgs.Items[0])^.FDataType;
    Node^.FScopeKind := skAgg;
  end
  else if  (CompareText(Node^.FData , 'COUNT') = 0) or
           (CompareText(Node^.FData , 'COUNT(*)') = 0) then
  begin
    Node^.FDataType := ftInteger;
    Node^.FScopeKind := skAgg;
  end
  else if (CompareText(Node^.FData , 'YEAR') = 0) or
          (CompareText(Node^.FData , 'MONTH') = 0) or
          (CompareText(Node^.FData , 'DAY') = 0) or
          (CompareText(Node^.FData , 'HOUR') = 0) or
          (CompareText(Node^.FData , 'MINUTE') = 0) or
          (CompareText(Node^.FData , 'SECOND') = 0 ) then
  begin
    Node^.FDataType := ftInteger;
    Node^.FScopeKind := PNLOExprNode(Node^.FArgs.Items[0])^.FScopeKind;
  end
  else if CompareText(Node^.FData , 'GETDATE') = 0  then
  begin
    Node^.FDataType := ftDateTime;
    Node^.FScopeKind := skConst;
  end
  else if CompareText(Node^.FData , 'DATE') = 0  then
  begin
    Node^.FDataType := ftDate;
    Node^.FScopeKind := PNLOExprNode(Node^.FArgs.Items[0])^.FScopeKind;
  end
  else if CompareText(Node^.FData , 'TIME') = 0  then
  begin
    Node^.FDataType := ftTime;
    Node^.FScopeKind := PNLOExprNode(Node^.FArgs.Items[0])^.FScopeKind;
  end;
end;

function To8ExprParser.TokenName: string;
begin
  if FSourcePtr = FTokenPtr then Result := SExprNothing else
  begin
    SetString(Result, FTokenPtr, FSourcePtr - FTokenPtr);
    Result := '''' + Result + '''';
  end;
end;

function To8ExprParser.TokenSymbolIs(const S: string): Boolean;
begin
  Result := (FToken = etSymbol) and (CompareText(FTokenString, S) = 0);
end;


function To8ExprParser.TokenSymbolIsFunc(const S: string) : Boolean;
begin
  Result := (CompareText(S, 'UPPER') = 0) or
            (CompareText(S, 'LOWER') = 0) or
            (CompareText(S, 'SUBSTRING') = 0) or
            (CompareText(S, 'SUBSTR') = 0) or
            (CompareText(S, 'TRIM') = 0) or
            (CompareText(S, 'TRIMLEFT') = 0) or
            (CompareText(S, 'LTRIM') = 0) or
            (CompareText(S, 'TRIMRIGHT') = 0) or
            (CompareText(S, 'RTRIM') = 0) or
            (CompareText(S, 'DTOS') = 0) or
            (CompareText(S, 'DTTOS') = 0) or
            (CompareText(S, 'STR') = 0) or
            (CompareText(S, 'YEAR') = 0) or
            (CompareText(S, 'MONTH') = 0) or
            (CompareText(S, 'DAY') = 0) or
            (CompareText(S, 'HOUR') = 0) or
            (CompareText(S, 'MINUTE') = 0) or
            (CompareText(S, 'SECOND') = 0) or
            (CompareText(S, 'GETDATE') = 0) or
            (CompareText(S, 'DATE') = 0) or
            (CompareText(S, 'TIME') = 0) or
            (CompareText(S, 'SUM') = 0) or
            (CompareText(S, 'MIN') = 0) or
            (CompareText(S, 'MAX') = 0) or
            (CompareText(S, 'AVG') = 0) or
            (CompareText(S, 'COUNT') = 0);

end;

procedure To8ExprParser.TypeCheckArithOp(Node: PNLOExprNode);
begin
  with Node^ do
  begin
    if IsNumeric(FLeft.FDataType) and IsNumeric(FRight.FDataType)  then
      FDataType := ftFloat
    else if (FLeft.FDataType in StringFieldTypes) and
       (FRight.FDataType in StringFieldTypes) and (FOperator = coADD) then
      FDataType := ftString
    else if IsTemporal(FLeft.FDataType) and IsNumeric(FRight.FDataType) and
       (FOperator = coADD) then
      FDataType := ftDateTime
    else if IsTemporal(FLeft.FDataType) and IsNumeric(FRight.FDataType) and
       (FOperator = coSUB) then
      FDataType := FLeft.FDataType
    else if IsTemporal(FLeft.FDataType) and IsTemporal(FRight.FDataType) and
       (FOperator = coSUB) then
      FDataType := ftFloat
    else if (FLeft.FDataType in StringFieldTypes) and IsTemporal(FRight.FDataType) and
       (FOperator = coSUB) then
    begin
      FLeft.FDataType := FRight.FDataType;
      FDataType := ftFloat;
    end
    else if ( FLeft.FDataType in StringFieldTypes) and  IsNumeric(FRight.FDataType )and
         (FLeft.FKind = enConst)  then
      FLeft.FDataType := ftDateTime
    else
      DatabaseError(SExprTypeMis);
  end;
end;

function To8ExprParser.Execute(Root: PNLOExprNode): Variant;

    function EvaluteNodeValueComplex(ANode: PNLOExprNode): Variant;
    var
      V: PNLOExprNode;
      l, r, Vr: Variant;
      i, j, k: Integer;
      S, S1: String;
      Year, Month, Day: Word;
      Hour, Min, Sec, MSec: Word;
      dt: TDateTime;
      ff: boolean;

      function VarIsString(const V: Variant): Boolean;
      var
          tp: Integer;
      begin
          tp := VarType(l);
          Result := (tp = varString) or (tp = varOleStr);
      end;

      procedure UpperCaseLR;
      begin
          if VarIsString(l) then
              l := AnsiUpperCase(l);
          if VarIsString(r) then
              r := AnsiUpperCase(r);
      end;

      function PartialEQ(AForce: Boolean): Boolean;
      var
          sL, sR: String;
          ln, lnL, lnR: Integer;
          partial: Boolean;
      begin
          if VarIsString(l) and VarIsString(r) then begin
              sL := l;
              sR := r;
              lnL := Length(sL);
              lnR := Length(sR);
              if l <> '' then begin
                  partial := False;
                  if sL[lnL] = '*' then begin
                      partial := True;
                      Dec(lnL);
                  end;
                  if r <> '' then begin
                      if sR[lnR] = '*' then begin
                          partial := True;
                          Dec(lnR);
                      end;
                      if partial or AForce then begin
                          ln := lnR;
                          if ln > lnL then
                              ln := lnL;
                          if (foCaseInsensitive in FOptions) then
                              Result := AnsiStrLIComp(PChar(sL), PChar(sR), ln) = 0
                          else
                              Result := AnsiStrLComp(PChar(sL), PChar(sR), ln) = 0;
                          Exit;
                      end;
                  end;
              end;
              if (foCaseInsensitive in FOptions) then
                  Result := AnsiCompareText(sL, sR) = 0
              else
                  Result := sL = Sr;
          end
          else begin
              UpperCaseLR;
              Result := l = r;
          end;
      end;

    begin
      Result := Unassigned;
      case ANode^.FKind of
      enField:
        begin
          if not FKeyFromValues then
            Result := (ANode^.FField).Value
          else begin
            ff := false;
            for i:=0 to FFields.Count - 1 do
              if TField(FFields[i]).FieldName = (ANode^.FField).FieldName then begin
                if VarIsArray(FKeyValues) then
                  Result := FKeyValues[i]
                else
                  Result := FKeyValues;
                ff := true;
                if (Result <> Null) and (ANode^.FDataType = ftString) and (Length(Result) < ANode^.FDataLen) then
                  Result := Result + StringOfChar(' ', ANode^.FDataLen - Length(Result));
                break;
              end;
            if not ff then Result := Null;
          end;
          if FIndexKeyValue then begin
            if ANode^.FDataType in [ftFloat, ftCurrency] then begin
              ANode^.FDataLen := TSmartDBF((ANode^.FField).DataSet).GetLen(ANode^.FField);
              ANode^.FDataPrec := TSmartDBF((ANode^.FField).DataSet).GetPrec(ANode^.FField);
            end else begin
              ANode^.FDataLen := TSmartDBF((ANode^.FField).DataSet).GetLen(ANode^.FField);
              ANode^.FDataPrec := 0;
            end;
            if Result = Null then begin
              case ANode^.FDataType of
                ftString: Result := StringOfChar(FFC, ANode^.FDataSize);
                ftFloat, ftInteger, ftWord, ftCurrency, ftBCD, ftSmallint: Result := 0;
                ftBoolean: Result := false;
                //ftDateTime: Result := StringOfChar(FFC, 8);
              end;
            end else
              if ANode^.FDataType = ftString then
                if Length(Result) < ANode^.FDataSize then
                  Result := Result + StringOfChar(FFC, ANode^.FDataSize - Length(Result));
          end;
        end;
      enConst:
        Result := ANode^.FData;
      enOperator:
        case ANode^.FOperator of
        coNOTDEFINED:;
        coASSIGN:
          begin
            (ANode^.FRight^.FField).Value :=
                EvaluteNodeValueComplex(ANode^.FLeft);
            Result := (ANode^.FRight^.FField).Value;
          end;
        coOR:
            Result := Boolean(EvaluteNodeValueComplex(ANode^.FLeft)) or
                      Boolean(EvaluteNodeValueComplex(ANode^.FRight));
        coAND:
            Result := Boolean(EvaluteNodeValueComplex(ANode^.FLeft)) and
                      Boolean(EvaluteNodeValueComplex(ANode^.FRight));
        coNOT:
            Result := not Boolean(EvaluteNodeValueComplex(ANode^.FLeft));
        coEQ, coNE, coGE, coLE, coGT, coLT:
          begin
            l := EvaluteNodeValueComplex(ANode^.FLeft);
            r := EvaluteNodeValueComplex(ANode^.FRight);
            if (foCaseInsensitive in FOptions) and
               not (foNoPartialCompare in FOptions) then
                UpperCaseLR;
            case ANode^.FOperator of
                coEQ:
                  if foNoPartialCompare in FOptions then
                    Result := l = r
                  else
                    Result := PartialEQ(ANode^.FPartial);
                coNE: Result := l <> r;
                coGE: Result := l >= r;
                coLE: Result := l <= r;
                coGT: Result := l > r;
                coLT: Result := l < r;
            end;
          end;
        coLIKE:
            begin
                Result := LikeOperator( EvaluteNodeValueComplex(ANode^.FLeft), EvaluteNodeValueComplex(ANode^.FRight),
                                        foCaseInsensitive in FOptions, '%', '_');
            end;
        coISBLANK, coNOTBLANK:
            begin
                if ANode^.FLeft^.FKind = enField then
                    Result := ANode^.FLeft^.FField.IsNull
                else
                    //Result := StrIsNull(EvaluteNodeValue(ANode^.FLeft));
                    Result := VarIsNull(EvaluteNodeValueComplex(ANode^.FLeft));
                if ANode^.FOperator = coNOTBLANK then
                    Result := not Result;
            end;
        coIN:
            begin
                Result := False;
                l := EvaluteNodeValueComplex(ANode^.FLeft);
                for i := 0 to ANode^.FArgs.Count - 1 do
                begin
                  r := EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[i]));
                  if foNoPartialCompare in FOptions then
                    Result := l = r
                  else
                    Result := PartialEQ(ANode^.FPartial);
                  if Result then Break;
                end;
            end;
        coADD:
          begin
            Result := EvaluteNodeValueComplex(ANode^.FLeft) + EvaluteNodeValueComplex(ANode^.FRight);
            if FIndexKeyValue then begin
              if ANode^.FLeft.FDataLen > ANode^.FRight.FDataLen then
                ANode^.FDataLen := ANode^.FLeft.FDataLen + 1
              else
                ANode^.FDataLen := ANode^.FRight.FDataLen + 1;
              if ANode^.FLeft.FDataPrec > ANode^.FRight.FDataPrec then
                ANode^.FDataPrec := ANode^.FLeft.FDataPrec
              else
                ANode^.FDataPrec := ANode^.FRight.FDataPrec;
              if ANode^.FDataLen > 14 then ANode^.FDataLen := 14;
              if ANode^.FDataPrec + 1 > ANode^.FDataLen then ANode^.FDataPrec := ANode^.FDataLen - 3;
            end;
          end;
        coSUB:
          begin
            Result := EvaluteNodeValueComplex(ANode^.FLeft) - EvaluteNodeValueComplex(ANode^.FRight);
            if FIndexKeyValue then begin
              if ANode^.FLeft.FDataLen > ANode^.FRight.FDataLen then
                ANode^.FDataLen := ANode^.FLeft.FDataLen
              else
                ANode^.FDataLen := ANode^.FRight.FDataLen;
              if ANode^.FLeft.FDataPrec > ANode^.FRight.FDataPrec then
                ANode^.FDataPrec := ANode^.FLeft.FDataPrec
              else
                ANode^.FDataPrec := ANode^.FRight.FDataPrec;
              if ANode^.FDataLen > 14 then ANode^.FDataLen := 14;
              if ANode^.FDataPrec + 1 > ANode^.FDataLen then ANode^.FDataPrec := ANode^.FDataLen - 3;
            end;
          end;
        coMUL:
          begin
            Result := EvaluteNodeValueComplex(ANode^.FLeft) * EvaluteNodeValueComplex(ANode^.FRight);
            if FIndexKeyValue then begin
              ANode^.FDataLen := ANode^.FLeft.FDataLen + ANode^.FRight.FDataLen;
              ANode^.FDataPrec := ANode^.FLeft.FDataPrec + ANode^.FRight.FDataPrec;
              if ANode^.FDataLen > 14 then ANode^.FDataLen := 14;
              if ANode^.FDataPrec + 1 > ANode^.FDataLen then ANode^.FDataPrec := ANode^.FDataLen - 3;
            end;
          end;
        coDIV:
          begin
            Result := EvaluteNodeValueComplex(ANode^.FLeft) / EvaluteNodeValueComplex(ANode^.FRight);
            if FIndexKeyValue then begin
              if ANode^.FLeft.FDataLen > ANode^.FRight.FDataLen then
                ANode^.FDataLen := ANode^.FLeft.FDataLen
              else
                ANode^.FDataLen := ANode^.FRight.FDataLen;
              if ANode^.FLeft.FDataPrec > ANode^.FRight.FDataPrec then
                ANode^.FDataPrec := ANode^.FLeft.FDataPrec
              else
                ANode^.FDataPrec := ANode^.FRight.FDataPrec;
              if ANode^.FDataLen > 14 then ANode^.FDataLen := 14;
              if ANode^.FDataPrec + 1 > ANode^.FDataLen then ANode^.FDataPrec := ANode^.FDataLen - 3;
            end;
          end;
        end;
      enFunc:
        begin
          S := AnsiUpperCase(ANode^.FData);
          if (CompareText(S, 'UPPER') = 0) then
            Result := AnsiUpperCase(VarToStr(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))))
          else if (CompareText(S, 'LOWER') = 0) then
            Result := AnsiLowerCase(VarToStr(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))))
          else if (CompareText(S, 'DTOS') = 0) then begin
            Vr := EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]));
            if VarType(Vr) = varDate then begin
              if Vr <> Null then
                Result := DtoS(VarToDateTime(Vr))
              else
                Result := StringOfChar(FFC, 8);
            end else
              Result := StringOfChar(FFC, 8);
          end else if (CompareText(S, 'DTTOS') = 0) then begin
            Vr := EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]));
            if VarType(Vr) = varDate then begin
              if Vr <> Null then
                Result := DTtoS(VarToDateTime(Vr))
              else
                Result := StringOfChar(FFC, 14);
            end else
              Result := StringOfChar(FFC, 14);
          end else if (CompareText(S, 'STR') = 0) then begin
            V := PNLOExprNode(ANode^.FArgs.Items[0]);
            Vr := EvaluteNodeValueComplex(V);
            if Vr <> Null then begin
              S1 := '';
              case ANode^.FArgs.Count of
                1: Str(Vr:V.FDataLen:V.FDataPrec, S1);
                2:
                  begin
                    j := EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[1]));
                    Str(Vr:j:V.FDataPrec, S1);
                  end;
                3:
                  begin
                    j := EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[1]));
                    k := EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[2]));
                    Str(Vr:j:k, S1);
                  end;
              end;
            end else
              S1 := StringOfChar(FFC, V.FDataLen);
            Result := S1;
          end else if ((CompareText(S, 'SUBSTRING') = 0) or (CompareText(S, 'SUBSTR') = 0)) then
          begin
            S1 := VarToStr(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0])));
            Result := Copy(S1, Integer(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[1]))),
                                Integer(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[2]))));
          end else if (CompareText(S, 'TRIM') = 0) then
            Result := Trim(VarToStr(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))))
          else if ((CompareText(S, 'TRIMLEFT') = 0) or (CompareText(S, 'LTRIM') = 0)) then
            Result := TrimLeft(VarToStr(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))))
          else if ((CompareText(S, 'TRIMRIGHT') = 0) or (CompareText(S, 'RTRIM') = 0)) then
            Result := TrimRight(VarToStr(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))))
          else if (CompareText(S, 'YEAR') = 0) then
          begin
            DecodeDate(VarToDateTime(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))), Year, Month, Day);
            Result := Year;
          end else if (CompareText(S, 'MONTH') = 0) then
          begin
            DecodeDate(VarToDateTime(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))), Year, Month, Day);
            Result := Month;
          end else if (CompareText(S, 'DAY') = 0) then
          begin
            DecodeDate(VarToDateTime(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))), Year, Month, Day);
            Result := Day;
          end else if (CompareText(S, 'HOUR') = 0) then
          begin
            DecodeTime(VarToDateTime(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))), Hour, Min, Sec, MSec);
            Result := Hour;
          end else if (CompareText(S, 'MINUTE') = 0) then
          begin
            DecodeTime(VarToDateTime(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))), Hour, Min, Sec, MSec);
            Result := Min;
          end else if (CompareText(S, 'SECOND') = 0) then
          begin
            DecodeTime(VarToDateTime(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))), Hour, Min, Sec, MSec);
            Result := Sec;
          end else if (CompareText(S, 'GETDATE') = 0) then
            Result := StrToDate(VarToStr(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0]))))
          else if (CompareText(S, 'DATE') = 0) then
            Result := Integer(Trunc(VarToDateTime(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0])))))
          else if (CompareText(S, 'TIME') = 0) then
          begin
            dt := VarToDateTime(EvaluteNodeValueComplex(PNLOExprNode(ANode^.FArgs.Items[0])));
            Result := dt - Trunc(dt);
          end;
          {(CompareText(S, 'SUM') = 0)
          (CompareText(S, 'MIN') = 0)
          (CompareText(S, 'MAX') = 0)
          (CompareText(S, 'AVG') = 0)
          (CompareText(S, 'COUNT') = 0)}
        end;
      else
        Result := Null;
      end;
    end;
begin
    Result := EvaluteNodeValueComplex(Root);
end;

function To8ExprParser.Execute: Variant;
begin
  FValue := Null;
  if FLastRoot <> nil then
    FValue := Execute(FLastRoot);
  Result := FValue;
end;

function To8ExprParser.EvaluteKey: String;
var
  sign: boolean;
begin
  FValue := Null;
  if FLastRoot <> nil then
    FValue := Execute(FLastRoot);
  if (VarType(FValue) in [varDouble, varInteger, varSmallint, varSingle, varCurrency] ) then begin
    if FValue >= 0 then
      sign := false
    else begin
      sign := true;
      FValue := -FValue;
    end;
    Str(FValue:FLastRoot.FDataLen:FLastRoot.FDataPrec, Result);
    ReplBlanks(Result);
    if sign then
      ReplSign(Result);
  end else if (VarType(FValue) = varBoolean) then begin
    if FValue then
      Result := 'T'
    else
      Result := 'F';
  end else if (VarType(FValue) = varDate) then begin
    case FLastRoot.FDataType of
      ftDate: Result := DtoS(VarToDateTime(FValue));
      ftTime: Result := TtoS(VarToDateTime(FValue));
      ftDateTime: Result := DTtoS(VarToDateTime(FValue));
    else
      Result := DtoS(VarToDateTime(FValue))
    end;
  end else if ((VarType(FValue) = varEmpty) or (VarType(FValue) = varNull)) then begin
    Result := StringOfChar(FFC, FLastRoot.FDataLen);
  end else
    Result := VarToStr(FValue);
  FKey := Result;
end;

function To8ExprParser.SuiteFieldList(fl: String; out m: Integer): Integer;
var
  fs, fn: String;
  fc: Integer;
  q: boolean;

  procedure SuiteFieldListInternal(ANode: PNLOExprNode);
  var
    i: Integer;
  begin
    case ANode^.FKind of
    enField:
      begin
        Inc(m);
        if not q then begin
          fn := UpperCase((ANode^.FField).FieldName);
          if Pos(fn, fs) <> 0 then
            Inc(fc)
          else
            q := true;
        end;
      end;
    enOperator, enFunc:
      begin
        if ANode^.FLeft <> nil then
          SuiteFieldListInternal(ANode^.FLeft);
        if ANode^.FRight <> nil then
          SuiteFieldListInternal(ANode^.FRight);
        if ANode^.FArgs <> nil then
          for i := 0 to ANode^.FArgs.Count - 1 do
            if ANode^.FArgs.Items[i] <> nil then
              SuiteFieldListInternal(PNLOExprNode(ANode^.FArgs.Items[i]));
      end;
    end;
  end;

begin
  fs := UpperCase(fl);
  fc := 0;
  m := 0;
  q := false;
  if FLastRoot <> nil then SuiteFieldListInternal(FLastRoot);
  Result := fc;
end;

function To8ExprParser.EvaluteKey(const KeyFields: string;
  const KeyValues: Variant; const CF: Char = #$20): String;
begin
  if CF <> #$20 then FFC := CF;
  FFields := TList.Create;
  FKeyValues := KeyValues;
  FKeyFromValues := true;
  try
    FDataSet.GetFieldList(FFields, KeyFields);
    Result := EvaluteKey;
  finally
    FKeyFromValues := false;
    FFC := #$20;
    FFields.Free;
    FFields := nil;
  end;
end;

function To8ExprParser.GetDataLen: Integer;
begin
  Result := FLastRoot.FDataLen;
end;

function To8ExprParser.GetDataPrec: Integer;
begin
  Result := FLastRoot.FDataPrec;
end;

end.
