{Copyright:      Vlad Karpov  mailto:KarpovVV@protek.ru
 Author:         Vlad Karpov
 Remarks:        freeware, but this Copyright must be included

 Description:    To8DataSet - read onle DataSet component for ORACLE SQL

 Version:        1.0.0

 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}
unit o8RODataSet;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, o8Stmt, o8Env, o8Session, o8StmtOrd, OCI, Math, o8Filter, DBcommon;

type

  {$A-}
  TRecInfo = packed record
     RecordRowID: ub4;
     UpdateStatus: TUpdateStatus;
     BookmarkFlag: TBookmarkFlag;
  end;
  {$A+}
  pTRecInfo = ^TRecInfo;
  ppTRecInfo = ^pTRecInfo;

  pDouble = ^Double;
  pInteger = ^Integer;

  To8RODataSet = class;

  {TVKDL2}
  TVKDL2 = class(TDataLink)
  private
    Fo8DataSet: To8RODataSet;
  protected
    procedure ActiveChanged; override;
    procedure RecordChanged(Field: TField); override;
    procedure CheckBrowseMode; override;
  public
    property o8DataSet: To8RODataSet read Fo8DataSet write Fo8DataSet;
  end;

  {To8RODataSet}
  To8RODataSet = class(TDataSet)
  private
    { Private declarations }
    FSTMT: To8Stmt;
    FFilterParser: To8ExprParser;
    FStreamedActive: boolean;
    FCurrentIndex: sb4;
    FTempRecord: pChar;
    FFilterRecord: pChar;
    FLocateType: TLocateType;
    FFetchAll: boolean;
    FIsLookup: Boolean;
    FDataLink: TVKDL2;
    procedure SetFSTMT(oStatment: To8Stmt);
    function GetRecordBufferSize: Integer;
    property RecordBufferSize: Integer read GetRecordBufferSize;
    function GetActiveRecBuf(var RecBuf: PChar): Boolean;
    function GetRecordByBuffer(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
    function AcceptRecord: Boolean;
    procedure SetParamsFromCursor;
    procedure SetDataSource(const Value: TDataSource);
    procedure ReopenFromSrc;
  protected
    { Protected declarations }
    procedure SetActive(Value: boolean); override;
    procedure Loaded; override;

    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function GetRecordSize: Word; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;

    function GetRecordCount: Integer; override;
    function GetRecNo: Integer; override;

    function GetFldType(rs: pRStruct): TFieldType;

    function LocateRecord(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions; SyncCursor: Boolean): Boolean;
    function FindRecord(Restart, GoForward: Boolean): Boolean; override;

    function GetStateFieldValue(State: TDataSetState; Field: TField): Variant; override;
    function CompareLocateField(const Fields: TList; const KeyValues: Variant; Options: TLocateOptions): Integer;

    procedure SetFiltered(Value: Boolean); override;

    function GetDataSource: TDataSource; override;

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function Locate(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions): Boolean; override;
    function Lookup(const KeyFields: string; const KeyValues: Variant; const ResultFields: string): Variant; override;
    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    function GetAsRowID(Field: TField): pOCIRowid;
    function GetAsResultSet(Field: TField): pOCIResult;

    procedure FetchAll;

    property IsLookup: boolean read FIsLookup write FIsLookup;
    property CurrentIndex: sb4 read FCurrentIndex;

  published
    { Published declarations }
    property Active;
    property STMT: To8Stmt read FSTMT write SetFSTMT;
    property LocateType: TLocateType read FLocateType write FLocateType;

    property DataSource: TDataSource read GetDataSource write SetDataSource;

    property Filter;
    property Filtered;
    property FilterOptions;

    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;

  end;

procedure Register;

implementation

//******************************************************************************
procedure Register;
begin
  RegisterComponents('VK NLO', [To8RODataSet]);
end;

{ To8RODataSet }

function To8RODataSet.AcceptRecord: Boolean;
begin
  Result := true;
  if Assigned(OnFilterRecord) then
    OnFilterRecord(self, Result)
  else if Filter <> '' then
    Result := FFilterParser.Execute;
end;

function To8RODataSet.AllocRecordBuffer: PChar;
begin
  Result := AllocMem(RecordBufferSize + SizeOf(TRecInfo));
end;

function To8RODataSet.CompareLocateField(const Fields: TList;
  const KeyValues: Variant; Options: TLocateOptions): Integer;
var
  FieldCount: Integer;
  Field: TField;
  KeyVal: Variant;
  v1, v2: String;
  i1, i2: Integer;
  w1, w2: Word;
  b1, b2: boolean;
  f1, f2: double;
  i: Integer;
begin
  Result := 1;
  FieldCount := Fields.Count;
  if FieldCount = 1 then
  begin
    try
      Field := TField(Fields.First);
    except
      Field := nil;
    end;
    try
      if VarIsArray(KeyValues) then
         KeyVal := KeyValues[0]
      else
         KeyVal := KeyValues;
    except
      KeyVal := NULL;
    end;
    case Field.DataType of
      ftFixedChar, ftWideString, ftString:
      begin
        v1 := Field.AsString;
        if not VarIsNull(KeyVal) then v2 := KeyVal else v2 := '';
        if loCaseInsensitive in Options then
          Result := AnsiCompareText(v1, v2)
        else
          Result := AnsiCompareStr(v1, v2);
      end;
      ftLargeint, ftSmallint, ftInteger:
      begin
        i1 := Field.AsInteger;
        if not VarIsNull(KeyVal) then i2 := KeyVal else i2 := 0;
        Result := i1 - i2;
      end;
      ftWord:
      begin
        w1 := Field.AsInteger;
        if not VarIsNull(KeyVal) then w2 := KeyVal else w2 := 0;
        Result := w1 - w2;
      end;
      ftBoolean:
      begin
        b1 := Field.AsBoolean;
        if not VarIsNull(KeyVal) then b2 := KeyVal else b2 := false;
        if (not b1) and b2 then
          Result := -1;
        if b1 = b2 then
          Result := 0;
        if b1 and (not b2) then
          Result := 1;
      end;
      ftFloat, ftCurrency:
      begin
        f1 := Field.AsFloat;
        if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
        if f1 < f2 then
          Result := -1;
        if f1 = f2 then
          Result := 0;
        if f1 > f2 then
          Result := 1;
      end;
      ftDate, ftTime, ftDateTime:
      begin
        f1 := Field.AsDateTime;
        if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
        if f1 < f2 then
          Result := -1;
        if f1 = f2 then
          Result := 0;
        if f1 > f2 then
          Result := 1;
      end;
    end;
  end else begin
    for i := 0 to FieldCount - 1 do
    begin
      //
      try
        Field := TField(Fields.Items[i]);
      except
        Field := nil;
      end;
      try
        if VarIsArray(KeyValues) then
           KeyVal := KeyValues[i]
        else
           KeyVal := NULL;
      except
        KeyVal := NULL;
      end;
      case Field.DataType of
        ftFixedChar, ftWideString, ftString:
        begin
          v1 := Field.AsString;
          if not VarIsNull(KeyVal) then v2 := KeyVal else v2 := '';
          if loCaseInsensitive in Options then
            Result := AnsiCompareText(v1, v2)
          else
            Result := AnsiCompareStr(v1, v2);
        end;
        ftLargeint, ftSmallint, ftInteger:
        begin
          i1 := Field.AsInteger;
          if not VarIsNull(KeyVal) then i2 := KeyVal else i2 := 0;
          Result := i1 - i2;
        end;
        ftWord:
        begin
          w1 := Field.AsInteger;
          if not VarIsNull(KeyVal) then w2 := KeyVal else w2 := 0;
          Result := w1 - w2;
        end;
        ftBoolean:
        begin
          b1 := Field.AsBoolean;
          if not VarIsNull(KeyVal) then b2 := KeyVal else b2 := false;
          if (not b1) and b2 then
            Result := -1;
          if b1 = b2 then
            Result := 0;
          if b1 and (not b2) then
            Result := 1;
        end;
        ftFloat, ftCurrency:
        begin
          f1 := Field.AsFloat;
          if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
          if f1 < f2 then
            Result := -1;
          if f1 = f2 then
            Result := 0;
          if f1 > f2 then
            Result := 1;
        end;
        ftDate, ftTime, ftDateTime:
        begin
          f1 := Field.AsDateTime;
          if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
          if f1 < f2 then
            Result := -1;
          if f1 = f2 then
            Result := 0;
          if f1 > f2 then
            Result := 1;
        end;
      end;
      //
      if Result <> 0 then
        Exit;
    end;
  end;
end;

constructor To8RODataSet.Create(AOwner: TComponent);
var
  FieldMap: TFieldMap;
begin
  inherited Create(AOwner);
  FFilterParser := To8ExprParser.Create(self, '', [], [poExtSyntax], '', nil, FieldMap);
  FStreamedActive := false;
  FLocateType := ltLoop;
  FFetchAll := false;
  FIsLookup := false;
  FDataLink := TVKDL2.Create;
  FDataLink.o8DataSet := self;
end;

function To8RODataSet.CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
var
  j: Integer;
  rr: pRStruct;
  pInd:   sb2;
  ActiveBuf: PChar;
begin
  j := Field.FieldNo-1;
  rr := STMT.RawFieldStruct[j];
  if GetActiveRecBuf(ActiveBuf) then
  begin
    pInd  := sb2((psb2(ActiveBuf + STMT.RecordSize + j*sizeof(sb2)))^);
  end else begin
    ActiveBuf := nil;
    pInd  := -1;
  end;
  if (ActiveBuf <> nil) and (pInd <> -1) then
    Result := self.STMT.GetBlobStream(Pointer(ActiveBuf + rr.S_OFFSET), rr.S_INERNALTYPE)
  else
    Result := TOraEmptyStream.Create(Pointer(ActiveBuf + rr.S_OFFSET), self.STMT, rr.S_INERNALTYPE);
end;

destructor To8RODataSet.Destroy;
begin
  FDataLink.o8DataSet := nil;
  FDataLink.Destroy;
  FDataLink := nil;
  FFilterParser.Destroy;
  if STMT <> nil then STMT.RemoveDataSet(self);
  inherited Destroy;
end;

procedure To8RODataSet.FetchAll;
begin
  if not FFetchAll then
    STMT.FetchAll;
  FFetchAll := true;
end;

procedure To8RODataSet.FreeRecordBuffer(var Buffer: PChar);
begin
  FreeMem(Buffer);
end;

function To8RODataSet.GetActiveRecBuf(var RecBuf: PChar): Boolean;
begin
  if not FIsLookup then
    case State of
      dsBrowse: if IsEmpty then RecBuf := nil else RecBuf := ActiveBuffer;
      dsEdit, dsInsert: RecBuf := ActiveBuffer;
      dsFilter: RecBuf := FFilterRecord;
      dsNewValue, dsOldValue, dsCurValue: RecBuf := ActiveBuffer;
      dsCalcFields: RecBuf := CalcBuffer;
    else
      RecBuf := nil;
    end
  else begin
    InternalInitRecord(FTempRecord);
    pTRecInfo(FTempRecord + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).BookmarkFlag := bfCurrent;
    pTRecInfo(FTempRecord + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).RecordRowID := STMT.RecNo;
    pTRecInfo(FTempRecord + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).UpdateStatus := usUnmodified;
    Move(STMT.RawBufferGet^, FTempRecord^, STMT.RecordSize);
    Move(STMT.RawIndGet^   , (FTempRecord + STMT.RecordSize)^, STMT.FieldCount*sizeof(sb2));
    Move(STMT.RawRCodeGet^ , (FTempRecord + STMT.RecordSize + STMT.FieldCount*sizeof(sb2))^, STMT.FieldCount*sizeof(sb2));
    Move(STMT.RawRlenGet^  , (FTempRecord + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*2)^, STMT.FieldCount*sizeof(sb2));
    RecBuf := FTempRecord;
  end;
  Result := RecBuf <> nil;
end;

function To8RODataSet.GetAsResultSet(Field: TField): pOCIResult;
var
  j: Integer;
  rr: pRStruct;
  pInd:   sb2;
  ActiveBuf: PChar;
begin
  j := Field.FieldNo-1;
  rr := STMT.RawFieldStruct[j];
  if GetActiveRecBuf(ActiveBuf) then
  begin
    pInd  := sb2((psb2(ActiveBuf + STMT.RecordSize + j*sizeof(sb2)))^);
  end else begin
    ActiveBuf := nil;
    pInd  := -1;
  end;
  if (rr.S_INERNALTYPE = SQLT_RSET) and (ActiveBuf <> nil) and (pInd <> -1) then
    Result := ppOCIResult(ActiveBuf + rr.S_OFFSET)^
  else
    Result := nil;
end;

function To8RODataSet.GetAsRowID(Field: TField): pOCIRowid;
var
  j: Integer;
  rr: pRStruct;
  pInd:   sb2;
  ActiveBuf: PChar;
begin
  j := Field.FieldNo-1;
  rr := STMT.RawFieldStruct[j];
  if GetActiveRecBuf(ActiveBuf) then
  begin
    pInd  := sb2((psb2(ActiveBuf + STMT.RecordSize + j*sizeof(sb2)))^);
  end else begin
    ActiveBuf := nil;
    pInd  := -1;
  end;
  if (rr.S_INERNALTYPE = SQLT_RDD) and (ActiveBuf <> nil) and (pInd <> -1) then
    Result := ppOCIRowid(ActiveBuf + rr.S_OFFSET)^
  else
    Result := nil;
end;

procedure To8RODataSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  ub4(Data^) := pTRecInfo(Buffer + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).RecordRowID;
end;

function To8RODataSet.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := pTRecInfo(Buffer + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).BookmarkFlag;
end;

function To8RODataSet.GetFieldData(Field: TField;
  Buffer: Pointer): Boolean;
var
    rr: pRStruct;
    pInd:   sb2;
    RLen:   ub2;
    ActiveBuf: PChar;
    j: Integer;
    oDate: pORA_Date;
    oDateTime: TDateTime;
    wYear: Integer;
    dd: double;
    sTS: TTimeStamp;
    c: PChar;
    oDS: TDataSet;
    LookupResult: Variant;
begin
    Result := false;
    case Field.FieldKind of
    fkData:
      begin
        j := Field.FieldNo-1;
        rr := STMT.RawFieldStruct[j];
        if GetActiveRecBuf(ActiveBuf) then
        begin
          pInd  := sb2((psb2(ActiveBuf + STMT.RecordSize + j*sizeof(sb2)))^);
          RLen  := ub2((pub2(ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(ub2)*2 + j*sizeof(ub2))^));
        end else begin
          ActiveBuf := nil;
          pInd  := -1;
          RLen  := 0;
        end;
        if Buffer <> nil then
        begin
          if ActiveBuf <> nil then
          begin
            case Field.DataType of
              ftString, ftInteger, ftFloat, ftBytes, ftVarBytes:
                begin
                  case pInd of
                    0:
                      begin
                        c := ActiveBuf + rr.S_OFFSET;
                        if RLen <> 0 then Move(c^, Buffer^, RLen) else FillChar(Buffer^, rr.S_LENGTH, 0);
                        if  Field.DataType in [ftString]  then pChar(Buffer)[RLen] := Char(0);
                        Result := true;
                      end;
                    -1: Result := false;
                  end;
                end;
              ftBlob,ftMemo:
                begin
                   if rr.S_INERNALTYPE in [SQLT_BLOB, SQLT_CLOB, SQLT_CFILE, SQLT_BFILE] then
                   begin
                    if pInd = -1 then
                      Result := false
                    else begin
                      c := ActiveBuf + rr.S_OFFSET;
                      if RLen <> 0 then Move(c^, Buffer^, RLen) else FillChar(Buffer^, rr.S_LENGTH, 0);
                      Result := true;
                    end;
                   end;
                   if rr.S_INERNALTYPE in [SQLT_LNG, SQLT_LBI] then
                   begin
                    if pInd = -1 then
                      Result := false
                    else begin
                      c := ActiveBuf + rr.S_OFFSET;
                      if RLen <> 0 then Move(c^, Buffer^, RLen) else FillChar(Buffer^, rr.S_LENGTH, 0);
                      pChar(Buffer)[RLen] := Char(0);
                      Result := true;
                    end;
                   end;
                end;
              ftCursor:
                begin
                  if pInd = -1 then
                    Result := false
                  else begin
                    c := ActiveBuf + rr.S_OFFSET;
                    if RLen <> 0 then Move(c^, Buffer^, RLen) else FillChar(Buffer^, rr.S_LENGTH, 0);
                    pChar(Buffer)[RLen] := Char(0);
                    Result := true;
                  end;
                end;
              ftDateTime:
                begin
                  case pInd of
                    0:
                      begin
                        oDate := pORA_Date(ActiveBuf + rr.S_OFFSET);
                        try
                          wYear := (oDate.cCentury-100)*100+(oDate.cYear-100);
                          if (wYear > 0) and ( wYear < 9999) and ( oDate.cMonth >= 1) and ( oDate.cMonth <= 12) then
                             oDateTime := EncodeDate(wYear, oDate.cMonth, oDate.cDay) +
                                      EncodeTime(oDate.cHour-1, oDate.cMinute-1, oDate.cSecond-1, 0)
                          else
                             oDateTime := 0;
                          sTS := DateTimeToTimeStamp(oDateTime);
                          dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                          double(Buffer^) := dd;
                          Result := true;
                        except
                          Result := false;
                        end;
                      end;
                    -1: Result := false;
                  end;
                end;
            end;
          end else
            Result := false;
        end else begin
          if pInd = -1 then
            Result := false
          else
            Result := true;
        end;
      end;
    fkCalculated:
      begin
        if GetActiveRecBuf(ActiveBuf) then
        begin
          pInd  := sb2(psb2(ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3 + sizeof(TRecInfo) + ub4(CalcFieldsSize) + Field.Offset)^);
        end else begin
          pInd  := -1;
        end;
        if Buffer <> nil then
        begin
          if (ActiveBuf <> nil) and not (csDesigning in ComponentState) then
          begin
            case pInd of
              0:
                begin
                  c := ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3 + sizeof(TRecInfo) + Field.Offset;
                  Move(c^, Buffer^, Field.DataSize);
                  if Field.DataType in [ftString, ftBytes, ftVarBytes]  then pChar(Buffer)[Field.DataSize] := Char(0);
                  Result := true;
                end;
              -1: Result := false;
            end;
          end else begin
            FillChar(Buffer^, Field.DataSize, 0);
            Result := false;
          end;
        end else
          Result := false;
      end;
    fkLookup:
      begin
        Result := false;
        if Buffer <> nil then
        begin
          if GetActiveRecBuf(ActiveBuf) then
          begin
            oDS := Field.LookupDataSet;
            if (oDS <> nil) and oDS.Active then
            begin
              LookupResult := oDS.Lookup(Field.LookupKeyFields, FieldValues[Field.KeyFields], Field.LookupResultField);
              if (not VarIsEmpty(LookupResult)) and (not VarIsNull(LookupResult)) then
              begin
                if Field.DataType = ftString then
                begin
                  c := TVarData(LookupResult).VPointer;
                  Move(c^, Buffer^, Length(c) + 1);
                end else if Field.DataType = ftDateTime then
                begin
                  sTS := DateTimeToTimeStamp(LookupResult);
                  dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                  double(Buffer^) := dd;
                end else begin
                  c := @TVarData(LookupResult).VAny;
                  Move(c^, Buffer^, Field.DataSize);
                end;
                Result := true;
              end;
            end;
          end else begin
            FillChar(Buffer^, Field.DataSize, 0);
            Result := false;
          end;
        end;
      end;
    end;
end;

function To8RODataSet.GetFldType(rs: pRStruct): TFieldType;
begin
  case rs.S_EXTERNALTYPE of
    SQLT_CHR:   Result := ftString;
    SQLT_AFC:   Result := ftString;
    SQLT_INT:   Result := ftInteger;
    SQLT_FLT:   Result := ftFloat;
    SQLT_RDD:   Result := ftBytes;
    SQLT_DAT:   Result := ftDateTime;
    SQLT_BIN:   Result := ftBytes;
    SQLT_LNG:   Result := ftMemo;
    SQLT_LBI:   Result := ftBlob;
    SQLT_CLOB:  Result := ftMemo;
    SQLT_BLOB:  Result := ftBlob;
    SQLT_CFILE: Result := ftMemo;
    SQLT_BFILE: Result := ftBlob;
    SQLT_RSET:  Result := ftDataSet;
    SQLT_NTY:   Result := ftADT;
  else
    Result := ftUnknown;
  end;
end;

function To8RODataSet.GetRecNo: Integer;
var
  ActiveBuf: pChar;
begin
  if GetActiveRecBuf(ActiveBuf) then
    Result := pTRecInfo(ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).RecordRowID
  else
    Result := -1;
end;

function To8RODataSet.FindRecord(Restart, GoForward: Boolean): Boolean;
var
  SaveState: TDataSetState;
  Accept: Boolean;
  Ret: TGetResult;
begin
  if (not Filtered) and (Filter <> '') then
      FFilterParser.SetExprParams(Filter, FilterOptions, [poExtSyntax], '');
  CheckBrowseMode;
  DoBeforeScroll;
  SetFound(False);
  UpdateCursorPos;
  CursorPosChanged;
  if GoForward then
  begin
    if Restart then InternalFirst;
    repeat
      Ret := GetRecordByBuffer(FFilterRecord, gmNext, false);
      SaveState := SetTempState(dsFilter);
      Accept := AcceptRecord;
      RestoreState(SaveState);
    until Accept or (Ret <> grOK);
  end else
  begin
    if Restart then InternalLast;
    repeat
      Ret := GetRecordByBuffer(FFilterRecord, gmPrior, false);
      SaveState := SetTempState(dsFilter);
      Accept := AcceptRecord;
      RestoreState(SaveState);
    until Accept or (Ret <> grOK);
  end;
  if Ret = grOK then
  begin
    Resync([rmExact, rmCenter]);
    SetFound(True);
  end else
    InternalSetToRecord(ActiveBuffer);
  Result := Found;
  if Result then DoAfterScroll;
end;

function To8RODataSet.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var
  SaveState: TDataSetState;
  Accept: Boolean;
begin
  if not Filtered then
    Result := GetRecordByBuffer(Buffer, GetMode, DoCheck)
  else begin
    if GetMode <> gmCurrent then
    begin
      repeat
        Result := GetRecordByBuffer(FFilterRecord, GetMode, DoCheck);
        SaveState := SetTempState(dsFilter);
        Accept := AcceptRecord;
        RestoreState(SaveState);
      until Accept or (Result <> grOK);
      if Accept then
        Move(FFilterRecord^, Buffer^, RecordBufferSize)
      else
        Move(FTempRecord^, Buffer^, RecordBufferSize);
    end else
      Result := GetRecordByBuffer(Buffer, GetMode, DoCheck);
  end;
end;

function To8RODataSet.GetRecordBufferSize: Integer;
begin
  Result := STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3 + sizeof(TRecInfo) + ub4(CalcFieldsSize)*2;
end;

function To8RODataSet.GetRecordByBuffer(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
var
  cc: PChar;
begin
  Result := grOK;
  case GetMode of
    gmCurrent:
      begin
        if FCurrentIndex <> -1 then
        begin
          if FCurrentIndex < sb4(STMT.FetchRowsCount) then
          begin
            cc := pChar(Buffer);
            InternalInitRecord(cc);
            pTRecInfo(cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).BookmarkFlag := bfCurrent;
            pTRecInfo(cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).RecordRowID := STMT.RecNo;
            pTRecInfo(cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).UpdateStatus := usUnmodified;
            Move(STMT.RawBufferGet^, cc^, STMT.RecordSize);
            Move(STMT.RawIndGet^   , (cc + STMT.RecordSize)^, STMT.FieldCount*sizeof(sb2));
            Move(STMT.RawRCodeGet^ , (cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2))^, STMT.FieldCount*sizeof(sb2));
            Move(STMT.RawRlenGet^  , (cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*2)^, STMT.FieldCount*sizeof(sb2));
          end else begin
            InternalInitRecord(FTempRecord);
            Move(FTempRecord^, Buffer^, RecordBufferSize);
            Result := grError;
          end;
        end else begin
          InternalInitRecord(FTempRecord);
          Move(FTempRecord^, Buffer^, RecordBufferSize);
          Result := grError;
        end;
      end;
    gmNext:
      begin
        if FCurrentIndex < sb4(STMT.FetchRowsCount) then
        begin
          if FCurrentIndex <> -1 then STMT.Next;
          Inc(FCurrentIndex);
          if not STMT.Eof then
          begin
            cc := pChar(Buffer);
            InternalInitRecord(cc);
            pTRecInfo(cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).BookmarkFlag := bfCurrent;
            pTRecInfo(cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).RecordRowID := STMT.RecNo;
            pTRecInfo(cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).UpdateStatus := usUnmodified;
            Move(STMT.RawBufferGet^, cc^, STMT.RecordSize);
            Move(STMT.RawIndGet^   , (cc + STMT.RecordSize)^, STMT.FieldCount*sizeof(sb2));
            Move(STMT.RawRCodeGet^ , (cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2))^, STMT.FieldCount*sizeof(sb2));
            Move(STMT.RawRlenGet^  , (cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*2)^, STMT.FieldCount*sizeof(sb2));
          end else begin
            Dec(FCurrentIndex);
            InternalInitRecord(FTempRecord);
            Move(FTempRecord^, Buffer^, RecordBufferSize);
            Result := grEOF;
          end;
        end else begin
          FCurrentIndex := STMT.FetchRowsCount;
          InternalInitRecord(FTempRecord);
          Move(FTempRecord^, Buffer^, RecordBufferSize);
          Result := grEOF;
        end;
      end;
    gmPrior:
      begin
        if ( ub4(FCurrentIndex) <> STMT.FetchRowsCount ) and ( not STMT.Bof ) then
          STMT.Prior;
        Dec(FCurrentIndex);
        if not STMT.Bof then
        begin
          cc := pChar(Buffer);
          InternalInitRecord(cc);
          pTRecInfo(cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).BookmarkFlag := bfCurrent;
          pTRecInfo(cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).RecordRowID := STMT.RecNo;
          pTRecInfo(cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).UpdateStatus := usUnmodified;
          Move(STMT.RawBufferGet^, cc^, STMT.RecordSize);
          Move(STMT.RawIndGet^   , (cc + STMT.RecordSize)^, STMT.FieldCount*sizeof(sb2));
          Move(STMT.RawRCodeGet^ , (cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2))^, STMT.FieldCount*sizeof(sb2));
          Move(STMT.RawRlenGet^  , (cc + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*2)^, STMT.FieldCount*sizeof(sb2));
        end else begin
          FCurrentIndex := -1;
          InternalInitRecord(FTempRecord);
          Move(FTempRecord^, Buffer^, RecordBufferSize);
          Result := grBOF;
        end;
      end;
  end;
  if Result = grOK then
    GetCalcFields(Buffer);
end;

function To8RODataSet.GetRecordCount: Integer;
begin
  Result := STMT.FetchRowsCount;
end;

function To8RODataSet.GetRecordSize: Word;
begin
  Result := Word(STMT.RecordSize);
end;

function To8RODataSet.GetStateFieldValue(State: TDataSetState;
  Field: TField): Variant;
begin
  Result := NULL;
  if State in [dsNewValue, dsCurValue, dsOldValue] then
    Result := inherited GetStateFieldValue(State, Field);
end;

procedure To8RODataSet.InternalAddRecord(Buffer: Pointer; Append: Boolean);
begin
  //
end;

procedure To8RODataSet.InternalClose;
begin
  FreeRecordBuffer(FFilterRecord);
  FreeRecordBuffer(FTempRecord);
  BindFields(false);
  if DefaultFields then DestroyFields;
end;

procedure To8RODataSet.InternalDelete;
begin
  //
end;

procedure To8RODataSet.InternalFirst;
begin
  FCurrentIndex := -1;
  STMT.First;
end;

procedure To8RODataSet.InternalGotoBookmark(Bookmark: Pointer);
begin
  STMT.RecNo := pTRecInfo(Bookmark).RecordRowID;
  FCurrentIndex := STMT.RecNo - 1;
end;

procedure To8RODataSet.InternalHandleException;
begin
  Application.HandleException(self);
end;

procedure To8RODataSet.InternalInitFieldDefs;
var
   i: Integer;
   RS: pRStruct;
begin
  FieldDefs.BeginUpdate;
  FieldDefs.Clear;
  for i := 0 to STMT.FieldCount -1 do
  begin
    with FieldDefs.AddFieldDef do begin
      Name := STMT.FieldName[i+1];
      RS := STMT.RawFieldStruct[i];
      DataType := GetFldType(RS);
      case DataType of
        ftString, ftBytes:
          Size := Word(RS.S_LENGTH + 1);
      else
        Size := 0;
      end;
      Precision := STMT.RawFieldStruct[i].S_PREC;
      if RS.S_NULLOK = 1 then
        Required := false
      else
        Required := true;
    end;
  end;
  FieldDefs.EndUpdate;
end;

procedure To8RODataSet.InternalInitRecord(Buffer: PChar);
begin
  if Buffer <> nil then FillChar(Buffer^, RecordBufferSize, -1);
end;

procedure To8RODataSet.InternalLast;
begin
  FetchAll;
  STMT.MoveToFast(STMT.FetchRowsCount);
  FCurrentIndex := STMT.RecNo;
end;

procedure To8RODataSet.InternalOpen;
begin
  FCurrentIndex := -1;
  FFetchAll := false;
  ObjectView := true;
  InternalInitFieldDefs();
  if DefaultFields then CreateFields;
  BindFields(true);
  BookmarkSize := sizeof(ub4);
  FTempRecord := AllocRecordBuffer;
  FFilterRecord := AllocRecordBuffer;
  STMT.First;
  if Filtered and (Filter <> '') then
    FFilterParser.SetExprParams(Filter, FilterOptions, [poExtSyntax], '');
end;

procedure To8RODataSet.InternalPost;
var
  ActiveBuf: pChar;
  OldR: ub4;
begin
  CheckActive;
  if State = dsEdit then
  begin
    OldR := STMT.RecNo;
    ActiveBuf := ActiveBuffer;
    STMT.RecNo := pTRecInfo(ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).RecordRowID;
    Move(ActiveBuf^, (STMT.RawBufferGet)^, STMT.RecordSize);
    Move((ActiveBuf + STMT.RecordSize)^, STMT.RawIndGet^   , STMT.FieldCount*sizeof(sb2));
    Move((ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(sb2))^, STMT.RawRCodeGet^ , STMT.FieldCount*sizeof(sb2));
    Move((ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*2)^, STMT.RawRlenGet^, STMT.FieldCount*sizeof(sb2));
    STMT.Post;
    STMT.RecNo := OldR;
  end
end;

procedure To8RODataSet.InternalSetToRecord(Buffer: PChar);
begin
  STMT.RecNo := pTRecInfo(Buffer + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).RecordRowID;
  FCurrentIndex := STMT.RecNo - 1;
end;

function To8RODataSet.IsCursorOpen: Boolean;
begin
  if Assigned(STMT) then Result := STMT.Active else Result := false;
end;

procedure To8RODataSet.Loaded;
begin
  inherited Loaded();
  if FStreamedActive then Active := true;
end;

function To8RODataSet.Locate(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions): Boolean;
begin
  DoBeforeScroll;
  FIsLookup := true;
  o8Env.o8CrSection.Acquire;
  try
    Result := LocateRecord(KeyFields, KeyValues, Options, True);
  finally
    FIsLookup := false;
    o8Env.o8CrSection.Release;
  end;
  if Result then
  begin
    Resync([]);
    DoAfterScroll;
  end;
end;

function To8RODataSet.LocateRecord(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions; SyncCursor: Boolean): Boolean;
var
  Fields: TList;
  i: Integer;
  bm: String;
  LowR, HighR, l: sb4;
begin
  Result := false;
  if not VarIsArray(KeyValues) then
  begin
    if VarIsNull(KeyValues) or VarIsEmpty(KeyValues) then
    begin
      Result := false;
      Exit;
    end;
  end else
    for i := 0 to VarArrayHighBound(KeyValues, 1) do
      if VarIsNull(KeyValues[i]) or VarIsEmpty(KeyValues[i]) then
      begin
        Result := false;
        Exit;
      end;
  if FLocateType = ltLoop then
  begin
    DisableControls;
    if not FFetchAll then FetchAll;
    if State <> dsBrowse then Post;
    bm := Bookmark;
    Result := false;
    Fields := TList.Create;
    try
      GetFieldList(Fields, KeyFields);
      FCurrentIndex := 0;
      STMT.MoveToFast(FCurrentIndex + 1);
      while ub4(FCurrentIndex) < STMT.FetchRowsCount do
      begin
        l := CompareLocateField(Fields, KeyValues, Options);
        if l = 0 then
        begin
          Result := true;
          Break;
        end;
        Inc(FCurrentIndex);
        STMT.MoveToFast(FCurrentIndex + 1);
      end;
      if not Result then Bookmark := bm;
    finally
      Fields.Free;
      CursorPosChanged;
      EnableControls;
    end;
  end else if FLocateType = ltOrdered then begin
    DisableControls;
    if not FFetchAll then FetchAll;
    if State <> dsBrowse then Post;
    bm := Bookmark;
    Result := false;
    Fields := TList.Create;
    try
      GetFieldList(Fields, KeyFields);
      //
      LowR := 0;
      HighR := STMT.FetchRowsCount;
      while LowR < HighR do
      begin
        FCurrentIndex := LowR + Trunc((HighR - LowR)/2);
        STMT.MoveToFast(FCurrentIndex + 1);
        l := CompareLocateField(Fields, KeyValues, Options);
        if l < 0 then
        begin
          if LowR <> FCurrentIndex then
          begin
            LowR := FCurrentIndex;
            continue;
          end else
            Break;
        end;
        if l = 0 then
        begin
          Result := true;
          Break;
        end;
        if l > 0 then
        begin
          if HighR <> FCurrentIndex then
          begin
            HighR := FCurrentIndex;
            continue;
          end else
            Break;
        end;
      end;
      //
      if not Result then Bookmark := bm;
    finally
      Fields.Free;
      CursorPosChanged;
      EnableControls;
    end;
  end else if FLocateType = ltOrderedDesc then begin
    DisableControls;
    if not FFetchAll then FetchAll;
    if State <> dsBrowse then Post;
    bm := Bookmark;
    Result := false;
    Fields := TList.Create;
    try
      GetFieldList(Fields, KeyFields);
      //
      LowR := 0;
      HighR := STMT.FetchRowsCount;
      while LowR < HighR do
      begin
        FCurrentIndex := LowR + Trunc((HighR - LowR)/2);
        STMT.MoveToFast(FCurrentIndex + 1);
        l := CompareLocateField(Fields, KeyValues, Options);
        if l < 0 then
        begin
          if HighR <> FCurrentIndex then
          begin
            HighR := FCurrentIndex;
            continue;
          end else
            Break;
        end;
        if l = 0 then
        begin
          Result := true;
          Break;
        end;
        if l > 0 then
        begin
          if  LowR <> FCurrentIndex then
          begin
            LowR := FCurrentIndex;
            continue;
          end else
            Break;
        end;
      end;
      //
      if not Result then Bookmark := bm;
    finally
      Fields.Free;
      CursorPosChanged;
      EnableControls;
    end;
  end;
end;

function To8RODataSet.Lookup(const KeyFields: string;
  const KeyValues: Variant; const ResultFields: string): Variant;
begin
  Result := Null;
  FIsLookup := true;
  o8Env.o8CrSection.Acquire;
  try
    if LocateRecord(KeyFields, KeyValues, [], false) then
    begin
      SetTempState(dsCalcFields);
      try
        Result := FieldValues[ResultFields];
      finally
        RestoreState(dsBrowse);
      end;
    end;
  finally
    FIsLookup := false;
    o8Env.o8CrSection.Release;
  end;
end;

procedure To8RODataSet.SetActive(Value: boolean);
begin
  if csReading in ComponentState then
    FStreamedActive := Value
  else
    if FSTMT <> nil then begin
      if FSTMT.Active then
      begin
        inherited SetActive(Value);
      end else inherited SetActive(false);
    end else
      inherited SetActive(false);
end;

procedure To8RODataSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  pTRecInfo(Buffer + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).RecordRowID := ub4(Data^);
end;

procedure To8RODataSet.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  pTRecInfo(Buffer + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3).BookmarkFlag := Value;
end;

procedure To8RODataSet.SetFieldData(Field: TField; Buffer: Pointer);
var
    rr: pRStruct;
    pInd:   psb2;
    RLen:   pub2;
    ActiveBuf: PChar;
    j: Integer;
    oDate: pORA_Date;
    OraDate: ORA_Date;
    c: PChar;
    dd: double;
    sTS: TTimeStamp;
begin
  case Field.FieldKind of
  fkData:
    begin
      j := Field.FieldNo-1;
      rr := STMT.RawFieldStruct[j];
      GetActiveRecBuf(ActiveBuf);
      if (Buffer <> nil) and (ActiveBuf <> nil) then
      begin
        pInd  := psb2((psb2(ActiveBuf + STMT.RecordSize + j*sizeof(sb2))));
        RLen  := pub2((pub2(ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(ub2)*2 + j*sizeof(ub2))));
        Field.Validate(Buffer);
        case Field.DataType of
          ftString, ftInteger, ftFloat, ftBytes, ftVarBytes:
            begin
              c := ActiveBuf + rr.S_OFFSET;
              Move(Buffer^, c^, rr.S_LENGTH);
              pInd^ := 0;
              if  Field.DataType in [ftString, ftBytes, ftVarBytes] then
              begin
                RLen^ := Length(pChar(Buffer));
                if RLen^ > rr.S_LENGTH then RLen^ := rr.S_LENGTH;
              end else
                RLen^ := rr.S_LENGTH;
              if  Field.DataType in [ftString, ftBytes, ftVarBytes]  then pChar(Buffer)[RLen^] := Char(0);
              SetModified(true);
            end;
          ftDateTime:
            begin
              oDate := pORA_Date(ActiveBuf + rr.S_OFFSET);
              dd := double(Buffer^);
              sTS.Date := Trunc(dd/(3600.0*24*1000));
              sTS.Time := Trunc(dd - sTS.Date*((3600.0*24*1000)));
              OraDate := DateTimeToOraDate(TimeStampToDateTime(sTS));
              Move(OraDate, oDate^, rr.S_LENGTH);
              RLen^ := rr.S_LENGTH;
              pInd^ := 0;
              SetModified(true);
            end;
        end;
      end;
    end;
  fkCalculated:
    begin
      GetActiveRecBuf(ActiveBuf);
      if (Buffer <> nil) and (ActiveBuf <> nil) then
      begin
        pInd  := psb2(ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3 + sizeof(TRecInfo) + ub4(CalcFieldsSize) + Field.Offset);
        c := ActiveBuf + STMT.RecordSize + STMT.FieldCount*sizeof(sb2)*3 + sizeof(TRecInfo) + Field.Offset;
        Move(Buffer^, c^, Field.DataSize);
        pInd^ := 0;
      end;
    end;
  end;
  if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
    DataEvent(deFieldChange, Longint(Field));
end;

procedure To8RODataSet.SetFiltered(Value: Boolean);
begin
  if Active then
  begin
    CheckBrowseMode;
    if ((not Filtered) and Value) and (Filter <> '') then
      FFilterParser.SetExprParams(Filter, FilterOptions, [poExtSyntax], '');
    if Filtered <> Value then
      inherited SetFiltered(Value);
    First;
  end else
    inherited SetFiltered(Value);
end;

procedure To8RODataSet.SetFSTMT(oStatment: To8Stmt);
begin
  if FSTMT <> nil then FSTMT.RemoveDataSet(self);
  Active := false;
  if oStatment <> nil then
  begin
    FSTMT := oStatment;
    FSTMT.AddDataSet(self);
  end else FSTMT := nil;
end;

function To8RODataSet.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure To8RODataSet.ReopenFromSrc;
var
  o8: To8StmtOrd;
begin
  o8 := nil;
  if FSTMT is To8StmtOrd then
    o8 := To8StmtOrd(FSTMT);
  if Active then begin
    SetParamsFromCursor;
    if o8 <> nil then o8.ReExecute;
    First;
  end else begin
    if o8 <> nil then begin
      o8.Active := false;
      SetParamsFromCursor;
      o8.Active := true;
    end;
    Active := true;
  end;
end;

procedure To8RODataSet.SetDataSource(const Value: TDataSource);
begin
  if IsLinkedTo(Value) then DatabaseError('SCircularDataLink', Self);
  FDataLink.DataSource := Value;
end;

procedure To8RODataSet.SetParamsFromCursor;
var
  i: Integer;
  o8: To8StmtOrd;
  DataSet: TDataSet;
begin
  if FDataLink.DataSource <> nil then
  begin
    DataSet := FDataLink.DataSource.DataSet;
    if DataSet <> nil then
    begin
      if FSTMT is To8StmtOrd then begin
        o8 := To8StmtOrd(FSTMT);
        DataSet.FieldDefs.Update;
        for i := 0 to o8.Params.Count - 1 do
          o8.Params[i].ParamValue := DataSet.FieldByName(o8.Params[i].Name).AsVariant;
      end;
    end;
  end;
end;

{ TVKDL2 }

procedure TVKDL2.ActiveChanged;
begin
  o8DataSet.ReopenFromSrc
end;

procedure TVKDL2.CheckBrowseMode;
begin
  if o8DataSet.Active then o8DataSet.CheckBrowseMode;
end;

procedure TVKDL2.RecordChanged(Field: TField);
begin
  if (Field = nil) and o8DataSet.Active then o8DataSet.ReopenFromSrc;
end;

end.
