{Copyright:      Vlad Karpov  mailto:KarpovVV@protek.ru
 Author:         Vlad Karpov
 Remarks:        freeware, but this Copyright must be included

 Description:    To8Session component for manage sessions handlers

 Version:        1.0.0

 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}
unit o8Session;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  {bde, dbtables,} OCI, o8Env;

type

  To8Session = class(TComponent)
  private
    { Private declarations }

    FEnv: To8Env;
    //FBDEDataBase: TDatabase;

    envhp: pOCIEnv; { Environment handle }

    srvhp: pOCIServer; { the server handle }
    usrhp: pOCISession; {user session handle}
    svchp: pOCISvcCtx; {server context handle}
    errhp: pOCIError; { the error handle }
    AllocHandlerServ: boolean;
    AllocHandlerSession: boolean;
    AllocHandlerSerC: boolean;
    AllocHandlerErr: boolean;

    sUser: String;
    sPass: String;
    sServ: String;

    Factive: boolean;
    FStreamedActive: boolean;

    errcode: sb4;
    buf: array [1..MaxErrorMessage] of char;

    FNonBlocking: Boolean;

    StatmentList: TList;

    function GetActive: boolean;
    procedure SetActive(bActive: boolean);

    procedure SetoEnv(Env: To8Env);
    //procedure SetBDEDataBase(const Value: TDatabase);

    function GetErrMessage: String;
    function GetErrCode: sb4;

    procedure FreeHandlers;
    procedure AllocHandlers;

    function GetEnvhp: pOCIEnv;
    function GetSrvhp: pOCIServer;
    procedure SetNonBlocking(const Value: Boolean);
    procedure SetInterNonBlocking;
    procedure SetPass(const Value: String);
    procedure SetServ(const Value: String);
    procedure SetUser(const Value: String);
    function GetServerVersion: String;


  protected
    { Protected declarations }

    procedure Loaded; override;
    procedure GetError;

  public
    { Public declarations }

    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    property hEnv: pOCIEnv read GetEnvhp;
    property hServ: pOCIServer read GetSrvhp;
    property hSess: pOCISession read usrhp;
    property hSvc: pOCISvcCtx read svchp;
    property hErr: pOCIError read errhp;

    property ErrMes: String read GetErrMessage;
    property ErrNum: sb4 read GetErrCode;

    property ServerVersion: String read GetServerVersion;

    procedure AddStatment(oOBJ: TObject);
    procedure RemoveStatment(oOBJ: TObject);

    procedure TransStart;
    procedure TransCommit;
    procedure TransRollback;

    //function TransPrepare: sword;
    //procedure TransDetach;
    //procedure TransForget;

    procedure BreakNonBlockingExecution;

  published
    { Published declarations }

    property Environment: To8Env read FEnv write SetoEnv;
    //property BDEDataBase: TDatabase read FBDEDataBase write SetBDEDataBase;
    property User: String read  sUser write SetUser;
    property Password: String read  sPass write SetPass;
    property Server: String read sServ write SetServ;
    property Active: boolean read GetActive write  SetActive;
    property NonBlocking: Boolean read FNonBlocking write SetNonBlocking;

  end;

procedure Register;

implementation

uses
   o8Stmt, o8Script;

//******************************************************************************
procedure Register;
begin
  RegisterComponents('VK NLO', [To8Session]);
end;

//******************************************************************************
constructor To8Session.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  StatmentList := TList.Create;
  FEnv := nil;
  envhp := nil;
  srvhp := nil;
  usrhp := nil;
  svchp := nil;
  errhp := nil;
  sUser := '';
  sPass := '';
  Active := false;
  FStreamedActive := false;
  AllocHandlerServ := false;
  AllocHandlerSession := false;
  AllocHandlerSerC := false;
  AllocHandlerErr  := false;
  errcode := 0;
  fillchar(buf, MaxErrorMessage, 0);
end;

//******************************************************************************
destructor To8Session.Destroy;
var
   oStmt: To8Stmt;
   oScr: To8Script;
begin
   Active := false;
   while StatmentList.Count > 0 do
   begin
    if TObject(StatmentList.Items[0]).InheritsFrom(To8Stmt) then
    begin
      oStmt := To8Stmt(StatmentList.Items[0]);
      oStmt.oraSession := nil;
    end;
   end;
   while StatmentList.Count > 0 do
   begin
    if TObject(StatmentList.Items[0]).InheritsFrom(To8Script) then
    begin
      oScr := To8Script(StatmentList.Items[0]);
      oScr.oraSession := nil;
    end;
   end;
   StatmentList.Destroy;
   if FEnv <> nil then FEnv.RemoveSession(self);
   FreeHandlers;
   inherited Destroy;
end;

//******************************************************************************
procedure To8Session.AllocHandlers;
begin
  FreeHandlers;
  if hEnv <> nil then
  begin
    if ( OCIHandleAlloc(    hEnv,
                            @srvhp,
                            OCI_HTYPE_SERVER,
                            0,
                            nil) = OCI_SUCCESS ) then AllocHandlerServ := true;
    if ( OCIHandleAlloc(    hEnv,
                            @svchp,
                            OCI_HTYPE_SVCCTX,
                            0,
                            nil) = OCI_SUCCESS ) then AllocHandlerSerC := true;
    if ( OCIHandleAlloc(    hEnv,
                            @usrhp,
                            OCI_HTYPE_SESSION,
                            0,
                            nil) = OCI_SUCCESS ) then AllocHandlerSession := true;
    if ( OCIHandleAlloc(    hEnv,
                            @errhp,
                            OCI_HTYPE_ERROR,
                            0,
                            nil) = OCI_SUCCESS ) then AllocHandlerErr := true;
  end;
end;

//******************************************************************************
procedure To8Session.FreeHandlers;
begin
  //if FEnv <> nil then
  //begin
    if AllocHandlerErr then begin
      OCIHandleFree(errhp, OCI_HTYPE_ERROR);
      errhp := nil;
      AllocHandlerErr := false;
    end;
    if AllocHandlerSession then begin
      OCIHandleFree(usrhp, OCI_HTYPE_SESSION);
      usrhp := nil;
      AllocHandlerSession := false;
    end;
    if AllocHandlerSerC then begin
      OCIHandleFree(svchp, OCI_HTYPE_SVCCTX);
      svchp := nil;
      AllocHandlerSerC := false;
    end;
    if AllocHandlerServ then begin
      OCIHandleFree(srvhp, OCI_HTYPE_SERVER);
      srvhp := nil;
      AllocHandlerServ := false;
    end;
  //end;
end;

//******************************************************************************
function To8Session.GetActive: boolean;
begin
   Result := FActive;
end;

//******************************************************************************
procedure To8Session.SetActive(bActive: boolean);
var
  i: Integer;
  oStmt: To8Stmt;
  //ErrCode: sb4;
  //ErrMes: String;
  //iLen: Word;
begin
  if ( csReading in ComponentState ) then FStreamedActive := bActive
  else
  begin
    if Environment <> nil then
    begin
      if Environment.Active then
      begin
        if bActive then
        begin
          if not FActive then
          begin
            AllocHandlers;
            {$B+}
            if AllocHandlerSession and AllocHandlerSerC and AllocHandlerErr  and AllocHandlerServ then
            {$B-}
            begin
              if OCIServerAttach(srvhp, errhp, PChar(sServ), Length(sServ), OCI_DEFAULT) = OCI_SUCCESS then
              begin
                {set the server context in the service context}
                OCIAttrSet(svchp, OCI_HTYPE_SVCCTX, hServ, ub4(0), OCI_ATTR_SERVER, errhp);
                {set username attribute in user session handle}
                OCIAttrSet(usrhp, OCI_HTYPE_SESSION, PChar(User), ub4(Length(User)), OCI_ATTR_USERNAME, errhp);
                {set password attribute in user session handle}
                OCIAttrSet(usrhp, OCI_HTYPE_SESSION, PChar(Password), ub4(Length(Password)), OCI_ATTR_PASSWORD, errhp);
                if OCISessionBegin(svchp, errhp, usrhp, OCI_CRED_RDBMS, OCI_DEFAULT) = OCI_SUCCESS then
                begin
                  {set the user session in the service context}
                  if OCIAttrSet(svchp, OCI_HTYPE_SVCCTX, usrhp, ub4(0), OCI_ATTR_SESSION, errhp) = OCI_SUCCESS then
                  begin
                    FActive := true;
                    SetInterNonBlocking;
                  end else if csDesigning in ComponentState then ShowMessage(ErrMes);
                end else if csDesigning in ComponentState then ShowMessage(ErrMes);
              end else if csDesigning in ComponentState then ShowMessage(ErrMes);
            end else if csDesigning in ComponentState then ShowMessage(ErrMes);
          end;
        end else begin
          for i := 0 to StatmentList.Count - 1 do
            if TObject(StatmentList.Items[i]).InheritsFrom(To8Stmt) then
            begin
              oStmt := To8Stmt(StatmentList.Items[i]);
              oStmt.Active := false;
            end;
          OCISessionEnd(svchp, errhp, usrhp, OCI_DEFAULT);
          OCIServerDetach(srvhp, errhp, OCI_DEFAULT);
          FActive := false;
          FreeHandlers;
        end;
      end else begin
        GetError;
        FActive := false;
        for i := 0 to StatmentList.Count - 1 do
          if TObject(StatmentList.Items[i]).InheritsFrom(To8Stmt) then
          begin
            oStmt := To8Stmt(StatmentList.Items[i]);
            oStmt.Active := false;
          end;
      end;
    end else begin
    {
      if (FBDEDataBase <> nil) and  FBDEDataBase.Connected then
      begin
        if  bActive then
        begin
          if not FActive then
          begin

            raise Exception.Create('Capabilities not supported!');

            if envhp <> nil then OCIHandleFree(envhp, OCI_HTYPE_ENV);
            envhp := nil;
            OCIEnvInit(@envhp, OCI_DEFAULT, 0, nil);

            AllocHandlers;

            iLen := 0;
            Check(DbiGetProp(hDBIObj(FBDEDataBase.Handle), dbNATIVEHNDL, svchp, 66, iLen));
            OCIAttrGet(svchp, OCI_HTYPE_SVCCTX, usrhp, nil, OCI_ATTR_SESSION, errhp);
            OCIAttrGet(svchp, OCI_HTYPE_SVCCTX, srvhp, nil, OCI_ATTR_SERVER, errhp);

            //To8Env.GetError(errhp, true, ErrCode, ErrMes);

            FActive := true;

          end;
        end else begin
          for i := 0 to StatmentList.Count - 1 do
            if TObject(StatmentList.Items[i]).InheritsFrom(To8Stmt) then
            begin
              oStmt := To8Stmt(StatmentList.Items[i]);
              oStmt.Active := false;
            end;
          FActive := false;
        end;
      end else begin
        GetError;
        FActive := false;
        for i := 0 to StatmentList.Count - 1 do
        if TObject(StatmentList.Items[i]).InheritsFrom(To8Stmt) then
        begin
          oStmt := To8Stmt(StatmentList.Items[i]);
          oStmt.Active := false;
        end;
      end;
    }
    end;
  end;
end;

//******************************************************************************
procedure To8Session.SetoEnv(Env: To8Env);
begin
   Active := false;
   if FEnv <> nil then FEnv.RemoveSession(self);
   FEnv := Env;
   if FEnv <> nil then FEnv.AddSession(self);
end;

//******************************************************************************
procedure To8Session.Loaded;
begin
    inherited Loaded;
    if FStreamedActive then Active := true;
end;

//******************************************************************************
procedure To8Session.GetError;
begin
  errcode := 0;
  fillchar(buf, MaxErrorMessage, 0);
  {$B+}
  if (FEnv <> nil) and AllocHandlerSession and AllocHandlerSerC and AllocHandlerErr and AllocHandlerServ then
  {$B-}
  begin
    if Environment.Active then
    begin
       OCIErrorGet   (   errhp,
                         1,
                         nil,
                         @errcode,
                         @buf[1],
                         MaxErrorMessage,
                         OCI_HTYPE_ERROR);
    end
    else
    begin
      errcode := 2000006;
      move('  Environment.', buf, 23);
    end;
  end
  else
  begin
    if not AllocHandlerErr then
    begin
      errcode := 2000001;
      move('  Error .', buf, 33);
    end;
    if not AllocHandlerSerC then
    begin
      errcode := 2000002;
      move('  Server  .', buf, 43);
    end;
    if not AllocHandlerSession then
    begin
      errcode := 2000003;
      move('  Session .', buf, 35);
    end;
    if FEnv = nil then
    begin
      errcode := 2000004;
      move('   Environment.', buf, 38);
    end;
    if not AllocHandlerServ then
    begin
      errcode := 2000004;
      move('  Server handler.', buf, 34);
    end;
  end;
end;

//******************************************************************************
function To8Session.GetErrMessage: String;
begin
   GetError;
   if errcode <> 0 then Result := Trim(String(buf)) else Result := '';
   //if errcode <> 0 then Result := IntToStr(errcode) else Result := 'Session errcode = 0';
end;

//******************************************************************************
function To8Session.GetErrCode: sb4;
begin
   GetError;
   Result := errcode;
end;

//******************************************************************************
procedure To8Session.AddStatment(oOBJ: TObject);
begin
    if Assigned(StatmentList) then
    begin
      StatmentList.Add(oOBJ);
    end;
end;

//******************************************************************************
procedure To8Session.RemoveStatment(oOBJ: TObject);
begin
   if Assigned(StatmentList) then
   begin
      StatmentList.Remove(oOBJ);
      StatmentList.Pack;
   end;
end;

//procedure To8Session.Commit;
//begin
//  if OCITransPrepare( svchp,
//                      errhp,
//                      OCI_DEFAULT) = OCI_SUCCESS then
//    OCITransCommit( svchp,
//                    errhp,
//                    OCI_DEFAULT);
//end;

procedure To8Session.TransCommit;
begin
  OCITransCommit( svchp,
                  errhp,
                  OCI_DEFAULT);
end;

//procedure To8Session.TransDetach;
//begin
//  OCITransDetach( svchp,
//                  errhp,
//                  OCI_DEFAULT);
//end;

//procedure To8Session.TransForget;
//begin
//  OCITransForget( svchp,
//                  errhp,
//                  OCI_DEFAULT);
//end;

//function To8Session.TransPrepare: sword;
//begin
//  Result := OCITransPrepare(  svchp,
//                              errhp,
//                              OCI_DEFAULT);
//end;

procedure To8Session.TransRollback;
begin
  OCITransRollback  ( svchp,
                      errhp,
                      OCI_DEFAULT);
end;

procedure To8Session.TransStart;
begin
  OCITransStart(  svchp,
                  errhp,
                  0,
                  OCI_TRANS_NEW);
end;

//procedure To8Session.SetBDEDataBase(const Value: TDatabase);
//begin
//  Active := false;
//  FBDEDataBase := Value;
//end;

function To8Session.GetEnvhp: pOCIEnv;
begin
  if (FEnv <> nil) and FEnv.Active then
    Result := FEnv.hEnv
  else
    Result := envhp;
end;

function To8Session.GetSrvhp: pOCIServer;
begin
  Result := srvhp;
end;

procedure To8Session.SetNonBlocking(const Value: Boolean);
begin
  FNonBlocking := Value;
  if Active then
    SetInterNonBlocking;
end;

procedure To8Session.SetInterNonBlocking;
var
   SetResult: sword;
   NONBLOCKING_MODE: ub1;
   ErrC: sb4;
   ErrM: String;
begin
  NONBLOCKING_MODE := 200;
  SetResult := OCIAttrGet(hSvc, OCI_HTYPE_SVCCTX, @NONBLOCKING_MODE, nil, OCI_ATTR_NONBLOCKING_MODE, errhp);
  if FNonBlocking and (NONBLOCKING_MODE = 0) then
    SetResult := OCIAttrSet(hSvc, OCI_HTYPE_SVCCTX, nil, 0, OCI_ATTR_NONBLOCKING_MODE, errhp)
  else if (not FNonBlocking) and (NONBLOCKING_MODE <> 0) then
    SetResult := OCIAttrSet(hSvc, OCI_HTYPE_SVCCTX, nil, 0, OCI_ATTR_NONBLOCKING_MODE, errhp);
  if not (SetResult in [OCI_SUCCESS, OCI_SUCCESS_WITH_INFO]) then
    To8Env.GetError(errhp, true, ErrC, ErrM);
end;

procedure To8Session.SetPass(const Value: String);
begin
  Active := false;
  sPass := Value;
end;

procedure To8Session.SetServ(const Value: String);
begin
  Active := false;
  sServ := Value;
end;

procedure To8Session.SetUser(const Value: String);
begin
  Active := false;
  sUser := Value;
end;

procedure To8Session.BreakNonBlockingExecution;
var
   BreakResult: sword;
   ErrC: sb4;
   ErrM: String;
begin
  BreakResult := OCIBreak(svchp, errhp);
  if not (BreakResult in [OCI_SUCCESS, OCI_SUCCESS_WITH_INFO]) then
    To8Env.GetError(errhp, true, ErrC, ErrM);
end;

function To8Session.GetServerVersion: String;
const
  OCI_MAX_ATTR_LEN = 1024;
var
  buff: array[0 .. OCI_MAX_ATTR_LEN-1] of char;
begin
  if Active then
  begin
    OCIServerVersion(srvhp, errhp, @buff, OCI_MAX_ATTR_LEN, OCI_HTYPE_SERVER);
    Result := buff;
  end else
    Result := '';
end;

end.
