{Copyright:      Vlad Karpov  mailto:KarpovVV@protek.ru
 Author:         Vlad Karpov
 Remarks:        freeware, but this Copyright must be included

 Description:    To8Stmt - base class for execute SQL mechanism

 Version:        1.0.0

 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}
unit o8Stmt;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  OCI, o8Env, o8Session, db, MemMgr, contnrs;

const

  MAX_COLUMN = 255;
  MAX_NAME = 256;
  DefMaxLengthStringBuffer = 256;

type

  pInteger = ^Integer;
  pDouble = ^double;

  TResultSetStream = (rsUndef, rsToFile, rsToMemory);

{$A-}
//******************************************************************************
   RStruct = packed record
      S_POS: sword;
      S_DBSIZE: sb4;
      S_INERNALTYPE: sb2;
      S_COLNAME: array [0..MAX_NAME-1] of char;
      S_COLNAMELENGTH: sb4;
      S_DISPLAYSIZE: sb4;
      S_PREC: sb2;
      S_SCALE: sb2;
      S_NULLOK: sb2;
      S_EXTERNALTYPE: sb2;
      S_OFFSET: ub4;
      S_LENGTH: ub4;
      S_TYPE_NAME: array [0..MAX_NAME-1] of char;
      S_TYPENAMELENGTH: sb4;
      S_TDO: pOCIType;
   end;

   pRStruct = ^RStruct;
   ppRStruct = ^pRStruct;

//******************************************************************************
   InS = packed record
      S_OFFSET: sb4;
      S_LENGTH: sb4;
   end;

   pInS = ^InS;
   ppInS = ^pInS;

//******************************************************************************
   SelectListItem = packed record
      uValue: Pointer;
      uValueType: sb2;
      uValueLength: sb4;
      uValueDBsize: sb4;
      uValueInd: psb2;
      uValueLen: pub2;
      uValueRCode: pub2;
   end;

   pSelectListItem = ^SelectListItem;
   ppSelectListItem = ^pSelectListItem;

//******************************************************************************
   ORA_Date = packed record
      cCentury: ub1;
      cYear: ub1;
      cMonth: ub1;
      cDay: ub1;
      cHour: ub1;
      cMinute: ub1;
      cSecond: ub1;
   end;

   pORA_Date = ^ORA_Date;
   ppORA_Date = ^pORA_Date;

{$A+}

  TOraAfterExecuteEvent = procedure(Sender: TObject; var sExec: sword) of object;

  TLocateType = (ltLoop, ltOrdered, ltOrderedDesc);

  TOraLobStream = class;
  TOraParam = class;
  To8Stmt = class;

  TOraException = class(Exception)
  private

    FErrorCode: sb4;

  public

    constructor Create(nErr: sb4; sMess: String);

    property ErrorCode: sb4 read FErrorCode write FErrorCode;

  end;

  TRawParamType = (ptUnknown, ptString, ptInteger, ptFloat, ptDateTime, ptCLob, ptBLob, ptCFile, ptBFile, ptCursor, ptRowID);

  TOraParams = class(TOwnedCollection)
  private
    function GetCollectionOwner: TPersistent;
    function GetItem(Index: Integer): TOraParam;
    procedure SetItem(Index: Integer; const Value: TOraParam);
  public
    procedure AssignValues(Value: TOraParams);
    function FindParam(const Value: string): TOraParam;
    function IsEqual(Value: TOraParams): Boolean;
    function CreateParam(DataType: TRawParamType; ParamName: String; IsNull: boolean; ParamValue: variant; MaxStrLen :Integer = DefMaxLengthStringBuffer): TOraParam;
    procedure DescriptorsAlloc;
    property Owner: TPersistent read GetCollectionOwner;
    function ParseSQL(SQL: String; DoCreate: Boolean): String;
    property Items[Index: Integer]: TOraParam read GetItem write SetItem; default;
  end;

  TOraParam = class(TCollectionItem)
  private

    BufInt:                   Integer;
    BufFloat:                 double;
    BufDate:                  ORA_Date;
    BufStr:                   pChar;
    BufLob:                   pOCILobLocator;
    BufFile:                  pOCILobLocator;
    BufCursor:                pOCIResult;
    BufRowId:                 pOCIRowid;
    SetedRowId:               pOCIRowid;

    FOraParams:               TOraParams;

    envhp: pOCIEnv; { the environment handle }
    srvhp: pOCIServer; { the server handle }
    errhp: pOCIError; { the error handle }
    usrhp: pOCISession; {user session handle}
    svchp: pOCISvcCtx; {server context handle}

    FParName:                 String;
    FDataType:                TRawParamType;
    FExternalType:            ub2;
    FValue:                   Variant;
    FMaxLengthStringBuffer:   Integer;
    FParInd:                  sb2;
    FParRealLen:              ub2;
    FParRCode:                ub2;
    FParOffset: Integer;

    function GetIsNull: boolean;
    function GetValue: Variant;
    procedure SetIsNull(const Value: boolean);
    procedure SetMaxLengthStringBuffer(const Value: Integer);
    procedure SetDataType(const Value: TRawParamType);
    function GetBuffer: Pointer;
    procedure SetValue(const Value: Variant);
    function GetAsDateTime: TDateTime;
    function GetAsFloat: Double;
    function GetAsInteger: Integer;
    function GetAsString: String;
    procedure SetAsDateTime(const Value: TDateTime);
    procedure SetAsFloat(const Value: Double);
    procedure SetAsInteger(const Value: Integer);
    procedure SetAsString(const Value: String);
    function GetLobStream: TOraLobStream;
    function GetParamSize: Integer;
    function GetParBindpp: ppOCIBind;
    procedure SetBufRowId(const Value: pOCIRowid);
    function GetBufRowId: pOCIRowid;

  protected

    FParBindp:                pOCIBind;

    procedure AssignParam(Param: TOraParam);
    function GetDisplayName: string; override;

  public

    MaxLength:                sb4;

    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;

    procedure DescriptorAlloc;

    procedure Assign(Source: TPersistent); override;
    function IsEqual(Value: TOraParam): Boolean;
    procedure AssignFieldValue(Field: TField; const Value: Variant);

    property ExternalType: ub2 read FExternalType;
    property ParInd: sb2 read FParInd write FParInd;
    property ParRealLen: ub2 read FParRealLen write FParRealLen;
    property ParRCode: ub2 read FParRCode write FParRCode;
    property Buffer: Pointer read GetBuffer;
    property AsString: String read GetAsString write SetAsString;
    property AsInteger: Integer read GetAsInteger write SetAsInteger;
    property AsFloat: Double read GetAsFloat write SetAsFloat;
    property AsDateTime: TDateTime read GetAsDateTime write SetAsDateTime;
    property LobStream: TOraLobStream read GetLobStream;

    property hEnv: pOCIEnv read envhp;
    property hSess: pOCISession read usrhp;
    property hServ: pOCIServer read srvhp;
    property hSvc: pOCISvcCtx read svchp;
    property hErr: pOCIError read errhp;

    property ParBindp: pOCIBind read FParBindp;
    property ParBindpp: ppOCIBind read GetParBindpp;

    property LobLob: pOCIBlobLocator read BufLob;
    property LobFile: pOCIBlobLocator read BufFile;
    property Cursor: pOCIResult read BufCursor;
    property RowId: pOCIRowid read GetBufRowId write SetBufRowId;

    property ParSize: Integer read GetParamSize;
    property ParOffset: Integer read FParOffset write FParOffset;

  published

    property Name: String read FParName write FParName;
    property DataType: TRawParamType read FDataType write SetDataType;
    property MaxLengthStringBuffer: Integer read FMaxLengthStringBuffer write SetMaxLengthStringBuffer;
    property IsNull: boolean read GetIsNull write SetIsNull;
    property ParamValue: Variant read GetValue write SetValue;

  end;

  TOraField = class
  private

    FOraStmt: To8Stmt;
    FIndex: Integer;

    envhp: pOCIEnv; { the environment handle }
    srvhp: pOCIServer; { the server handle }
    errhp: pOCIError; { the error handle }
    usrhp: pOCISession; {user session handle}
    svchp: pOCISvcCtx; {server context handle}

    Fpsli: pSelectListItem;
    FValue: Variant;
    FFieldName: String;
    function GetSelectListItem: pSelectListItem;
    function GetAsDateTime: TDateTime;
    function GetAsFloat: double;
    function GetAsInteger: Integer;
    function GetAsString: String;
    function GetIsNull: boolean;
    function GetAsVariant: Variant;
    function GetLobStream: TOraLobStream;
    function GetResultSet: To8Stmt;
    function GetCursor: pOCIResult;
    function GetOraFieldType: Integer;
    function GetRowId: pOCIRowid;
    function GetComplexObject: pOCIComplexObjectComp;
    procedure SetAsDateTime(const Value: TDateTime);
    procedure SetAsFloat(const Value: double);
    procedure SetAsInteger(const Value: Integer);
    procedure SetAsString(const Value: String);
    procedure SetAsVariant(const Value: Variant);
    procedure SetIsNull(const Value: boolean);

  public

    constructor Create(OraStmt: To8Stmt; Index: Integer);

    property SelectListItem: pSelectListItem read GetSelectListItem;
    property OraStmt: To8Stmt read FOraStmt;

    property Value: Variant read GetAsVariant write SetAsVariant;
    property AsString: String read GetAsString write SetAsString;
    property AsInteger: Integer read GetAsInteger write SetAsInteger;
    property AsFloat: double read GetAsFloat write SetAsFloat;
    property AsDateTime: TDateTime read GetAsDateTime write SetAsDateTime;
    property IsNull: boolean read GetIsNull write SetIsNull;
    property LobStream: TOraLobStream read GetLobStream;
    property ResultSet: To8Stmt read GetResultSet;

    property hEnv: pOCIEnv read envhp;
    property hSess: pOCISession read usrhp;
    property hServ: pOCIServer read srvhp;
    property hSvc: pOCISvcCtx read svchp;
    property hErr: pOCIError read errhp;

    property Cursor: pOCIResult read GetCursor;
    property RowId: pOCIRowid read GetRowId;
    property ComplexObject: pOCIComplexObjectComp read GetComplexObject;

    property FieldName: String read FFieldName;
    property OraFieldType: Integer read GetOraFieldType;

  end;

  TOraLobStream = class(TStream)
  private

    envhp: pOCIEnv; { the environment handle }
    srvhp: pOCIServer; { the server handle }
    errhp: pOCIError; { the error handle }
    usrhp: pOCISession; {user session handle}
    svchp: pOCISvcCtx; {server context handle}

    LongField: Boolean;
    pLob: pOCILobLocator;
    FField: TOraField;
    FParam: TOraParam;
    Fo8Stmt: To8Stmt;

    FPosition: ub4;
    procedure SetNewSize(const Value: ub4);

  protected

    function GetSize: ub4; virtual;
    procedure SetSize(NewSize: Longint); override;

  public

    constructor Create(oField: TOraField); overload;
    constructor Create(oPar: TOraParam); overload;
    constructor Create(ParamLob: Pointer; oStmt: To8Stmt; b_type: sb2); overload;
    destructor Destroy; override;

    function Read(var Buffer; Count: Longint): Longint; override;
    function Write(const Buffer; Count: Longint): Longint; override;
    function Seek(Offset: Longint; Origin: Word): Longint; override;

    property Size: ub4 read GetSize write SetNewSize;

    property hEnv: pOCIEnv read envhp;
    property hSess: pOCISession read usrhp;
    property hServ: pOCIServer read srvhp;
    property hSvc: pOCISvcCtx read svchp;
    property hErr: pOCIError read errhp;

  end;

  TOraEmptyStream = class(TOraLobStream)
  protected
    function GetSize: ub4; override;
  end;

  To8Stmt = class(TComponent)
  private
    { Private declarations }
    FParams: TOraParams;
    FSession: To8Session;
    FActive: boolean;
    FExec: boolean;
    FStreamedActive: boolean;
    FStreamedExec: boolean;
    DataSetList: TList;
    FArrayOfStructureLength: Integer;
    FOnBeforeExecute: TNotifyEvent;
    FOnAfterExecute: TOraAfterExecuteEvent;
    FCursorParam: TOraParam;
    FCursorIndex: Integer;
    FCursorSource: To8Stmt;
    FNestedStmtClass: TClass;
    FDataSetFieldName: String;
    FDataSetFieldNameStreamed: String;
    FDataSetSource: To8Stmt;
    FDataSetField: TOraField;
    FStreamedArrayOfPar: Boolean;
    FStreamedResultSetStream: TResultSetStream;
    FState: TDataSetState;
    FResultSet: pOCIResult;
    function GetActive: boolean;
    procedure SetActive(bActive: boolean);
    function GetSession: To8Session;
    procedure SetSession(oSession: To8Session);
    function GetFieldName(Index: Integer): String;
    function GetParamByName(sIndex: String): TOraParam;
    function GetParamByNum(iIndex: Integer): TOraParam;
    procedure ReadParamData(Reader: TReader);
    procedure WriteParamData(Writer: TWriter);
    procedure SetParamList(const Value: TOraParams);
    function GetOraFieldByNum(Index: Integer): TOraField;
    function GetOraFieldByName(Index: String): TOraField;
    procedure SetCursorIndex(const Value: Integer);
    procedure SetCursorSource(const Value: To8Stmt);
    procedure SetExec(const Value: boolean);
    procedure SetDataSetFieldName(const Value: String);
    procedure SetDataSetSource(const Value: To8Stmt);
    function GetDataSetFieldName: String;
    function GetDataSetField: TOraField;
    procedure SetState(const Value: TDataSetState);
    function GetParamNumByName(sIndex: String): Integer;
  protected
    { Protected declarations }
    stmthp: pOCIStmt;
    errhp: pOCIError;
    AllocHandlerStmt: boolean;

    AllocHandlerErr: boolean;
    errcode: sb4;
    buf: array [1..MaxErrorMessage] of char;
    FFieldCount: ub4;
    Feof: boolean;
    Fbof: boolean;
    FRecNo: ub4;
    FieldDescribe: TList;
    FRecordSize: ub2;
    SLI: SelectListItem;
    PrepareStatus: boolean;
    FDescriptors: TList;

    FArrayOfPar: Boolean;
    FParamBuffer: Pointer;
    FParamIndBuffer: psb2;
    FParamLenBuffer: pub2;
    FParamRCodeBuffer: pub2;
    FParRecordSize: ub2;
    FLastParamRow: ub4;

    FOraFields: TObjectList;

    procedure FreeHandlers;
    procedure AllocHandlers;
    
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Loaded; override;
    procedure InternalExecute; virtual;
    procedure DescriptorsAlloc; virtual; abstract;
    procedure DescriptorsFree; virtual; abstract;
    function GetRawBuffer: PChar; virtual; abstract;
    function GetRawInd: psb2; virtual; abstract;
    function GetRawRCode: pub2; virtual; abstract;
    function GetRawRlen: pub2; virtual; abstract;
    procedure FlashCurrentRecord; virtual; abstract;
    function GetFetchRowsCount: ub4; virtual; abstract;

    function GetErrMessage: String;
    function GetErrCode: sb4;
    procedure GetError(bError: boolean = false);

    procedure SetArrayOfPar(const Value: Boolean); virtual;
    function GetArrayParValue(Index1, Index2: Integer): Variant; virtual;
    procedure SetArrayParValue(Index1, Index2: Integer; const Value: Variant); virtual;
    function GetArrayParString(Index1, Index2: Integer): String; virtual;
    procedure SetArrayParString(Index1, Index2: Integer; const Value: String); virtual;
    function GetArrayParInteger(Index1, Index2: Integer): Integer; virtual;
    procedure SetArrayParInteger(Index1, Index2: Integer; const Value: Integer); virtual;
    function GetArrayParFloat(Index1, Index2: Integer): double; virtual;
    procedure SetArrayParFloat(Index1, Index2: Integer; const Value: double); virtual;
    function GetArrayParDateTime(Index1, Index2: Integer): TDateTime; virtual;
    procedure SetArrayParDateTime(Index1, Index2: Integer; const Value: TDateTime); virtual;
    function GetArrayParNull(Index1, Index2: Integer): Boolean; virtual;
    procedure SetArrayParNull(Index1, Index2: Integer; const Value: Boolean); virtual;
    function GetArrayParRCode(Index1, Index2: Integer): ub2; virtual;
    function GetResultSetStream: TResultSetStream; virtual;
    procedure SetResultSetStream(const Value: TResultSetStream); virtual;

    property RecNoInternal: ub4 read FRecno write FRecno;
    property EofInternal: boolean read Feof write Feof;
    property BofInternal: boolean read Fbof write Fbof;

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    constructor CreateFromStmtField(AOwner: TComponent; oField: TOraField);
    destructor Destroy; override;

    procedure DefineProperties(Filer: TFiler); override;

    procedure Open; virtual;
    procedure Close; virtual;
    procedure Prepare; virtual;
    procedure ExecSql; virtual;
    procedure FetchAll; virtual; abstract;
    procedure MoveTo(NewRec: ub4); virtual; abstract;
    procedure MoveToFast(NewRec: ub4); virtual; abstract;
    procedure MoveBy(SkipRec: sb4); virtual; abstract;
    procedure First; virtual; abstract;
    procedure Next; virtual; abstract;
    procedure Prior; virtual; abstract;
    procedure Last; virtual; abstract;
    procedure BreakExecut; virtual;
    procedure Edit; virtual;
    procedure Post; virtual;
    function RawFieldGet(position: sword): pSelectListItem; virtual; abstract;
    function RawFieldStructGet(position: sword): pRStruct; virtual; abstract;
    function GetBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; overload; virtual; abstract;
    function GetBlobStream(ParamLob: pOCILobLocator; b_type: sb2): TStream; overload; virtual; abstract;
    function Locate(sFieldName: String; vFind: Variant): boolean;
    function FieldByName(sFieldName: String): TOraField;
    function FieldByNum(FieldNum: Integer): TOraField;

    procedure AddDataSet(oOBJ: TObject);
    procedure RemoveDataSet(oOBJ: TObject);

    property hStmt: pOCIStmt read stmthp;
    property hErr: pOCIError read errhp;
    property ErrMes: String read GetErrMessage;
    property ErrNum: sb4 read GetErrCode;

    property Params: TOraParams read FParams write SetParamList stored False;
    property ParamByName[sIndex: String]: TOraParam read GetParamByName;
    property ParamByNum[iIndex: Integer]: TOraParam read GetParamByNum;
    property ParamNum[sIndex: String]: Integer read GetParamNumByName;
    property ArrayOfPar: Boolean read FArrayOfPar write SetArrayOfPar;
    property ArrayParValue[Index1, Index2: Integer]: Variant read GetArrayParValue write SetArrayParValue;
    property ArrayParAsString[Index1, Index2: Integer]: String read GetArrayParString write SetArrayParString;
    property ArrayParAsInteger[Index1, Index2: Integer]: Integer read GetArrayParInteger write SetArrayParInteger;
    property ArrayParAsFloat[Index1, Index2: Integer]: double read GetArrayParFloat write SetArrayParFloat;
    property ArrayParAsDateTime[Index1, Index2: Integer]: TDateTime read GetArrayParDateTime write SetArrayParDateTime;
    property ArrayParNull[Index1, Index2: Integer]: Boolean read GetArrayParNull write SetArrayParNull;
    property ArrayParRCode[Index1, Index2: Integer]: ub2 read GetArrayParRCode;
    property LastParamRow: ub4 read FLastParamRow write FLastParamRow;

    property FieldCount: ub4 read FFieldCount;
    property RecordSize: ub2 read FRecordSize;
    property ParamRecordSize: ub2 read FParRecordSize;

    property Eof: boolean read Feof;
    property Bof: boolean read Fbof;

    property RecNo: ub4 read FRecno write MoveTo;
    property FetchRowsCount: ub4 read GetFetchRowsCount;

    property FieldName[Index: Integer]: String read GetFieldName;
    property RawFieldStruct[position: sword]: pRStruct read RawFieldStructGet;
    property RawField[position: sword]: pSelectListItem read RawFieldGet;
    property RawBufferGet: PChar read GetRawBuffer;
    property RawIndGet: psb2 read GetRawInd;
    property RawRCodeGet: pub2 read GetRawRCode;
    property RawRlenGet: pub2 read GetRawRlen;

    property OraFieldByNum[Index: Integer]: TOraField read GetOraFieldByNum;
    property OraFieldByName[Index: String]: TOraField read GetOraFieldByName;

    property Active: boolean read GetActive write SetActive;
    property Exec: boolean read FExec write SetExec;
    property oraSession: To8Session read GetSession write SetSession;

    property NestedStmtClass: TClass read FNestedStmtClass write FNestedStmtClass;

    property CursorParam: TOraParam read FCursorParam;
    property CursorSource: To8Stmt read FCursorSource write SetCursorSource;
    property CursorIndex: Integer read  FCursorIndex  write SetCursorIndex;

    property DataSetField: TOraField read GetDataSetField;
    property DataSetSource: To8Stmt read FDataSetSource write SetDataSetSource;
    property DataSetFieldName: String read GetDataSetFieldName write SetDataSetFieldName;

    property ResultSet: pOCIResult read FResultSet write FResultSet;

    property OnBeforeExecute: TNotifyEvent read FOnBeforeExecute write FOnBeforeExecute;
    property OnAfterExecute: TOraAfterExecuteEvent read FOnAfterExecute write FOnAfterExecute;

    property ResultSetStream: TResultSetStream read GetResultSetStream write SetResultSetStream;

    property State: TDataSetState read FState write SetState;

    function GetLastErrMessage: String;

  end;

function DateTimeToOraDate(oDate: TDateTime): ORA_Date;
function OraDateToDateTime(oDate: ORA_Date): TDateTime;

implementation

uses
   o8DataSet;

function DateTimeToOraDate(oDate: TDateTime): ORA_Date;
var
  Year, Month, Day: Word;
  Hour, Min, Sec, MSec: Word;
begin
  DecodeDate(oDate, Year, Month, Day);
  DecodeTime(oDate, Hour, Min, Sec, MSec);
  Result.cCentury   := ub1(Trunc(Year / 100) + 100);
  Result.cYear      := ub1(Year - Trunc(Year / 100) * 100 + 100);
  Result.cMonth     := ub1(Month);
  Result.cDay       := ub1(Day);
  Result.cHour      := ub1(Hour + 1);
  Result.cMinute    := ub1(Min + 1);
  Result.cSecond    := ub1(Sec + 1);
end;

function OraDateToDateTime(oDate: ORA_Date): TDateTime;
var
  Year: Word;
begin
  Year := (oDate.cCentury - 100) * 100 + (oDate.cYear - 100);
  try
    Result := EncodeDate(Year, oDate.cMonth, oDate.cDay) + EncodeTime(oDate.cHour - 1, oDate.cMinute - 1, oDate.cSecond - 1, 0);
  except
    Result := 0;
  end;
end;

//******************************************************************************
constructor To8Stmt.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FieldDescribe := TList.Create;
  DataSetList := TList.Create;
  FParams := TOraParams.Create(self, TOraParam);
  stmthp := nil;
  errhp := nil;
  FActive := false;
  FStreamedActive := false;
  AllocHandlerStmt := false;
  AllocHandlerErr  := false;
  errcode := 0;
  fillchar(buf, MaxErrorMessage, 0);
  FFieldCount := 0;
  FArrayOfStructureLength := 1;
  FDescriptors := TList.Create;
  FCursorParam := nil;
  FCursorSource := nil;
  FCursorIndex := -1;
  FNestedStmtClass := Self.ClassType;
  FDataSetFieldName := '';
  FDataSetSource := nil;
  FDataSetField := nil;
  FResultSet := nil;
  FDataSetFieldNameStreamed := '';
  FArrayOfPar := false;
  FParamBuffer := nil;
  FParamIndBuffer := nil;
  FParamLenBuffer := nil;
  FParamRCodeBuffer := nil;
  FParRecordSize := 0;
  FLastParamRow := 0;
  FStreamedResultSetStream := rsUndef;
  FOraFields := TObjectList.Create;
end;

//******************************************************************************
destructor To8Stmt.Destroy;
var
   position: Integer;
   oDataSet: To8DataSet;
begin
//   Active := false;
   while DataSetList.Count > 0 do
   begin
      oDataSet := To8DataSet(DataSetList.Items[0]);
      oDataSet.STMT := nil;
   end;
   DataSetList.Destroy;
   if oraSession <> nil then oraSession.RemoveStatment(self);
   FreeHandlers;
   for position := 0 to FieldDescribe.Count - 1 do MemMgr.oMem.FreeMem(PChar(FieldDescribe.Items[position]));
   FieldDescribe.Clear;
   FieldDescribe.Destroy;
   FParams.Destroy;
   DescriptorsFree;
   FDescriptors.Destroy;
   if FParamBuffer <> nil then MemMgr.oMem.FreeMem(FParamBuffer);
   if FParamIndBuffer   <> nil then MemMgr.oMem.FreeMem(PChar(FParamIndBuffer));
   if FParamLenBuffer   <> nil then MemMgr.oMem.FreeMem(PChar(FParamLenBuffer));
   if FParamRCodeBuffer <> nil then MemMgr.oMem.FreeMem(PChar(FParamRCodeBuffer));
   FOraFields.Destroy;
   inherited Destroy;
end;

//******************************************************************************
procedure To8Stmt.AllocHandlers;
begin
  if ( oraSession <> nil ) and ( oraSession.Active ) then
  begin
    FreeHandlers;
    PrepareStatus := false;
    if (CursorParam = nil) and (DataSetField = nil) and (ResultSet = nil) then
    begin
      if not AllocHandlerStmt then
        if ( OCIHandleAlloc(  oraSession.hEnv,
                              @stmthp,
                              OCI_HTYPE_STMT,
                              0,
                              nil) = OCI_SUCCESS ) then AllocHandlerStmt := true;
    end else begin
      if DataSetField <> nil then
      begin
        stmthp := DataSetField.Cursor;
        AllocHandlerStmt := true;
      end;
      if CursorParam <> nil then
      begin
        stmthp := CursorParam.Cursor;
        AllocHandlerStmt := true;
      end;
      if ResultSet <> nil then
      begin
        stmthp := ResultSet;
        AllocHandlerStmt := true;
      end;
    end;
    if not AllocHandlerErr then
      if ( OCIHandleAlloc(  oraSession.hEnv,
                            @errhp,
                            OCI_HTYPE_ERROR,
                            0,
                            nil) = OCI_SUCCESS ) then AllocHandlerErr := true;
  end;
end;

//******************************************************************************
procedure To8Stmt.FreeHandlers;
begin
  if oraSession <> nil then
  begin
    if oraSession.hEnv <> nil then
    begin
      DescriptorsFree;
      if AllocHandlerErr then
      begin
        OCIHandleFree(errhp, OCI_HTYPE_ERROR);
        errhp := nil;
        AllocHandlerErr := false;
      end;
      if AllocHandlerStmt then
      begin
        OCIHandleFree(stmthp, OCI_HTYPE_STMT);
        stmthp := nil;
        AllocHandlerStmt := false;
      end;
      PrepareStatus := false;
    end;
  end;
end;

//******************************************************************************
procedure To8Stmt.Loaded;
begin
  inherited Loaded;
  if FStreamedResultSetStream <> rsUndef then
    ResultSetStream := FStreamedResultSetStream;
  if FDataSetFieldNameStreamed <> '' then
    DataSetFieldName := FDataSetFieldNameStreamed;
  if FStreamedArrayOfPar then ArrayOfPar := true;
  if FStreamedExec then Exec := true;
  if FStreamedActive then Active := true;
end;

//******************************************************************************
function To8Stmt.GetActive: boolean;
begin
   Result := FActive;
end;

//******************************************************************************
procedure To8Stmt.SetActive(bActive: boolean);
var
   i: Integer;
begin
  if ( csReading in ComponentState ) then
    FStreamedActive := bActive
  else begin
    if oraSession <> nil then
    begin
      if oraSession.Active then
      begin
        if  bActive then
        begin
          if not FActive then
          begin
            //{$B+}
            //if not (AllocHandlerStmt and AllocHandlerErr) then AllocHandlers;
            //if AllocHandlerStmt and AllocHandlerErr then
            //{$B-}
              Open
            //else GetError(true);
          end;
        end else begin
          for i := 0 to DataSetList.Count - 1 do To8DataSet(DataSetList.Items[i]).Active := false;
          Close;
          FreeHandlers;
        end;
      end else begin
        for i := 0 to DataSetList.Count - 1 do To8DataSet(DataSetList.Items[i]).Active := false;
        if FActive then GetError(true);
        FActive := false;
      end;
    end else begin
       for i := 0 to DataSetList.Count - 1 do To8DataSet(DataSetList.Items[i]).Active := false;
       if FActive then GetError(true);
       FActive := false;
    end;
  end;
end;

//******************************************************************************
procedure To8Stmt.Open;
begin
   FActive := True;
   //PrepareStatus := false;
end;

//******************************************************************************
procedure To8Stmt.Close;
begin
   FActive := false;
   PrepareStatus := false;
   FOraFields.Clear;
end;

//******************************************************************************
function To8Stmt.GetSession: To8Session;
begin
   Result := FSession;
end;

//******************************************************************************
procedure To8Stmt.SetSession(oSession: To8Session);
begin
  if Active then
    Active := false;
  if oraSession <> nil then oraSession.RemoveStatment(self);
  FSession := oSession;
  //if oraSession <> nil then AllocHandlers;
  if oraSession <> nil then oraSession.AddStatment(self);
end;

//******************************************************************************
procedure To8Stmt.GetError(bError: boolean);
begin
  errcode := 0;
  fillchar(buf, MaxErrorMessage, 0);
  if oraSession <> nil then
  begin
    {$B+}
    if (oraSession.hEnv <> nil) and AllocHandlerStmt and AllocHandlerErr then
    {$B-}
    begin
       if oraSession.Active then
       begin
         OCIErrorGet   (   errhp,
                           1,
                           nil,
                           @errcode,
                           @buf[1],
                           MaxErrorMessage,
                           OCI_HTYPE_ERROR);
       end
       else
       begin
          errcode := 3000006;
          move('  Session.', buf, 20);
       end;
    end
    else
    begin
       if not AllocHandlerErr then
       begin
          errcode := 3000001;
          move('  Error .', buf, 33);
       end;
       if not AllocHandlerStmt then
       begin
          errcode := 3000002;
          move('  Server  .', buf, 43);
       end;
    end;
  end
  else
  begin
    errcode := 3000004;
    move('   Session.', buf, 33);
  end;
  //if (not(csDesigning in ComponentState)) then
  if (not(csReading in ComponentState)) then
    if bError and (errcode <> 0) then
         raise TOraException.Create(errcode, Trim(String(buf)))
end;

//******************************************************************************
function To8Stmt.GetErrMessage: String;
begin
   GetError;
   if errcode <> 0 then Result := Trim(String(buf)) else Result := '';
end;

//******************************************************************************
function To8Stmt.GetLastErrMessage: String;
begin
  Result := Trim(String(buf));
end;

//******************************************************************************
function To8Stmt.GetErrCode: sb4;
begin
   GetError;
   Result := errcode;
end;

//******************************************************************************
function To8Stmt.GetFieldName(Index: Integer): String;
var
   aCol: pRStruct;
begin
    aCol := pRStruct(FieldDescribe.Items[Index-1]);
    Result := copy(aCol.S_COLNAME, 1, aCol.S_COLNAMELENGTH);
end;

//******************************************************************************
procedure To8Stmt.AddDataSet(oOBJ: TObject);
begin
    if Assigned(DataSetList) then
    begin
      DataSetList.Add(oOBJ);
    end;
end;

//******************************************************************************
procedure To8Stmt.RemoveDataSet(oOBJ: TObject);
begin
   if Assigned(DataSetList) then
   begin
      DataSetList.Remove(oOBJ);
      DataSetList.Pack;
   end;
end;

procedure To8Stmt.Prepare;
begin
  //
end;

procedure To8Stmt.ExecSql;
begin
  if oraSession <> nil then
  begin
    if oraSession.Active then
    begin
      {$B+}
      if not (AllocHandlerStmt and AllocHandlerErr) then AllocHandlers;
      if AllocHandlerStmt and AllocHandlerErr then
      {$B-}
      begin
        InternalExecute;
        //FreeHandlers;
      end else GetError(true);
    end else GetError(true);
  end else GetError(true);
end;

function To8Stmt.GetParamNumByName(sIndex: String): Integer;
var
  i: Integer;
begin
  Result := -1;
  for i := 0 to Params.Count - 1 do
    if AnsiUpperCase(TOraParam(Params.Items[i]).Name) = AnsiUpperCase(sIndex) then
    begin
      Result := i;
      break;
    end;
end;

function To8Stmt.GetParamByName(sIndex: String): TOraParam;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to Params.Count - 1 do
    if AnsiUpperCase(TOraParam(Params.Items[i]).Name) = AnsiUpperCase(sIndex) then
    begin
      Result := TOraParam(Params.Items[i]);
      break;
    end;
  if Result = nil then
    Result := Params.CreateParam(ptUnknown, sIndex, false, Null,  5);
end;

function To8Stmt.GetParamByNum(iIndex: Integer): TOraParam;
begin
  try
    Result := TOraParam(Params.Items[iIndex]);
  except
    Result := nil;
  end;
end;

procedure To8Stmt.ReadParamData(Reader: TReader);
begin
  Reader.ReadValue;
  Reader.ReadCollection(Params);
end;

procedure To8Stmt.WriteParamData(Writer: TWriter);
begin
  Writer.WriteCollection(Params);
end;

procedure To8Stmt.DefineProperties(Filer: TFiler);

  function WriteData: Boolean;
  begin
    if Filer.Ancestor <> nil then
      Result := not FParams.IsEqual(To8Stmt(Filer.Ancestor).FParams) else
      Result := FParams.Count > 0;
  end;

begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('ParamData', ReadParamData, WriteParamData, WriteData);
end;

procedure To8Stmt.SetParamList(const Value: TOraParams);
begin
  FParams.Assign(Value);
end;

function To8Stmt.GetOraFieldByNum(Index: Integer): TOraField;
begin
  Result := TOraField.Create(self, Index);
end;

function To8Stmt.FieldByNum(FieldNum: Integer): TOraField;
begin
  //Result := TOraField.Create(self, FieldNum);
  Result := TOraField(FOraFields.Items[FieldNum - 1]);
end;

function To8Stmt.GetOraFieldByName(Index: String): TOraField;
var
  i: Integer;
  bFound: boolean;
  OraF: TOraField;
begin
  Result := nil;
  bFound := false;
  for i := 0 to FieldCount - 1 do begin
    OraF := TOraField(FOraFields.Items[i]);
    if AnsiUpperCase(OraF.FFieldName) = AnsiUpperCase(Index) then
    begin
      Result := OraF;
      bFound := true;
      break;
    end;
  end;
  if not bFound then
    raise TOraException.Create(3000005, '  ' + Index);
end;

function To8Stmt.Locate(sFieldName: String; vFind: Variant): boolean;
var
  OldRec: Integer;
begin
  OldRec := RecNo;
  Result := false;
  First;
  while not Eof do
  begin
    if OraFieldByName[sFieldName].Value = vFind then
    begin
      Result := true;
      break;
    end;
    Next;
  end;
  if not Result then
    MoveTo(OldRec);
end;

procedure To8Stmt.InternalExecute;
begin
  if Params <> nil then Params.DescriptorsAlloc;
  if Assigned(OnBeforeExecute) then OnBeforeExecute(self);
end;

constructor To8Stmt.CreateFromStmtField(AOwner: TComponent;
  oField: TOraField);
begin
  Create(AOwner);
  DataSetSource := oField.OraStmt;
  DataSetFieldName := oField.FieldName;
end;

function To8Stmt.GetDataSetFieldName: String;
begin
  Result := FDataSetFieldName;
end;

function To8Stmt.GetDataSetField: TOraField;
begin
  if (FDataSetSource <> nil) then
    try
      if FDataSetSource.Active then
      begin
        FDataSetField := FDataSetSource.OraFieldByName[FDataSetFieldName];
        if FDataSetField.OraFieldType <> SQLT_RSET then
          FDataSetField := nil;
      end else
        FDataSetField := nil;
    except
      FDataSetField := nil;
    end;
  Result := FDataSetField;
end;

procedure To8Stmt.SetDataSetFieldName(const Value: String);
begin
  if ( csReading in ComponentState ) then
    FDataSetFieldNameStreamed := Value
  else
    FDataSetFieldName := Value;
end;

procedure To8Stmt.SetDataSetSource(const Value: To8Stmt);
begin
  FDataSetSource := Value;
  if (FDataSetSource <> nil) then
    try
      FDataSetField := FDataSetSource.OraFieldByName[FDataSetFieldName];
      if FDataSetField.OraFieldType <> SQLT_RSET then
        FDataSetField := nil;
    except
      FDataSetField := nil
    end
  else
    FDataSetField := nil;
end;

procedure To8Stmt.SetCursorIndex(const Value: Integer);
begin
  FCursorIndex := Value;
  if  (FCursorIndex >= 0) and
      (FCursorSource <> nil) and
      (FCursorSource.FParams.Count > FCursorIndex) and
      (TOraParam(FCursorSource.FParams.Items[FCursorIndex]).DataType = ptCursor) then
    FCursorParam := TOraParam(FCursorSource.FParams.Items[FCursorIndex])
  else
    FCursorParam := nil;
end;

procedure To8Stmt.SetCursorSource(const Value: To8Stmt);
begin
  FCursorSource := Value;
  if  (FCursorIndex >= 0) and
      (FCursorSource <> nil) and
      (FCursorSource.FParams.Count > FCursorIndex) and
      (TOraParam(FCursorSource.FParams.Items[FCursorIndex]).DataType = ptCursor) then
    FCursorParam := TOraParam(FCursorSource.FParams.Items[FCursorIndex])
  else
    FCursorParam := nil;
end;

procedure To8Stmt.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if Operation = (opRemove) then
    if AComponent.InheritsFrom(To8Stmt) then
    begin
      if AComponent = FCursorSource then
      begin
        FCursorSource := nil;
        FCursorIndex := -1;
      end;
      if AComponent = FDataSetSource then
      begin
        FDataSetSource := nil;
        FDataSetFieldName := '';
      end;
    end;
end;

procedure To8Stmt.SetExec(const Value: boolean);
begin
  if ( csReading in ComponentState ) then
    FStreamedExec := Value
  else begin
    if not FActive then
    begin
      FExec := Value;
      if FExec then
        ExecSql;
    end;
  end;
end;

procedure To8Stmt.SetArrayOfPar(const Value: Boolean);
begin
  if ( csReading in ComponentState ) then FStreamedArrayOfPar := Value
  else
  begin
    FArrayOfPar := Value;
    if FParamBuffer <> nil then MemMgr.oMem.FreeMem(FParamBuffer);
    if FParamIndBuffer    <> nil then MemMgr.oMem.FreeMem(PChar(FParamIndBuffer  ));
    if FParamLenBuffer    <> nil then MemMgr.oMem.FreeMem(PChar(FParamLenBuffer  ));
    if FParamRCodeBuffer  <> nil then MemMgr.oMem.FreeMem(PChar(FParamRCodeBuffer));
    FLastParamRow := 0;
  end;
end;

function To8Stmt.GetArrayParNull(Index1, Index2: Integer): Boolean;
begin
  Result := true;
end;

function To8Stmt.GetArrayParRCode(Index1, Index2: Integer): ub2;
begin
  Result := 0;
end;

function To8Stmt.GetArrayParValue(Index1, Index2: Integer): Variant;
begin
  //
end;

procedure To8Stmt.SetArrayParNull(Index1, Index2: Integer;
  const Value: Boolean);
begin
  //
end;

procedure To8Stmt.SetArrayParValue(Index1, Index2: Integer;
  const Value: Variant);
begin
  //
end;

function To8Stmt.GetArrayParString(Index1, Index2: Integer): String;
begin
  //
end;

procedure To8Stmt.SetArrayParString(Index1, Index2: Integer;
  const Value: String);
begin
  //
end;

function To8Stmt.GetArrayParInteger(Index1, Index2: Integer): Integer;
begin
  Result := 0;
end;

procedure To8Stmt.SetArrayParInteger(Index1, Index2: Integer; const Value: Integer);
begin
  //
end;

function To8Stmt.GetArrayParFloat(Index1, Index2: Integer): double;
begin
  Result := 0;
end;

procedure To8Stmt.SetArrayParFloat(Index1, Index2: Integer; const Value: double);
begin
  //
end;

function To8Stmt.GetArrayParDateTime(Index1, Index2: Integer): TDateTime;
begin
  Result := 0;
end;

procedure To8Stmt.SetArrayParDateTime(Index1, Index2: Integer; const Value: TDateTime);
begin
  //
end;

function To8Stmt.FieldByName(sFieldName: String): TOraField;
begin
  Result := GetOraFieldByName(sFieldName);
end;

function To8Stmt.GetResultSetStream: TResultSetStream;
begin
  Result := rsUndef;
end;

procedure To8Stmt.SetResultSetStream(const Value: TResultSetStream);
begin
  if ( csReading in ComponentState ) then
    FStreamedResultSetStream := Value;
end;

procedure To8Stmt.BreakExecut;
begin
  FSession.BreakNonBlockingExecution;
end;

procedure To8Stmt.Edit;
begin
  //Nothing
end;

procedure To8Stmt.Post;
begin
  //Nothing
end;

procedure To8Stmt.SetState(const Value: TDataSetState);
begin
  FState := Value;
end;

{ TOraParam }

constructor TOraParam.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FOraParams := TOraParams(Collection);
  FParBindp := nil;
  FMaxLengthStringBuffer := DefMaxLengthStringBuffer;
  BufStr := MemMgr.oMem.GetMem(self, FMaxLengthStringBuffer);
  FillChar(BufStr^, FMaxLengthStringBuffer, #0);
  FDataType := ptUnknown;
  FParInd := 0;
  FParRealLen := 0;
  FParRCode := 0;
  BufInt := 0;
  BufFloat := 0;
  with BufDate do
  begin
    cCentury := 100;
    cYear := 100;
    cMonth := 1;
    cDay := 1;
    cHour := 1;
    cMinute := 1;
    cSecond := 1;
  end;
  BufLob := nil;
  BufFile := nil;
  BufCursor := nil;
  BufRowId := nil;
  SetedRowId := nil;
  envhp := nil;
  srvhp := nil;
  errhp := nil;
  usrhp := nil;
  svchp := nil;
  FValue := NULL;
end;

destructor TOraParam.Destroy;
var
  IsDscInit: bool;
begin
  if  (envhp <> nil) and
      (errhp <> nil) and
      ((BufLob <> nil) or
      (BufFile <> nil) or
      (BufCursor <> nil)) then
  begin
    if (BufLob <> nil) then
    begin
      IsDscInit := 0;
      OCILobLocatorIsInit(  envhp,
                            errhp,
                           BufLob,
                            @IsDscInit);
      if IsDscInit <> 0 then
        OCIDescriptorFree(BufLob, OCI_DTYPE_LOB);
    end;
    if (BufFile <> nil) then
    begin
      IsDscInit := 0;
      OCILobLocatorIsInit(  envhp,
                            errhp,
                            BufFile,
                            @IsDscInit);
      if IsDscInit <> 0 then
        OCIDescriptorFree(BufFile, OCI_DTYPE_FILE);
    end;
    if (BufRowId <> nil) then
    begin
      //IsDscInit := 0;
      //OCILobLocatorIsInit(  envhp,
      //                      errhp,
      //                      BufRowId,
      //                      @IsDscInit);
      //if IsDscInit <> 0 then
        OCIDescriptorFree(BufFile, OCI_DTYPE_ROWID);
    end;
    if (BufCursor <> nil) then
      OCIHandleFree(BufCursor, OCI_HTYPE_STMT);
  end;
  MemMgr.oMem.FreeMem(BufStr);
  inherited Destroy;
end;

function TOraParam.GetBuffer: Pointer;
begin
  case FDataType of
    ptString:         Result := Pointer(BufStr);
    ptInteger:        Result := Pointer(@BufInt);
    ptFloat:          Result := Pointer(@BufFloat);
    ptDateTime:       Result := Pointer(@BufDate);
    ptCLob,ptBLob:    Result := Pointer(@BufLob);
    ptCFile,ptBFile:  Result := Pointer(@BufFile);
    ptCursor:         Result := Pointer(@BufCursor);
    ptRowId:
      begin
        if SetedRowId <> nil then
          Result := Pointer(@SetedRowId)
        else
          Result := Pointer(@BufRowID);
      end;
  else
    Result := Pointer(nil);
  end;
end;

function TOraParam.GetIsNull: boolean;
begin
  Result := (FParInd = -1);
end;

procedure TOraParam.SetIsNull(const Value: boolean);
begin
  if Value then
    FParInd := -1
  else
    FParInd := 0;
end;

procedure TOraParam.SetMaxLengthStringBuffer(const Value: Integer);
begin
  if FMaxLengthStringBuffer <> Value then
  begin
    FMaxLengthStringBuffer := Value;
    MemMgr.oMem.FreeMem(BufStr);
    BufStr := MemMgr.oMem.GetMem(self, FMaxLengthStringBuffer);
    FillChar(BufStr^, FMaxLengthStringBuffer, #0);
  end;
end;

procedure TOraParam.SetDataType(const Value: TRawParamType);
begin
  FDataType := Value;
  case Value of
    ptString:         FExternalType := SQLT_STR;
    ptInteger:        FExternalType := SQLT_INT;
    ptFloat:          FExternalType := SQLT_FLT;
    ptDateTime:       FExternalType := SQLT_DAT;
    ptCLob:           FExternalType := SQLT_CLOB;
    ptBLob:           FExternalType := SQLT_BLOB;
    ptBFile:          FExternalType := SQLT_BFILE;
    ptCFile:          FExternalType := SQLT_CFILE;
    ptCursor:         FExternalType := SQLT_RSET;
    ptRowId:
      begin
        FExternalType := SQLT_RDD;
        FParRealLen := SizeOf(pOCIRowid);
        MaxLength := SizeOf(pOCIRowid);
        FParInd := 0;
      end;
  end;
end;

procedure TOraParam.SetValue(const Value: Variant);
var
  ss: String;
begin
  IsNull := false;
  ParRCode := 0;
  if VarIsNull(Value) or VarIsEmpty(Value) then begin
    SetIsNull(true);
    Exit;
  end;
  case FDataType of
    ptUnknown:
      case VarType(Value) of
        varEmpty, varNull: SetIsNull(true);
        varSmallint, varInteger:
          begin
            DataType := ptInteger;
            SetValue(Value);
          end;
        varSingle, varDouble, varCurrency:
          begin
            DataType := ptFloat;
            SetValue(Value);
          end;
        varDate:
          begin
            DataType := ptDateTime;
            SetValue(Value);
          end;
        varString:
          begin
            DataType := ptString;
            ss := Value;
            SetMaxLengthStringBuffer(Length(ss) + 10);
            SetValue(Value);
          end;
      end;
    ptString:
      begin
        ss := Value;
        FParRealLen := Length(ss);
        if FParRealLen > MaxLengthStringBuffer then
          SetMaxLengthStringBuffer(FParRealLen + 10);
        StrMove(BufStr, pChar(ss), FParRealLen);
        BufStr[FParRealLen] := #0;
        Inc(FParRealLen);
        MaxLength := FMaxLengthStringBuffer - 1;
      end;
    ptInteger:
      begin
        BufInt := Value;
        FParRealLen := SizeOf(Integer);
        MaxLength := FParRealLen;
      end;
    ptFloat:
      begin
        BufFloat := Value;
        FParRealLen := SizeOf(double);
        MaxLength := FParRealLen;
      end;
    ptDateTime:
      begin
        BufDate := DateTimeToOraDate(Value);
        FParRealLen := SizeOf(ORA_Date);
        MaxLength := FParRealLen;
      end;
  end;
  FValue := Value;
end;

function TOraParam.GetValue: Variant;
var
  ss: String;
  td: TDateTime;
begin
  if not IsNull then
    case FDataType of
      ptString:
        begin
          ss := String(BufStr);
          FValue := ss;
        end;
      ptInteger:
        begin
          FValue := BufInt;
        end;
      ptFloat:
        begin
          FValue := BufFloat;
        end;
      ptDateTime:
        begin
          FValue := OraDateToDateTime(BufDate);
        end;
    end
  else
    case FDataType of
      ptString: FValue := '';
      ptInteger:
        begin
          BufInt := 0;
          FValue := BufInt;
        end;
      ptFloat:
        begin
          BufFloat := 0;
          FValue := BufFloat;
        end;
      ptDateTime:
        begin
          td := 0;
          FValue := td;
        end;
    end;
  Result := FValue;
end;

function TOraParam.GetAsDateTime: TDateTime;
var
  CurrValue: Variant;
begin
  CurrValue := GetValue;
  Result := CurrValue;
end;

function TOraParam.GetAsFloat: Double;
var
  CurrValue: Variant;
begin
  CurrValue := GetValue;
  Result := CurrValue;
end;

function TOraParam.GetAsInteger: Integer;
var
  CurrValue: Variant;
begin
  CurrValue := GetValue;
  Result := CurrValue;
end;

function TOraParam.GetAsString: String;
var
  CurrValue: Variant;
begin
  CurrValue := GetValue;
  Result := CurrValue;
end;

procedure TOraParam.SetAsDateTime(const Value: TDateTime);
begin
  SetValue(Value);
end;

procedure TOraParam.SetAsFloat(const Value: Double);
begin
  SetValue(Value);
end;

procedure TOraParam.SetAsInteger(const Value: Integer);
begin
  SetValue(Value);
end;

procedure TOraParam.SetAsString(const Value: String);
begin
  SetValue(Value);
end;

function TOraParam.IsEqual(Value: TOraParam): Boolean;
begin
  Result := (VarType(FValue) = VarType(Value.ParamValue)) and
    (VarIsEmpty(FValue) or (FValue = Value.ParamValue)) and
    (Name = Value.Name) and (DataType = Value.DataType) and
    (IsNull = Value.IsNull);
end;

function TOraParam.GetLobStream: TOraLobStream;
begin
  Result := nil;
  if FDataType in [ptCLob, ptBLob, ptCFile, ptBFile] then
    Result := TOraLobStream.Create(self);
end;

procedure TOraParam.DescriptorAlloc;
var
  IsDscInit: bool;
begin
  if  (envhp <> nil) and
      (errhp <> nil) and
      ((BufLob <> nil) or
      (BufFile <> nil) or
      (BufCursor <> nil)) then
  begin
    if DataType in [ptCLob, ptBLob, ptCFile, ptBFile, ptCursor, ptRowId] then
    begin
      if (BufLob <> nil) then
      begin
        IsDscInit := 0;
        OCILobLocatorIsInit(  envhp,
                              errhp,
                              BufLob,
                              @IsDscInit);
        if IsDscInit <> 0 then
          OCIDescriptorFree(BufLob, OCI_DTYPE_LOB);
      end;
      if (BufFile <> nil) then
      begin
        IsDscInit := 0;
        OCILobLocatorIsInit(  envhp,
                              errhp,
                              BufFile,
                              @IsDscInit);
        if IsDscInit <> 0 then
          OCIDescriptorFree(BufFile, OCI_DTYPE_FILE);
      end;
      if (BufRowId <> nil) then
      begin
        //IsDscInit := 0;
        //OCILobLocatorIsInit(  envhp,
        //                      errhp,
        //                      BufRowId,
        //                      @IsDscInit);
        //if IsDscInit <> 0 then
          OCIDescriptorFree(BufRowId, OCI_DTYPE_ROWID);
      end;
    end;
  end;
  envhp := To8Stmt(FOraParams.Owner).oraSession.hEnv;
  srvhp := To8Stmt(FOraParams.Owner).oraSession.hServ;
  errhp := To8Stmt(FOraParams.Owner).oraSession.hErr;
  usrhp := To8Stmt(FOraParams.Owner).oraSession.hSess;
  svchp := To8Stmt(FOraParams.Owner).oraSession.hSvc;
  if DataType in [ptCLob, ptBLob, ptCFile, ptBFile, ptCursor, ptRowId] then
  begin
    OCIDescriptorAlloc( envhp,
                        @BufLob,
                        OCI_DTYPE_LOB,
                        0,
                        nil);
    OCIAttrSet(BufLob, OCI_DTYPE_LOB, nil, ub4(0), OCI_ATTR_LOBEMPTY, errhp);
    OCIDescriptorAlloc( envhp,
                        @BufFile,
                        OCI_DTYPE_FILE,
                        0,
                        nil);
    OCIAttrSet(BufFile, OCI_DTYPE_FILE, nil, ub4(0), OCI_ATTR_LOBEMPTY, errhp);
    OCIDescriptorAlloc( envhp,
                        @BufRowId,
                        OCI_DTYPE_ROWID,
                        0,
                        nil);
    OCIHandleAlloc (  envhp,
                      @BufCursor,
                      OCI_HTYPE_STMT,
                      0,
                      nil);
  end;
end;

function TOraParam.GetParamSize: Integer;
begin
  Result := 0;
  case DataType of
    ptString:       Result := MaxLengthStringBuffer;
    ptInteger:      Result := SizeOf(Integer);
    ptFloat:        Result := SizeOf(Double);
    ptDateTime:     Result := SizeOf(ORA_Date);
    ptCLob:         Result := SizeOf(pOCILobLocator);
    ptBLob:         Result := SizeOf(pOCILobLocator);
    ptCFile:        Result := SizeOf(pOCILobLocator);
    ptBFile:        Result := SizeOf(pOCILobLocator);
    ptCursor:       Result := SizeOf(pOCIResult);
    ptRowId:        Result := SizeOf(pOCIRowid);
  end;
end;

procedure TOraParam.AssignFieldValue(Field: TField; const Value: Variant);
begin
  case Field.DataType of
  ftString, ftFixedChar, ftWideString:
    begin
      DataType := ptString;
      IsNull := VarIsEmpty(Value) or VarIsNull(Value);
      if not IsNull then
        AsString := Value
      else
        AsString := '';
    end;
  ftBoolean, ftSmallint, ftInteger, ftWord, ftLargeint:
    begin
      DataType := ptInteger;
      MaxLengthStringBuffer := 5;
      IsNull := VarIsEmpty(Value) or VarIsNull(Value);
      if not IsNull then
        AsInteger := Value;
    end;
  ftFloat, ftCurrency:
    begin
      DataType := ptFloat;
      MaxLengthStringBuffer := 5;
      IsNull := VarIsEmpty(Value) or VarIsNull(Value);
      if not IsNull then
        AsFloat := Value;
    end;
  ftDate, ftTime, ftDateTime:
    begin
      DataType := ptDateTime;
      MaxLengthStringBuffer := 5;
      IsNull := VarIsEmpty(Value) or VarIsNull(Value);
      if not IsNull then
        AsDateTime := Value;
    end;
  end;
end;

procedure TOraParam.Assign(Source: TPersistent);
begin
  if Source is TOraParam then
    AssignParam(TOraParam(Source))
//  else if Source is TOraField then
//    AssignField(TOraField(Source))
//  else if Source is TField then
//    AssignField(TField(Source))
//  else if Source is TStrings then
//    LoadFromStrings(TStrings(Source))
  else
    inherited Assign(Source);
end;

procedure TOraParam.AssignParam(Param: TOraParam);
begin
  if Param <> nil then
  begin
    DataType := Param.DataType;
    IsNull := Param.IsNull;
    MaxLengthStringBuffer := Param.MaxLengthStringBuffer;
    ParamValue := Param.ParamValue;
    Name := Param.Name;
  end;
end;

function TOraParam.GetParBindpp: ppOCIBind;
begin
  Result := @FParBindp;
end;

procedure TOraParam.SetBufRowId(const Value: pOCIRowid);
begin
  SetedRowId := Value;
  FParRealLen := SizeOf(pOCIRowid);
  MaxLength := SizeOf(pOCIRowid);
end;

function TOraParam.GetDisplayName: string;
begin
  if Name <> '' then
    Result := Name
  else
    Result := inherited GetDisplayName;
end;

function TOraParam.GetBufRowId: pOCIRowid;
begin
  if SetedRowId <> nil then
    Result := SetedRowId
  else
    Result := BufRowID;
end;

{ TOraParams }

procedure TOraParams.AssignValues(Value: TOraParams);
var
  I: Integer;
  P: TOraParam;
begin
  for I := 0 to Value.Count - 1 do
  begin
    P := FindParam(Value[I].Name);
    if P <> nil then
      P.Assign(Value[I]);
  end;
end;

function TOraParams.CreateParam(DataType: TRawParamType; ParamName: String; IsNull: boolean;
  ParamValue: variant; MaxStrLen :Integer): TOraParam;
begin
  Result := Add as TOraParam;
  Result.DataType := DataType;
  Result.Name := ParamName;
  Result.IsNull := IsNull;
  Result.ParamValue := ParamValue;
  Result.MaxLengthStringBuffer := MaxStrLen;
end;

procedure TOraParams.DescriptorsAlloc;
var
  i: Integer;
begin
  for i := 0 to Count - 1 do
    TOraParam(Items[i]).DescriptorAlloc;
end;

function TOraParams.FindParam(const Value: string): TOraParam;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
  begin
    Result := TOraParam(inherited Items[I]);
    if AnsiCompareText(Result.Name, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TOraParams.GetCollectionOwner: TPersistent;
begin
  Result := GetOwner;
end;

function TOraParams.GetItem(Index: Integer): TOraParam;
begin
  Result := TOraParam(inherited Items[Index]);
end;

function TOraParams.IsEqual(Value: TOraParams): Boolean;
var
  I: Integer;
begin
  Result := Count = Value.Count;
  if Result then
    for I := 0 to Count - 1 do
    begin
      Result := TOraParam(Items[I]).IsEqual(TOraParam(Value.Items[I]));
      if not Result then Break;
    end
end;

function TOraParams.ParseSQL(SQL: String; DoCreate: Boolean): String;
const
  Literals = ['''', '"', '`'];
var
  Value, CurPos, StartPos: PChar;
  CurChar: Char;
  Literal: Boolean;
  EmbeddedLiteral: Boolean;
  Name: string;
  i: Integer;
  bExist: boolean;


  function NameDelimiter: Boolean;
  begin
    Result := CurChar in [' ', ',', ';', ')', #13, #10];
  end;

  function IsLiteral: Boolean;
  begin
    Result := CurChar in Literals;
  end;

  function StripLiterals(Buffer: PChar): string;
  var
    Len: Word;
    TempBuf: PChar;

    procedure StripChar;
    begin
      if TempBuf^ in Literals then
        StrMove(TempBuf, TempBuf + 1, Len - 1);
      if TempBuf[StrLen(TempBuf) - 1] in Literals then
        TempBuf[StrLen(TempBuf) - 1] := #0;
    end;

  begin
    Len := StrLen(Buffer) + 1;
    TempBuf := AllocMem(Len);
    Result := '';
    try
      StrCopy(TempBuf, Buffer);
      StripChar;
      Result := StrPas(TempBuf);
    finally
      FreeMem(TempBuf, Len);
    end;
  end;

begin
  Result := SQL;
  Value := PChar(Result);
  if DoCreate then Clear;
  CurPos := Value;
  Literal := False;
  EmbeddedLiteral := False;
  repeat
    while (CurPos^ in LeadBytes) do Inc(CurPos, 2);
    CurChar := CurPos^;
    if (CurChar = ':') and not Literal and ((CurPos + 1)^ <> '=') then
    begin
      StartPos := CurPos;
      while (CurChar <> #0) and (Literal or not NameDelimiter) do
      begin
        Inc(CurPos);
        while (CurPos^ in LeadBytes) do Inc(CurPos, 2);
        CurChar := CurPos^;
        if IsLiteral then
        begin
          Literal := Literal xor True;
          if CurPos = StartPos + 1 then EmbeddedLiteral := True;
        end;
      end;
      CurPos^ := #0;
      if EmbeddedLiteral then
      begin
        Name := StripLiterals(StartPos + 1);
        EmbeddedLiteral := False;
      end
      else Name := StrPas(StartPos + 1);
      if DoCreate then
      begin
        bExist := false;
        for i := 0 to self.Count - 1 do
          if AnsiUpperCase(TOraParam(self.Items[i]).Name) = AnsiUpperCase(Name) then
            bExist := true;
        if not bExist then TOraParam(Add).Name := Name;
      end;
      CurPos^ := CurChar;
      StartPos^ := '?';
      Inc(StartPos);
      StrMove(StartPos, CurPos, StrLen(CurPos) + 1);
      CurPos := StartPos;
    end
    else if (CurChar = ':') and not Literal and ((CurPos + 1)^ = ':') then
      StrMove(CurPos, CurPos + 1, StrLen(CurPos) + 1)
    else if IsLiteral then Literal := Literal xor True;
    Inc(CurPos);
  until CurChar = #0;
end;

procedure TOraParams.SetItem(Index: Integer; const Value: TOraParam);
begin
  inherited SetItem(Index, TCollectionItem(Value));
end;

{ TOraException }

constructor TOraException.Create(nErr: sb4; sMess: String);
begin
  inherited Create(sMess);
  FErrorCode := nErr;
end;

{ TOraField }

constructor TOraField.Create(OraStmt: To8Stmt; Index: Integer);
begin
  inherited Create;
  FOraStmt := OraStmt;
  FIndex := Index;
  envhp := OraStmt.oraSession.hEnv;
  srvhp := OraStmt.oraSession.hServ;
  errhp := OraStmt.oraSession.hErr;
  usrhp := OraStmt.oraSession.hSess;
  svchp := OraStmt.oraSession.hSvc;
  FFieldName := FOraStmt.FieldName[FIndex];
end;

function TOraField.GetAsDateTime: TDateTime;
var
  sCont: String;
  iCont: Integer;
  fCont: double;
  oDate: pORA_Date;
  wYear: Integer;
begin
  if not IsNull then
  begin
    case Fpsli.uValueType of
      SQLT_CHR, SQLT_AFC:
        begin
          SetString(sCont, pChar(Fpsli.uValue), Fpsli.uValueLen^);
          try
            Result := StrToDate(sCont);
          except
            Result := 0;
          end
        end;
      SQLT_INT:
        begin
          iCont := pInteger(Fpsli.uValue)^;
          Result := iCont;
        end;
      SQLT_FLT:
        begin
          fCont := pDouble(Fpsli.uValue)^;
          Result := fCont;
        end;
      SQLT_DAT:
        begin
          oDate := pORA_Date(Fpsli.uValue);
          try
            wYear := (oDate.cCentury-100)*100+(oDate.cYear-100);
            if (wYear > 0) and ( wYear < 9999) and ( oDate.cMonth >= 1) and ( oDate.cMonth <= 12) then
              Result := EncodeDate(wYear, oDate.cMonth, oDate.cDay) + EncodeTime(oDate.cHour-1, oDate.cMinute-1, oDate.cSecond-1, 0)
             else
               Result := 0;
          except
              Result := 0;
          end;
        end;
    else
      Result := 0;
    end;
  end else
    Result := 0;
end;

function TOraField.GetAsFloat: double;
var
  sCont: String;
  iCont: Integer;
  fCont: double;
  oDate: pORA_Date;
  wYear: Integer;
begin
  if not IsNull then
  begin
    case Fpsli.uValueType of
      SQLT_CHR, SQLT_AFC:
        begin
          SetString(sCont, pChar(Fpsli.uValue), Fpsli.uValueLen^);
          try
            Result := StrToFloat(sCont);
          except
            Result := 0;
          end
        end;
      SQLT_INT:
        begin
          iCont := pInteger(Fpsli.uValue)^;
          Result := iCont;
        end;
      SQLT_FLT:
        begin
          fCont := pDouble(Fpsli.uValue)^;
          Result := fCont;
        end;
      SQLT_DAT:
        begin
          oDate := pORA_Date(Fpsli.uValue);
          try
            wYear := (oDate.cCentury-100)*100+(oDate.cYear-100);
            if (wYear > 0) and ( wYear < 9999) and ( oDate.cMonth >= 1) and ( oDate.cMonth <= 12) then
              Result := EncodeDate(wYear, oDate.cMonth, oDate.cDay) + EncodeTime(oDate.cHour-1, oDate.cMinute-1, oDate.cSecond-1, 0)
             else
               Result := 0;
          except
              Result := 0;
          end;
        end;
    else
      Result := 0;
    end;
  end else
    Result := 0;
end;

function TOraField.GetAsInteger: Integer;
var
  sCont: String;
  fCont: double;
  oDate: pORA_Date;
  wYear: Integer;
begin
  if not IsNull then
  begin
    case Fpsli.uValueType of
      SQLT_CHR, SQLT_AFC:
        begin
          SetString(sCont, pChar(Fpsli.uValue), Fpsli.uValueLen^);
          try
            Result := StrToInt(sCont);
          except
            Result := 0;
          end
        end;
      SQLT_INT:
        begin
          Result := pInteger(Fpsli.uValue)^;
        end;
      SQLT_FLT:
        begin
          fCont := pDouble(Fpsli.uValue)^;
          Result := Round(fCont);
        end;
      SQLT_DAT:
        begin
          oDate := pORA_Date(Fpsli.uValue);
          try
            wYear := (oDate.cCentury-100)*100+(oDate.cYear-100);
            if (wYear > 0) and ( wYear < 9999) and ( oDate.cMonth >= 1) and ( oDate.cMonth <= 12) then
              Result := Round(EncodeDate(wYear, oDate.cMonth, oDate.cDay) + EncodeTime(oDate.cHour-1, oDate.cMinute-1, oDate.cSecond-1, 0))
             else
               Result := 0;
          except
              Result := 0;
          end;
        end;
    else
      Result := 0;
    end;
  end else
    Result := 0;
end;

function TOraField.GetAsString: String;
var
  iCont: Integer;
  fCont: double;
  oDate: pORA_Date;
  wYear: Integer;
begin
  if not IsNull then
  begin
    case Fpsli.uValueType of
      SQLT_CHR, SQLT_AFC, SQLT_LNG: SetString(Result, pChar(Fpsli.uValue), Fpsli.uValueLen^);
      SQLT_INT:
        begin
          iCont := pInteger(Fpsli.uValue)^;
          Result := IntToStr(iCont);
        end;
      SQLT_FLT:
        begin
          fCont := pDouble(Fpsli.uValue)^;
          Result := FloatToStr(fCont);
        end;
      SQLT_DAT:
        begin
          oDate := pORA_Date(Fpsli.uValue);
          try
            wYear := (oDate.cCentury-100)*100+(oDate.cYear-100);
            if (wYear > 0) and ( wYear < 9999) and ( oDate.cMonth >= 1) and ( oDate.cMonth <= 12) then
              Result := DateToStr(EncodeDate(wYear, oDate.cMonth, oDate.cDay) + EncodeTime(oDate.cHour-1, oDate.cMinute-1, oDate.cSecond-1, 0))
             else
               Result := '';
          except
              Result := '';
          end;
        end;
    else
      Result := '';
    end;
  end else
    Result := '';
end;

function TOraField.GetAsVariant: Variant;
begin
  FValue := AsString;
  Result := FValue;
end;

function TOraField.GetComplexObject: pOCIComplexObjectComp;
begin
  if not IsNull then
    Result := ppOCIComplexObjectComp(Fpsli.uValue)^
  else
    Result := nil;
end;

function TOraField.GetCursor: pOCIResult;
begin
  if not IsNull then
    Result := ppOCIResult(Fpsli.uValue)^
  else
    Result := nil;
end;

function TOraField.GetIsNull: boolean;
begin
  GetSelectListItem;
  Result := (Fpsli.uValueInd^ = -1);
end;

function TOraField.GetLobStream: TOraLobStream;
begin
  Result := nil;
  if not IsNull then
  begin
    case Fpsli.uValueType of
      SQLT_BLOB, SQLT_CLOB, SQLT_CFILE, SQLT_BFILE:
      begin
        Result := TOraLobStream.Create(self);
      end;
    end;
  end;
end;

function TOraField.GetOraFieldType: Integer;
begin
  if not IsNull then
    Result := Fpsli.uValueType
  else
    Result := 0;
end;

function TOraField.GetResultSet: To8Stmt;
begin
  Result := nil;
  if not IsNull then
  begin
    case Fpsli.uValueType of
      SQLT_RSET:
      begin
        Result := To8Stmt(self.OraStmt.NestedStmtClass.NewInstance);
        Result.CreateFromStmtField(self.OraStmt, self);
        try
          //Result.ObjectView := True;
          //Result.DataSetField := DataSetField;
        except
          Result.Free;
          raise;
        end;
      end;
    end;
  end;
end;

function TOraField.GetRowId: pOCIRowId;
begin
  if not IsNull then
    Result := ppOCIRowid(Fpsli.uValue)^
  else
    Result := nil;
end;

function TOraField.GetSelectListItem: pSelectListItem;
begin
  Fpsli := FOraStmt.RawFieldGet(FIndex);
  Result := Fpsli;
end;

procedure TOraField.SetAsDateTime(const Value: TDateTime);
var
  sValue: String;
begin
  IsNull := false;
  case Fpsli.uValueType of
    SQLT_CHR, SQLT_AFC:
      begin
        sValue := DateTimeToStr(Value);
        if Length(sValue) >= Fpsli.uValueLength then
          Fpsli.uValueLen^ := Fpsli.uValueLength
        else
          Fpsli.uValueLen^ := Length(sValue);
        FillChar(pChar(Fpsli.uValue)^, Fpsli.uValueLength, 0);
        StrMove(pChar(Fpsli.uValue), pChar(sValue), Fpsli.uValueLen^);
      end;
    SQLT_INT: pInteger(Fpsli.uValue)^ := Integer(Trunc(Value));
    SQLT_FLT: pDouble(Fpsli.uValue)^ := Value;
    SQLT_DAT: pORA_Date(Fpsli.uValue)^ := DateTimeToOraDate(Value);
  else
    raise Exception.Create('Not editable field.');
  end;
end;

procedure TOraField.SetAsFloat(const Value: double);
var
  sValue: String;
begin
  IsNull := false;
  case Fpsli.uValueType of
    SQLT_CHR, SQLT_AFC:
      begin
        sValue := FloatToStr(Value);
        if Length(sValue) >= Fpsli.uValueLength then
          Fpsli.uValueLen^ := Fpsli.uValueLength
        else
          Fpsli.uValueLen^ := Length(sValue);
        FillChar(pChar(Fpsli.uValue)^, Fpsli.uValueLength, 0);
        StrMove(pChar(Fpsli.uValue), pChar(sValue), Fpsli.uValueLen^);
      end;
    SQLT_INT: pInteger(Fpsli.uValue)^ := Integer(Trunc(Value));
    SQLT_FLT: pDouble(Fpsli.uValue)^ := Value;
    SQLT_DAT: pORA_Date(Fpsli.uValue)^ := DateTimeToOraDate(Value);
  else
    raise Exception.Create('Not editable field.');
  end;
end;

procedure TOraField.SetAsInteger(const Value: Integer);
var
  sValue: String;
  dValue: double;
begin
  dValue := 0;
  IsNull := false;
  case Fpsli.uValueType of
    SQLT_CHR, SQLT_AFC:
      begin
        sValue := IntToStr(Value);
        if Length(sValue) >= Fpsli.uValueLength then
          Fpsli.uValueLen^ := Fpsli.uValueLength
        else
          Fpsli.uValueLen^ := Length(sValue);
        FillChar(pChar(Fpsli.uValue)^, Fpsli.uValueLength, 0);
        StrMove(pChar(Fpsli.uValue), pChar(sValue), Fpsli.uValueLen^);
      end;
    SQLT_INT: pInteger(Fpsli.uValue)^ := Value;
    SQLT_FLT: pDouble(Fpsli.uValue)^ := dValue + Value;
    SQLT_DAT: pORA_Date(Fpsli.uValue)^ := DateTimeToOraDate(dValue + Value);
  else
    raise Exception.Create('Not editable field.');
  end;
end;

procedure TOraField.SetAsString(const Value: String);
begin
  IsNull := false;
  case Fpsli.uValueType of
    SQLT_CHR, SQLT_AFC:
      begin
        if Length(Value) >= Fpsli.uValueLength then
          Fpsli.uValueLen^ := Fpsli.uValueLength
        else
          Fpsli.uValueLen^ := Length(Value);
        FillChar(pChar(Fpsli.uValue)^, Fpsli.uValueLength, 0);
        StrMove(pChar(Fpsli.uValue), pChar(Value), Fpsli.uValueLen^);
      end;
    SQLT_INT: pInteger(Fpsli.uValue)^ := StrToInt(Value);
    SQLT_FLT: pDouble(Fpsli.uValue)^ := StrToFloat(Value);
    SQLT_DAT: pORA_Date(Fpsli.uValue)^ := DateTimeToOraDate(StrToDateTime(Value));
  else
    raise Exception.Create('Not editable field.');
  end;
end;

procedure TOraField.SetAsVariant(const Value: Variant);
begin
  AsString := Value;
end;

procedure TOraField.SetIsNull(const Value: boolean);
begin
  if OraStmt.State = dsEdit then
  begin
    GetSelectListItem;
    if Value then
      Fpsli.uValueInd^ := -1
    else
      Fpsli.uValueInd^ := 0;
  end else
    raise Exception.Create('o8Stmt not in Edit mode.');
end;

{ TOraLobStream }

constructor TOraLobStream.Create(oField: TOraField);
begin
  inherited Create;
  FField := oField;
  FParam := nil;
  Fo8Stmt := nil;
  envhp := FField.hEnv;
  srvhp := FField.hServ;
  errhp := FField.hErr;
  usrhp := FField.hSess;
  svchp := FField.hSvc;
  pLob := nil;
  FPosition := 0;
  LongField := false;
  if oField.SelectListItem.uValueType in [SQLT_BLOB, SQLT_CLOB, SQLT_CFILE, SQLT_BFILE] then
    if oField.SelectListItem.uValueInd^ <> -1 then
    begin
      pLob := (ppOCILobLocator(oField.SelectListItem.uValue))^;
      if oField.SelectListItem.uValueType in [SQLT_CFILE, SQLT_BFILE] then
        OCILobFileOpen (   svchp,
                           errhp,
                           pLob,
                           OCI_FILE_READONLY);
    end;
  if oField.SelectListItem.uValueType in [SQLT_LNG, SQLT_LBI] then
    if oField.SelectListItem.uValueInd^ <> -1 then
      LongField := true;
end;

constructor TOraLobStream.Create(oPar: TOraParam);
begin
  inherited Create;
  FField := nil;
  Fo8Stmt := nil;
  FParam := oPar;
  envhp := FParam.hEnv;
  srvhp := FParam.hServ;
  errhp := FParam.hErr;
  usrhp := FParam.hSess;
  svchp := FParam.hSvc;
  pLob := nil;
  FPosition := 0;
  LongField := false;
  if FParam.DataType in [ptCLob, ptBLob] then
    pLob := FParam.LobLob;
  if FParam.DataType in [ptCFile, ptBFile] then
  begin
    pLob := FParam.LobFile;
    OCILobFileOpen (   svchp,
                       errhp,
                       pLob,
                       OCI_FILE_READONLY);
  end;
end;

constructor TOraLobStream.Create(ParamLob: Pointer; oStmt: To8Stmt;  b_type: sb2);
begin
  inherited Create;
  FField := nil;
  FParam := nil;
  Fo8Stmt := oStmt;
  envhp := Fo8Stmt.FSession.hEnv;
  srvhp := Fo8Stmt.FSession.hServ;
  errhp := Fo8Stmt.FSession.hErr;
  usrhp := Fo8Stmt.FSession.hSess;
  svchp := Fo8Stmt.FSession.hSvc;
  pLob := nil;
  FPosition := 0;
  LongField := false;
  if b_type in [SQLT_CLOB, SQLT_BLOB, SQLT_CFILE, SQLT_BFILE] then
  begin
    pLob := (ppOCILobLocator(ParamLob))^;
    if b_type in [SQLT_CFILE, SQLT_BFILE] then
      OCILobFileOpen (   svchp,
                         errhp,
                         pLob,
                         OCI_FILE_READONLY);
  end else begin
    if b_type in [SQLT_LNG, SQLT_LBI] then
    begin
      LongField := true;
      pLob := pOCILobLocator(ParamLob);
    end;
  end;
end;

destructor TOraLobStream.Destroy;
begin
  if FField <> nil then
    if FField.SelectListItem.uValueType in [SQLT_CFILE, SQLT_BFILE] then
      OCILobFileClose ( svchp,
                        errhp,
                        pLob);
  if FParam <> nil then
    if FParam.DataType in [ptCFile, ptBFile] then
      OCILobFileClose ( svchp,
                        errhp,
                        pLob);
  inherited Destroy;
end;

function TOraLobStream.GetSize: ub4;
var
  IsLobInit: bool;
  r: ub4;
begin
  Result := 0;
  if (not LongField) and (pLob <> nil) then
  begin
    IsLobInit := 0;
    OCILobLocatorIsInit(  hEnv,
                          errhp,
                          pLob,
                          @IsLobInit);
    if IsLobInit <> 0 then
    begin
      r := 0;
      OCILobGetLength(  hSvc,
                        errhp,
                        pLob,
                        @r);
      Result := r;
    end;
  end else if LongField then
  begin
    if FField <> nil then
      Result := FField.SelectListItem.uValueLen^
    else
      Result := Length(pChar(pLob));
  end;
end;

function TOraLobStream.Read(var Buffer; Count: Integer): Longint;
var
  IsLobInit: bool;
  pRead, ReadL: ub4;
begin
  Result := 0;
  if (not LongField) and (pLob <> nil) then
  begin
    IsLobInit := 0;
    OCILobLocatorIsInit(  hEnv,
                          errhp,
                          pLob,
                          @IsLobInit);
    if IsLobInit <> 0 then
    begin
      pRead := Count;
      ReadL := Count;
      OCILobRead( hSvc,
                  errhp,
                  pLob,
                  @pRead,
                  FPosition + 1,
                  Pointer(@Buffer),
                  ReadL,
                  nil,
                  nil,
                  0,
                  SQLCS_IMPLICIT);
      FPosition := FPosition + pRead;
      Result := pRead;
    end;// else
      //FField.OraStmt.GetError;
  end else if LongField then
  begin
    ReadL := Count;
    if (FPosition + ub4(Count)) > ub4(Size) then
      ReadL := Size - FPosition;
    if FField <> nil then
      Move(Pointer(ub4(FField.SelectListItem.uValue) + FPosition)^, Buffer, ReadL)
    else
      Move(Pointer(pChar(pLob) + FPosition)^, Buffer, ReadL);
    FPosition := FPosition + ub4(ReadL);
    Result := ReadL;
  end;
end;

function TOraLobStream.Seek(Offset: Integer; Origin: Word): Longint;
begin
  case Origin of
    soFromBeginning: FPosition := Offset;
    soFromCurrent: Inc(FPosition, Offset);
    soFromEnd: FPosition := Size + ub4(Offset);
  end;
  Result := FPosition;
end;

procedure TOraLobStream.SetNewSize(const Value: ub4);
begin
  SetSize(Integer(Value));
end;

procedure TOraLobStream.SetSize(NewSize: Integer);
var
  IsLobInit: bool;
  r: ub4;
begin
  if (not LongField) and (pLob <> nil) then
  begin
    IsLobInit := 0;
    OCILobLocatorIsInit(  hEnv,
                          errhp,
                          pLob,
                          @IsLobInit);
    if IsLobInit <> 0 then
    begin
      r := ub4(NewSize);
      OCILobTrim( hSvc,
                  errhp,
                  pLob,
                  r);
    end;
  end else if LongField then
  begin
    if FField <> nil then
      FField.SelectListItem.uValueLen^ := NewSize
    else
      pChar(pLob)[NewSize] := #0;
  end;
end;

function TOraLobStream.Write(const Buffer; Count: Integer): Longint;
var
  IsLobInit: bool;
  pRead, ReadL: ub4;
begin
  Result := 0;
  if (not LongField) and (pLob <> nil) then
  begin
    IsLobInit := 0;
    OCILobLocatorIsInit(  hEnv,
                          errhp,
                          pLob,
                          @IsLobInit);
    if IsLobInit <> 0 then
    begin
      pRead := Count;
      ReadL := Count;
      OCILobWrite(  hSvc,
                    errhp,
                    pLob,
                    @pRead,
                    FPosition + 1,
                    Pointer(@Buffer),
                    ReadL,
                    OCI_ONE_PIECE,
                    nil,
                    nil,
                    0,
                    SQLCS_IMPLICIT);
      FPosition := FPosition + pRead;
      Result := pRead;
    end else
      FField.OraStmt.GetError;
  end else if LongField then
  begin
    if (FPosition + ub4(Count)) > ub4(Size) then
      Size := FPosition + ub4(Count);
    if FField <> nil then
      Move(Buffer, Pointer(ub4(FField.SelectListItem.uValue) + FPosition)^, Count)
    else
      Move(Buffer, Pointer(pChar(pLob) + FPosition)^, Count);
    FPosition := FPosition + ub4(Count);
    Result := Count;
  end;
end;

{ TOraEmptyStream }

function TOraEmptyStream.GetSize: ub4;
begin
  Result := 0;
end;

end.
