{Copyright:      Vlad Karpov  mailto:KarpovVV@protek.ru
 Author:         Vlad Karpov
 Remarks:        freeware, but this Copyright must be included

 Description:    To8StmtOrd - class for execute SQL mechanism

 Version:        1.0.0

 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ''AS IS'' AND ANY EXPRESS
 * OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 * BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
}
unit o8StmtOrd;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, OCI, o8Env, o8Session, o8Stmt, o8Util, MemMgr, Streams;

const

   MAX_LONG_SIZE = 8*1024;

type

  TOnFetch = procedure (Sender: TObject; RowsFetched: Integer; var EOFflag: boolean) of object;

  TDescriptors = class
  public
    Descriptor: pOCIDescriptor;
    DescType: ub4;
  end;

  To8StmtOrd = class(To8Stmt)
  private
    { Private declarations }

    FMaxLongSize: ub4;

    FSQL: TStrings;

    FPortion: ub4;

    FetchBuffer: Pointer;
    IndBuffer: psb2;
    LenBuffer: pub2;
    RCodeBuffer: pub2;

    FResultSetStream: TResultSetStream;

    FFetchBuffer: String;
    FIndBuffer: String;
    FLenBuffer: String;
    FRCodeBuffer: String;

    FetchBufferStream: TAllStream;
    IndBufferStream: TAllStream;
    LenBufferStream: TAllStream;
    RCodeBufferStream: TAllStream;

    CurrGranule: ub4;
    LastGranule: ub4;
    RowsDone: ub4;
    FExec: boolean;
    FStmtType: ub4;

    FSQLText: String;
    FParamCheck: Boolean;

    FOnFetch: TOnFetch;

    LastSizeStream: Integer;

    procedure QueryChanged(Sender: TObject);
    procedure SetSQL(NewSQL: TStrings);
    procedure WriteBuffers(LastR: ub4);
    procedure Fetch;
    function GetSQL: TStrings;
    function GetLastRowId: pOCIRowid;

  protected
    { Protected declarations }

    procedure InternalExecute; override;
    procedure SetOrdExternalType(rs: pRStruct);

    procedure DescriptorsAlloc; override;
    procedure DescriptorsFree; override;

    function GetRawBuffer: PChar; override;
    function GetRawInd: psb2; override;
    function GetRawRCode: pub2; override;
    function GetRawRlen: pub2; override;
    procedure FlashCurrentRecord; override;
    procedure CheckEdit;

    procedure SetArrayOfPar(const Value: Boolean); override;
    function GetArrayParValue(Index1, Index2: Integer): Variant; override;
    procedure SetArrayParValue(Index1, Index2: Integer; const Value: Variant); override;
    function GetArrayParString(Index1, Index2: Integer): String; override;
    procedure SetArrayParString(Index1, Index2: Integer; const Value: String); override;
    function GetArrayParInteger(Index1, Index2: Integer): Integer; override;
    procedure SetArrayParInteger(Index1, Index2: Integer; const Value: Integer); override;
    function GetArrayParFloat(Index1, Index2: Integer): double; override;
    procedure SetArrayParFloat(Index1, Index2: Integer; const Value: double); override;
    function GetArrayParDateTime(Index1, Index2: Integer): TDateTime; override;
    procedure SetArrayParDateTime(Index1, Index2: Integer; const Value: TDateTime); override;
    function GetArrayParNull(Index1, Index2: Integer): Boolean; override;
    procedure SetArrayParNull(Index1, Index2: Integer; const Value: Boolean); override;
    function GetArrayParRCode(Index1, Index2: Integer): ub2; override;
    function GetResultSetStream: TResultSetStream; override;
    procedure SetResultSetStream(const Value: TResultSetStream); override;

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure ReExecute;
    procedure Open; override;
    procedure Prepare; override;
    procedure Close; override;
    procedure FetchAll; override;
    procedure MoveTo(NewRec: ub4); override;
    procedure MoveToFast(NewRec: ub4); override;
    procedure MoveBy(SkipRec: sb4); override;
    procedure First; override;
    procedure Next; override;
    procedure Prior; override;
    procedure Last; override;
    procedure Edit; override;
    procedure Post; override;
    function RawFieldGet(position: sword): pSelectListItem; override;
    function RawFieldStructGet(position: sword): pRStruct; override;
    function GetBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; overload; override;
    function GetBlobStream(ParamLob: Pointer; b_type: sb2): TStream; overload; override;
    function GetFetchRowsCount: ub4; override;

    property StmtType: ub4 read FStmtType;
    property LastRowId: pOCIRowid read GetLastRowId;

  published

    property Active;
    property oraSession;
    property SQL: TStrings read GetSQL write SetSQL;
    property Params;
    property ParamCheck: Boolean read FParamCheck write FParamCheck default True;
    property ArrayOfPar;
    property MaxLongSize: ub4 read FMaxLongSize write FMaxLongSize default MAX_LONG_SIZE;
    property Portion: ub4 read FPortion write FPortion default 100;
    property CursorSource;
    property CursorIndex;
    property DataSetSource;
    property DataSetFieldName;
    property Exec;
    property ResultSetStream;

    property OnBeforeExecute;
    property OnAfterExecute;
    property OnFetch: TOnFetch read FOnFetch write FOnFetch;

  end;

procedure Register;

implementation

uses
   o8DataSet;

//******************************************************************************
procedure Register;
begin
  RegisterComponents('VK NLO', [To8StmtOrd]);
end;

//******************************************************************************
constructor To8StmtOrd.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FSQL := TStringList.Create;
   TStringList(SQL).OnChange := QueryChanged;
   FResultSetStream := rsToMemory;
   FetchBufferStream  := nil; //TMemoryStream.Create;
   IndBufferStream    := nil; //TMemoryStream.Create;
   LenBufferStream    := nil; //TMemoryStream.Create;
   RCodeBufferStream  := nil; //TMemoryStream.Create;
   LastSizeStream := 0;
   FPortion := 100;
   FetchBuffer := nil;
   IndBuffer   := nil;
   LenBuffer   := nil;
   RCodeBuffer := nil;
   FExec := false;
   MaxLongSize := MAX_LONG_SIZE;
   PrepareStatus := false;
   FParamCheck := true;
   RowsDone := 0;
   ResultSetStream := FResultSetStream;
end;

//******************************************************************************
destructor To8StmtOrd.Destroy;
begin

  //
  LastSizeStream := 0;
  //
  //FetchBufferStream.Size := 0;
  //IndBufferStream.Size := 0;
  //LenBufferStream.Size := 0;
  //RCodeBufferStream.Size := 0;
  //FetchBufferStream.Clear;
  //IndBufferStream.Clear;
  //LenBufferStream.Clear;
  //RCodeBufferStream.Clear;

  FetchBufferStream.Destroy;
  IndBufferStream.Destroy;
  LenBufferStream.Destroy;
  RCodeBufferStream.Destroy;

  DeleteFile(FFetchBuffer);
  DeleteFile(FIndBuffer);
  DeleteFile(FLenBuffer);
  DeleteFile(FRCodeBuffer);

  if FetchBuffer <> nil then MemMgr.oMem.FreeMem(FetchBuffer);
  if IndBuffer   <> nil then MemMgr.oMem.FreeMem(PChar(IndBuffer));
  if LenBuffer   <> nil then MemMgr.oMem.FreeMem(PChar(LenBuffer));
  if RCodeBuffer <> nil then MemMgr.oMem.FreeMem(PChar(RCodeBuffer));
  FetchBuffer := nil;
  IndBuffer   := nil;
  LenBuffer   := nil;
  RCodeBuffer := nil;
  FSQL.Destroy;
  inherited Destroy;
end;

//******************************************************************************
procedure To8StmtOrd.Open;
begin
  {$B+}
  if not (AllocHandlerStmt and AllocHandlerErr) then AllocHandlers;
  if AllocHandlerStmt and AllocHandlerErr then begin
  {$B-}
    Prepare;
    InternalExecute;
    if ( GetErrCode = 0 ) or ( GetErrCode = 1403 ) then {No data found}
       if FStmtType = OCI_STMT_SELECT then
          inherited Open
       else
          inherited Close
    else
       inherited Close;
  end else GetError(true);
end;

//******************************************************************************
procedure To8StmtOrd.Close;
begin
  inherited Close;
  //
  DescriptorsFree;
  //
  //
  LastSizeStream := 0;
  //
  //FetchBufferStream.Size := 0;
  //IndBufferStream.Size := 0;
  //LenBufferStream.Size := 0;
  //RCodeBufferStream.Size := 0;
  //FetchBufferStream.Clear;
  //IndBufferStream.Clear;
  //LenBufferStream.Clear;
  //RCodeBufferStream.Clear;
  //
  if FetchBuffer <> nil then MemMgr.oMem.FreeMem(FetchBuffer);
  if IndBuffer   <> nil then MemMgr.oMem.FreeMem(PChar(IndBuffer));
  if LenBuffer   <> nil then MemMgr.oMem.FreeMem(PChar(LenBuffer));
  if RCodeBuffer <> nil then MemMgr.oMem.FreeMem(PChar(RCodeBuffer));
  FetchBuffer := nil;
  IndBuffer   := nil;
  LenBuffer   := nil;
  RCodeBuffer := nil;
  //
end;

//******************************************************************************
procedure To8StmtOrd.SetSQL(NewSQL: TStrings);
begin
   FSQL.Clear;
   FSQL.Assign(NewSQL);
   PrepareStatus := false;
end;

function To8StmtOrd.GetSQL: TStrings;
begin
  Result := FSQL;
  PrepareStatus := false;
end;

//******************************************************************************
procedure To8StmtOrd.MoveToFast(NewRec: ub4);
var
  CurBuf: ub4;
  RealRead: Longint;
begin
  if RowsDone > 0 then
  begin
    BofInternal := false;
    EofInternal := false;
    CheckEdit;
    //
    if NewRec > RowsDone then
      NewRec := RowsDone;
    CurBuf := (RecNoInternal - 1) div Portion;
    if CurBuf <> ((NewRec - 1) div Portion) then
    begin

      CurBuf := ((NewRec - 1) div Portion);

      FetchBufferStream.Seek(CurBuf * FRecordSize * Portion, soFromBeginning);
      IndBufferStream.Seek  (CurBuf * FFieldCount * sizeof(sb2) * Portion,  soFromBeginning);
      LenBufferStream.Seek  (CurBuf * FFieldCount * sizeof(ub2) * Portion,  soFromBeginning);
      RCodeBufferStream.Seek(CurBuf * FFieldCount * sizeof(ub2) * Portion,  soFromBeginning);

      RealRead := FetchBufferStream.Read(FetchBuffer^, Portion*FRecordSize);
      IndBufferStream.Read  (IndBuffer^,   Portion*FFieldCount*sizeof(sb2));
      LenBufferStream.Read  (LenBuffer^,   Portion*FFieldCount*sizeof(ub2));
      RCodeBufferStream.Read(RCodeBuffer^, Portion*FFieldCount*sizeof(ub2));

      LastGranule := RealRead div FRecordSize;

    end;
    CurrGranule := NewRec - (CurBuf * Portion) - 1;
    RecNoInternal := NewRec;
  end else
    RecNoInternal := 1;
end;

//******************************************************************************
procedure To8StmtOrd.MoveTo(NewRec: ub4);
begin
  if NewRec > RowsDone then
  begin
    BofInternal := false;
    EofInternal := false;
    CheckEdit;
    if NewRec < ub4(Abs(RecNoInternal - NewRec)) then
    begin
      First;
      while ( RecNoInternal < NewRec ) and not Eof do Next;
    end else begin
      if RecNoInternal > NewRec then
        while ( RecNoInternal > NewRec ) and not Bof do Prior
      else
        while ( RecNoInternal < NewRec ) and not Eof do Next;
    end;
  end else
    MoveToFast(NewRec);
end;

//******************************************************************************
procedure To8StmtOrd.MoveBy(SkipRec: sb4);
begin
  CheckEdit;
end;

//******************************************************************************
procedure To8StmtOrd.First;
var
   RealRead: LongInt;
begin
   BofInternal := false;
   EofInternal := false;
   CheckEdit;
   if oraSession <> nil then
   begin
      if oraSession.Active then
      begin
         {$B+}
         if AllocHandlerStmt and AllocHandlerErr then
         begin
            if FExec and ( FStmtType = OCI_STMT_SELECT ) then
            begin
         {$B-}
               FetchBufferStream.Seek(0, soFromBeginning);
               IndBufferStream.Seek  (0,  soFromBeginning);
               LenBufferStream.Seek  (0,  soFromBeginning);
               RCodeBufferStream.Seek(0,  soFromBeginning);
               RealRead := FetchBufferStream.Read(FetchBuffer^, Portion*FRecordSize);
               IndBufferStream.Read  (IndBuffer^,   Portion*FFieldCount*sizeof(sb2));
               LenBufferStream.Read  (LenBuffer^,   Portion*FFieldCount*sizeof(ub2));
               RCodeBufferStream.Read(RCodeBuffer^, Portion*FFieldCount*sizeof(ub2));
               LastGranule := RealRead div FRecordSize;
               if LastGranule <> 0 then
               begin
                  CurrGranule := 0;
                  RecNoInternal := 1;
               end else
               begin
                  EofInternal := true;
                  BofInternal := true;
               end;
            end else
            begin
               EofInternal := true;
               BofInternal := true;
            end;
         end else
         begin
            EofInternal := true;
            BofInternal := true;
         end;
      end else
      begin
         EofInternal := true;
         BofInternal := true;
      end;
   end else
   begin
      EofInternal := true;
      BofInternal := true;
   end;
end;

//******************************************************************************
procedure To8StmtOrd.Next;
var
   RealRead: LongInt;
begin
   BofInternal := false;
   EofInternal := false;
   CheckEdit;
   if oraSession <> nil then
   begin
      if oraSession.Active then
      begin
         {$B+}
         if AllocHandlerStmt and AllocHandlerErr then
         begin
            if FExec and ( FStmtType = OCI_STMT_SELECT ) then
            begin
         {$B-}
               CurrGranule := CurrGranule + 1;
               RecNoInternal := RecNoInternal + 1;
               if CurrGranule >= LastGranule then
               begin
                  if FetchBufferStream.Seek(0, soFromCurrent) < LastSizeStream then
                  begin
                     RealRead := FetchBufferStream.Read(FetchBuffer^, Portion*FRecordSize);
                     IndBufferStream.Read  (IndBuffer^,   Portion*FFieldCount*sizeof(sb2));
                     LenBufferStream.Read  (LenBuffer^,   Portion*FFieldCount*sizeof(ub2));
                     RCodeBufferStream.Read(RCodeBuffer^, Portion*FFieldCount*sizeof(ub2));
                     if RealRead <> 0 then
                     begin
                        CurrGranule := 0;
                        LastGranule := RealRead div FRecordSize;
                     end else Fetch;
                  end else Fetch;
               end;
            end else
            begin
               EofInternal := true;
               BofInternal := true;
            end;
         end else
         begin
            EofInternal := true;
            BofInternal := true;
         end;
      end else
      begin
         EofInternal := true;
         BofInternal := true;
      end;
   end
   else
   begin
      EofInternal := true;
      BofInternal := true;
   end;
end;

//******************************************************************************
procedure To8StmtOrd.FetchAll;
var
  fp1, fp2, fp3, fp4: Longint;
  b1: Pointer;
  b2: psb2;
  b3: pub2;
  b4: pub2;
  lg: ub4;
begin

  fp1 := FetchBufferStream.Seek(0, soFromCurrent);
  fp2 := IndBufferStream.Seek(0, soFromCurrent);
  fp3 := LenBufferStream.Seek(0, soFromCurrent);
  fp4 := RCodeBufferStream.Seek(0, soFromCurrent);

  b1 := Pointer(MemMgr.oMem.GetMem(self, FPortion*FRecordSize));
  b2 := psb2(MemMgr.oMem.GetMem(self, FPortion*FFieldCount*sizeof(sb2)));
  b3 := pub2(MemMgr.oMem.GetMem(self, FPortion*FFieldCount*sizeof(ub2)));
  b4 := pub2(MemMgr.oMem.GetMem(self, FPortion*FFieldCount*sizeof(ub2)));

  Move(FetchBuffer^ , b1^, FPortion*FRecordSize);
  Move(IndBuffer^   , b2^, FPortion*FFieldCount*sizeof(sb2));
  Move(LenBuffer^   , b3^, FPortion*FFieldCount*sizeof(ub2));
  Move(RCodeBuffer^ , b4^, FPortion*FFieldCount*sizeof(ub2));

  FetchBufferStream.Seek(0, soFromEnd);
  IndBufferStream.Seek(0, soFromEnd);
  LenBufferStream.Seek(0, soFromEnd);
  RCodeBufferStream.Seek(0, soFromEnd);

  //
  EofInternal := false;
  while not EofInternal do
  begin
    DescriptorsAlloc;
    if OCIStmtFetch(  stmthp,
                      errhp,
                      Portion,
                      OCI_FETCH_NEXT,
                      OCI_DEFAULT) in [OCI_SUCCESS, OCI_SUCCESS_WITH_INFO] then
    begin
      RowsDone := RowsDone + Portion;
      WriteBuffers(Portion);
    end else begin
      if ErrNum = 1403 then
      begin
        OCIAttrGet(stmthp, OCI_HTYPE_STMT, @lg, nil, OCI_ATTR_ROW_COUNT, errhp);
        lg := lg - RowsDone;
        RowsDone := RowsDone + lg;
        if lg = 0 then
          EofInternal := true
        else
          WriteBuffers(lg);
      end else
        EofInternal := true;
    end;
  end;
  //

  Move(b1^, FetchBuffer^ , FPortion*FRecordSize);
  Move(b2^, IndBuffer^   , FPortion*FFieldCount*sizeof(sb2));
  Move(b3^, LenBuffer^   , FPortion*FFieldCount*sizeof(ub2));
  Move(b4^, RCodeBuffer^ , FPortion*FFieldCount*sizeof(ub2));

  MemMgr.oMem.FreeMem(b1);
  MemMgr.oMem.FreeMem(PChar(b2));
  MemMgr.oMem.FreeMem(PChar(b3));
  MemMgr.oMem.FreeMem(PChar(b4));

  FetchBufferStream.Seek(fp1, soFromBeginning);
  IndBufferStream.Seek  (fp2, soFromBeginning);
  LenBufferStream.Seek  (fp3, soFromBeginning);
  RCodeBufferStream.Seek(fp4, soFromBeginning);

end;

//******************************************************************************
procedure To8StmtOrd.Fetch;
var
  EOFFlag: Boolean;
begin
  DescriptorsAlloc;
  if OCIStmtFetch(  stmthp,
                    errhp,
                    Portion,
                    OCI_FETCH_NEXT,
                    OCI_DEFAULT) in [OCI_SUCCESS, OCI_SUCCESS_WITH_INFO] then
  begin
    CurrGranule := 0;
    LastGranule := Portion;
    RowsDone := RowsDone + Portion;
    WriteBuffers(LastGranule);
  end else begin
    if ErrNum = 1403 then
    begin
      CurrGranule := 0;
      OCIAttrGet(stmthp, OCI_HTYPE_STMT, @LastGranule, nil, OCI_ATTR_ROW_COUNT, errhp);
      LastGranule := LastGranule - RowsDone;
      RowsDone := RowsDone + LastGranule;
      if LastGranule = 0 then
      begin
        LastGranule := Portion;
        CurrGranule := LastGranule - 1;
        RecNoInternal := RecNoInternal - 1;
        EofInternal := true;
      end else begin
        WriteBuffers(LastGranule);
      end;
    end else begin
      CurrGranule := CurrGranule - 1;
      RecNoInternal := RecNoInternal - 1;
      EofInternal := true;
    end;
  end;
  if Assigned(OnFetch) then
  begin
    EOFFlag := false;
    OnFetch(self, RowsDone, EOFFlag);
    if EOFFlag then
      EofInternal := true;
  end;
end;

//******************************************************************************
procedure To8StmtOrd.Prior;
var
   LastEOF, LastBOF: boolean;
   RealRead: LongInt;
begin
   LastEOF := EofInternal;
   LastBOF := BofInternal;
   BofInternal := false;
   EofInternal := false;
   CheckEdit;
   if oraSession <> nil then
   begin
      if oraSession.Active then
      begin
         {$B+}
         if AllocHandlerStmt and AllocHandlerErr then
         begin
            if FExec and ( FStmtType = OCI_STMT_SELECT ) then
            begin
         {$B-}
               if not LastEOF then
               begin
                  RecNoInternal := RecNoInternal - 1;
                  if RecNoInternal > 0 then
                  begin
                     if CurrGranule <> 0 then CurrGranule := CurrGranule - 1 else
                     begin

                        FetchBufferStream.Seek(((RecNoInternal - 1) div Portion) * FRecordSize * Portion, soFromBeginning);
                        IndBufferStream.Seek  (((RecNoInternal - 1) div Portion) * FFieldCount * sizeof(sb2) * Portion,  soFromBeginning);
                        LenBufferStream.Seek  (((RecNoInternal - 1) div Portion) * FFieldCount * sizeof(ub2) * Portion,  soFromBeginning);
                        RCodeBufferStream.Seek(((RecNoInternal - 1) div Portion) * FFieldCount * sizeof(ub2) * Portion,  soFromBeginning);

                        RealRead := FetchBufferStream.Read(FetchBuffer^, Portion*FRecordSize);
                        IndBufferStream.Read  (IndBuffer^,   Portion*FFieldCount*sizeof(sb2));
                        LenBufferStream.Read  (LenBuffer^,   Portion*FFieldCount*sizeof(ub2));
                        RCodeBufferStream.Read(RCodeBuffer^, Portion*FFieldCount*sizeof(ub2));

                        LastGranule := RealRead div FRecordSize;
                        CurrGranule := LastGranule - 1;

                     end;
                  end else begin
                     RecNoInternal := 1;
                     BofInternal := true;
                  end;
               end else
                  BofInternal := LastBOF;
            end else
            begin
               EofInternal := true;
               BofInternal := true;
            end;
         end else
         begin
            EofInternal := true;
            BofInternal := true;
         end;
      end else
      begin
         EofInternal := true;
         BofInternal := true;
      end;
   end else
   begin
      EofInternal := true;
      BofInternal := true;
   end;
end;

//******************************************************************************
procedure To8StmtOrd.Last;
begin
  CheckEdit;
  while not Eof do Next;
end;

//******************************************************************************
function To8StmtOrd.RawFieldGet(position: sword): pSelectListItem;
var
  aCol: pRStruct;
begin
  aCol := pRStruct(FieldDescribe.Items[position-1]);
  SLI.uValue := Pointer(PChar(FetchBuffer) + aCol.S_OFFSET + FRecordSize*CurrGranule);
  SLI.uValueType := aCol.S_EXTERNALTYPE;
  SLI.uValueLength := aCol.S_LENGTH;
  SLI.uValueDBsize := aCol.S_DBSIZE;
  SLI.uValueInd   := psb2(PChar(IndBuffer) + (FFieldCount*CurrGranule + ub4(position-1))*sizeof(sb2));
  SLI.uValueLen   := pub2(PChar(LenBuffer) + (FFieldCount*CurrGranule + ub4(position-1))*sizeof(ub2));
  SLI.uValueRCode := pub2(PChar(RCodeBuffer) + (FFieldCount*CurrGranule + ub4(position-1))*sizeof(ub2));
  Result :=  @SLI;
end;

//******************************************************************************
procedure To8StmtOrd.WriteBuffers(LastR: ub4);
begin
  FetchBufferStream.Write(FetchBuffer^, LastR*FRecordSize );
  LastSizeStream := LastSizeStream + Integer(LastR*FRecordSize);
  IndBufferStream.Write  (IndBuffer^,   LastR*FFieldCount*sizeof(sb2) );
  LenBufferStream.Write  (LenBuffer^,   LastR*FFieldCount*sizeof(ub2) );
  RCodeBufferStream.Write(RCodeBuffer^, LastR*FFieldCount*sizeof(ub2) );
end;

//******************************************************************************
function To8StmtOrd.RawFieldStructGet(position: sword): pRStruct;
begin
   Result := pRStruct(FieldDescribe.Items[position]);
end;

//******************************************************************************
function To8StmtOrd.GetRawBuffer: PChar;
begin
   Result := PChar(FetchBuffer) + FRecordSize*CurrGranule;
end;

//******************************************************************************
function To8StmtOrd.GetRawInd: psb2;
begin
   Result := psb2(PChar(IndBuffer) + FFieldCount*CurrGranule*sizeof(sb2));
end;

//******************************************************************************
function To8StmtOrd.GetRawRCode: pub2;
begin
   Result := pub2(PChar(RCodeBuffer) + FFieldCount*CurrGranule*sizeof(ub2));
end;

//******************************************************************************
function To8StmtOrd.GetRawRlen: pub2;
begin
   Result := pub2(PChar(LenBuffer) + FFieldCount*CurrGranule*sizeof(ub2));
end;

//******************************************************************************
function To8StmtOrd.GetBlobStream(Field: TField; Mode: TBlobStreamMode): TStream;
begin
   Result := TOraLobStream.Create(TOraField.Create(self, Field.FieldNo));
end;

//******************************************************************************
function To8StmtOrd.GetBlobStream(ParamLob: Pointer; b_type: sb2): TStream;
begin
  Result := TOraLobStream.Create(ParamLob, self, b_type);
end;

//******************************************************************************
procedure To8StmtOrd.SetOrdExternalType(rs: pRStruct);
begin
   case rs.S_INERNALTYPE of
      SQLT_CHR:       {1      (ORANET TYPE) character string }
         begin
            rs.S_LENGTH := rs.S_DBSIZE;
            rs.S_EXTERNALTYPE := SQLT_CHR;
         end;
      SQLT_NUM:       {2      (ORANET TYPE) oracle numeric }
         begin
            if (rs.S_SCALE = 0) and (rs.S_PREC = 0) then
            begin
               rs.S_LENGTH := sizeof(double);
               rs.S_EXTERNALTYPE := SQLT_FLT;
            end
            else
            begin
               if rs.S_SCALE = 0 then
               begin
                  rs.S_LENGTH := sizeof(ub4);
                  rs.S_EXTERNALTYPE := SQLT_INT;
               end
               else
               begin
                  rs.S_LENGTH := sizeof(double);
                  rs.S_EXTERNALTYPE := SQLT_FLT;
               end;
            end;
         end;
      SQLT_DAT:       {12  date in oracle format}
         begin
            rs.S_LENGTH := 7;
            rs.S_EXTERNALTYPE := SQLT_DAT;
         end;
      SQLT_BIN:       {23  binary data(DTYBIN)}
         begin
            rs.S_LENGTH := rs.S_DBSIZE;
            rs.S_EXTERNALTYPE := SQLT_BIN;
         end;
      SQLT_AFC:       {96  Ansi fixed char }
         begin
            rs.S_LENGTH := rs.S_DBSIZE;
            rs.S_EXTERNALTYPE := SQLT_AFC;
         end;
      SQLT_RDD:
         begin
            rs.S_LENGTH := sizeof(pOCIRowid);
            rs.S_EXTERNALTYPE := SQLT_RDD;
         end;
      SQLT_LNG:
         begin
            rs.S_LENGTH := MaxLongSize;
            rs.S_EXTERNALTYPE := SQLT_LNG;
         end;
      SQLT_LBI:
         begin
            rs.S_LENGTH := MaxLongSize;
            rs.S_EXTERNALTYPE := SQLT_LBI;
         end;
      SQLT_BLOB, SQLT_CLOB, SQLT_CFILE, SQLT_BFILE:
         begin
            rs.S_LENGTH := sizeof(pOCILobLocator);
            rs.S_EXTERNALTYPE := rs.S_INERNALTYPE;
         end;
      SQLT_RSET:
        begin
            rs.S_LENGTH := sizeof(pOCIResult);
            rs.S_EXTERNALTYPE := rs.S_INERNALTYPE;
        end;
      SQLT_NTY:
        begin
            rs.S_LENGTH := SizeOf(Pointer);
            rs.S_EXTERNALTYPE := rs.S_INERNALTYPE;
        end;
   end;
end;

procedure To8StmtOrd.Prepare;
var
  k: Integer;
begin
  inherited Prepare;
  if (CursorParam = nil) and (DataSetField = nil) and (ResultSet = nil) then
    if not PrepareStatus then
    begin
      FSQLText := SQL.Text;
      for k := 1 to Length(FSQLText) do if FSQLText[k] = #$D then FSQLText[k] := #$20;
      PrepareStatus := OCIStmtPrepare(  stmthp,
                                        errhp,
                                        PChar(FSQLText),
                                        Length(FSQLText),
                                        OCI_NTV_SYNTAX,
                                        OCI_DEFAULT) in [OCI_SUCCESS, OCI_SUCCESS_WITH_INFO];
    end;
end;

procedure To8StmtOrd.InternalExecute;
var
   i: sword;
   j: sword;
   Par: TOraParam;
   position: sword;
   colhd: pOCIParam;
   defhp: pOCIDefine;
   aCol: pRStruct;
   cc: PChar;
   pDef: Pointer;
   pInd: Pointer;
   //pvszsp: pub4;
   ExecStatus: boolean;
   LastExec: boolean;
   DescriptorsIsHear: boolean;
   FetchResult: sword;
   sExec: sword;
begin
  //
  inherited InternalExecute;
  //
  //
  LastSizeStream := 0;
  //
  FetchBufferStream.qq;
  IndBufferStream.qq;
  LenBufferStream.qq;
  RCodeBufferStream.qq;

  //
  colhd := nil;
  defhp := nil;
  {$B+}
  if AllocHandlerStmt and AllocHandlerErr then
  {$B-}
  begin
    if not PrepareStatus then Prepare;
    //
    if (CursorParam = nil) and (DataSetField = nil) and (ResultSet = nil) then
      if FArrayOfPar then
      begin
        for j := 0 to Params.Count - 1 do
        begin
         Par := TOraParam(Params.Items[j]);
         OCIBindByName( stmthp,
                        Par.ParBindpp,
                        errhp,
                        pChar(':' + Par.Name),
                        Length(Par.Name) + 1,
                        Pointer(pChar(FParamBuffer) + Par.ParOffset),
                        Par.ParSize,  //Par.MaxLength,
                        Par.ExternalType,
                        Pointer(pChar(FParamIndBuffer) + j*sb4(FPortion)*sizeof(sb2)),
                        pub2(pChar(FParamLenBuffer) + j*sb4(FPortion)*sizeof(ub2)),
                        pub2(pChar(FParamRCodeBuffer) + j*sb4(FPortion)*sizeof(ub2)),
                        0,
                        nil,
                        OCI_DEFAULT);
        end;
      end else begin
        for j := 0 to Params.Count - 1 do
        begin
         Par := TOraParam(Params.Items[j]);
         sExec := OCIBindByName( stmthp,
                        Par.ParBindpp,  //@bindp,
                        errhp,
                        pChar(':' + Par.Name),
                        Length(Par.Name) + 1,
                        Par.Buffer,
                        Par.ParSize,  //Par.MaxLength,
                        Par.ExternalType,
                        Pointer(@Par.ParInd),
                        pub2(@Par.ParRealLen),
                        pub2(@Par.ParRCode),
                        0,
                        nil,
                        OCI_DEFAULT);
        end;
      end;
    //
    FStmtType := 0;
    OCIAttrGet(stmthp, OCI_HTYPE_STMT, @FStmtType, nil, OCI_ATTR_STMT_TYPE, errhp);
    if (CursorParam = nil) and (DataSetField = nil) and (ResultSet = nil) then
    begin
      case FStmtType of
        OCI_STMT_SELECT:
           ExecStatus := OCIStmtExecute(  oraSession.hSvc,
                                          stmthp,
                                          errhp,
                                          0,
                                          0,
                                          nil,
                                          nil,
                                          OCI_DESCRIBE_ONLY) in [OCI_SUCCESS, OCI_SUCCESS_WITH_INFO];
      else

        ExecStatus := false;
        while true do
        begin
          if FArrayOfPar then
          begin
            sExec := OCIStmtExecute(  oraSession.hSvc,
                                      stmthp,
                                      errhp,
                                      FLastParamRow + 1,
                                      0,
                                      nil,
                                      nil,
                                      OCI_DEFAULT);
            ExecStatus := (sExec = OCI_SUCCESS)
                          or (sExec = OCI_SUCCESS_WITH_INFO)
                          or (sExec = OCI_STILL_EXECUTING);
            FLastParamRow := 0;
          end else begin
            sExec := OCIStmtExecute(  oraSession.hSvc,
                                      stmthp,
                                      errhp,
                                      1,
                                      0,
                                      nil,
                                      nil,
                                      OCI_DEFAULT);
            ExecStatus := (sExec = OCI_SUCCESS)
                          or (sExec = OCI_SUCCESS_WITH_INFO)
                          or (sExec = OCI_STILL_EXECUTING);
          end;
          if (sExec = OCI_STILL_EXECUTING) and Assigned(OnAfterExecute) then
          begin
            OnAfterExecute(self, sExec);
          end;
          if sExec <> OCI_STILL_EXECUTING then Break;
        end;

      end;
    end else
      ExecStatus := true;
    if ExecStatus then
    begin
      FOraFields.Clear;
      FRecordSize := 0;
      FFieldCount := 0;
      for position := 0 to FieldDescribe.Count - 1 do MemMgr.oMem.FreeMem(PChar(FieldDescribe.Items[position]));
      FieldDescribe.Clear;
      OCIAttrGet(stmthp, OCI_HTYPE_STMT, @FFieldCount, nil, OCI_ATTR_PARAM_COUNT, errhp);
      DescriptorsIsHear := false;
      for position := 1 to FFieldCount do
      begin
         OCIParamGet(stmthp, OCI_HTYPE_STMT, errhp, @colhd, position);
         i := FieldDescribe.Add(pRStruct(MemMgr.oMem.GetMem(self, sizeof(RStruct))));
         aCol := pRStruct(FieldDescribe.Items[i]);
         aCol.S_POS := position - 1;
         aCol.S_DBSIZE := 0;
         aCol.S_INERNALTYPE := 0;
         fillchar(aCol.S_COLNAME, MAX_NAME, 0);
         aCol.S_COLNAMELENGTH := MAX_NAME - 1;
         aCol.S_DISPLAYSIZE := 0;
         aCol.S_PREC := 0;
         aCol.S_SCALE := 0;
         aCol.S_NULLOK := 0;
         aCol.S_EXTERNALTYPE := 0;
         aCol.S_OFFSET := 0;
         aCol.S_LENGTH := 0;
         fillchar(aCol.S_TYPE_NAME, MAX_NAME, 0);
         aCol.S_TDO := nil;
         aCol.S_TYPENAMELENGTH := MAX_NAME - 1;
         OCIAttrGet(colhd, OCI_DTYPE_PARAM, @aCol.S_DBSIZE, nil, OCI_ATTR_DATA_SIZE, errhp);
         OCIAttrGet(colhd, OCI_DTYPE_PARAM, @aCol.S_INERNALTYPE, nil, OCI_ATTR_DATA_TYPE, errhp);
         OCIAttrGet(colhd, OCI_DTYPE_PARAM, @cc, @aCol.S_COLNAMELENGTH, OCI_ATTR_NAME, errhp);
         move(cc^, aCol.S_COLNAME, aCol.S_COLNAMELENGTH);
         OCIAttrGet(colhd, OCI_DTYPE_PARAM, @aCol.S_DISPLAYSIZE, nil, OCI_ATTR_DISP_SIZE, errhp);
         OCIAttrGet(colhd, OCI_DTYPE_PARAM, @aCol.S_PREC, nil, OCI_ATTR_PRECISION, errhp);
         OCIAttrGet(colhd, OCI_DTYPE_PARAM, @aCol.S_SCALE, nil, OCI_ATTR_SCALE, errhp);
         OCIAttrGet(colhd, OCI_DTYPE_PARAM, @aCol.S_NULLOK, nil, OCI_ATTR_IS_NULL, errhp);
         OCIAttrGet(colhd, OCI_DTYPE_PARAM, @cc, @aCol.S_TYPENAMELENGTH, OCI_ATTR_TYPE_NAME, errhp);
         move(cc^, aCol.S_TYPE_NAME, aCol.S_TYPENAMELENGTH);
         if aCol.S_INERNALTYPE = SQLT_NTY then
          OCITypeByName  (  oraSession.hEnv,
                            hErr,
                            oraSession.hSvc,
                            nil, 0,
                            aCol.S_TYPE_NAME, aCol.S_TYPENAMELENGTH,
                            nil, 0,
                            OCI_DURATION_DEFAULT,
                            ord(OCI_TYPEGET_ALL),
                            @aCol.S_TDO);
         aCol.S_OFFSET := FRecordSize;

         SetOrdExternalType(aCol);
         FRecordSize := FRecordSize + aCol.S_LENGTH;
         if aCol.S_EXTERNALTYPE in [  SQLT_BLOB,
                                      SQLT_CLOB,
                                      SQLT_CFILE,
                                      SQLT_BFILE,
                                      SQLT_RSET,
                                      SQLT_RDD] then DescriptorsIsHear := true;

         OCIDescriptorFree(colhd, OCI_DTYPE_PARAM);
         colhd := nil;
         FOraFields.Add(TOraField.Create(self, position));
      end;
      if FFieldCount > 0 then
      begin
         if FetchBuffer <> nil then MemMgr.oMem.FreeMem(FetchBuffer);
         if IndBuffer   <> nil then MemMgr.oMem.FreeMem(PChar(IndBuffer));
         if LenBuffer   <> nil then MemMgr.oMem.FreeMem(PChar(LenBuffer));
         if RCodeBuffer <> nil then MemMgr.oMem.FreeMem(PChar(RCodeBuffer));
         FetchBuffer := nil;
         IndBuffer   := nil;
         LenBuffer   := nil;
         RCodeBuffer := nil;
         if DescriptorsIsHear then Portion := 1;
         FetchBuffer := Pointer(MemMgr.oMem.GetMem(self, FPortion*FRecordSize));
         IndBuffer   := psb2(MemMgr.oMem.GetMem(self, FPortion*FFieldCount*sizeof(sb2)));
         LenBuffer   := pub2(MemMgr.oMem.GetMem(self, FPortion*FFieldCount*sizeof(ub2)));
         RCodeBuffer := pub2(MemMgr.oMem.GetMem(self, FPortion*FFieldCount*sizeof(ub2)));
         for position := 1 to FFieldCount do
         begin
            aCol := pRStruct(FieldDescribe.Items[position-1]);
            pDef := Pointer(PChar(FetchBuffer)+aCol.S_OFFSET);
            pInd := Pointer(PChar(IndBuffer)+aCol.S_POS*sizeof(sb2));
            OCIDefineByPos(   stmthp,
                              @defhp,
                              errhp,
                              aCol.S_POS + 1,
                              pDef,
                              aCol.S_LENGTH,
                              aCol.S_EXTERNALTYPE,
                              pInd,
                              pub2(PChar(LenBuffer)+aCol.S_POS*sizeof(ub2)),
                              pub2(PChar(RCodeBuffer)+aCol.S_POS*sizeof(ub2)),
                              OCI_DEFAULT);
            if aCol.S_INERNALTYPE = SQLT_NTY then
              OCIDefineObject  (   defhp,
                                   errhp,
                                   aCol.S_TDO,
                                   @pDef,
                                   nil,
                                   nil,
                                   nil);
            OCIDefineArrayOfStruct( defhp,
                                    errhp,
                                    ub4(FRecordSize),
                                    ub4(sizeof(sb2)*FFieldCount),
                                    ub4(sizeof(ub2)*FFieldCount),
                                    ub4(sizeof(ub2)*FFieldCount));
         end;
         EofInternal := false;
         BofInternal := false;
         DescriptorsAlloc;
         if (CursorParam = nil) and (DataSetField = nil) and (ResultSet = nil) then
          LastExec := OCIStmtExecute(  oraSession.hSvc,
                                       stmthp,
                                       errhp,
                                       Portion,
                                       0,
                                       nil,
                                       nil,
                                       OCI_DEFAULT) in [OCI_SUCCESS, OCI_SUCCESS_WITH_INFO, OCI_NO_DATA]
         else begin
          FetchResult := OCIStmtFetch(  stmthp,
                            errhp,
                            Portion,
                            OCI_FETCH_FIRST,  //OCI_FETCH_NEXT,
                            OCI_DEFAULT);
          LastExec := FetchResult in [OCI_SUCCESS, OCI_SUCCESS_WITH_INFO, OCI_NO_DATA];
         end;
         if LastExec then
         begin
           CurrGranule := 0;
           LastGranule := 0;
           OCIAttrGet(stmthp, OCI_HTYPE_STMT, @LastGranule, nil, OCI_ATTR_ROW_COUNT, errhp);
           RowsDone := LastGranule;
           FExec := true;
           if LastGranule = 0 then
           begin
            BofInternal := true;
            EofInternal := true;
           end else begin
            WriteBuffers(LastGranule);
            EofInternal := false;
            RecNoInternal := 1;
           end;
         end else begin
          BofInternal := true;
          EofInternal := true;
          GetError(true);
         end;
      end;
    end else GetError(true);
  end else GetError(true);
end;

procedure To8StmtOrd.DescriptorsAlloc;
var
  position, i: sword;
  aCol: pRStruct;
  pDef: pPointer;
  DescType: ub4;
  oDscChr: TDescriptors;
begin
  for position := 1 to FFieldCount do
  begin
    aCol := pRStruct(FieldDescribe.Items[position-1]);
    if aCol.S_EXTERNALTYPE in [ SQLT_BLOB,
                                SQLT_CLOB,
                                SQLT_CFILE,
                                SQLT_BFILE,
                                SQLT_RSET,
                                SQLT_RDD] then
    begin
      DescType := 0;
      if aCol.S_EXTERNALTYPE in [SQLT_BLOB, SQLT_CLOB] then DescType := OCI_DTYPE_LOB;
      if aCol.S_EXTERNALTYPE in [SQLT_CFILE, SQLT_BFILE] then DescType := OCI_DTYPE_FILE;
      if aCol.S_EXTERNALTYPE = SQLT_RSET then DescType := OCI_DTYPE_RSET;
      if aCol.S_EXTERNALTYPE = SQLT_RDD then DescType := OCI_DTYPE_ROWID;
      for i := 0 to Portion - 1 do
      begin
        pDef := pPointer(PChar(FetchBuffer) + aCol.S_OFFSET + (i*FRecordSize));
        if OCIDescriptorAlloc(  oraSession.hEnv,
                                pDef,
                                DescType,
                                0,
                                nil) = OCI_INVALID_HANDLE then
          ShowMessage('   !')
        else begin
          oDscChr := TDescriptors.Create;
          oDscChr.Descriptor := pDef^;
          oDscChr.DescType := DescType;
          FDescriptors.Add(oDscChr);
        end;
      end;
    end;
  end;
end;

procedure To8StmtOrd.DescriptorsFree;
var
  IsDscInit: bool;
  i: Integer;
  pDesc: pOCIDescriptor;
  oDscChr: TDescriptors;
begin
  for i := 0 to FDescriptors.Count - 1 do
  begin
    oDscChr := TDescriptors(FDescriptors.Items[i]);
    pDesc := oDscChr.Descriptor;
    if oDscChr.DescType in [  SQLT_BLOB,
                              SQLT_CLOB,
                              SQLT_CFILE,
                              SQLT_BFILE,
                              SQLT_RSET,
                              SQLT_RDD] then
    begin
      IsDscInit := 0;
      OCILobLocatorIsInit(  oraSession.hEnv,
                            errhp,
                            pDesc,
                            @IsDscInit);
      if IsDscInit <> 0 then
        OCIDescriptorFree(pDesc, oDscChr.DescType);
    end;
    oDscChr.Destroy;
  end;
  FDescriptors.Clear;
end;

procedure To8StmtOrd.SetArrayOfPar(const Value: Boolean);
var
  i: Integer;
  FColOffset: ub4;
begin
  inherited SetArrayOfPar(Value);
  if FArrayOfPar then
  begin
    FParRecordSize := 0;
    FColOffset := 0;
    for i := 0 to Params.Count - 1 do
    begin
      TOraParam(Params.Items[i]).ParOffset := FColOffset;
      FColOffset := FColOffset + ub4(TOraParam(Params.Items[i]).ParSize) * FPortion;
      FParRecordSize := FParRecordSize + TOraParam(Params.Items[i]).ParSize;
    end;
    FParamBuffer := Pointer(MemMgr.oMem.GetMem(self, FPortion*FParRecordSize));
    FParamIndBuffer   := psb2(MemMgr.oMem.GetMem(self, FPortion*ub2(Params.Count)*sizeof(sb2)));
    FParamLenBuffer   := pub2(MemMgr.oMem.GetMem(self, FPortion*ub2(Params.Count)*sizeof(ub2)));
    FParamRCodeBuffer := pub2(MemMgr.oMem.GetMem(self, FPortion*ub2(Params.Count)*sizeof(ub2)));
  end;
end;

function To8StmtOrd.GetArrayParNull(Index1, Index2: Integer): Boolean;
var
  Ind: sb2;
begin
  Result := true;
  if FArrayOfPar then
  begin
    Ind := psb2(pChar(FParamIndBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
    //Ind := psb2(pChar(FParamIndBuffer) + (ub4(Index1) * ub2(Params.Count) *  SizeOf(ub2)) + (ub4(Index2) * SizeOf(ub2)))^;
    if Ind = -1 then
      Result := true
    else
      Result := false;
  end;
end;

function To8StmtOrd.GetArrayParRCode(Index1, Index2: Integer): ub2;
begin
  Result := 0;
  if FArrayOfPar then
    Result := pub2(pChar(FParamRCodeBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
    //Result := pub2(pChar(FParamRCodeBuffer) + (ub4(Index1) * ub2(Params.Count) *  SizeOf(ub2)) + (ub4(Index2) * SizeOf(ub2)))^;
end;

function To8StmtOrd.GetArrayParValue(Index1, Index2: Integer): Variant;
type
  pInteger = ^Integer;
  pDouble = ^Double;
var
  Ind: sb2;
  lng: ub2;
  ValB: Pointer;
  oPar: TOraParam;
  RStr: String;
  RInt: Integer;
  RFloat: Double;
  RDateTime: TDateTime;
begin
  Result := Null;
  if FArrayOfPar then
  begin
    Ind := psb2(pChar(FParamIndBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
    //Ind := psb2(pChar(FParamIndBuffer) + (ub4(Index1) * ub2(Params.Count) *  SizeOf(ub2)) + (ub4(Index2) * SizeOf(ub2)))^;
    if Ind = -1 then
      Result := Null
    else begin
      oPar := TOraParam(Params.Items[Index2]);
      ValB := Pointer(pChar(FParamBuffer) + oPar.ParOffset + Index1 * oPar.ParSize);
      //ValB := Pointer(pChar(FParamBuffer) + (FParRecordSize * ub4(Index1)) + oPar.ParOffset);
      lng := pub2(pChar(FParamLenBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
      //lng := pub2(pChar(FParamLenBuffer) + (ub4(Index1) * ub2(Params.Count) *  SizeOf(ub2)) + (ub4(Index2) * SizeOf(ub2)))^;
      case oPar.DataType of
        ptString:
        begin
          SetString(RStr, pChar(ValB), lng);
          Result := RStr;
        end;
        ptInteger:
        begin
          RInt := pInteger(ValB)^;
          Result := RInt;
        end;
        ptFloat:
        begin
          RFloat := pDouble(ValB)^;
          Result := RFloat;
        end;
        ptDateTime:
        begin
          RDateTime := OraDateToDateTime(pORA_Date(ValB)^);
          Result := RDateTime;
        end;
        ptCLob, ptBLob, ptCFile, ptBFile, ptCursor:
        begin
          TVarData(Result).VType := varAny;
          TVarData(Result).VAny := Pointer(ValB);
        end;
      end;
    end;
  end;
end;

procedure To8StmtOrd.SetArrayParNull(Index1, Index2: Integer; const Value: Boolean);
var
  Ind: psb2;
begin
  if FArrayOfPar then
  begin
    Ind := psb2(pChar(FParamIndBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)));
    //Ind := psb2(pChar(FParamIndBuffer) + (ub4(Index1) * ub2(Params.Count) *  SizeOf(ub2)) + (ub4(Index2) * SizeOf(ub2)));
    if Value then
      Ind^ := -1
    else
      Ind^ := 0;
  end;
end;

procedure To8StmtOrd.SetArrayParValue(Index1, Index2: Integer; const Value: Variant);
type
  pInteger = ^Integer;
  pDouble = ^Double;
var
  lng: ub2;
  pLng: pub2;
  ValB: Pointer;
  oPar: TOraParam;
  RStr: String;
  RDateTime: TDateTime;
  ODT: ORA_Date;
  pRC: pub2;
begin
  if FArrayOfPar then
  begin
    if FLastParamRow < ub4(Index1) then FLastParamRow := Index1;
    if Value <> NULL then
    begin
      SetArrayParNull(Index1, Index2, false);
      oPar := TOraParam(Params.Items[Index2]);
      ValB := Pointer(pChar(FParamBuffer) + oPar.ParOffset + Index1 * oPar.ParSize);
      //ValB := Pointer(pChar(FParamBuffer) + (FParRecordSize * ub4(Index1)) + oPar.ParOffset);
      pLng := pub2(pChar(FParamLenBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)));
      //pLng := pub2(pChar(FParamLenBuffer) + (ub4(Index1) * ub2(Params.Count) *  SizeOf(ub2)) + (ub4(Index2) * SizeOf(ub2)));
      pRC := pub2(pChar(FParamRCodeBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)));
      //pRC := pub2(pChar(FParamRCodeBuffer) + (ub4(Index1) * ub2(Params.Count) *  SizeOf(ub2)) + (ub4(Index2) * SizeOf(ub2)));
      pRC^ := 0;
      case oPar.DataType of
        ptString:
        begin
          lng := Length(Value);
          oPar.MaxLength := oPar.MaxLengthStringBuffer - 1;
          RStr := Value;
          StrMove(pChar(ValB), pChar(RStr), lng);
          pChar(ValB)[lng] := #0;
          Inc(lng);
          pLng^ := lng;
        end;
        ptInteger:
        begin
          lng := SizeOf(Integer);
          Integer((pInteger(ValB))^) := Value;
          pLng^ := lng;
          oPar.MaxLength := lng;
        end;
        ptFloat:
        begin
          lng := SizeOf(Double);
          Double((pDouble(ValB))^) := Value;
          pLng^ := lng;
          oPar.MaxLength := lng;
        end;
        ptDateTime:
        begin
          lng := SizeOf(ORA_Date);
          RDateTime := Value;
          ODT := DateTimeToOraDate(RDateTime);
          ORA_Date(pORA_Date(ValB)^) := ODT;
          pLng^ := lng;
          oPar.MaxLength := lng;
        end;
        ptCLob, ptBLob, ptCFile, ptBFile, ptCursor:
        begin
          Pointer(ValB^) := TVarData(Value).VAny;
        end;
      end;
    end else
      SetArrayParNull(Index1, Index2, true);
  end;
end;

procedure To8StmtOrd.QueryChanged(Sender: TObject);
var
  List: TOraParams;
begin
  if not (csReading in ComponentState) then
  begin
    Active := false;
    if ParamCheck or (csDesigning in ComponentState) then
    begin
      List := TOraParams.Create(self, TOraParam);
      try
        List.ParseSQL(SQL.Text, True);
        List.AssignValues(Params);
        if List.Count > 0 then
        begin
          Params.Clear;
          Params := List;
        end;
      finally
        List.Free;
      end;
    end;
  end else
    Params.ParseSQL(SQL.Text, False);
end;

function To8StmtOrd.GetResultSetStream: TResultSetStream;
begin
  Result := FResultSetStream;
end;

procedure To8StmtOrd.SetResultSetStream(const Value: TResultSetStream);
begin
  inherited SetResultSetStream(Value);
  if not Active then
  begin
    FResultSetStream := Value;
    if Value = rsToMemory then
    begin
      FetchBufferStream.Free;
      IndBufferStream.Free;
      LenBufferStream.Free;
      RCodeBufferStream.Free;

      DeleteFile(FFetchBuffer);
      DeleteFile(FIndBuffer);
      DeleteFile(FLenBuffer);
      DeleteFile(FRCodeBuffer);

      FetchBufferStream  := TMemStream.Create;
      IndBufferStream    := TMemStream.Create;
      LenBufferStream    := TMemStream.Create;
      RCodeBufferStream  := TMemStream.Create;
    end;
    if Value = rsToFile then
    begin
      FetchBufferStream.Free;
      IndBufferStream.Free;
      LenBufferStream.Free;
      RCodeBufferStream.Free;

      DeleteFile(FFetchBuffer);
      DeleteFile(FIndBuffer);
      DeleteFile(FLenBuffer);
      DeleteFile(FRCodeBuffer);

      FFetchBuffer := GetTmpFileName;
      FIndBuffer := GetTmpFileName;
      FLenBuffer := GetTmpFileName;
      FRCodeBuffer := GetTmpFileName;

      FetchBufferStream  := TFStream.Create(FFetchBuffer, fmCreate);
      IndBufferStream    := TFStream.Create(FIndBuffer, fmCreate);
      LenBufferStream    := TFStream.Create(FLenBuffer, fmCreate);
      RCodeBufferStream  := TFStream.Create(FRCodeBuffer, fmCreate);
    end;
  end;
end;

procedure To8StmtOrd.FlashCurrentRecord;
var
  s1, s2, s3, s4: Longint;
begin
  s1 := FetchBufferStream.Seek(0, soFromCurrent);
  s2 := IndBufferStream.Seek(0, soFromCurrent);
  s3 := LenBufferStream.Seek(0, soFromCurrent);
  s4 := RCodeBufferStream.Seek(0, soFromCurrent);
  //
    FetchBufferStream.Seek((RecNoInternal - 1)  * FRecordSize, soFromBeginning);
    IndBufferStream.Seek  ((RecNoInternal - 1)  * FFieldCount * sizeof(sb2),  soFromBeginning);
    LenBufferStream.Seek  ((RecNoInternal - 1)  * FFieldCount * sizeof(ub2),  soFromBeginning);
    RCodeBufferStream.Seek((RecNoInternal - 1)  * FFieldCount * sizeof(ub2),  soFromBeginning);

    FetchBufferStream.Write(GetRawBuffer^, FRecordSize);
    IndBufferStream.Write  (GetRawInd^,   FFieldCount*sizeof(sb2));
    LenBufferStream.Write  (GetRawRlen^,   FFieldCount*sizeof(ub2));
    RCodeBufferStream.Write(GetRawRCode^, FFieldCount*sizeof(ub2));
  //
  FetchBufferStream.Seek(s1, soFromBeginning);
  IndBufferStream.Seek(s2, soFromBeginning);
  LenBufferStream.Seek(s3, soFromBeginning);
  RCodeBufferStream.Seek(s4, soFromBeginning);
end;

procedure To8StmtOrd.Post;
begin
  FlashCurrentRecord;
  State := dsBrowse;
end;

procedure To8StmtOrd.Edit;
begin
  State := dsEdit;
end;

procedure To8StmtOrd.CheckEdit;
begin
  if State = dsEdit then Post;
end;

function To8StmtOrd.GetFetchRowsCount: ub4;
begin
  Result := RowsDone;
end;

function To8StmtOrd.GetLastRowId: pOCIRowid;
begin
  Result := nil;
end;

procedure To8StmtOrd.ReExecute;
var
   sExec: sword;
   LastExec: boolean;
   ExecStatus: boolean;
begin
  case FStmtType of
    OCI_STMT_SELECT:
      begin

        //
        LastSizeStream := 0;
        //


        FetchBufferStream.qq;
        IndBufferStream.qq;
        LenBufferStream.qq;
        RCodeBufferStream.qq;

        //FetchBufferStream.Size := 0;
        //IndBufferStream.Size := 0;
        //LenBufferStream.Size := 0;
        //RCodeBufferStream.Size := 0;

        EofInternal := false;
        BofInternal := false;
        DescriptorsAlloc;
        LastExec := OCIStmtExecute( oraSession.hSvc,
                                    stmthp,
                                    errhp,
                                    Portion,
                                    0,
                                    nil,
                                    nil,
                                    OCI_DEFAULT) in [OCI_SUCCESS, OCI_SUCCESS_WITH_INFO, OCI_NO_DATA];
        if LastExec then begin
          CurrGranule := 0;
          LastGranule := 0;
          OCIAttrGet(stmthp, OCI_HTYPE_STMT, @LastGranule, nil, OCI_ATTR_ROW_COUNT, errhp);
          RowsDone := LastGranule;
          FExec := true;
          if LastGranule = 0 then
          begin
            BofInternal := true;
            EofInternal := true;
          end else begin
            WriteBuffers(LastGranule);
            EofInternal := false;
            RecNoInternal := 1;
          end;
        end else begin
          BofInternal := true;
          EofInternal := true;
          GetError(true);
        end;
      end;
  else
    while true do
    begin
      if FArrayOfPar then
      begin
        sExec := OCIStmtExecute(  oraSession.hSvc,
                                  stmthp,
                                  errhp,
                                  FLastParamRow + 1,
                                  0,
                                  nil,
                                  nil,
                                  OCI_DEFAULT);
        ExecStatus := (sExec = OCI_SUCCESS)
                      or (sExec = OCI_SUCCESS_WITH_INFO)
                      or (sExec = OCI_STILL_EXECUTING);
        FLastParamRow := 0;
      end else begin
        sExec := OCIStmtExecute(  oraSession.hSvc,
                                  stmthp,
                                  errhp,
                                  1,
                                  0,
                                  nil,
                                  nil,
                                  OCI_DEFAULT);
        ExecStatus := (sExec = OCI_SUCCESS)
                      or (sExec = OCI_SUCCESS_WITH_INFO)
                      or (sExec = OCI_STILL_EXECUTING);
      end;
      if not ExecStatus then GetError(true);
      if (sExec = OCI_STILL_EXECUTING) and Assigned(OnAfterExecute) then OnAfterExecute(self, sExec);
      if sExec <> OCI_STILL_EXECUTING then Break;
    end;
  end;
end;

function To8StmtOrd.GetArrayParString(Index1, Index2: Integer): String;
type
  pInteger = ^Integer;
  pDouble = ^Double;
var
  Ind: sb2;
  lng: ub2;
  ValB: Pointer;
  oPar: TOraParam;
  RStr: String;
  RInt: Integer;
  RFloat: Double;
  RDateTime: TDateTime;
begin
  Result := Null;
  if FArrayOfPar then
  begin
    Ind := psb2(pChar(FParamIndBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
    if Ind = -1 then
      Result := Null
    else begin
      oPar := TOraParam(Params.Items[Index2]);
      ValB := Pointer(pChar(FParamBuffer) + oPar.ParOffset + Index1 * oPar.ParSize);
      lng := pub2(pChar(FParamLenBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
      case oPar.DataType of
        ptString:
        begin
          SetString(RStr, pChar(ValB), lng);
          Result := RStr;
        end;
        ptInteger:
        begin
          RInt := pInteger(ValB)^;
          Str(RInt, Result);
        end;
        ptFloat:
        begin
          RFloat := pDouble(ValB)^;
          Result := FloatToStr(RFloat);
        end;
        ptDateTime:
        begin
          RDateTime := OraDateToDateTime(pORA_Date(ValB)^);
          Result := DateTimeToStr(RDateTime);
        end;
        ptCLob, ptBLob, ptCFile, ptBFile, ptCursor:
        begin
          Result := '';
        end;
      end;
    end;
  end;
end;

procedure To8StmtOrd.SetArrayParString(Index1, Index2: Integer; const Value: String);
type
  pInteger = ^Integer;
  pDouble = ^Double;
var
  lng: ub2;
  pLng: pub2;
  Code: Integer;
  ValB: Pointer;
  oPar: TOraParam;
  RDateTime: TDateTime;
  ODT: ORA_Date;
  pRC: pub2;
begin
  if FArrayOfPar then
  begin
    if FLastParamRow < ub4(Index1) then FLastParamRow := Index1;
    if Value <> '' then
    begin
      SetArrayParNull(Index1, Index2, false);
      oPar := TOraParam(Params.Items[Index2]);
      ValB := Pointer(pChar(FParamBuffer) + oPar.ParOffset + Index1 * oPar.ParSize);
      pLng := pub2(pChar(FParamLenBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)));
      pRC := pub2(pChar(FParamRCodeBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)));
      pRC^ := 0;
      case oPar.DataType of
        ptString:
        begin
          lng := Length(Value);
          oPar.MaxLength := oPar.MaxLengthStringBuffer - 1;
          StrMove(pChar(ValB), pChar(Value), lng);
          pChar(ValB)[lng] := #0;
          Inc(lng);
          pLng^ := lng;
        end;
        ptInteger:
        begin
          lng := SizeOf(Integer);
          Val(Value, (pInteger(ValB))^, Code);
          pLng^ := lng;
          oPar.MaxLength := lng;
        end;
        ptFloat:
        begin
          lng := SizeOf(Double);
          Val(Value, (pDouble(ValB))^, Code);
          pLng^ := lng;
          oPar.MaxLength := lng;
        end;
        ptDateTime:
        begin
          lng := SizeOf(ORA_Date);
          RDateTime := StrToDateTime(Value);
          ODT := DateTimeToOraDate(RDateTime);
          ORA_Date(pORA_Date(ValB)^) := ODT;
          pLng^ := lng;
          oPar.MaxLength := lng;
        end;
        ptCLob, ptBLob, ptCFile, ptBFile, ptCursor:
        begin
          Pointer(ValB^) := nil;
        end;
      end;
    end else
      SetArrayParNull(Index1, Index2, true);
  end;
end;

function To8StmtOrd.GetArrayParInteger(Index1, Index2: Integer): Integer;
type
  pInteger = ^Integer;
  pDouble = ^Double;
var
  Ind: sb2;
  //lng: ub2;
  ValB: Pointer;
  oPar: TOraParam;
  //RStr: String;
  Code: Integer;
  RInt: Integer;
  RFloat: Double;
  RDateTime: TDateTime;
begin
  Result := Null;
  if FArrayOfPar then
  begin
    Ind := psb2(pChar(FParamIndBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
    if Ind = -1 then
      Result := Null
    else begin
      oPar := TOraParam(Params.Items[Index2]);
      ValB := Pointer(pChar(FParamBuffer) + oPar.ParOffset + Index1 * oPar.ParSize);
      //lng := pub2(pChar(FParamLenBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
      case oPar.DataType of
        ptString:
        begin
          //SetString(RStr, pChar(ValB), lng);
          Val(pChar(ValB), RInt, Code);
          Result := RInt;
        end;
        ptInteger:
        begin
          Result := pInteger(ValB)^;
        end;
        ptFloat:
        begin
          RFloat := pDouble(ValB)^;
          Result := Trunc(RFloat);
        end;
        ptDateTime:
        begin
          RDateTime := OraDateToDateTime(pORA_Date(ValB)^);
          Result := Trunc(RDateTime);
        end;
        ptCLob, ptBLob, ptCFile, ptBFile, ptCursor:
        begin
          Result := 0;
        end;
      end;
    end;
  end;
end;

procedure To8StmtOrd.SetArrayParInteger(Index1, Index2: Integer; const Value: Integer);
type
  pInteger = ^Integer;
  pDouble = ^Double;
var
  lng: ub2;
  pLng: pub2;
  ValB: Pointer;
  oPar: TOraParam;
  RDateTime: TDateTime;
  ODT: ORA_Date;
  pRC: pub2;
  RStr: String;
begin
  if FArrayOfPar then
  begin
    if FLastParamRow < ub4(Index1) then FLastParamRow := Index1;
    SetArrayParNull(Index1, Index2, false);
    oPar := TOraParam(Params.Items[Index2]);
    ValB := Pointer(pChar(FParamBuffer) + oPar.ParOffset + Index1 * oPar.ParSize);
    pLng := pub2(pChar(FParamLenBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)));
    pRC := pub2(pChar(FParamRCodeBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)));
    pRC^ := 0;
    case oPar.DataType of
      ptString:
      begin
        RStr := IntToStr(Value);
        lng := Length(RStr);
        oPar.MaxLength := oPar.MaxLengthStringBuffer - 1;
        StrMove(pChar(ValB), pChar(RStr), lng);
        pChar(ValB)[lng] := #0;
        Inc(lng);
        pLng^ := lng;
      end;
      ptInteger:
      begin
        lng := SizeOf(Integer);
        pInteger(ValB)^ := Value;
        pLng^ := lng;
        oPar.MaxLength := lng;
      end;
      ptFloat:
      begin
        lng := SizeOf(Double);
        pDouble(ValB)^ := Value;
        pLng^ := lng;
        oPar.MaxLength := lng;
      end;
      ptDateTime:
      begin
        lng := SizeOf(ORA_Date);
        RDateTime := Value;
        ODT := DateTimeToOraDate(RDateTime);
        ORA_Date(pORA_Date(ValB)^) := ODT;
        pLng^ := lng;
        oPar.MaxLength := lng;
      end;
      ptCLob, ptBLob, ptCFile, ptBFile, ptCursor:
      begin
        Pointer(ValB^) := nil;
      end;
    end;
  end;
end;

function To8StmtOrd.GetArrayParFloat(Index1, Index2: Integer): double;
type
  pInteger = ^Integer;
  pDouble = ^Double;
var
  Ind: sb2;
  //lng: ub2;
  ValB: Pointer;
  oPar: TOraParam;
  //RStr: String;
  Code: Integer;
  //RInt: Integer;
  //RFloat: Double;
  //RDateTime: TDateTime;
begin
  Result := Null;
  if FArrayOfPar then
  begin
    Ind := psb2(pChar(FParamIndBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
    if Ind = -1 then
      Result := Null
    else begin
      oPar := TOraParam(Params.Items[Index2]);
      ValB := Pointer(pChar(FParamBuffer) + oPar.ParOffset + Index1 * oPar.ParSize);
      //lng := pub2(pChar(FParamLenBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
      case oPar.DataType of
        ptString:
        begin
          //SetString(RStr, pChar(ValB), lng);
          Val(pChar(ValB), Result, Code);
        end;
        ptInteger:
        begin
          Result := pInteger(ValB)^;
        end;
        ptFloat:
        begin
          Result := pDouble(ValB)^;
        end;
        ptDateTime:
        begin
          Result := OraDateToDateTime(pORA_Date(ValB)^);
        end;
        ptCLob, ptBLob, ptCFile, ptBFile, ptCursor:
        begin
          Result := 0;
        end;
      end;
    end;
  end;
end;

procedure To8StmtOrd.SetArrayParFloat(Index1, Index2: Integer; const Value: double);
type
  pInteger = ^Integer;
  pDouble = ^Double;
var
  lng: ub2;
  pLng: pub2;
  ValB: Pointer;
  oPar: TOraParam;
  RDateTime: TDateTime;
  ODT: ORA_Date;
  pRC: pub2;
  RStr: String;
begin
  if FArrayOfPar then
  begin
    if FLastParamRow < ub4(Index1) then FLastParamRow := Index1;
    SetArrayParNull(Index1, Index2, false);
    oPar := TOraParam(Params.Items[Index2]);
    ValB := Pointer(pChar(FParamBuffer) + oPar.ParOffset + Index1 * oPar.ParSize);
    pLng := pub2(pChar(FParamLenBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)));
    pRC := pub2(pChar(FParamRCodeBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)));
    pRC^ := 0;
    case oPar.DataType of
      ptString:
      begin
        RStr := FloatToStr(Value);
        lng := Length(RStr);
        oPar.MaxLength := oPar.MaxLengthStringBuffer - 1;
        StrMove(pChar(ValB), pChar(RStr), lng);
        pChar(ValB)[lng] := #0;
        Inc(lng);
        pLng^ := lng;
      end;
      ptInteger:
      begin
        lng := SizeOf(Integer);
        pInteger(ValB)^ := Trunc(Value);
        pLng^ := lng;
        oPar.MaxLength := lng;
      end;
      ptFloat:
      begin
        lng := SizeOf(Double);
        pDouble(ValB)^ := Value;
        pLng^ := lng;
        oPar.MaxLength := lng;
      end;
      ptDateTime:
      begin
        lng := SizeOf(ORA_Date);
        RDateTime := Value;
        ODT := DateTimeToOraDate(RDateTime);
        ORA_Date(pORA_Date(ValB)^) := ODT;
        pLng^ := lng;
        oPar.MaxLength := lng;
      end;
      ptCLob, ptBLob, ptCFile, ptBFile, ptCursor:
      begin
        Pointer(ValB^) := nil;
      end;
    end;
  end;
end;

function To8StmtOrd.GetArrayParDateTime(Index1, Index2: Integer): TDateTime;
type
  pInteger = ^Integer;
  pDouble = ^Double;
var
  Ind: sb2;
  //lng: ub2;
  ValB: Pointer;
  oPar: TOraParam;
  //RStr: String;
  //Code: Integer;
  //RInt: Integer;
  //RFloat: Double;
  //RDateTime: TDateTime;
begin
  Result := Null;
  if FArrayOfPar then
  begin
    Ind := psb2(pChar(FParamIndBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
    if Ind = -1 then
      Result := Null
    else begin
      oPar := TOraParam(Params.Items[Index2]);
      ValB := Pointer(pChar(FParamBuffer) + oPar.ParOffset + Index1 * oPar.ParSize);
      //lng := pub2(pChar(FParamLenBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)))^;
      case oPar.DataType of
        ptString:
        begin
          //SetString(RStr, pChar(ValB), lng);
          Result := StrToDateTime(pChar(ValB));
          //Val(pChar(ValB), Result, Code);
        end;
        ptInteger:
        begin
          Result := pInteger(ValB)^;
        end;
        ptFloat:
        begin
          Result := pDouble(ValB)^;
        end;
        ptDateTime:
        begin
          Result := OraDateToDateTime(pORA_Date(ValB)^);
        end;
        ptCLob, ptBLob, ptCFile, ptBFile, ptCursor:
        begin
          Result := 0;
        end;
      end;
    end;
  end;
end;

procedure To8StmtOrd.SetArrayParDateTime(Index1, Index2: Integer; const Value: TDateTime);
type
  pInteger = ^Integer;
  pDouble = ^Double;
var
  lng: ub2;
  pLng: pub2;
  ValB: Pointer;
  oPar: TOraParam;
  RDateTime: TDateTime;
  ODT: ORA_Date;
  pRC: pub2;
  RStr: String;
begin
  if FArrayOfPar then
  begin
    if FLastParamRow < ub4(Index1) then FLastParamRow := Index1;
    SetArrayParNull(Index1, Index2, false);
    oPar := TOraParam(Params.Items[Index2]);
    ValB := Pointer(pChar(FParamBuffer) + oPar.ParOffset + Index1 * oPar.ParSize);
    pLng := pub2(pChar(FParamLenBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)));
    pRC := pub2(pChar(FParamRCodeBuffer) + (FPortion * SizeOf(ub2) * ub4(Index2)) + (ub4(Index1) * SizeOf(ub2)));
    pRC^ := 0;
    case oPar.DataType of
      ptString:
      begin
        RStr := DateTimeToStr(Value);
        lng := Length(RStr);
        oPar.MaxLength := oPar.MaxLengthStringBuffer - 1;
        StrMove(pChar(ValB), pChar(RStr), lng);
        pChar(ValB)[lng] := #0;
        Inc(lng);
        pLng^ := lng;
      end;
      ptInteger:
      begin
        lng := SizeOf(Integer);
        pInteger(ValB)^ := Trunc(Value);
        pLng^ := lng;
        oPar.MaxLength := lng;
      end;
      ptFloat:
      begin
        lng := SizeOf(Double);
        pDouble(ValB)^ := Value;
        pLng^ := lng;
        oPar.MaxLength := lng;
      end;
      ptDateTime:
      begin
        lng := SizeOf(ORA_Date);
        RDateTime := Value;
        ODT := DateTimeToOraDate(RDateTime);
        ORA_Date(pORA_Date(ValB)^) := ODT;
        pLng^ := lng;
        oPar.MaxLength := lng;
      end;
      ptCLob, ptBLob, ptCFile, ptBFile, ptCursor:
      begin
        Pointer(ValB^) := nil;
      end;
    end;
  end;
end;

end.
