unit DBFDataSet;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Db, stdctrls, o8Filter, DBcommon, DBFIndex, DBFNTX, DBFCDX,
  DsgnIntf, EditProp, o8Util, MemMgr;

const
  LOCATE_BUFFER_SIZE = 4096;

type

  {$A-}
  DBF_HEAD = packed record
    dbf_id:             Byte;                     //0
    last_update:        array[0..2] of Byte;      //1
    last_rec:           Longint;                  //4
    data_offset:        Word;                     //8
    rec_size:           Word;                     //10
    filler:             array[0..15] of Byte;     //12
    prod_ind:           Shortint;                 //28
    lang:               Word;                     //29
    dummy:              Byte;                     //31
  end;

  num_size = packed record
    len: Byte;
    dec: Byte;
  end;

  len_info = packed record
    case Shortint of
      0: (char_len: Word);
      1: (num_len: num_size);
  end;

  TVKDBFType = (
    dbftS1,             //Shortint              (1 byte)
    dbftS1_N,           //Shortint with NULL    (1 byte ShortInt + 1 byte null/not null)
    dbftU1,             //Byte
    dbftU1_N,           //Byte  with NULL
    dbftS2,             //Smallint
    dbftS2_N,           //Smallint with NULL
    dbftU2,             //Word
    dbftU2_N,           //Word with NULL
    dbftS4,             //Longint
    dbftS4_N,           //Longint with NULL
    dbftU4,             //Longword
    dbftU4_N,           //Longword with NULL
    dbftS8,             //Int64
    dbftS8_N,           //Int64 with NULL
    dbftR4,             //Single
    dbftR4_N,           //Single with NULL
    dbftR6,             //Real48
    dbftR6_N,           //Real48 with NULL
    dbftR8,             //Double
    dbftR8_N,           //Double with NULL
    dbftR10,            //Extended                !!!!!!!Not yet realized
    dbftR10_N,          //Extended with NULL      !!!!!!!Not yet realized
    dbftD1,             //TDateTime
    dbftD1_N,           //TDateTime with NULL
    dbftD2,             //DataSet DateTime
    dbftD2_N,           //DataSet DateTime with NULL
    //
    dbftString,         //String
    dbftString_N,       //String witn NULL
    dbftFixedChar,      //FixedChar
    dbftWideString,     //WideString              !!!!!!!Not yet realized
    dbftCurrency,       //Currency
    dbftCurrency_N,     //Currency with NULL
    //
    dbftClob,           //Clob
    dbftBlob,           //Blob
    dbftGraphic,        //Graphic
    dbftFmtMemo,        //FmtMemo
    //
    dbftBCD,            //BCD (34 bytes VCL BCD structure TBCD)
    //
    dbftDate,            //Date  (Integer, 4 bytes)
    dbftDate_N,          //Date with NULL (Integer, 4 bytes + 1 null/not null byte)
    dbftTime,            //Time (Integer, 4 bytes)
    dbftTime_N,          //Time with NULL (Integer, 4 bytes + 1 null/not null byte)

    dbftD3,             //TDateTime as dbftR6 (6 bytes)
    dbftD3_N,           //TDateTime as dbftR6 (6 bytes) with NULL ( + 1 byte null/not null)

////////////////////////////////////////////////////////////////////////////////
/// This integer and real types with 1 null/not null bit instead of sign bit
////////////////////////////////////////////////////////////////////////////////
    dbftU1_NB,          //Byte              ( 0 - 127         )
    dbftU2_NB,          //Word              ( 0 - 32767       )
    dbftU4_NB,          //Longword          ( 0 - 2147483647  )
    dbftR4_NB,          //Single            ( Positive Single )
    dbftR6_NB,          //Real48            ( Positive Real48 )
    dbftR8_NB,          //Double            ( Positive Double )
    dbftD1_NB,          //TDateTime         ( Positive TDateTime  )  (8 bytes)
    dbftD2_NB,          //DataSet DateTime  ( Positive DataSet DateTime) (8 bytes)
    //
    dbftCurrency_NB,    //Currency
    //
    dbftDate_NB,        //Date
    dbftTime_NB,        //Time

    dbftD3_NB,          //TDateTime as dbftR6 (6 bytes) ( Positive Real48 )

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
///////////       This is Lob types with absolute pointer (dbftU4_NB) to a lob
///////////       file and not align 512 byte pages into the Lob
////////////////////////////////////////////////////////////////////////////////

    dbftClob_NB,           //Clob
    dbftBlob_NB,           //Blob
    dbftGraphic_NB,        //Graphic
    dbftFmtMemo_NB,        //FmtMemo

////////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////////
///////////       Not yet realized
////////////////////////////////////////////////////////////////////////////////

    dbftBytes,          //Bytes
    dbftVarBytes,       //VarBytes
    dbftTypedBinary,    //TypedBinary

    dbftADT,            //ADT
    dbftArray,          //Array
    dbftReference,      //Reference
    dbftDataSet,        //DataSet
    dbftVariant,        //Variant
    dbftInterface,      //Interface
    dbftIDispatch,      //IDispatch
    dbftGuid            //Guid

////////////////////////////////////////////////////////////////////////////////

  );

  FIELD_REC = packed record
    field_name:     array[0..10] of char;
    field_type:     char;                   //C N D L M     E - Extendes types
    extend_type:    TVKDBFType;             //use if field_type = 'E'
    dummy:          array[0..2] of char;
    lendth:         len_info;
    filler:         array [0..13] of char;
  end;

  TRecInfo = packed record
    RecordRowID: Longint;
    UpdateStatus: TUpdateStatus;
    BookmarkFlag: TBookmarkFlag;
  end;
  {$A+}

  TBufDirection = (bdFromTop, bdFromBottom);

  pTRecInfo = ^TRecInfo;
  ppTRecInfo = ^pTRecInfo;

  pDouble = ^Double;
  pInteger = ^Integer;
  pReal48 = ^Real48;

  TFieldInfo = packed record
    FName: ShortString;
    FType: Char;
    FExType: TVKDBFType;
    FLen:  Integer;
    FPrec: Integer;
    FOff: Integer;
  end;
  PFieldInfo = ^TFieldInfo;

  {TAccessMode}
  TAccessMode = class(TPersistent)
  private
    fOpenRead      : boolean;
    fOpenWrite     : boolean;
    fOpenReadWrite : boolean;
    fShareCompat   : boolean;
    fShareExclusive: boolean;
    fShareDenyWrite: boolean;
    fShareDenyRead : boolean;
    fShareDenyNone : boolean;
    function GetAccessMode: LongWord;
    procedure SetAccessMode(const Value: LongWord);
  public
    FLast: LongWord;
    constructor Create;
    function IsExists(const Value: LongWord): boolean;
  published
    property AccessMode: LongWord read GetAccessMode write SetAccessMode;
    property OpenRead       : boolean read fOpenRead       write fOpenRead      ;
    property OpenWrite      : boolean read fOpenWrite      write fOpenWrite     ;
    property OpenReadWrite  : boolean read fOpenReadWrite  write fOpenReadWrite ;
    property ShareCompat    : boolean read fShareCompat    write fShareCompat   ;
    property ShareExclusive : boolean read fShareExclusive write fShareExclusive;
    property ShareDenyWrite : boolean read fShareDenyWrite write fShareDenyWrite;
    property ShareDenyRead  : boolean read fShareDenyRead  write fShareDenyRead ;
    property ShareDenyNone  : boolean read fShareDenyNone  write fShareDenyNone ;
  end;

  {TDBFDataSet}
  TDBFDataSet = class(TDataSet)
  private
    { Private declarations }
    FStreamedActive: boolean;
    FCurrentIndex: Longint;
    FCurrGranul: Longint;
    FLastGranul: Longint;
    FTempRecord: pChar;
    FFilterRecord: pChar;
    FFilterParser: To8ExprParser;
    FBufferSize: Integer;
    BufferDBF: Pointer;
    FDBFFileName: String;
    DBFHandler: Integer;
    DBFHeader: DBF_HEAD;
    FRecordSize: Integer;
    FAccessMode: TAccessMode;
    FFieldInfo: TList;
    FOEM: Boolean;
    FSetDeleted: Boolean;
    function GetRecordBufferSize: Integer;
    property RecordBufferSize: Integer read GetRecordBufferSize;
    function GetActiveRecBuf(var RecBuf: PChar): Boolean;
    function GetPortion: Integer;
    procedure SetBufferSize(const Value: Integer);
    function GetPortionSize: Integer;
    function GetDeleted: Boolean;
    //function FindFieldInfo(FieldName: String): PFieldInfo;
    function GetRecordByBuffer(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
    function AcceptRecord: Boolean;
    procedure SetSetDeleted(const Value: Boolean);
  protected
    { Protected declarations }
    procedure SetActive(Value: boolean); override;
    procedure Loaded; override;

    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function GetRecordSize: Word; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;

    function FindRecord(Restart, GoForward: Boolean): Boolean; override;

    function GetRecordCount: Integer; override;
    function GetRecNo: Integer; override;
    procedure SetRecNo(Value: Integer); override;
    procedure SetRecNoInternal(Value: Integer);

    function GetStateFieldValue(State: TDataSetState; Field: TField): Variant; override;
    function CompareLocateField(const Fields: TList; const KeyValues: Variant; Options: TLocateOptions): Integer;

    procedure SetFiltered(Value: Boolean); override;

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    function Translate(Src, Dest: PChar; ToOem: Boolean): Integer; override;

    procedure CreateTable;

    property CurrentIndex: Longint read FCurrentIndex;
    property CurrentGranul: Longint read FCurrGranul;
    property Portion: Integer read GetPortion;
    property PortionSize: Integer read GetPortionSize;

    property Deleted: Boolean read GetDeleted;
    property Header: DBF_HEAD read DBFHeader;
    property FieldInfo: TList read FFieldInfo;

  published
    { Published declarations }
    property Active;
    property DBFFileName: String read FDBFFileName write FDBFFileName;
    property AccessMode: TAccessMode read FAccessMode write FAccessMode;
    property BufferSize: Integer read FBufferSize write SetBufferSize;
    property OEM: Boolean read FOEM write FOEM default false;
    property SetDeleted: Boolean read FSetDeleted write SetSetDeleted;

    property Filter;
    property Filtered;
    property FilterOptions;

    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;

  end;

  TSmartDBF = class;
  TDBFFieldDef = class;

  {TVKDataLink}
  TVKDataLink = class(TDataLink)
  private
    FDBFDataSet: TSmartDBF;
  protected
    procedure DataEvent(Event: TDataEvent; Info: Longint); override;
    procedure DataSetScrolled(Distance: Integer); override;
  public
    property DBFDataSet: TSmartDBF read FDBFDataSet write FDBFDataSet;
  end;

  {TDBFFieldDefs}
  TDBFFieldDefs = class(TOwnedCollection)
  private

    function GetCollectionOwner: TPersistent;
    function GetItem(Index: Integer): TDBFFieldDef;
    procedure SetItem(Index: Integer; const Value: TDBFFieldDef);

  public

    constructor Create(AOwner: TPersistent);
    procedure AssignValues(Value: TDBFFieldDefs);
    function FindIndex(const Value: string): TDBFFieldDef;
    function IsEqual(Value: TDBFFieldDefs): Boolean;
    property Owner: TPersistent read GetCollectionOwner;
    property Items[Index: Integer]: TDBFFieldDef read GetItem write SetItem; default;

  end;

  {TDBFFieldDef}
  TDBFFieldDef = class(TCollectionItem)
  private

    FieldRec: FIELD_REC;
    Fdec: Word;
    Flen: Word;
    function GetField: FIELD_REC;
    function GetDataSize: Word;

  protected

    function GetDisplayName: string; override;
    procedure SetDisplayName(const Value: string); override;
    procedure AssignTo(Dest: TPersistent); override;

  public

    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    function IsEqual(Value: TDBFFieldDef): Boolean; virtual;

    property Field: FIELD_REC read GetField;
    property DataSize: Word read GetDataSize;

  published

    property Name: string read GetDisplayName write SetDisplayName;
    property field_type: char read FieldRec.field_type write FieldRec.field_type;
    property extend_type: TVKDBFType read FieldRec.extend_type write FieldRec.extend_type;
    property len: Word read Flen write Flen;
    property dec: Word read Fdec write Fdec;

  end;

  {TDBTStream}
  TDBTStream = class(TMemoryStream)
  protected
    FModified: boolean;
    FSmartDBF: TSmartDBF;
    FField: TField;
  public
    constructor Create;
    constructor CreateDBTStream(dbf: TSmartDBF; field: TField);
    destructor Destroy; override;
    procedure Clear;
    procedure SetSize(NewSize: Longint); override;
    procedure SaveToDBT;
    function Write(const Buffer; Count: Longint): Longint; override;
    property SmartDBF: TSmartDBF read FSmartDBF write FSmartDBF;
    property Field: TField read FField write FField;
  end;

  {TSmartDBF}
  TSmartDBF = class(TDataSet)
  private

    FDataLink: TVKDataLink;
    FDBFFieldDefs: TDBFFieldDefs;
    FIndRecBuf: pChar;
    FIndState: boolean;
    FLocateBuffer: pChar;
    FMasterFields: String;
    FRange: boolean;
    ListMasterFields: TList;

    FAddBuffered: boolean;
    FAddBuffer: pChar;
    FAddBufferCount: Integer;
    FAddBufferCurrent: Integer;

    procedure SetRngInt;
    function GetRecBuf: pChar;
    function GetRecNoBuf: Longint;
    procedure SetDataSource(const Value: TDataSource);
    function GetCretateNow: Boolean;
    procedure SetCretateNow(const Value: Boolean);
    procedure SetMasterFields(const Value: String);
    function GetMasterFields: Variant;
    procedure SetDBFFieldDefs(const Value: TDBFFieldDefs);

  private
    { Private declarations }
    FStreamedActive: boolean;
    FTempRecord: pChar;
    FFilterRecord: pChar;
    FFilterParser: To8ExprParser;
    FDBFFileName: String;
    DBFHandler: Integer;
    LobHandler: Integer;
    DBFHeader: DBF_HEAD;
    FRecordSize: Integer;
    FAccessMode: TAccessMode;
    FFieldInfo: TList;
    FOEM: Boolean;
    FSetDeleted: Boolean;
    FIndexes: TIndexes;
    FTmpActive: boolean;
    FKeyCalk: boolean;
    FWaitBusyRes: Integer;

    FBufferSize: Integer;
    FRecordsPerBuf: Integer;
    FBuffer: pChar;
    FBufInd: pLongint;
    FBufCnt: Longint;
    FBufDir: TBufDirection;
    FCurInd: Integer;

    FBOF: boolean;
    FEOF: boolean;

    FFileLock: boolean;
    FLockRecords: TList;

    FFastPostRecord: boolean;

    function GetRecordBufferSize: Integer;
    property RecordBufferSize: Integer read GetRecordBufferSize;
    property RecBuf: pChar read GetRecBuf;
    property RecNoBuf: Longint read GetRecNoBuf;
    function GetActiveRecBuf(var pRecBuf: PChar): Boolean;
    function GetDeleted: Boolean;
    //function FindFieldInfo(FieldName: String): PFieldInfo;
    function GetRecordByBuffer(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
    function AcceptRecord: Boolean;
    procedure SetSetDeleted(const Value: Boolean);
    procedure ReadIndexData(Reader: TReader);
    procedure WriteIndexData(Writer: TWriter);
    procedure SetIndexList(const Value: TIndexes);
    function FileRead(Handle: Integer; var Buffer; Count: LongWord): Integer;
    function FileWrite(Handle: Integer; const Buffer; Count: LongWord): Integer;
  protected
    { Protected declarations }
    procedure SetActive(Value: boolean); override;
    procedure Loaded; override;

    function AllocRecordBuffer: PChar; override;
    procedure FreeRecordBuffer(var Buffer: PChar); override;
    procedure GetBookmarkData(Buffer: PChar; Data: Pointer); override;
    function GetBookmarkFlag(Buffer: PChar): TBookmarkFlag; override;
    function GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult; override;
    function GetRecordSize: Word; override;
    procedure InternalAddRecord(Buffer: Pointer; Append: Boolean); override;
    procedure InternalClose; override;
    procedure InternalDelete; override;
    procedure InternalFirst; override;
    procedure InternalGotoBookmark(Bookmark: Pointer); override;
    procedure InternalHandleException; override;
    procedure InternalInitFieldDefs; override;
    procedure InternalInitRecord(Buffer: PChar); override;
    procedure InternalLast; override;
    procedure InternalOpen; override;
    procedure InternalEdit; override;
    procedure InternalPost; override;
    procedure InternalSetToRecord(Buffer: PChar); override;
    function IsCursorOpen: Boolean; override;
    procedure SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag); override;
    procedure SetBookmarkData(Buffer: PChar; Data: Pointer); override;
    procedure SetFieldData(Field: TField; Buffer: Pointer); override;
    procedure InternalRefresh; override;


    function FindRecord(Restart, GoForward: Boolean): Boolean; override;
    function LocateRecord(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions): Integer;

    procedure SetRange(FieldList: String; FieldValues: array of const); overload; virtual;
    procedure SetRange(FieldList: String; FieldValues: variant); overload; virtual;
    procedure ClearRange; virtual;

    procedure NextIndexBuf;
    procedure PriorIndexBuf;

    function NextBuffer: Longint;
    function PriorBuffer: Longint;
    procedure GetBufferByRec(Rec: Longint);

    function GetRecordCount: Integer; override;
    function GetRecNo: Integer; override;
    procedure SetRecNo(Value: Integer); override;

    function GetStateFieldValue(State: TDataSetState; Field: TField): Variant; override;
    function CompareLocateField(const Fields: TList; const KeyValues: Variant; Options: TLocateOptions): Integer;

    procedure SetFiltered(Value: Boolean); override;

    function GetDataSource: TDataSource; override;

    function GetCurIndByRec(nRec: Longint): Integer;

    function LockHeader: boolean;
    function UnlockHeader: boolean;

  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    function GetFieldData(Field: TField; Buffer: Pointer): Boolean; override;
    function Translate(Src, Dest: PChar; ToOem: Boolean): Integer; override;

    function CreateBlobStream(Field: TField; Mode: TBlobStreamMode): TStream; override;
    procedure SaveToDBT(Source: TMemoryStream; Field: TField);

    procedure CreateTable;
    procedure Reindex;

    procedure DefineProperties(Filer: TFiler); override;

    property LockRecords: TList read FLockRecords;

    property Deleted: Boolean read GetDeleted;
    property Header: DBF_HEAD read DBFHeader;
    property Handle: Integer read DBFHandler;
    property FieldInfo: TList read FFieldInfo;
    property IndRecBuf: pChar read FIndRecBuf write FIndRecBuf;
    property IndState: boolean read FIndState write FIndState;

    function FirstByIndex(IndInd: Integer): TGetResult;
    function PriorByIndex(IndInd: Integer): TGetResult;
    function NextByIndex(IndInd: Integer): TGetResult;
    function LastByIndex(IndInd: Integer): TGetResult;

    procedure AddRecord(const Values: array of const); overload;
    procedure AddRecord(ne: TNotifyEvent); overload;
    procedure AddRecord(const Values: variant); overload;
    procedure BeginAddRecord;
    procedure EndAddRecord;
    procedure SetTmpRecord(nRec: DWORD);
    procedure CloseTmpRecord;

    procedure BeginAddBuffered(RecInBuffer: Integer);
    procedure FlushAddBuffer;
    procedure EndAddBuffered;

    function Locate(const KeyFields: string; const KeyValues: Variant;
      Options: TLocateOptions): Boolean; override;
    function Lookup(const KeyFields: string; const KeyValues: Variant;
      const ResultFields: string): Variant; override;

    function GetPrec(aField: TField): Integer;
    function GetLen(aField: TField): Integer;

    function RLock: Boolean; overload;
    function RLock(nRec: Integer): Boolean; overload;

    function RUnLock: Boolean; overload;
    function RUnLock(nRec: Integer): Boolean; overload;

    function FLock: Boolean;
    function UnLock: Boolean;

    procedure SetOrder(nOrd: Integer); overload;
    procedure SetOrder(sOrd: ShortString); overload;

    property Last_Rec: Longint read DBFHeader.last_rec;

  published

    { Published declarations }
    property Active;
    property DBFFileName: String read FDBFFileName write FDBFFileName;
    property AccessMode: TAccessMode read FAccessMode write FAccessMode;
    property BufferSize: Integer read FBufferSize write FBufferSize;
    property OEM: Boolean read FOEM write FOEM default false;
    property SetDeleted: Boolean read FSetDeleted write SetSetDeleted;
    property Indexes: TIndexes read FIndexes write SetIndexList stored false;
    property DBFFieldDefs: TDBFFieldDefs read FDBFFieldDefs write SetDBFFieldDefs stored false;
    property WaitBusyRes: Integer read FWaitBusyRes write FWaitBusyRes;
    property DataSource: TDataSource read GetDataSource write SetDataSource;
    property CretateNow: Boolean read GetCretateNow write SetCretateNow;
    property MasterFields: String read FMasterFields write SetMasterFields;

    property Filter;
    property Filtered;
    property FilterOptions;

    property BeforeOpen;
    property AfterOpen;
    property BeforeClose;
    property AfterClose;
    property BeforeInsert;
    property AfterInsert;
    property BeforeEdit;
    property AfterEdit;
    property BeforePost;
    property AfterPost;
    property BeforeCancel;
    property AfterCancel;
    property BeforeDelete;
    property AfterDelete;
    property BeforeScroll;
    property AfterScroll;
    property OnCalcFields;
    property OnDeleteError;
    property OnEditError;
    property OnFilterRecord;
    property OnNewRecord;
    property OnPostError;

  end;

  {TDBFNTX}
  TDBFNTX = class(TSmartDBF)
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure SetRange(FieldList: String; FieldValues: array of const); overload; override;
    procedure SetRange(FieldList: String; FieldValues: variant); overload; override;
    procedure ClearRange; override;

  end;

  {TDBFCDX}
  TDBFCDX = class(TSmartDBF)
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

procedure Wait(t: double; l: boolean = true);
function Space(iSize: Integer): String;
function Zerro(iSize: Integer): String;
function DoEncodeDate(Year, Month, Day: Word; var Date: TDateTime): Boolean;
function IsBlank(Buff: pChar; BufLen: Integer): boolean;
procedure Register;

implementation

//******************************************************************************
procedure Register;
begin
  RegisterComponents('VK DBF', [TDBFDataSet, TDBFNTX]);
  RegisterPropertyEditor(TypeInfo(String), TSmartDBF, 'DBFFileName', TEdOpenDBF);
  RegisterPropertyEditor(TypeInfo(String), TNTXIndex, 'NTXFileName', TEdOpenInd);
end;

//******************************************************************************
procedure Wait(t: double; l: boolean = true);
var
  t1: TDateTime;
begin
  t1 := Now;
  while (Now - t1) < (0.0000115741 * t) do
    if l then Application.ProcessMessages;
end;

//******************************************************************************
function Space(iSize: Integer): String;
var
  i: Integer;
begin
  Result := '';
  for i := 1 to iSize do
    Result := Result + ' ';
end;

//******************************************************************************
function Zerro(iSize: Integer): String;
var
  i: Integer;
begin
  Result := '';
  for i := 1 to iSize do
    Result := Result + #0;
end;

//******************************************************************************
function DoEncodeDate(Year, Month, Day: Word; var Date: TDateTime): Boolean;
var
  I: Integer;
  DayTable: PDayTable;
begin
  Result := False;
  DayTable := @MonthDays[IsLeapYear(Year)];
  if (Year >= 1) and (Year <= 9999) and (Month >= 1) and (Month <= 12) and
    (Day >= 1) and (Day <= DayTable^[Month]) then
  begin
    for I := 1 to Month - 1 do Inc(Day, DayTable^[I]);
    I := Year - 1;
    Date := I * 365 + I div 4 - I div 100 + I div 400 + Day - DateDelta;
    Result := True;
  end;
end;

//******************************************************************************
function IsBlank(Buff: pChar; BufLen: Integer): boolean;
var
  i, j: Integer;
begin
  Result := true;
  j := BufLen - 1;
  for i := 0 to j do
  begin
    if Buff[i] <> #32 then
    begin
      Result := false;
      break;
    end;
  end;
end;

{ TDBFDataSet }

function TDBFDataSet.AllocRecordBuffer: PChar;
begin
  Result := MemMgr.oMem.GetMem(self, RecordBufferSize);
end;

function TDBFDataSet.CompareLocateField(const Fields: TList;
  const KeyValues: Variant; Options: TLocateOptions): Integer;
begin
  Result := 0;
end;

constructor TDBFDataSet.Create(AOwner: TComponent);
var
  FieldMap: TFieldMap;
begin
  inherited Create(AOwner);
  FFilterParser := To8ExprParser.Create(self, '', [], [poExtSyntax], '', nil, FieldMap);
  FAccessMode := TAccessMode.Create;
  FFieldInfo := TList.Create;
  FOEM := false;
  DBFHandler := -1;
  FCurrentIndex := -1;
  FCurrGranul := 0;
  FDBFFileName := '';
  FBufferSize := 4096;    //4 Kb
  BufferDBF := MemMgr.oMem.GetMem(self, FBufferSize);
  FillChar(BufferDBF^, FBufferSize, #0);
end;

destructor TDBFDataSet.Destroy;
var
  i: Integer;
begin
  Active := false;
  MemMgr.oMem.FreeMem(BufferDBF);
  if FFieldInfo.Count <> 0 then
    for i := 0 to FFieldInfo.Count - 1 do
      MemMgr.oMem.FreeMem(FFieldInfo.Items[i]);
  FFieldInfo.Clear;
  FFieldInfo.DeStroy;
  FAccessMode.Destroy;
  FFilterParser.Destroy;
  inherited Destroy;
end;

procedure TDBFDataSet.SetBufferSize(const Value: Integer);
begin
  if not Active then
  begin
    if FBufferSize <> Value then
    begin
      FBufferSize := Value;
      MemMgr.oMem.FreeMem(BufferDBF);
      BufferDBF := MemMgr.oMem.GetMem(self, FBufferSize);
      FillChar(BufferDBF^, FBufferSize, #0);
    end;
  end;
end;

procedure TDBFDataSet.FreeRecordBuffer(var Buffer: PChar);
begin
  MemMgr.oMem.FreeMem(Buffer);
end;

function TDBFDataSet.GetActiveRecBuf(var RecBuf: PChar): Boolean;
begin
  case State of
    dsBrowse: if IsEmpty then RecBuf := nil else RecBuf := ActiveBuffer;
    dsEdit, dsInsert: RecBuf := ActiveBuffer;
    dsFilter: RecBuf := FFilterRecord;
    dsNewValue, dsOldValue, dsCurValue: RecBuf := ActiveBuffer;
    dsCalcFields: RecBuf := CalcBuffer;
  else
    RecBuf := nil;
  end;
  Result := RecBuf <> nil;
end;

procedure TDBFDataSet.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  Longword(Data^) := pTRecInfo(Buffer + FRecordSize).RecordRowID;
end;

function TDBFDataSet.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := pTRecInfo(Buffer + FRecordSize).BookmarkFlag;
end;

function TDBFDataSet.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  iCode, dInt: Integer;
  dFloat: Double;
  ww: Extended;
  dBool: WordBool;
  dDate: TDateTime;
  sTS: TTimeStamp;
  dd: double;
  Year, Month, Day: Word;
  ss: pChar;
  ss1: array [0..255] of Char;
  ActiveBuf: pChar;
  qq: PFieldInfo;
  oDS: TDataSet;
  LookupResult: Variant;
  LastSetp: Char;
  //sStr: String;
  Len: Integer;
begin
  Result := false;
  case Field.FieldKind of
  fkData:
    begin
      qq := PFieldInfo(FFieldInfo.Items[Field.FieldNo - 1]);
      GetActiveRecBuf(ActiveBuf);
      if ActiveBuf <> nil then
      begin
        if Buffer <> nil then
        begin
          ss := ActiveBuf + qq.FOff;
          if IsBlank(ss, qq.FLen) then
          begin
            Result := false;
            exit;
          end;
          case Field.DataType of
            ftString:
              begin
                //sStr := String(ss);
                //SetString(sStr, ss, qq.FLen);
                //SetLength(sStr, qq.FLen);
                //sStr := TrimRight(sStr);
                //Len := Length(sStr);
                //Move(pChar(sStr)^, Buffer^, qq.FLen);
                Move(ss^, Buffer^, qq.FLen);
                Len := qq.FLen;
                repeat
                  pChar(Buffer)[Len] := #0;
                  Dec(Len);
                until not ((Len >= 0) and (pChar(Buffer)[Len] = ' '));
                Result := true;
              end;
            ftInteger:
              begin
                Move(ss^, ss1, qq.FLen);
                ss1[qq.FLen] := #0;
                //SetString(ss1, ss, qq.FLen);
                Val(ss1, dInt, iCode);
                if iCode = 0 then
                begin
                  Integer(Buffer^) := dInt;
                  Result := true;
                end else
                  Result := false;
              end;
            ftFloat:
              begin
                Move(ss^, ss1, qq.FLen);
                ss1[qq.FLen] := #0;
                //SetString(ss1, ss, qq.FLen);
                LastSetp := DecimalSeparator;
                DecimalSeparator := '.';
                if TextToFloat(ss1, ww, fvExtended) then
                begin
                  dFloat := ww;
                  double(Buffer^) := dFloat;
                  Result := true;
                end else
                  Result := false;
                DecimalSeparator := LastSetp;
              end;
            ftMemo:
              begin
                Result := false;
              end;
            ftDateTime:
              begin
                Year := (Byte(ss[0]) - $30) * 1000 + (Byte(ss[1]) - $30) * 100 + (Byte(ss[2]) - $30) * 10 + (Byte(ss[3]) - $30);
                Month := (Byte(ss[4]) - $30) * 10 + (Byte(ss[5]) - $30);
                Day := (Byte(ss[6]) - $30) * 10 + (Byte(ss[7]) - $30);
                if DoEncodeDate(Year, Month, Day, dDate) then
                begin
                  sTS := DateTimeToTimeStamp(dDate);
                  dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                  double(Buffer^) := dd;
                  Result := true;
                end else
                  Result := false;
              end;
            ftBoolean:
              begin
                if ss[0] = 'T' then
                  dBool := true
                else
                  dBool := false;
                WordBool(Buffer^) := dBool;
                Result := true;
              end;
          end;
        end;  // else
          //Result := false;
      end else
        Result := false;
    end;
  fkCalculated:
    begin
      GetActiveRecBuf(ActiveBuf);
      if Buffer <> nil then
      begin
        if (ActiveBuf <> nil) and not (csDesigning in ComponentState) then
        begin
          ss := ActiveBuf + FRecordSize + sizeof(TRecInfo) + Field.Offset;
          Move(ss^, Buffer^, Field.DataSize);
          if Field.DataType in [ftString, ftBytes, ftVarBytes]  then pChar(Buffer)[Field.DataSize] := Char(0);
          Result := true;
        end else begin
          FillChar(Buffer^, Field.DataSize, 0);
          Result := false;
        end;
      end else
        Result := false;
    end;
  fkLookup:
    begin
      Result := false;
      if Buffer <> nil then
      begin
        if GetActiveRecBuf(ActiveBuf) then
        begin
          oDS := Field.LookupDataSet;
          if (oDS <> nil) and oDS.Active then
          begin
            LookupResult := oDS.Lookup(Field.LookupKeyFields, FieldValues[Field.KeyFields], Field.LookupResultField);
            if (not VarIsEmpty(LookupResult)) and (not VarIsNull(LookupResult)) then
            begin
              if Field.DataType = ftString then
              begin
                ss := TVarData(LookupResult).VPointer;
                Move(ss^, Buffer^, Length(ss) + 1);
              end else begin
                ss := TVarData(LookupResult).VPointer;
                Move(ss^, Buffer^, Field.DataSize);
              end;
              Result := true;
            end;
          end;
        end else begin
          FillChar(Buffer^, Field.DataSize, 0);
          Result := false;
        end;
      end;
    end;
  end;
end;

function TDBFDataSet.GetPortion: Integer;
begin
  Result := FBufferSize div FRecordSize;
end;

function TDBFDataSet.GetPortionSize: Integer;
begin
  Result := Portion * FRecordSize;
end;

function TDBFDataSet.GetRecNo: Integer;
var
  ActiveBuf: pChar;
begin
  Result := -1;
  GetActiveRecBuf(ActiveBuf);
  if ActiveBuf <> nil then
    Result := pTRecInfo(ActiveBuf + RecordSize).RecordRowID;
end;

procedure TDBFDataSet.SetRecNo(Value: Integer);
begin
  CheckBrowseMode;
  CursorPosChanged;
  DoBeforeScroll;
  SetRecNoInternal(Value);
  Resync([rmExact, rmCenter]);
  DoAfterScroll;
end;

procedure TDBFDataSet.SetRecNoInternal(Value: Integer);
var
  RealRead: Integer;
  CurBuf: Integer;
begin
  if Value < 1 then Value := 1;
  if Value > DBFHeader.last_rec then Value := DBFHeader.last_rec;
  CurBuf := (RecNo - 1) div Portion;
  if CurBuf <> ((Value - 1) div Portion) then
  begin
    CurBuf := (Value - 1) div Portion;
    FileSeek(DBFHandler, DBFHeader.data_offset + LongWord(CurBuf * Portion * FRecordSize), 0);
    RealRead := FileRead(DBFHandler, BufferDBF^, PortionSize);
    FLastGranul := RealRead div FRecordSize;
  end;
  FCurrGranul := Value - (CurBuf * Portion) - 1;
  FCurrentIndex := Value - 1;
end;

function TDBFDataSet.GetRecord(Buffer: PChar; GetMode: TGetMode; DoCheck: Boolean): TGetResult;
var
  SaveState: TDataSetState;
  Accept: Boolean;
begin
  if not Filtered then
  begin
    if not FSetDeleted then
      Result := GetRecordByBuffer(Buffer, GetMode, DoCheck)
    else begin
      if GetMode <> gmCurrent then
      begin
        repeat
          Result := GetRecordByBuffer(FFilterRecord, GetMode, DoCheck);
          SaveState := SetTempState(dsFilter);
          Accept := not Deleted;
          RestoreState(SaveState);
        until Accept or (Result <> grOK);
        if Accept then
          Move(FFilterRecord^, Buffer^, RecordBufferSize)
        else
          Move(FTempRecord^, Buffer^, RecordBufferSize);
      end else
        Result := GetRecordByBuffer(Buffer, GetMode, DoCheck);
    end;
  end else begin
    if GetMode <> gmCurrent then
    begin
      repeat
        Result := GetRecordByBuffer(FFilterRecord, GetMode, DoCheck);
        SaveState := SetTempState(dsFilter);
        Accept := AcceptRecord;
        RestoreState(SaveState);
      until Accept or (Result <> grOK);
      if Accept then
        Move(FFilterRecord^, Buffer^, RecordBufferSize)
      else
        Move(FTempRecord^, Buffer^, RecordBufferSize);
    end else
      Result := GetRecordByBuffer(Buffer, GetMode, DoCheck);
  end;
end;

function TDBFDataSet.GetRecordByBuffer(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
var
  cc: PChar;
  RealRead: Integer;
  CurBuf: Integer;
begin
  Result := grOK;
  case GetMode of
    gmCurrent:
      begin
        if FCurrentIndex <> -1 then
        begin
          if FCurrentIndex < Longint(DBFHeader.last_rec) then
          begin
            cc := pChar(Buffer);
            InternalInitRecord(cc);
            pTRecInfo(cc + FRecordSize).BookmarkFlag := bfCurrent;
            pTRecInfo(cc + FRecordSize).RecordRowID := FCurrentIndex + 1;
            pTRecInfo(cc + FRecordSize).UpdateStatus := usUnmodified;
            Move((pChar(BufferDBF) + FCurrGranul * FRecordSize)^, cc^, FRecordSize);
          end else begin
            InternalInitRecord(FTempRecord);
            Move(FTempRecord^, Buffer^, RecordBufferSize);
            Result := grError;
          end;
        end else begin
          InternalInitRecord(FTempRecord);
          Move(FTempRecord^, Buffer^, RecordBufferSize);
          Result := grError;
        end;
      end;
    gmNext:
      begin
        if FCurrentIndex < Longint(DBFHeader.last_rec) then
        begin
          if FCurrentIndex <> -1 then
          begin
            Inc(FCurrGranul);
            if FCurrGranul = FLastGranul then
            begin
              RealRead := FileRead(DBFHandler, BufferDBF^, PortionSize);
              FLastGranul := RealRead div FRecordSize;
              FCurrGranul := 0;
            end;
          end;
          Inc(FCurrentIndex);
          if not (FCurrentIndex = Longint(DBFHeader.last_rec)) then
          begin
            cc := pChar(Buffer);
            InternalInitRecord(cc);
            pTRecInfo(cc + FRecordSize).BookmarkFlag := bfCurrent;
            pTRecInfo(cc + FRecordSize).RecordRowID := FCurrentIndex + 1;
            pTRecInfo(cc + FRecordSize).UpdateStatus := usUnmodified;
            Move((pChar(BufferDBF) + FCurrGranul * FRecordSize)^, cc^, FRecordSize);
          end else begin
            Dec(FCurrentIndex);
            InternalInitRecord(FTempRecord);
            Move(FTempRecord^, Buffer^, RecordBufferSize);
            Result := grEOF;
          end;
        end else begin
          FCurrentIndex := Longint(DBFHeader.last_rec);
          InternalInitRecord(FTempRecord);
          Move(FTempRecord^, Buffer^, RecordBufferSize);
          Result := grEOF;
        end;
      end;
    gmPrior:
      begin
        if FCurrentIndex <> Longint(DBFHeader.last_rec) then
        begin
          Dec(FCurrGranul);
          if FCurrGranul = -1 then
          begin
            CurBuf := (FCurrentIndex - 1) div Portion;
            FileSeek(DBFHandler, DBFHeader.data_offset + LongWord(CurBuf * Portion * FRecordSize), 0);
            RealRead := FileRead(DBFHandler, BufferDBF^, PortionSize);
            FLastGranul := RealRead div FRecordSize;
            FCurrGranul := FLastGranul - 1;
          end;
        end;
        if not (FCurrentIndex = -1) then Dec(FCurrentIndex);
        if not (FCurrentIndex = -1) then
        begin
          cc := pChar(Buffer);
          InternalInitRecord(cc);
          pTRecInfo(cc + FRecordSize).BookmarkFlag := bfCurrent;
          pTRecInfo(cc + FRecordSize).RecordRowID := FCurrentIndex + 1;
          pTRecInfo(cc + FRecordSize).UpdateStatus := usUnmodified;
          Move((pChar(BufferDBF) + FCurrGranul * FRecordSize)^, cc^, FRecordSize);
        end else begin
          FCurrentIndex := -1;
          InternalInitRecord(FTempRecord);
          Move(FTempRecord^, Buffer^, RecordBufferSize);
          Result := grBOF;
        end;
      end;
  end;
  if Result = grOK then
    GetCalcFields(Buffer);
end;

function TDBFDataSet.GetRecordBufferSize: Integer;
begin
  Result := FRecordSize + sizeof(TRecInfo) + CalcFieldsSize;
end;

function TDBFDataSet.GetRecordCount: Integer;
begin
  Result := DBFHeader.last_rec;
end;

function TDBFDataSet.GetRecordSize: Word;
begin
  Result := FRecordSize;
end;

function TDBFDataSet.GetStateFieldValue(State: TDataSetState;
  Field: TField): Variant;
begin
  Result := NULL;
  if State in [dsNewValue, dsCurValue, dsOldValue] then
    Result := inherited GetStateFieldValue(State, Field);
end;

procedure TDBFDataSet.InternalAddRecord(Buffer: Pointer; Append: Boolean);
var
  RealRead: Integer;
  lpMsgBuf: array [0..500] of Char;
  le: DWORD;
begin
  CheckActive;
  pTRecInfo(pChar(Buffer) + FRecordSize).RecordRowID := DBFHeader.last_rec + 1;
  FileSeek(DBFHandler, 0, 2);
  FileSeek(DBFHandler, DBFHeader.data_offset + LongWord(DBFHeader.last_rec * FRecordSize), 0);
  RealRead := FileWrite(DBFHandler, Buffer^, DBFHeader.rec_size);
  if RealRead = -1 then
  begin
    le := GetLastError();
    FormatMessage(
        FORMAT_MESSAGE_FROM_SYSTEM,
        nil,
        le,
        0, // Default language
        lpMsgBuf,
        500,
        nil
    );
    raise Exception.Create('TDBFDataSet: ' + lpMsgBuf);
  end else begin
    Inc(DBFHeader.last_rec);
    FileSeek(DBFHandler, 0, 0); //go to the begin
    RealRead := FileWrite(DBFHandler, DBFHeader, SizeOf(DBFHeader));
    if RealRead = -1 then
    begin
      le := GetLastError();
      FormatMessage(
          FORMAT_MESSAGE_FROM_SYSTEM,
          nil,
          le,
          0, // Default language
          lpMsgBuf,
          500,
          nil
      );
      raise Exception.Create('TDBFDataSet: ' + lpMsgBuf);
    end else begin
      InternalSetToRecord(Buffer);
      Move(Buffer^, (pChar(BufferDBF) + FCurrGranul * FRecordSize)^, FRecordSize);
    end;
  end;
end;

procedure TDBFDataSet.InternalDelete;
begin
  //
end;

procedure TDBFDataSet.InternalFirst;
var
  RealRead: Integer;
begin
  FileSeek(DBFHandler, DBFHeader.data_offset, 0);
  RealRead := FileRead(DBFHandler, BufferDBF^, PortionSize);
  FCurrGranul := 0;
  FLastGranul := RealRead div FRecordSize;
  FCurrentIndex := -1;
end;

procedure TDBFDataSet.InternalGotoBookmark(Bookmark: Pointer);
begin
  SetRecNoInternal(pTRecInfo(Bookmark).RecordRowID);
end;

procedure TDBFDataSet.InternalHandleException;
begin
  Application.HandleException(self);
end;

procedure TDBFDataSet.InternalInitFieldDefs;
var
  DBFField: FIELD_REC;
  dbSize: Integer;
  dbOffset: Integer;
  i, j: Integer;
  qq: PFieldInfo;
begin
  FieldDefs.BeginUpdate;
  FieldDefs.Clear;
  if FFieldInfo.Count <> 0 then
    for i := 0 to FFieldInfo.Count - 1 do
      MemMgr.oMem.FreeMem(FFieldInfo.Items[i]);
  FFieldInfo.Clear;
  FileSeek(DBFHandler, SizeOf(DBFHeader), 0);
  dbOffset := 1;
  j := 0;
  while true do
  begin
    FileRead(DBFHandler, DBFField, SizeOf(FIELD_REC));
    if DBFField.field_name[0] = #13 then break;
    dbSize := 0;
    qq := PFieldInfo(MemMgr.oMem.GetMem(self, SizeOf(TFieldInfo)));
    FieldDefs.AddFieldDef;
    FieldDefs.Items[j].Name := UpperCase(Trim(DBFField.field_name));
    FieldDefs.Items[j].Required := false;
    case DBFField.field_type of
      'C':
        begin
          FieldDefs.Items[j].DataType := ftString;
          FieldDefs.Items[j].Size := DBFField.lendth.char_len;
          dbSize := DBFField.lendth.char_len;
          FieldDefs.Items[j].Precision := 0;
        end;
      'N':
        begin
          if DBFField.lendth.num_len.dec = 0 then
          begin
            FieldDefs.Items[j].DataType := ftInteger;
            FieldDefs.Items[j].Size := 0;
            dbSize := DBFField.lendth.num_len.len;
            FieldDefs.Items[j].Precision := 0;
          end else begin
            FieldDefs.Items[j].DataType := ftFloat;
            FieldDefs.Items[j].Size := 0;
            dbSize := DBFField.lendth.num_len.len;
            FieldDefs.Items[j].Precision := DBFField.lendth.num_len.dec;
          end;
        end;
      'D':
        begin
          FieldDefs.Items[j].DataType := ftDateTime;
          FieldDefs.Items[j].Size := 0;
          dbSize := 8;
          FieldDefs.Items[j].Precision := 0;
        end;
      'L':
        begin
          FieldDefs.Items[j].DataType := ftBoolean;
          FieldDefs.Items[j].Size := 0;
          dbSize := 1;
          FieldDefs.Items[j].Precision := 0;
        end;
      'M':
        begin
          FieldDefs.Items[j].DataType := ftMemo;
          FieldDefs.Items[j].Size := 0;
          dbSize := 10;
          FieldDefs.Items[j].Precision := 0;
        end;
    end;
    qq.FName  := FieldDefs.Items[j].Name;
    qq.FType  := DBFField.field_type;
    qq.FLen   := dbSize;
    qq.FPrec  := FieldDefs.Items[j].Precision;
    qq.FOff   := dbOffset;
    dbOffset  := dbOffset + dbSize;
    FFieldInfo.Add(qq);
    Inc(j);
  end;
  FieldDefs.EndUpdate;
end;

procedure TDBFDataSet.InternalInitRecord(Buffer: PChar);
begin
  if Buffer <> nil then FillChar(Buffer^, RecordBufferSize, #32);
end;

procedure TDBFDataSet.InternalLast;
begin
  FileSeek(DBFHandler, 0, 0);
  FileRead(DBFHandler, DBFHeader, SizeOf(DBF_HEAD));
  SetRecNoInternal(DBFHeader.last_rec);
  FCurrentIndex := DBFHeader.last_rec;
end;

procedure TDBFDataSet.InternalOpen;
begin
  DBFHandler := FileOpen(DBFFileName, AccessMode.AccessMode);
  if DBFHandler <= 0 then
    raise Exception.Create('TDBFDataSet: Open error');
  FileSeek(DBFHandler, 0, 0);
  FileRead(DBFHandler, DBFHeader, SizeOf(DBF_HEAD));
  if not ((DBFHeader.dbf_id = $03) or (DBFHeader.dbf_id = $07) or (DBFHeader.dbf_id = $83) or (DBFHeader.dbf_id = $8B)) then
    raise Exception.Create('TDBFDataSet: File is not DBF file');
  FRecordSize := DBFHeader.rec_size;
  FCurrentIndex := -1;
  ObjectView := false;
  InternalInitFieldDefs();
  if DefaultFields then CreateFields;
  BindFields(true);
  BookmarkSize := sizeof(Longword);
  FTempRecord := AllocRecordBuffer;
  FFilterRecord := AllocRecordBuffer;
  InternalFirst;
end;

procedure TDBFDataSet.InternalClose;
begin
  FreeRecordBuffer(FTempRecord);
  FreeRecordBuffer(FFilterRecord);
  BindFields(false);
  if DefaultFields then DestroyFields;
  FileClose(DBFHandler);
  DBFHandler := -1;
end;

procedure TDBFDataSet.InternalPost;
var

  l: Integer;
  fOffset: Integer;
  ActiveBuf: pChar;
  RealRead: Integer;

  lpMsgBuf: array [0..500] of Char;
  le: DWORD;

begin
  CheckActive;
  GetActiveRecBuf(ActiveBuf);
  l := pTRecInfo(ActiveBuf + RecordSize).RecordRowID;
  if State = dsEdit then
  begin
    fOffset := FileSeek(DBFHandler, 0, 1);
    try
      FileSeek(DBFHandler, DBFHeader.data_offset + LongWord((l - 1) * FRecordSize), 0);
      RealRead := FileWrite(DBFHandler, ActiveBuf^, DBFHeader.rec_size);
      if RealRead = -1 then
      begin
        le := GetLastError();
        FormatMessage(
            FORMAT_MESSAGE_FROM_SYSTEM,
            nil,
            le,
            0, // Default language
            lpMsgBuf,
            500,
            nil
        );
        raise Exception.Create('TDBFDataSet: ' + lpMsgBuf);
      end else
        Move(ActiveBuf^, (pChar(BufferDBF) + FCurrGranul * FRecordSize)^, FRecordSize);
    finally
      FileSeek(DBFHandler, fOffset, 0);
    end;
  end else begin
    InternalAddRecord(ActiveBuf, true);
  end;
end;

procedure TDBFDataSet.InternalSetToRecord(Buffer: PChar);
begin
  SetRecNoInternal(pTRecInfo(Buffer + RecordSize).RecordRowID);
end;

function TDBFDataSet.IsCursorOpen: Boolean;
begin
  Result := (DBFHandler > 0);
end;

procedure TDBFDataSet.Loaded;
begin
  inherited Loaded();
  if FStreamedActive then Active := true;
end;

procedure TDBFDataSet.SetActive(Value: boolean);
begin
  if csReading in ComponentState then
    FStreamedActive := Value
  else
    inherited SetActive(Value);
end;

procedure TDBFDataSet.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  pTRecInfo(Buffer + RecordSize).RecordRowID := Longword(Data^);
end;

procedure TDBFDataSet.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  pTRecInfo(Buffer + RecordSize).BookmarkFlag := Value;
end;

procedure TDBFDataSet.SetFieldData(Field: TField; Buffer: Pointer);
var
  ss, ActiveBuf: pChar;
  qq: PFieldInfo;
  dd: double;
  sTS: TTimeStamp;
  Year, Month, Day: Word;
  dInt: Integer;
  dFloat: double;
  dBool: WordBool;
  q: String;
  p0, p1, p2, p3: byte;
  wE: boolean;
  w: char;
  i: Integer;
begin
  case Field.FieldKind of
  fkData:
    begin
      qq := PFieldInfo(FFieldInfo.Items[Field.FieldNo - 1]);
      GetActiveRecBuf(ActiveBuf);
      if (Buffer <> nil) and (ActiveBuf <> nil) then
      begin
        ss := ActiveBuf + qq.FOff;
        case Field.DataType of
          ftString:
            begin
              wE := false;
              for i := 0 to qq.FLen - 1 do
              begin
                w := pChar(Buffer)[i];
                if w = #0 then
                  wE := true;
                if not wE then
                  ss[i] := w
                else
                  ss[i] := ' ';
              end;
            end;
          ftDateTime, ftDate:
            begin
              dd := double(Buffer^);
              sTS.Date := Trunc(dd/(3600.0*24*1000));
              sTS.Time := Trunc(dd - sTS.Date*((3600.0*24*1000)));
              dd := TimeStampToDateTime(sTS);
              DecodeDate(dd, Year, Month, Day);
              p0 := Year div 1000;
              ss[0] := char( p0 + $30 );
              p1 := (Year - p0 * 1000) div 100;
              ss[1] := Char( p1 + $30 );
              p2 := (Year - p0 * 1000 - p1 * 100) div 10;
              ss[2] := Char( p2 + $30 );
              p3 := (Year - p0 * 1000 - p1 * 100 - p2 * 10);
              ss[3] := Char( p3 + $30 );
              ss[4] := Char( (Month div 10) + $30 );
              ss[5] := Char( (Month - (Month div 10) * 10 ) + $30 );
              ss[6] := Char( (Day div 10) + $30 );
              ss[7] := Char( (Day - (Day div 10) * 10) + $30 );
            end;
          ftInteger:
            begin
              dInt := Integer(Buffer^);
              Str(dInt:qq.FLen, q);
              Move(pChar(q)^, ss^, qq.FLen);
            end;
          ftFloat:
            begin
              dFloat := Double(Buffer^);
              Str(dFloat:qq.FLen:qq.FPrec, q);
              Move(pChar(q)^, ss^, qq.FLen);
            end;
          ftBoolean:
            begin
              dBool := WordBool(Buffer^);
              if dBool then
                 ss[0] := 'T'
              else
                ss[0] := 'F';
            end;
        end;
      end;
    end;
  fkCalculated:
    begin
      GetActiveRecBuf(ActiveBuf);
      if (Buffer <> nil) and (ActiveBuf <> nil) then
      begin
        ss := ActiveBuf + FRecordSize + sizeof(TRecInfo) + Field.Offset;
        Move(Buffer^, ss^, Field.DataSize);
      end;
    end;
  end;
  if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
    DataEvent(deFieldChange, Longint(Field));
end;

function TDBFDataSet.GetDeleted: Boolean;
var
  ActBuff: pChar;
begin
  Result := false;
  GetActiveRecBuf(ActBuff);
  if ActBuff <> nil then
  begin
    if ActBuff[0] = #42 then
      Result := true;
  end;
end;

function TDBFDataSet.Translate(Src, Dest: PChar; ToOem: Boolean): Integer;
begin
  if FOEM then
  begin
    if not ToOem then begin
      if (Src <> nil) then
      begin
        if OemToChar(Src, Dest) then
          Result := StrLen(Dest)
        else
          Result := 0;
      end else
        Result := 0;
    end else begin
      if (Src <> nil) then
      begin
        if CharToOem(Src, Dest) then
          Result := StrLen(Dest)
        else
          Result := 0;
      end else
        Result := 0;
    end;
  end else
    Result := inherited Translate(Src, Dest, ToOem);
end;

{function TDBFDataSet.FindFieldInfo(FieldName: String): PFieldInfo;
var
  i: Integer;
begin
  Result := nil;
  for i := 0 to FFieldInfo.Count - 1 do
  begin
    Result := PFieldInfo(FFieldInfo.Items[i]);
    if Result.FName = FieldName then Break;
  end;
end;
}

procedure TDBFDataSet.SetFiltered(Value: Boolean);
begin
  if Active then
  begin
    CheckBrowseMode;
    if ((not Filtered) and Value) and (Filter <> '') then
      FFilterParser.SetExprParams(Filter, FilterOptions, [poExtSyntax], '');
    if Filtered <> Value then
      inherited SetFiltered(Value);
    First;
  end else
    inherited SetFiltered(Value);
end;

function TDBFDataSet.FindRecord(Restart, GoForward: Boolean): Boolean;
var
  SaveState: TDataSetState;
  Accept: Boolean;
  Ret: TGetResult;
begin
  if (not Filtered) and (Filter <> '') then
      FFilterParser.SetExprParams(Filter, FilterOptions, [poExtSyntax], '');
  CheckBrowseMode;
  DoBeforeScroll;
  SetFound(False);
  UpdateCursorPos;
  CursorPosChanged;
  if GoForward then
  begin
    if Restart then InternalFirst;
    repeat
      Ret := GetRecordByBuffer(FFilterRecord, gmNext, false);
      SaveState := SetTempState(dsFilter);
      Accept := AcceptRecord;
      RestoreState(SaveState);
    until Accept or (Ret <> grOK);
  end else
  begin
    if Restart then InternalLast;
    repeat
      Ret := GetRecordByBuffer(FFilterRecord, gmPrior, false);
      SaveState := SetTempState(dsFilter);
      Accept := AcceptRecord;
      RestoreState(SaveState);
    until Accept or (Ret <> grOK);
  end;
  if Ret = grOK then
  begin
    Resync([rmExact, rmCenter]);
    SetFound(True);
  end else
    InternalSetToRecord(ActiveBuffer);
  Result := Found;
  if Result then DoAfterScroll;
end;

function TDBFDataSet.AcceptRecord: Boolean;
begin
  Result := true;
  if Assigned(OnFilterRecord) then
    OnFilterRecord(self, Result)
  else if Filter <> '' then
    Result := FFilterParser.Execute;
  if FSetDeleted then
    Result := Result and Deleted;
end;

procedure TDBFDataSet.SetSetDeleted(const Value: Boolean);
begin
  if Active then
  begin
    CheckBrowseMode;
    if FSetDeleted <> Value then
      FSetDeleted := Value;
    First;
  end else
    FSetDeleted := Value;
end;

procedure TDBFDataSet.CreateTable;
var
  cf: Integer;
  cHeader: DBF_HEAD;
  cField: FIELD_REC;
  i, j: Integer;
  Year, Month, Day: Word;
  qq: byte;
begin
  if not Active then
  begin
    cf := FileCreate(DBFFileName);
    if cf > 0 then
    begin
      DecodeDate(Now, Year, Month, Day);
      with cHeader do
      begin
        dbf_id := $03;
        last_update[0] := Byte(Year);
        last_update[1] := Byte(Month);
        last_update[2] := Byte(Day);
        last_rec := 0;
        data_offset := 1;
        rec_size := 1;
        for i := 0 to 15 do
          filler[i] := 0;
        prod_ind := 0;
        lang := 0;
        dummy := 0;
      end;
      FileSeek(cf, 0, 0);
      FileWrite(cf, cHeader, SizeOf(DBF_HEAD));
      cHeader.data_offset := cHeader.data_offset + SizeOf(DBF_HEAD);
      for i := 0 to FieldDefs.Count - 1 do
      begin
        for j := 0 to 10 do
          cField.field_name[j] := #0;
        j := 1;
        while j <= Length(FieldDefs.Items[i].Name) do
        begin
          cField.field_name[j - 1] := FieldDefs.Items[i].Name[j];
          Inc(j);
        end;
        case FieldDefs.Items[i].DataType of
          ftString:
            begin
              cField.field_type := 'C';
              cField.lendth.char_len := FieldDefs.Items[i].Size;
              cHeader.rec_size := cHeader.rec_size + FieldDefs.Items[i].Size;
            end;
          ftInteger:
            begin
              cField.field_type := 'N';
              if FieldDefs.Items[i].Size <> 0 then
                cField.lendth.num_len.len := FieldDefs.Items[i].Size
              else
                cField.lendth.num_len.len := 15;
              cField.lendth.num_len.dec := 0;
              cHeader.rec_size := cHeader.rec_size + cField.lendth.num_len.len;
            end;
          ftFloat:
            begin
              cField.field_type := 'N';
              if FieldDefs.Items[i].Size <> 0 then
                cField.lendth.num_len.len := FieldDefs.Items[i].Size
              else
                cField.lendth.num_len.len := 20;
              cField.lendth.num_len.dec := FieldDefs.Items[i].Precision;
              cHeader.rec_size := cHeader.rec_size + cField.lendth.num_len.len;
            end;
          ftBCD:
            begin
              cField.field_type := 'N';
              cField.lendth.num_len.len := FieldDefs.Items[i].Size;
              cField.lendth.num_len.dec := FieldDefs.Items[i].Precision;
              cHeader.rec_size := cHeader.rec_size + cField.lendth.num_len.len;
            end;
          ftDateTime, ftDate:
            begin
              cField.field_type := 'D';
              cField.lendth.num_len.len := 8;
              cField.lendth.num_len.dec := 0;
              cHeader.rec_size := cHeader.rec_size + 8;
            end;
          ftBoolean:
            begin
              cField.field_type := 'L';
              cField.lendth.num_len.len := 1;
              cField.lendth.num_len.dec := 0;
              cHeader.rec_size := cHeader.rec_size + 1;
            end;
          ftMemo:
            begin
              cField.field_type := 'M';
              cField.lendth.num_len.len := 10;
              cField.lendth.num_len.dec := 0;
              cHeader.rec_size := cHeader.rec_size + 10;
              cHeader.dbf_id := $83;
            end;
          else raise Exception.Create('TDBFDataSet: Can not create table with this datatype');
        end;
        FileWrite(cf, cField, SizeOf(FIELD_REC));
        cHeader.data_offset := cHeader.data_offset + SizeOf(FIELD_REC);
      end;
      qq := $D;
      FileWrite(cf, qq, 1);
      cHeader.data_offset := cHeader.data_offset + 1;
      FileSeek(cf, 0, 0);
      FileWrite(cf, cHeader, SizeOf(DBF_HEAD));
      FileClose(cf);
    end else raise Exception.Create('TDBFDataSet: Create error');
  end else raise Exception.Create('TDBFDataSet: Can not create table while dataset is open');
end;

{ TAccessMode }

constructor TAccessMode.Create;
begin
  inherited Create;
  fOpenRead         := true;
  fOpenWrite        := false;
  fOpenReadWrite    := false;
  fShareCompat      := false;
  fShareExclusive   := true;
  fShareDenyWrite   := false;
  fShareDenyRead    := false;
  fShareDenyNone    := false;
end;

function TAccessMode.GetAccessMode: LongWord;
begin
  Result := 0;
  if fOpenRead       then Result := Result or fmOpenRead      ;
  if fOpenWrite      then Result := Result or fmOpenWrite     ;
  if fOpenReadWrite  then Result := Result or fmOpenReadWrite ;
  if fShareCompat    then Result := Result or fmShareCompat   ;
  if fShareExclusive then Result := Result or fmShareExclusive;
  if fShareDenyWrite then Result := Result or fmShareDenyWrite;
  if fShareDenyRead  then Result := Result or fmShareDenyRead ;
  if fShareDenyNone  then Result := Result or fmShareDenyNone ;
  FLast := Result;
end;

function TAccessMode.IsExists(const Value: LongWord): boolean;
begin
  Result := ( (AccessMode and Value) = Value );
end;

procedure TAccessMode.SetAccessMode(const Value: LongWord);
begin
  fOpenRead       := false;
  fOpenWrite      := false;
  fOpenReadWrite  := false;
  fShareCompat    := false;
  fShareExclusive := false;
  fShareDenyWrite := false;
  fShareDenyRead  := false;
  fShareDenyNone  := false;
  if (Value and fmOpenRead       ) = fmOpenRead       then fOpenRead       := true;
  if (Value and fmOpenWrite      ) = fmOpenWrite      then fOpenWrite      := true;
  if (Value and fmOpenReadWrite  ) = fmOpenReadWrite  then fOpenReadWrite  := true;
  if (Value and fmShareCompat    ) = fmShareCompat    then fShareCompat    := true;
  if (Value and fmShareExclusive ) = fmShareExclusive then fShareExclusive := true;
  if (Value and fmShareDenyWrite ) = fmShareDenyWrite then fShareDenyWrite := true;
  if (Value and fmShareDenyRead  ) = fmShareDenyRead  then fShareDenyRead  := true;
  if (Value and fmShareDenyNone  ) = fmShareDenyNone  then fShareDenyNone  := true;
end;

{ TSmartDBF }

function TSmartDBF.AcceptRecord: Boolean;
begin
  Result := true;
  if Assigned(OnFilterRecord) then
    OnFilterRecord(self, Result)
  else if Filter <> '' then
    Result := FFilterParser.Execute;
  if FSetDeleted then
    Result := Result and Deleted;
end;

function TSmartDBF.AllocRecordBuffer: PChar;
begin
  Result := MemMgr.oMem.GetMem(self, RecordBufferSize);
end;

function TSmartDBF.CompareLocateField(const Fields: TList;
  const KeyValues: Variant; Options: TLocateOptions): Integer;
var
  FieldCount: Integer;
  Field: TField;
  KeyVal: Variant;
  v1, v2: String;
  i1, i2: Integer;
  w1, w2: Word;
  b1, b2: boolean;
  f1, f2: double;
  i: Integer;
begin
  Result := 1;
  FieldCount := Fields.Count;
  if FieldCount = 1 then
  begin
    try
      Field := TField(Fields.First);
    except
      Field := nil;
    end;
    try
      if VarIsArray(KeyValues) then
         KeyVal := KeyValues[0]
      else
         KeyVal := KeyValues;
    except
      KeyVal := NULL;
    end;
    case Field.DataType of
      ftFixedChar, ftWideString, ftString, ftMemo:
      begin
        v1 := Field.AsString;
        if not VarIsNull(KeyVal) then v2 := KeyVal else v2 := '';
        if loCaseInsensitive in Options then
          Result := AnsiCompareText(v1, v2)
        else
          Result := AnsiCompareStr(v1, v2);
      end;
      ftLargeint, ftSmallint, ftInteger:
      begin
        i1 := Field.AsInteger;
        if not VarIsNull(KeyVal) then i2 := KeyVal else i2 := 0;
        Result := i1 - i2;
      end;
      ftWord:
      begin
        w1 := Field.AsInteger;
        if not VarIsNull(KeyVal) then w2 := KeyVal else w2 := 0;
        Result := w1 - w2;
      end;
      ftBoolean:
      begin
        b1 := Field.AsBoolean;
        if not VarIsNull(KeyVal) then b2 := KeyVal else b2 := false;
        if (not b1) and b2 then
          Result := -1;
        if b1 = b2 then
          Result := 0;
        if b1 and (not b2) then
          Result := 1;
      end;
      ftFloat, ftCurrency, ftBCD:
      begin
        f1 := Field.AsFloat;
        if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
        if f1 < f2 then
          Result := -1;
        if f1 = f2 then
          Result := 0;
        if f1 > f2 then
          Result := 1;
      end;
      ftDate, ftTime, ftDateTime:
      begin
        f1 := Field.AsDateTime;
        if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
        if f1 < f2 then
          Result := -1;
        if f1 = f2 then
          Result := 0;
        if f1 > f2 then
          Result := 1;
      end;
    end;
  end else begin
    for i := 0 to FieldCount - 1 do
    begin
      //
      try
        Field := TField(Fields.Items[i]);
      except
        Field := nil;
      end;
      try
        if VarIsArray(KeyValues) then
           KeyVal := KeyValues[i]
        else
           KeyVal := NULL;
      except
        KeyVal := NULL;
      end;
      case Field.DataType of
        ftFixedChar, ftWideString, ftString:
        begin
          v1 := Field.AsString;
          if not VarIsNull(KeyVal) then v2 := KeyVal else v2 := '';
          if loCaseInsensitive in Options then
            Result := AnsiCompareText(v1, v2)
          else
            Result := AnsiCompareStr(v1, v2);
        end;
        ftLargeint, ftSmallint, ftInteger:
        begin
          i1 := Field.AsInteger;
          if not VarIsNull(KeyVal) then i2 := KeyVal else i2 := 0;
          Result := i1 - i2;
        end;
        ftWord:
        begin
          w1 := Field.AsInteger;
          if not VarIsNull(KeyVal) then w2 := KeyVal else w2 := 0;
          Result := w1 - w2;
        end;
        ftBoolean:
        begin
          b1 := Field.AsBoolean;
          if not VarIsNull(KeyVal) then b2 := KeyVal else b2 := false;
          if (not b1) and b2 then
            Result := -1;
          if b1 = b2 then
            Result := 0;
          if b1 and (not b2) then
            Result := 1;
        end;
        ftFloat, ftCurrency, ftBCD:
        begin
          f1 := Field.AsFloat;
          if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
          if f1 < f2 then
            Result := -1;
          if f1 = f2 then
            Result := 0;
          if f1 > f2 then
            Result := 1;
        end;
        ftDate, ftTime, ftDateTime:
        begin
          f1 := Field.AsDateTime;
          if not VarIsNull(KeyVal) then f2 := KeyVal else f2 := 0;
          if f1 < f2 then
            Result := -1;
          if f1 = f2 then
            Result := 0;
          if f1 > f2 then
            Result := 1;
        end;
      end;
      //
      if Result <> 0 then
        Exit;
    end;
  end;
end;

constructor TSmartDBF.Create(AOwner: TComponent);
var
  FieldMap: TFieldMap;
begin
  inherited Create(AOwner);
  FFilterParser := To8ExprParser.Create(self, '', [], [poExtSyntax], '', nil, FieldMap);
  FAccessMode := TAccessMode.Create;
  FFieldInfo := TList.Create;
  FLockRecords := TList.Create;
  FDBFFieldDefs := TDBFFieldDefs.Create(self);
  FOEM := false;
  FDBFFileName := '';
  FTmpActive := false;
  FKeyCalk := false;
  FBufferSize := 4096;
  FRecordsPerBuf := 0;
  FBuffer := nil;
  FBufInd := nil;
  FBufCnt := 0;
  FCurInd := -1;
  FBufDir := bdFromTop;;
  FBOF := false;
  FEOF := false;
  FWaitBusyRes := 3000;  //3 sec. waiting for a locking resource

  FDataLink := TVKDataLink.Create;
  FDataLink.DBFDataSet := self;

  FFileLock := false;

  FIndRecBuf := nil;
  FIndState := false;

  FLocateBuffer := MemMgr.oMem.GetMem(self, LOCATE_BUFFER_SIZE);

  FMasterFields := '';
  FRange := false;
  ListMasterFields := TList.Create;

  FFastPostRecord := false;

  FAddBuffered := false;
  FAddBuffer := nil;
  FAddBufferCount := -1;
  FAddBufferCurrent := -1;

end;

procedure TSmartDBF.CreateTable;
var
  cf: Integer;
  cHeader: DBF_HEAD;
  cField: FIELD_REC;
  i: Integer;
  Year, Month, Day: Word;
  qq: byte;
begin
  if not Active then
  begin
    cf := SysUtils.FileCreate(DBFFileName);
    if cf > 0 then
    begin
      DecodeDate(Now, Year, Month, Day);
      with cHeader do
      begin
        dbf_id := $03;
        last_update[0] := Byte(Year);
        last_update[1] := Byte(Month);
        last_update[2] := Byte(Day);
        last_rec := 0;
        data_offset := 1;
        rec_size := 1;
        for i := 0 to 15 do
          filler[i] := 0;
        prod_ind := 0;
        lang := 0;
        dummy := 0;
      end;
      SysUtils.FileSeek(cf, 0, 0);
      SysUtils.FileWrite(cf, cHeader, SizeOf(DBF_HEAD));
      cHeader.data_offset := cHeader.data_offset + SizeOf(DBF_HEAD);
      for i := 0 to FDBFFieldDefs.Count - 1 do
      begin
        cHeader.rec_size := cHeader.rec_size + FDBFFieldDefs[i].DataSize;
        cField := FDBFFieldDefs[i].Field;
        if  ( cField.field_type = 'M' ) or
            ( ( cField.field_type = 'E' ) and ( cField.extend_type in
              [dbftClob, dbftBlob, dbftGraphic, dbftFmtMemo,
              dbftClob_NB, dbftBlob_NB, dbftGraphic_NB, dbftFmtMemo_NB]  ) ) then
          cHeader.dbf_id := $83;
        SysUtils.FileWrite(cf, cField, SizeOf(FIELD_REC));
        cHeader.data_offset := cHeader.data_offset + SizeOf(FIELD_REC);
      end;
      qq := $D;
      SysUtils.FileWrite(cf, qq, 1);
      cHeader.data_offset := cHeader.data_offset + 1;
      SysUtils.FileSeek(cf, 0, 0);
      SysUtils.FileWrite(cf, cHeader, SizeOf(DBF_HEAD));
      SysUtils.FileClose(cf);
      if cHeader.dbf_id = $83 then begin
        cf := SysUtils.FileCreate(ChangeFileExt(DBFFileName, '.dbt'));
        SysUtils.FileWrite(cf, 'This is Lob!', 12);
        SysUtils.FileClose(cf);
      end;
    end else raise Exception.Create('TDBFDataSet: Create error');
  end else raise Exception.Create('TDBFDataSet: Can not create table while dataset is open');
end;

destructor TSmartDBF.Destroy;
var
  i: Integer;
begin
  EndAddBuffered;
  Active := false;
  FLockRecords.Destroy;
  if FFieldInfo.Count <> 0 then
    for i := 0 to FFieldInfo.Count - 1 do
      MemMgr.oMem.FreeMem(FFieldInfo.Items[i]);
  FFieldInfo.Clear;
  FFieldInfo.Destroy;
  FAccessMode.Destroy;
  FFilterParser.Destroy;
  MemMgr.oMem.FreeMem(FLocateBuffer);
  FDataLink.Destroy;
  FDBFFieldDefs.Destroy;
  ListMasterFields.Free;
  inherited Destroy;
end;

function TSmartDBF.FindRecord(Restart, GoForward: Boolean): Boolean;
var
  SaveState: TDataSetState;
  Accept: Boolean;
  Ret: TGetResult;
begin
  if (not Filtered) and (Filter <> '') then
      FFilterParser.SetExprParams(Filter, FilterOptions, [poExtSyntax], '');
  CheckBrowseMode;
  DoBeforeScroll;
  SetFound(False);
  UpdateCursorPos;
  CursorPosChanged;
  if GoForward then
  begin
    if Restart then InternalFirst;
    repeat
      Ret := GetRecordByBuffer(FFilterRecord, gmNext, false);
      SaveState := SetTempState(dsFilter);
      Accept := AcceptRecord;
      RestoreState(SaveState);
    until Accept or (Ret <> grOK);
  end else
  begin
    if Restart then InternalLast;
    repeat
      Ret := GetRecordByBuffer(FFilterRecord, gmPrior, false);
      SaveState := SetTempState(dsFilter);
      Accept := AcceptRecord;
      RestoreState(SaveState);
    until Accept or (Ret <> grOK);
  end;
  if Ret = grOK then
  begin
    Resync([rmExact, rmCenter]);
    SetFound(True);
  end else
    InternalSetToRecord(ActiveBuffer);
  Result := Found;
  if Result then DoAfterScroll;
end;

procedure TSmartDBF.FreeRecordBuffer(var Buffer: PChar);
begin
  MemMgr.oMem.FreeMem(Buffer)
end;

function TSmartDBF.GetActiveRecBuf(var pRecBuf: PChar): Boolean;
begin
  if FIndState then begin
    pRecBuf := FIndRecBuf;
  end else begin
    if FKeyCalk then begin
      pRecBuf := RecBuf
    end else begin
      if FTmpActive then
        pRecBuf := FTempRecord
      else begin
        case State of
          dsBrowse: if IsEmpty then pRecBuf := nil else pRecBuf := ActiveBuffer;
          dsEdit, dsInsert: pRecBuf := ActiveBuffer;
          dsFilter: pRecBuf := FFilterRecord;
          dsNewValue, dsOldValue, dsCurValue: pRecBuf := ActiveBuffer;
          dsCalcFields: pRecBuf := CalcBuffer;
        else
          pRecBuf := nil;
        end;
      end;
    end;
  end;
  Result := pRecBuf <> nil;
end;

procedure TSmartDBF.GetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  Longword(Data^) := pTRecInfo(Buffer + FRecordSize).RecordRowID;
end;

function TSmartDBF.GetBookmarkFlag(Buffer: PChar): TBookmarkFlag;
begin
  Result := pTRecInfo(Buffer + FRecordSize).BookmarkFlag;
end;

function TSmartDBF.GetDeleted: Boolean;
var
  ActBuff: pChar;
begin
  Result := false;
  GetActiveRecBuf(ActBuff);
  if ActBuff <> nil then
  begin
    if ActBuff[0] = #42 then
      Result := true;
  end;
end;

function TSmartDBF.GetFieldData(Field: TField; Buffer: Pointer): Boolean;
var
  iCode, dInt: Integer;
  dFloat: Double;
  ww: Extended;
  dBool: WordBool;
  dDate: TDateTime;
  sTS: TTimeStamp;
  dd: double;
  Year, Month, Day: Word;
  ss: pChar;
  ss1: array [0..255] of char;
  ActiveBuf: pChar;
  qq: PFieldInfo;
  oDS: TDataSet;
  LookupResult: Variant;
  LastSetp: Char;
  SLen: WORD;
begin
  Result := false;
  case Field.FieldKind of
  fkData:
    begin
      qq := PFieldInfo(FFieldInfo.Items[Field.FieldNo - 1]);
      GetActiveRecBuf(ActiveBuf);
      if ActiveBuf <> nil then
      begin
        ss := ActiveBuf + qq.FOff;
        if Buffer <> nil then
        begin
          case Field.DataType of
            ftTime:
              begin
                case qq.FExType of
                  dbftTime:
                    begin
                      Integer(Buffer^) := pInteger(ss)^;
                      Result := true;
                    end;
                  dbftTime_N:
                    begin
                      Result := (ss[0] <> ' ');  //if ' ' then NULL
                      if Result then
                        Integer(Buffer^) := pInteger(ss + 1)^;
                    end;
                  dbftTime_NB:
                    begin
                      Result := ( ( pLongword(ss)^ and $80000000 ) = $80000000 );
                      if Result then
                        Longword(Buffer^) := ( pLongword(ss)^ and $7FFFFFFF );
                    end;
                end;
              end;
            ftDate:
              begin
                if qq.FType <> 'E' then begin
                  Result := not IsBlank(ss, qq.FLen);
                  if Result then begin
                    Year := (Byte(ss[0]) - $30) * 1000 + (Byte(ss[1]) - $30) * 100 + (Byte(ss[2]) - $30) * 10 + (Byte(ss[3]) - $30);
                    Month := (Byte(ss[4]) - $30) * 10 + (Byte(ss[5]) - $30);
                    Day := (Byte(ss[6]) - $30) * 10 + (Byte(ss[7]) - $30);
                    if DoEncodeDate(Year, Month, Day, dDate) then
                    begin
                      sTS := DateTimeToTimeStamp(dDate);
                      //dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                      //double(Buffer^) := dd;
                      Integer(Buffer^) := sTS.Date;
                    end else
                      Result := false;
                  end;
                end else begin
                  case qq.FExType of
                    dbftDate:
                      begin
                        Integer(Buffer^) := pInteger(ss)^;
                        Result := true;
                      end;
                    dbftDate_NB:
                      begin
                        Result := ( ( pLongword(ss)^ and $80000000 ) = $80000000 );
                        if Result then
                          Longword(Buffer^) := ( pLongword(ss)^ and $7FFFFFFF );
                      end;
                    dbftDate_N:
                      begin
                        Result := (ss[0] <> ' ');  //if ' ' then NULL
                        if Result then
                          Integer(Buffer^) := pInteger(ss + 1)^;
                      end;
                  end;
                end;
              end;
            ftBCD:
              begin
                Result := (( Byte(ss[1]) and $40 ) <> $00);
                if Result  then begin
                  Tbcd(Buffer^) := Pbcd(ss)^;
                  Tbcd(Buffer^).SignSpecialPlaces := (Tbcd(Buffer^).SignSpecialPlaces or $40);
                end;
              end;
            ftCurrency:
              begin
                case qq.FExType of
                  dbftCurrency:
                    begin
                      Currency(Buffer^) := pCurrency(ss)^;
                      Result := true;
                    end;
                  dbftCurrency_N:
                    begin
                      Result := (ss[0] <> ' ');  //if ' ' then NULL
                      if Result then
                        Currency(Buffer^) := pCurrency(ss + 1)^;
                    end;
                  dbftCurrency_NB:
                    begin
                      Result := ( ( pByte(ss + 7)^ and $80 ) = $80 );
                      if Result then begin
                        pByte(ss + 7)^ := ( pByte(ss + 7)^ and $7F );
                        Currency(Buffer^) := Currency( Pointer(ss)^ );
                        pByte(ss + 7)^ := ( pByte(ss + 7)^ or $80 );
                      end;
                    end;
                end;
              end;
            ftWideString:
              begin
                //WideString(Buffer^) := pWideString(ss)^;
                Result := false;
              end;
            ftString:
              begin
                if qq.FType <> 'E' then begin
                  Result := not IsBlank(ss, qq.FLen);
                  if Result then begin
                    Move(ss^, Buffer^, qq.FLen);
                    pChar(Buffer)[qq.FLen] := #0;
                  end;
                end else
                  case qq.FExType of
                    dbftString:       //
                      begin
                        SLen := pWORD(ss)^;
                        if SLen < 8224 then begin
                          ss := ss + SizeOf(WORD);
                          Move(ss^, Buffer^, SLen);
                          pChar(Buffer)[SLen] := #0;
                          Result := true;
                        end else
                          Result := false;
                      end;
                    dbftString_N:     //
                      begin
                        Result := (ss[0] <> ' ');  //if ' ' then NULL
                        if Result then begin
                          ss := ss + 1;
                          SLen := pWORD(ss)^;
                          ss := ss + SizeOf(WORD);
                          Move(ss^, Buffer^, SLen);
                          pChar(Buffer)[SLen] := #0;
                        end;
                      end;
                    dbftFixedChar:
                      begin
                        Result := not IsBlank(ss, qq.FLen + 1);
                        if Result then
                          Move(ss^, Buffer^, qq.FLen + 1);
                      end;
                  else
                    Result := false;
                  end;
              end;
            ftSmallint:
              begin
                case qq.FExType of
                  dbftS1:       //Shortint
                    begin
                      Smallint(Buffer^) := pShortint(ss)^;
                      Result := true;
                    end;
                  dbftS2:       //Smallint
                    begin
                      Smallint(Buffer^) := pSmallint(ss)^;
                      Result := true;
                    end;
                  dbftS1_N:     //Shortint with NULL
                    begin
                      Result := (ss[0] <> ' ');  //if ' ' then NULL
                      if Result then
                        Smallint(Buffer^) := pShortint(ss + 1)^;
                    end;
                  dbftS2_N:     //Smallint with NULL
                    begin
                      Result := (ss[0] <> ' ');  //if ' ' then NULL
                      if Result then
                        Smallint(Buffer^) := pSmallint(ss + 1)^;
                    end;
                else
                  Result := false;
                end;
              end;
            ftWord:
              begin
                case qq.FExType of
                  dbftU1:     //Byte
                    begin
                      Word(Buffer^) := pByte(ss)^;
                      Result := true;
                    end;
                  dbftU2:     //Word
                    begin
                      Word(Buffer^) := pWord(ss)^;
                      Result := true;
                    end;
                  dbftU1_N:   //Byte with NULL
                    begin
                      Result := (ss[0] <> ' ');  //if ' ' then NULL
                      if Result then
                        Word(Buffer^) := pByte(ss + 1)^;
                    end;
                  dbftU2_N:   //Word with NULL
                    begin
                      Result := (ss[0] <> ' ');  //if ' ' then NULL
                      if Result then
                        Word(Buffer^) := pWord(ss + 1)^;
                    end;
                  dbftU1_NB:  //Positive byte with NULL bit instead of sign bit
                    begin
                      Result := ( ( Byte(ss[0]) and $80 ) = $80 );
                      if Result then
                        Word(Buffer^) := ( Byte(ss[0]) and $7F );
                    end;
                  dbftU2_NB:  //Positive word with NULL bit instead of sign bit
                    begin
                      Result := ( ( pWord(ss)^ and $8000 ) = $8000 );
                      if Result then
                        Word(Buffer^) := ( pWord(ss)^ and $7FFF );
                    end;
                else
                  Result := false;
                end;
              end;
            ftInteger:
              begin
                if qq.FType <> 'E' then begin
                  Move(ss^, ss1, qq.FLen);
                  ss1[qq.FLen] := #0;
                  Val(ss1, dInt, iCode);
                  if iCode = 0 then
                  begin
                    Integer(Buffer^) := dInt;
                    Result := true;
                  end else
                    Result := false;
                end else begin
                  case qq.FExType of
                    dbftS4, dbftU4:       //Longint, Longword
                      begin
                        Integer(Buffer^) := pInteger(ss)^;
                        Result := true;
                      end;
                    dbftS4_N, dbftU4_N:     //Longint with NULL, Longword with NULL
                      begin
                        Result := (ss[0] <> ' ');  //if ' ' then NULL
                        if Result then
                          Integer(Buffer^) := pInteger(ss + 1)^;
                      end;
                    dbftU4_NB:
                      begin
                        Result := ( ( pLongword(ss)^ and $80000000 ) = $80000000 );
                        if Result then
                          Longword(Buffer^) := ( pLongword(ss)^ and $7FFFFFFF );
                      end;
                  else
                    Result := false;
                  end;
                end;
              end;
            ftLargeint:
              begin
                case qq.FExType of
                  dbftS8:    //Int64
                    begin
                      Int64(Buffer^) := pInt64(ss)^;
                      Result := true;
                    end;
                  dbftS8_N:  //Int64 with NULL
                    begin
                      Result := (ss[0] <> ' ');  //if ' ' then NULL
                      if Result then
                        Int64(Buffer^) := pInt64(ss + 1)^;
                    end;
                else
                  Result := false;
                end;
              end;
            ftFloat:
              begin
                if qq.FType <> 'E' then begin
                  Result := not IsBlank(ss, qq.FLen);
                  if Result then begin
                    Move(ss^, ss1, qq.FLen);
                    ss1[qq.FLen] := #0;
                    LastSetp := DecimalSeparator;
                    DecimalSeparator := '.';
                    if TextToFloat(ss1, ww, fvExtended) then
                    begin
                      dFloat := ww;
                      double(Buffer^) := dFloat;
                    end else
                      Result := false;
                    DecimalSeparator := LastSetp;
                  end;
                end else begin
                  case qq.FExType of
                    dbftR4:
                      begin
                        double(Buffer^) := pSingle(ss)^;
                        Result := true;
                      end;
                    dbftR6:
                      begin
                        double(Buffer^) := pReal48(ss)^;
                        Result := true;
                      end;
                    dbftR8:
                      begin
                        double(Buffer^) := pDouble(ss)^;
                        Result := true;
                      end;
                    dbftR10:
                      begin
                        Extended(Buffer^) := pExtended(ss)^;
                        Result := true;
                      end;
                    dbftR4_N:
                      begin
                        Result := (ss[0] <> ' ');  //if ' ' then NULL
                        if Result then
                          double(Buffer^) := pSingle(ss + 1)^;
                      end;
                    dbftR4_NB:
                      begin
                        Result := ( ( pByte(ss + 3)^ and $80 ) = $80 );
                        if Result then begin
                          pByte(ss + 3)^ := ( pByte(ss + 3)^ and $7F );
                          double(Buffer^) := Single( Pointer(ss)^ );
                          pByte(ss + 3)^ := ( pByte(ss + 3)^ or $80 );
                        end;
                      end;
                    dbftR6_NB:
                      begin
                        Result := ( ( pByte(ss + 5)^ and $80 ) = $80 );
                        if Result then begin
                          pByte(ss + 5)^ := ( pByte(ss + 5)^ and $7F );
                          double(Buffer^) := Real48( Pointer(ss)^ );
                          pByte(ss + 5)^ := ( pByte(ss + 5)^ or $80 );
                        end;
                      end;
                    dbftR8_NB:
                      begin
                        Result := ( ( pByte(ss + 7)^ and $80 ) = $80 );
                        if Result then begin
                          pByte(ss + 7)^ := ( pByte(ss + 7)^ and $7F );
                          double(Buffer^) := double( Pointer(ss)^ );
                          pByte(ss + 7)^ := ( pByte(ss + 7)^ or $80 );
                        end;
                      end;
                    dbftR6_N:
                      begin
                        Result := (ss[0] <> ' ');  //if ' ' then NULL
                        if Result then
                          double(Buffer^) := pReal48(ss + 1)^;
                      end;
                    dbftR8_N:
                      begin
                        Result := (ss[0] <> ' ');  //if ' ' then NULL
                        if Result then
                          double(Buffer^) := pDouble(ss + 1)^;
                      end;
                    dbftR10_N:
                      begin
                        Result := (ss[0] <> ' ');  //if ' ' then NULL
                        if Result then
                          Extended(Buffer^) := pExtended(ss + 1)^;
                      end;
                  else
                    Result := false;
                  end;
                end;
              end;
            ftMemo, ftBlob, ftFmtMemo, ftGraphic:
              begin
                Result := false;
              end;
            ftDateTime:
              begin
                case qq.FExType of
                  dbftD1:
                    begin
                      sTS := DateTimeToTimeStamp(pDouble(ss)^);
                      dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                      double(Buffer^) := dd;
                      Result := true;
                    end;
                  dbftD1_NB:
                    begin
                      Result := ( ( pByte(ss + 7)^ and $80 ) = $80 );
                      if Result then begin
                        pByte(ss + 7)^ := ( pByte(ss + 7)^ and $7F );
                        sTS := DateTimeToTimeStamp(pDouble(ss)^);
                        pByte(ss + 7)^ := ( pByte(ss + 7)^ or $80 );
                        dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                        double(Buffer^) := dd;
                      end;
                    end;
                  dbftD2:
                    begin
                      double(Buffer^) := pDouble(ss)^;
                      Result := true;
                    end;
                  dbftD2_NB:
                    begin
                      Result := ( ( pByte(ss + 7)^ and $80 ) = $80 );
                      if Result then begin
                        pByte(ss + 7)^ := ( pByte(ss + 7)^ and $7F );
                        double(Buffer^) := pDouble(ss)^;
                        pByte(ss + 7)^ := ( pByte(ss + 7)^ or $80 );
                      end;
                    end;
                  dbftD3:
                    begin
                      sTS := DateTimeToTimeStamp(pReal48(ss)^);
                      dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                      double(Buffer^) := dd;
                      Result := true;
                    end;
                  dbftD3_NB:
                    begin
                      Result := ( ( pByte(ss + 5)^ and $80 ) = $80 );
                      if Result then begin
                        pByte(ss + 5)^ := ( pByte(ss + 5)^ and $7F );
                        double(Buffer^) := pReal48(ss)^;
                        pByte(ss + 5)^ := ( pByte(ss + 5)^ or $80 );
                      end;
                    end;
                  dbftD1_N:
                    begin
                      Result := (ss[0] <> ' ');  //if ' ' then NULL
                      if Result then begin
                        sTS := DateTimeToTimeStamp(pDouble(ss + 1)^);
                        dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                        double(Buffer^) := dd;
                      end;
                    end;
                  dbftD2_N:
                    begin
                      Result := (ss[0] <> ' ');  //if ' ' then NULL
                      if Result then
                        double(Buffer^) := pDouble(ss + 1)^;
                    end;
                  dbftD3_N:
                    begin
                      Result := (ss[0] <> ' ');  //if ' ' then NULL
                      if Result then begin
                        sTS := DateTimeToTimeStamp(pReal48(ss + 1)^);
                        dd := 3600.0*24*1000*sTS.Date + sTS.Time;
                        double(Buffer^) := dd;
                      end;
                    end;
                end;
              end;
            ftBoolean:
              begin
                case ss[0] of
                  'T':
                    begin
                      dBool := true;
                      Result := true;
                    end;
                  'F':
                    begin
                      dBool := false;
                      Result := true;
                    end;
                  ' ':
                    begin
                      dBool := false;
                      Result := false;
                    end;
                else
                  dBool := false;
                  Result := false;
                end;
                WordBool(Buffer^) := dBool;
              end;
          end;
        end else begin
          if qq.FType <> 'E' then begin
            Result := not IsBlank(ss, qq.FLen);
          end else begin
            case qq.FExType of
              dbftS1_N,     //Shortint with NULL
              dbftU1_N,     //Byte  with NULL
              dbftS2_N,     //Smallint with NULL
              dbftU2_N,     //Word with NULL
              dbftS4_N,     //Longint with NULL
              dbftU4_N,     //Longword with NULL
              dbftS8_N,     //Int64 with NULL
              dbftR4_N,     //Single with NULL
              dbftR6_N,     //Real48 with NULL
              dbftR8_N,     //Double with NULL
              dbftR10_N,    //Extended with NULL
              dbftD1_N,
              dbftD2_N,
              dbftD3_N,
              dbftString_N, //String with NULL
              dbftCurrency_N, //Currency with NULL
              dbftDate_N,
              dbftTime_N: Result := not (ss[0] = ' ');  //if ' ' then NULL
              dbftU1_NB: Result := not ( ( Byte(ss[0]) and $80 ) = $80 );
              dbftU2_NB: Result := ( ( pWord(ss)^ and $8000 ) = $8000 );
              dbftU4_NB: Result := ( ( pLongword(ss)^ and $80000000 ) = $80000000 );
              dbftR4_NB: Result := ( ( pByte(ss + 3)^ and $80 ) = $80 );
              dbftR6_NB: Result := ( ( pByte(ss + 5)^ and $80 ) = $80 );
              dbftR8_NB: Result := ( ( pByte(ss + 7)^ and $80 ) = $80 );
              dbftCurrency_NB: Result := ( ( pByte(ss + 7)^ and $80 ) = $80 );
              dbftD1_NB: Result := ( ( pByte(ss + 7)^ and $80 ) = $80 );
              dbftD2_NB: Result := ( ( pByte(ss + 7)^ and $80 ) = $80 );
              dbftD3_NB: Result := ( ( pByte(ss + 5)^ and $80 ) = $80 );
              dbftDate_NB: Result := ( ( pLongword(ss)^ and $80000000 ) = $80000000 );
              dbftTime_NB: Result := ( ( pLongword(ss)^ and $80000000 ) = $80000000 );
              dbftClob, dbftFmtMemo, dbftBlob, dbftGraphic:
                Result := not IsBlank(ss, qq.FLen);
              dbftClob_NB, dbftFmtMemo_NB, dbftBlob_NB, dbftGraphic_NB:
                Result := ( ( pLongword(ss)^ and $80000000 ) = $80000000 );
              dbftString: Result := not ( pWORD(ss)^ = 8224 );
              dbftFixedChar: Result := not IsBlank(ss, qq.FLen + 1);
              dbftS1,
              dbftU1,
              dbftS2,
              dbftU2,
              dbftS4,
              dbftU4,
              dbftS8,
              dbftR4,
              dbftR6,
              dbftR8,
              dbftR10,
              dbftD1,
              dbftD2,
              dbftD3,
              dbftCurrency,
              dbftDate,
              dbftTime: Result := true;
              dbftBCD: Result := not ( ( Byte(ss[1]) and $40 ) = $00 );
            else
              Result := false;
            end;
          end;
        end;
      end else
        Result := false;
    end;
  fkCalculated:
    begin
      GetActiveRecBuf(ActiveBuf);
      if Buffer <> nil then
      begin
        if (ActiveBuf <> nil) and not (csDesigning in ComponentState) then
        begin
          ss := ActiveBuf + FRecordSize + sizeof(TRecInfo) + Field.Offset;
          Move(ss^, Buffer^, Field.DataSize);
          if Field.DataType in [ftString, ftBytes, ftVarBytes]  then pChar(Buffer)[Field.DataSize] := Char(0);
          Result := true;
        end else begin
          FillChar(Buffer^, Field.DataSize, 0);
          Result := false;
        end;
      end else
        Result := false;
    end;
  fkLookup:
    begin
      Result := false;
      if Buffer <> nil then
      begin
        if GetActiveRecBuf(ActiveBuf) then
        begin
          oDS := Field.LookupDataSet;
          if (oDS <> nil) and oDS.Active then
          begin
            LookupResult := oDS.Lookup(Field.LookupKeyFields, FieldValues[Field.KeyFields], Field.LookupResultField);
            if (not VarIsEmpty(LookupResult)) and (not VarIsNull(LookupResult)) then
            begin
              if Field.DataType = ftString then
              begin
                ss := TVarData(LookupResult).VPointer;
                Move(ss^, Buffer^, Length(ss) + 1);
              end else begin
                ss := TVarData(LookupResult).VPointer;
                Move(ss^, Buffer^, Field.DataSize);
              end;
              Result := true;
            end;
          end;
        end else begin
          FillChar(Buffer^, Field.DataSize, 0);
          Result := false;
        end;
      end;
    end;
  end;
end;

function TSmartDBF.GetRecNo: Integer;
var
  ActiveBuf: pChar;
begin
  Result := -1;
  GetActiveRecBuf(ActiveBuf);
  if ActiveBuf <> nil then
    Result := pTRecInfo(ActiveBuf + RecordSize).RecordRowID;
end;

function TSmartDBF.GetRecord(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
var
  SaveState: TDataSetState;
  Accept: Boolean;
begin
  if not Filtered then
  begin
    if not FSetDeleted then
      Result := GetRecordByBuffer(Buffer, GetMode, DoCheck)
    else begin
      if GetMode <> gmCurrent then
      begin
        repeat
          Result := GetRecordByBuffer(FFilterRecord, GetMode, DoCheck);
          SaveState := SetTempState(dsFilter);
          Accept := not Deleted;
          RestoreState(SaveState);
        until Accept or (Result <> grOK);
        if Accept then
          Move(FFilterRecord^, Buffer^, RecordBufferSize)
        else
          Move(FTempRecord^, Buffer^, RecordBufferSize);
      end else
        Result := GetRecordByBuffer(Buffer, GetMode, DoCheck);
    end;
  end else begin
    if GetMode <> gmCurrent then
    begin
      repeat
        Result := GetRecordByBuffer(FFilterRecord, GetMode, DoCheck);
        SaveState := SetTempState(dsFilter);
        Accept := AcceptRecord;
        RestoreState(SaveState);
      until Accept or (Result <> grOK);
      if Accept then
        Move(FFilterRecord^, Buffer^, RecordBufferSize)
      else
        Move(FTempRecord^, Buffer^, RecordBufferSize);
    end else
      Result := GetRecordByBuffer(Buffer, GetMode, DoCheck);
  end;
end;

function TSmartDBF.GetRecordBufferSize: Integer;
begin
  Result := FRecordSize + sizeof(TRecInfo) + CalcFieldsSize;
end;

function TSmartDBF.GetRecordByBuffer(Buffer: PChar; GetMode: TGetMode;
  DoCheck: Boolean): TGetResult;
var
  cc: PChar;
begin
  Result := grOK;
  cc := pChar(Buffer);
  case GetMode of
    gmCurrent:
      if ( not FBOF ) and ( not FEOF ) then
      begin
        InternalInitRecord(cc);
        pTRecInfo(cc + FRecordSize).BookmarkFlag := bfCurrent;
        pTRecInfo(cc + FRecordSize).RecordRowID := RecNoBuf;
        pTRecInfo(cc + FRecordSize).UpdateStatus := usUnmodified;
        Move(RecBuf^, cc^, FRecordSize);
      end else begin
        InternalInitRecord(FTempRecord);
        Move(FTempRecord^, Buffer^, RecordBufferSize);
        Result := grError;
      end;
    gmNext:
      begin
        NextIndexBuf;
        if not FEOF then
        begin
          InternalInitRecord(cc);
          pTRecInfo(cc + FRecordSize).BookmarkFlag := bfCurrent;
          pTRecInfo(cc + FRecordSize).RecordRowID := RecNoBuf;
          pTRecInfo(cc + FRecordSize).UpdateStatus := usUnmodified;
          Move(RecBuf^, cc^, FRecordSize);
        end else begin
          InternalInitRecord(FTempRecord);
          Move(FTempRecord^, Buffer^, RecordBufferSize);
          Result := grEOF;
        end;
      end;
    gmPrior:
      begin
        PriorIndexBuf;
        if not FBOF then
        begin
          InternalInitRecord(cc);
          pTRecInfo(cc + FRecordSize).BookmarkFlag := bfCurrent;
          pTRecInfo(cc + FRecordSize).RecordRowID := RecNoBuf;
          pTRecInfo(cc + FRecordSize).UpdateStatus := usUnmodified;
          Move(RecBuf^, cc^, FRecordSize);
        end else begin
          InternalInitRecord(FTempRecord);
          Move(FTempRecord^, Buffer^, RecordBufferSize);
          Result := grBOF;
        end;
      end;
  end;
  if Result = grOK then
    GetCalcFields(Buffer);
end;

function TSmartDBF.GetRecordCount: Integer;
begin
  if LockHeader then
    try
      DBFHeader.last_rec := ( (FileSeek(DBFHandler, 0, 2) - DBFHeader.data_offset) div DBFHeader.rec_size );
    finally
      UnLockHeader;
    end;
  Result := DBFHeader.last_rec;
end;

function TSmartDBF.GetRecordSize: Word;
begin
  Result := FRecordSize;
end;

function TSmartDBF.GetStateFieldValue(State: TDataSetState;
  Field: TField): Variant;
begin
  Result := NULL;
  if State in [dsNewValue, dsCurValue, dsOldValue] then
    Result := inherited GetStateFieldValue(State, Field);
end;

procedure TSmartDBF.CloseTmpRecord;
begin
  FTmpActive := false;
  FFastPostRecord := false;
end;

procedure TSmartDBF.SetTmpRecord(nRec: DWORD);
begin
  FileSeek(DBFHandler, DBFHeader.data_offset + (nRec - 1) * DWORD(FRecordSize), soFromBeginning);
  FileRead(DBFHandler, FTempRecord^, FRecordSize);
  SetBookmarkData(FTempRecord, @nRec);
  SetBookmarkFlag(FTempRecord, bfCurrent);
  FTmpActive := true;
  FFastPostRecord := true;
end;

procedure TSmartDBF.AddRecord(const Values: variant);
var
  i, j: Integer;
begin
  InternalInitRecord(FTempRecord);
  FTmpActive := true;
  FFastPostRecord := true;
  try
    j := VarArrayHighBound(Values, 1);
    for i := 0 to j - 1 do
      Fields[i].AsVariant := Values[i];
    InternalAddRecord(FTempRecord, true);
  finally
    FTmpActive := false;
    FFastPostRecord := false;
  end;
end;

procedure TSmartDBF.AddRecord(ne: TNotifyEvent);
begin
  InternalInitRecord(FTempRecord);
  FTmpActive := true;
  FFastPostRecord := true;
  try
    if Assigned(ne) then ne(self);
    InternalAddRecord(FTempRecord, true);
  finally
    FTmpActive := false;
    FFastPostRecord := false;
  end;
end;

procedure TSmartDBF.BeginAddRecord;
begin
  InternalInitRecord(FTempRecord);
  FTmpActive := true;
  FFastPostRecord := true;
end;

procedure TSmartDBF.EndAddRecord;
begin
  InternalAddRecord(FTempRecord, true);
  FTmpActive := false;
  FFastPostRecord := false;
end;

procedure TSmartDBF.AddRecord(const Values: array of const);
begin
  InternalInitRecord(FTempRecord);
  FTmpActive := true;
  FFastPostRecord := true;
  try
    SetFields(Values);
    InternalAddRecord(FTempRecord, true);
  finally
    FTmpActive := false;
    FFastPostRecord := false;
  end;
end;

procedure TSmartDBF.InternalAddRecord(Buffer: Pointer; Append: Boolean);
var
  i, RealRead: Integer;
  lpMsgBuf: array [0..500] of Char;
  le: DWORD;
  NewR: Longint;
  NewKey: String;
  b: boolean;
  cc: pChar;
begin
  if FAddBuffered then begin
    if FAddBufferCurrent = FAddBufferCount - 1 then FlushAddBuffer;
    Inc(FAddBufferCurrent);
    cc := FAddBuffer + FAddBufferCurrent * DBFHeader.rec_size;
    Move(Buffer^, cc^, DBFHeader.rec_size);
  end else begin
    b := false;
    CheckActive;
    if LockHeader then begin
      try
        DBFHeader.last_rec := ( (FileSeek(DBFHandler, 0, 2) - DBFHeader.data_offset) div DBFHeader.rec_size );
        NewR := DBFHeader.last_rec + 1;
        if RLock(NewR) then begin
          try
            pTRecInfo(pChar(Buffer) + FRecordSize).RecordRowID := NewR;
            FileSeek(DBFHandler, 0, 2);
            FileSeek(DBFHandler, DBFHeader.data_offset + LongWord(DBFHeader.last_rec * FRecordSize), 0);
            RealRead := FileWrite(DBFHandler, Buffer^, DBFHeader.rec_size);
            if RealRead = -1 then
            begin
              le := GetLastError();
              FormatMessage(
                  FORMAT_MESSAGE_FROM_SYSTEM,
                  nil,
                  le,
                  0, // Default language
                  lpMsgBuf,
                  500,
                  nil
              );
              raise Exception.Create('TSmartDBF: ' + lpMsgBuf);
            end else begin
              Inc(DBFHeader.last_rec);
              FileSeek(DBFHandler, 0, 0); //go to the begin
              RealRead := FileWrite(DBFHandler, DBFHeader, SizeOf(DBFHeader));
              if RealRead = -1 then
              begin
                le := GetLastError();
                FormatMessage(
                    FORMAT_MESSAGE_FROM_SYSTEM,
                    nil,
                    le,
                    0, // Default language
                    lpMsgBuf,
                    500,
                    nil
                );
                raise Exception.Create('TSmartDBF: ' + lpMsgBuf);
              end else begin
                for i := 0 to Indexes.Count - 1 do begin
                  NewKey := Indexes[i].EvaluteKeyExpr;

                  if  (Indexes.ActiveObject <> nil) and
                      (Indexes.ActiveObject = Indexes[i]) and
                      (Indexes[i].IsRanged) and
                      (not Indexes[i].InRange(NewKey)) then b := true;

                  if  not (
                    ( (FAccessMode.FLast and fmShareExclusive) = fmShareExclusive ) or
                    ( (FAccessMode.FLast and fmShareDenyWrite) = fmShareDenyWrite ) or
                    FFileLock
                      ) then begin
                    if Indexes[i].FLock then
                      try
                        Indexes[i].StartUpdate(false);
                        Indexes[i].AddKey(NewKey, DBFHeader.last_rec);
                      finally
                        Indexes[i].Flush;
                        Indexes[i].FUnLock;
                      end
                    else
                      raise Exception.Create('TSmartDBF.InternalAddRecord: Can not add key to index file (FLock is false).');
                  end else begin
                    if Indexes[i].FLock then
                      try
                        Indexes[i].AddKey(NewKey, DBFHeader.last_rec);
                      finally
                        Indexes[i].FUnLock;
                      end
                    else
                      raise Exception.Create('TSmartDBF.InternalAddRecord: Can not add key to index file (FLock is false).');
                  end;

                end;
                if not FFastPostRecord then begin
                  if b then
                    InternalLast
                  else
                    GetBufferByRec(DBFHeader.last_rec);
                end;
              end;
            end;
          finally
            RUnLock(NewR);
          end
        end else
          raise Exception.Create('TSmartDBF.InternalAddRecord: Can not lock DBF record.');
      finally
        UnLockHeader;
      end
    end else
      raise Exception.Create('TSmartDBF.InternalAddRecord: Can not lock DBF header.');
  end;
end;

procedure TSmartDBF.InternalClose;
var
  i: Integer;
begin
  if Indexes <> nil then
    for i := 0 to Indexes.Count - 1 do Indexes[i].Flush;
  FreeRecordBuffer(FBuffer);
  FRecordsPerBuf := 0;
  FBuffer := nil;
  MemMgr.oMem.FreeMem(FBufInd);
  FBufInd := nil;
  FBufCnt := 0;
  FBufDir := bdFromTop;
  FreeRecordBuffer(FTempRecord);
  FreeRecordBuffer(FFilterRecord);
  BindFields(false);
  if DefaultFields then DestroyFields;
  if FIndexes <> nil then FIndexes.CloseAll;
  if (LobHandler > 0) then begin
    FileClose(LobHandler);
    LobHandler := -1;
  end;
  if (DBFHandler > 0) then begin
    FileClose(DBFHandler);
    DBFHandler := -1;
  end;
end;

procedure TSmartDBF.InternalDelete;
begin
 //
end;

procedure TSmartDBF.InternalFirst;
var
  i, RealRead: Integer;
begin
  FBOF := true;
  FEOF := false;
  FBufDir := bdFromTop;
  FCurInd := -1;
  if FIndexes.ActiveObject = nil then begin
    FileSeek(DBFHandler, DBFHeader.data_offset, soFromBeginning);
    RealRead := FileRead(DBFHandler, FBuffer^, FRecordsPerBuf * FRecordSize);
    FBufCnt := RealRead div FRecordSize;
    for i := 0 to FBufCnt - 1 do pLongint(pChar(FBufInd) + i * SizeOf(LongInt))^ := i + 1;
    if FBufCnt = 0 then begin
      FBOF := true;
      FEOF := true;
    end;
  end else begin
    if FIndexes.ActiveObject.FLock then
      try
        FBufCnt := FIndexes.ActiveObject.FillFirstBufRecords(DBFHandler, FBuffer, FRecordsPerBuf, FRecordSize, FBufInd, DBFHeader.data_offset);
      finally
        FIndexes.ActiveObject.FUnLock;
      end
    else
      raise Exception.Create('TDBFDataSet: Can not read from index file (FLock is false).');
  end;
end;

procedure TSmartDBF.InternalGotoBookmark(Bookmark: Pointer);
var
  i : Longint;
begin
  i := pTRecInfo(Bookmark).RecordRowID;
  GetBufferByRec(i);
end;

procedure TSmartDBF.InternalHandleException;
begin
  Application.HandleException(self);
end;

procedure TSmartDBF.InternalInitFieldDefs;
var
  DBFField: FIELD_REC;
  dbSize: Integer;
  dbOffset: Integer;
  i, j: Integer;
  qq: PFieldInfo;
begin
  FieldDefs.BeginUpdate;
  FieldDefs.Clear;
  if FFieldInfo.Count <> 0 then
    for i := 0 to FFieldInfo.Count - 1 do
      MemMgr.oMem.FreeMem(FFieldInfo.Items[i]);
  FFieldInfo.Clear;
  FileSeek(DBFHandler, SizeOf(DBFHeader), soFromBeginning);
  dbOffset := 1;
  j := 0;
  while true do
  begin
    FileRead(DBFHandler, DBFField, SizeOf(FIELD_REC));
    if DBFField.field_name[0] = #13 then break;
    dbSize := 0;
    qq := PFieldInfo(MemMgr.oMem.GetMem(self, SizeOf(TFieldInfo)));
    FieldDefs.AddFieldDef;
    FieldDefs.Items[j].Name := UpperCase(Trim(DBFField.field_name));
    FieldDefs.Items[j].Required := false;
    case DBFField.field_type of
      'C':
        begin
          FieldDefs.Items[j].DataType := ftString;
          FieldDefs.Items[j].Size := DBFField.lendth.char_len;
          dbSize := DBFField.lendth.char_len;
          FieldDefs.Items[j].Precision := 0;
          qq.FLen    := dbSize;
          qq.FPrec   := 0;
        end;
      'N':
        begin
          if DBFField.lendth.num_len.dec = 0 then
          begin
            FieldDefs.Items[j].DataType := ftInteger;
            FieldDefs.Items[j].Size := 0;
            dbSize := DBFField.lendth.num_len.len;
            FieldDefs.Items[j].Precision := 0;
          end else begin
            FieldDefs.Items[j].DataType := ftFloat;
            FieldDefs.Items[j].Size := 0;
            dbSize := DBFField.lendth.num_len.len;
            FieldDefs.Items[j].Precision := DBFField.lendth.num_len.dec;
          end;
          qq.FLen    := dbSize;
          qq.FPrec   := FieldDefs.Items[j].Precision;
        end;
      'D':
        begin
          FieldDefs.Items[j].DataType := ftDate;
          FieldDefs.Items[j].Size := 0;
          dbSize := 8;
          FieldDefs.Items[j].Precision := 0;
          qq.FLen    := 8;
          qq.FPrec   := 0;
        end;
      'L':
        begin
          FieldDefs.Items[j].DataType := ftBoolean;
          FieldDefs.Items[j].Size := 0;
          dbSize := 1;
          FieldDefs.Items[j].Precision := 0;
          qq.FLen    := 1;
          qq.FPrec   := 0;
        end;
      'M':
        begin
          FieldDefs.Items[j].DataType := ftMemo;
          FieldDefs.Items[j].Size := 0;
          dbSize := 10;
          FieldDefs.Items[j].Precision := 0;
          qq.FLen    := 10;
          qq.FPrec   := 0;
        end;
      'E':    //Extended types
        begin
          case DBFField.extend_type of
            dbftS1:       //Shortint
              begin
                FieldDefs.Items[j].DataType := ftSmallint;
                dbSize := 1;
                qq.FLen    := 4;
                qq.FPrec   := 0;
              end;
            dbftU1:       //Byte
              begin
                FieldDefs.Items[j].DataType := ftWord;
                dbSize := 1;
                qq.FLen    := 4;
                qq.FPrec   := 0;
              end;
            dbftU1_NB:       //Byte with null bit instead of sign bit
              begin
                FieldDefs.Items[j].DataType := ftWord;
                dbSize := 1;
                qq.FLen    := 3;
                qq.FPrec   := 0;
              end;
            dbftU2_NB:       //Byte with null bit instead of sign bit
              begin
                FieldDefs.Items[j].DataType := ftWord;
                dbSize := 2;
                qq.FLen    := 5;
                qq.FPrec   := 0;
              end;
            dbftS2:       //Smallint
              begin
                FieldDefs.Items[j].DataType := ftSmallint;
                dbSize := 2;
                qq.FLen    := 6;
                qq.FPrec   := 0;
              end;
            dbftU2:       //Word
              begin
                FieldDefs.Items[j].DataType := ftWord;
                dbSize := 2;
                qq.FLen    := 6;
                qq.FPrec   := 0;
              end;
            dbftS4:       //Longint
              begin
                FieldDefs.Items[j].DataType := ftInteger;
                dbSize := 4;
                qq.FLen    := 11;
                qq.FPrec   := 0;
              end;
            dbftU4:       //Longword
              begin
                FieldDefs.Items[j].DataType := ftInteger;
                dbSize := 4;
                qq.FLen    := 11;
                qq.FPrec   := 0;
              end;
            dbftU4_NB:       //
              begin
                FieldDefs.Items[j].DataType := ftInteger;
                dbSize := 4;
                qq.FLen    := 10;
                qq.FPrec   := 0;
              end;
            dbftS8:       //Int64
              begin
                FieldDefs.Items[j].DataType := ftLargeint;
                dbSize := 8;
                qq.FLen    := 21;
                qq.FPrec   := 0;
              end;
            dbftR4:       //Single
              begin
                FieldDefs.Items[j].DataType := ftFloat;
                dbSize := 4;
                qq.FLen    := 18;
                qq.FPrec   := 8;
              end;
            dbftR4_NB:
              begin
                FieldDefs.Items[j].DataType := ftFloat;
                dbSize := 4;
                qq.FLen    := 18;
                qq.FPrec   := 8;
              end;
            dbftR6_NB:
              begin
                FieldDefs.Items[j].DataType := ftFloat;
                dbSize := 6;
                qq.FLen    := 26;
                qq.FPrec   := 12;
              end;
            dbftR8_NB:
              begin
                FieldDefs.Items[j].DataType := ftFloat;
                dbSize := 8;
                qq.FLen    := 34;
                qq.FPrec   := 16;
              end;
            dbftR6:       //Real48
              begin
                FieldDefs.Items[j].DataType := ftFloat;
                dbSize := 6;
                qq.FLen    := 26;
                qq.FPrec   := 12;
              end;
            dbftR8:       //Double
              begin
                FieldDefs.Items[j].DataType := ftFloat;
                dbSize := 8;
                qq.FLen    := 34;
                qq.FPrec   := 16;
              end;
            dbftR10:      //Extended
              begin
                FieldDefs.Items[j].DataType := ftFloat;
                dbSize := 10;
                qq.FLen    := 42;
                qq.FPrec   := 20;
              end;
            dbftD1:
              begin
                FieldDefs.Items[j].DataType := ftDateTime;
                dbSize := 8;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftD1_NB:
              begin
                FieldDefs.Items[j].DataType := ftDateTime;
                dbSize := 8;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftD2:
              begin
                FieldDefs.Items[j].DataType := ftDateTime;
                dbSize := 8;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftD2_NB:
              begin
                FieldDefs.Items[j].DataType := ftDateTime;
                dbSize := 8;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftD3:
              begin
                FieldDefs.Items[j].DataType := ftDateTime;
                dbSize := 6;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftD3_NB:
              begin
                FieldDefs.Items[j].DataType := ftDateTime;
                dbSize := 6;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftS1_N:     //Shortint with NULL
              begin
                FieldDefs.Items[j].DataType := ftSmallint;
                dbSize := 2;
                qq.FLen    := 4;
                qq.FPrec   := 0;
              end;
            dbftU1_N:     //Byte  with NULL
              begin
                FieldDefs.Items[j].DataType := ftWord;
                dbSize := 2;
                qq.FLen    := 4;
                qq.FPrec   := 0;
              end;
            dbftS2_N:     //Smallint with NULL
              begin
                FieldDefs.Items[j].DataType := ftSmallint;
                dbSize := 3;
                qq.FLen    := 6;
                qq.FPrec   := 0;
              end;
            dbftU2_N:     //Word with NULL
              begin
                FieldDefs.Items[j].DataType := ftWord;
                dbSize := 3;
                qq.FLen    := 6;
                qq.FPrec   := 0;
              end;
            dbftS4_N:     //Longint with NULL
              begin
                FieldDefs.Items[j].DataType := ftInteger;
                dbSize := 5;
                qq.FLen    := 11;
                qq.FPrec   := 0;
              end;
            dbftU4_N:     //Longword with NULL
              begin
                FieldDefs.Items[j].DataType := ftInteger;
                dbSize := 5;
                qq.FLen    := 11;
                qq.FPrec   := 0;
              end;
            dbftS8_N:     //Int64 with NULL
              begin
                FieldDefs.Items[j].DataType := ftLargeint;
                dbSize := 9;
                qq.FLen    := 21;
                qq.FPrec   := 0;
              end;
            dbftR4_N:     //Single with NULL
              begin
                FieldDefs.Items[j].DataType := ftFloat;
                dbSize := 5;
                qq.FLen    := 18;
                qq.FPrec   := 8;
              end;
            dbftR6_N:     //Real48 with NULL
              begin
                FieldDefs.Items[j].DataType := ftFloat;
                dbSize := 7;
                qq.FLen    := 26;
                qq.FPrec   := 12;
              end;
            dbftR8_N:     //Double with NULL
              begin
                FieldDefs.Items[j].DataType := ftFloat;
                dbSize := 9;
                qq.FLen    := 34;
                qq.FPrec   := 16;
              end;
            dbftR10_N:     //Extended with NULL
              begin
                FieldDefs.Items[j].DataType := ftFloat;
                FieldDefs.Items[j].Size := 11;
                dbSize := 11;
                qq.FLen    := 42;
                qq.FPrec   := 20;
              end;
            dbftD1_N:
              begin
                FieldDefs.Items[j].DataType := ftDateTime;
                dbSize := 9;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftD2_N:
              begin
                FieldDefs.Items[j].DataType := ftDateTime;
                dbSize := 9;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftD3_N:
              begin
                FieldDefs.Items[j].DataType := ftDateTime;
                dbSize := 7;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftDate:
              begin
                FieldDefs.Items[j].DataType := ftDate;
                dbSize := 4;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftDate_N:
              begin
                FieldDefs.Items[j].DataType := ftDate;
                dbSize := 5;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftDate_NB:
              begin
                FieldDefs.Items[j].DataType := ftDate;
                dbSize := 4;
                qq.FLen    := 8;
                qq.FPrec   := 0;
              end;
            dbftTime:
              begin
                FieldDefs.Items[j].DataType := ftTime;
                dbSize := 4;
                qq.FLen    := 6;
                qq.FPrec   := 0;
              end;
            dbftTime_NB:
              begin
                FieldDefs.Items[j].DataType := ftTime;
                dbSize := 4;
                qq.FLen    := 6;
                qq.FPrec   := 0;
              end;
            dbftTime_N:
              begin
                FieldDefs.Items[j].DataType := ftTime;
                dbSize := 5;
                qq.FLen    := 6;
                qq.FPrec   := 0;
              end;
            dbftClob, dbftFmtMemo:
              begin
                FieldDefs.Items[j].DataType := ftMemo;
                FieldDefs.Items[j].Size := 0;
                dbSize := 10;
                FieldDefs.Items[j].Precision := 0;
                qq.FLen    := 10;
                qq.FPrec   := 0;
              end;
            dbftBlob, dbftGraphic:
              begin
                FieldDefs.Items[j].DataType := ftBlob;
                FieldDefs.Items[j].Size := 0;
                dbSize := 10;
                FieldDefs.Items[j].Precision := 0;
                qq.FLen    := 10;
                qq.FPrec   := 0;
              end;
            dbftClob_NB, dbftFmtMemo_NB:
              begin
                FieldDefs.Items[j].DataType := ftMemo;
                FieldDefs.Items[j].Size := 0;
                dbSize := 4;
                FieldDefs.Items[j].Precision := 0;
                qq.FLen    := 4;
                qq.FPrec   := 0;
              end;
            dbftBlob_NB, dbftGraphic_NB:
              begin
                FieldDefs.Items[j].DataType := ftBlob;
                FieldDefs.Items[j].Size := 0;
                dbSize := 4;
                FieldDefs.Items[j].Precision := 0;
                qq.FLen    := 4;
                qq.FPrec   := 0;
              end;
            dbftString:
              begin
                FieldDefs.Items[j].DataType := ftString;
                FieldDefs.Items[j].Size := DBFField.lendth.char_len;
                dbSize := DBFField.lendth.char_len + 2;
                FieldDefs.Items[j].Precision := 0;
                qq.FLen    := DBFField.lendth.char_len;
                qq.FPrec   := 0;
              end;
            dbftString_N:
              begin
                FieldDefs.Items[j].DataType := ftString;
                FieldDefs.Items[j].Size := DBFField.lendth.char_len;
                dbSize := DBFField.lendth.char_len + 3;
                FieldDefs.Items[j].Precision := 0;
                qq.FLen    := DBFField.lendth.char_len;
                qq.FPrec   := 0;
              end;
            dbftFixedChar:
              begin
                FieldDefs.Items[j].DataType := ftFixedChar;
                FieldDefs.Items[j].Size := DBFField.lendth.char_len;
                dbSize := DBFField.lendth.char_len + 1;
                FieldDefs.Items[j].Precision := 0;
                qq.FLen    := DBFField.lendth.char_len;
                qq.FPrec   := 0;
              end;
            dbftWideString:
              begin
                FieldDefs.Items[j].DataType := ftWideString;
                FieldDefs.Items[j].Size := DBFField.lendth.char_len;
                dbSize := DBFField.lendth.char_len * 2 + 4;
                FieldDefs.Items[j].Precision := 0;
                qq.FLen    := DBFField.lendth.char_len;
                qq.FPrec   := 0;
              end;
            dbftCurrency:
              begin
                FieldDefs.Items[j].DataType := ftCurrency;
                dbSize := 8;
                qq.FLen    := 25;
                qq.FPrec   := 4;
              end;
            dbftCurrency_N:
              begin
                FieldDefs.Items[j].DataType := ftCurrency;
                dbSize := 9;
                qq.FLen    := 25;
                qq.FPrec   := 4;
              end;
            dbftCurrency_NB:
              begin
                FieldDefs.Items[j].DataType := ftCurrency;
                dbSize := 8;
                qq.FLen    := 25;
                qq.FPrec   := 4;
              end;
            dbftBCD:
              begin
                FieldDefs.Items[j].DataType := ftBCD;
                dbSize := DBFField.lendth.num_len.len shr 1;
                if ( DBFField.lendth.num_len.len and $01 ) = $01 then Inc(dbSize);
                Inc(dbSize);
                qq.FLen := 25;
                qq.FPrec := 4;
              end;
          end;
        end;
    end;
    qq.FName   := FieldDefs.Items[j].Name;
    qq.FType   := DBFField.field_type;
    qq.FExType := DBFField.extend_type;
    //qq.FLen    := dbSize;
    //qq.FPrec   := FieldDefs.Items[j].Precision;
    qq.FOff    := dbOffset;
    dbOffset   := dbOffset + dbSize;
    FFieldInfo.Add(qq);
    Inc(j);
  end;
  FieldDefs.EndUpdate;
end;

procedure TSmartDBF.InternalInitRecord(Buffer: PChar);
begin
  if Buffer <> nil then FillChar(Buffer^, RecordBufferSize, #32);
end;

procedure TSmartDBF.InternalLast;
var
  j: Longint;
  i: Integer;
  LOff: Integer;
begin
  FBOF := false;
  FEOF := true;
  FBufDir := bdFromBottom;
  FCurInd := FRecordsPerBuf;
  if FIndexes.ActiveObject = nil then begin
    LOff := FileSeek(DBFHandler, 0, 2);
    if LockHeader then
      try
        DBFHeader.last_rec := ( (LOff - DBFHeader.data_offset) div DBFHeader.rec_size );
      finally
        UnLockHeader;
      end;
    if DBFHeader.last_rec <> 0 then begin
      j := DBFHeader.last_rec - FRecordsPerBuf + 1;
      if j < 1 then j := 1;
      FBufCnt := DBFHeader.last_rec - j + 1;
      FileSeek(DBFHandler, DBFHeader.data_offset + ((j - 1) * FRecordSize), soFromBeginning);
      FileRead(DBFHandler, (FBuffer + (FRecordsPerBuf - FBufCnt) * FRecordSize)^, FBufCnt * FRecordSize);
      for i := 0 to FBufCnt - 1 do pLongint(pChar(FBufInd) + (FRecordsPerBuf - i - 1)*SizeOf(Longint))^ := DBFHeader.last_rec - i;
    end else begin
      FBOF := true;
      FEOF := true;
    end;
  end else begin
    if FIndexes.ActiveObject.FLock then
      try
        FBufCnt := FIndexes.ActiveObject.FillLastBufRecords(DBFHandler, FBuffer, FRecordsPerBuf, FRecordSize, FBufInd, DBFHeader.data_offset);
      finally
        FIndexes.ActiveObject.FUnLock;
      end
    else
      raise Exception.Create('TDBFDataSet: Can not read from index file (FLock is false).');
  end;
end;

procedure TSmartDBF.InternalOpen;
var
  i: Integer;
  b: boolean;
begin
  LobHandler := -1;
  DBFHandler := FileOpen(DBFFileName, AccessMode.AccessMode);
  if DBFHandler <= 0 then begin
    if (csDesigning in ComponentState) then
      ShowMessage('TDBFDataSet: Open error');
  end else begin
    FileSeek(DBFHandler, 0, 0);
    FileRead(DBFHandler, DBFHeader, SizeOf(DBF_HEAD));
    if not ((DBFHeader.dbf_id = $03) or (DBFHeader.dbf_id = $07) or (DBFHeader.dbf_id = $83) or (DBFHeader.dbf_id = $8B)) then begin
      FileClose(DBFHandler);
      DBFHandler := -1;
      if (csDesigning in ComponentState) then
        ShowMessage('TDBFDataSet: File is not DBF file');
    end else begin
      if DBFHeader.dbf_id = $83 then
        LobHandler := FileOpen(ChangeFileExt(DBFFileName, '.dbt'), AccessMode.AccessMode);
      FRecordSize := DBFHeader.rec_size;
      FBuffer := MemMgr.oMem.GetMem(self, FBufferSize + 10);
      FRecordsPerBuf := FBufferSize div FRecordSize;
      FBufCnt := 0;
      FBufDir := bdFromTop;;
      FBufInd := MemMgr.oMem.GetMem(self, FRecordsPerBuf * SizeOf(LongInt));
      InternalInitFieldDefs();
      if DefaultFields then CreateFields;
      BindFields(true);
      b := true;
      for i := 0 to FIndexes.Count - 1 do begin
        if not FIndexes.Items[i].Open then begin
          b := false;
          FreeRecordBuffer(FBuffer);
          FRecordsPerBuf := 0;
          FBuffer := nil;
          MemMgr.oMem.FreeMem(FBufInd);
          FBufInd := nil;
          FBufCnt := 0;
          FBufDir := bdFromTop;
          BindFields(false);
          if DefaultFields then DestroyFields;
          FileClose(DBFHandler);
          DBFHandler := -1;
          FIndexes.CloseAll;
          break;
        end;
      end;
      if b then begin
        SetRngInt;
        InternalFirst;
        FTempRecord := AllocRecordBuffer;
        FFilterRecord := AllocRecordBuffer;
      end;
    end;
  end;
  FBOF := true;
  ObjectView := false;
  BookmarkSize := sizeof(Longword);
end;

procedure TSmartDBF.InternalPost;
var

  i, l: Integer;
  fOffset: Integer;
  ActiveBuf: pChar;
  RealRead: Integer;

  lpMsgBuf: array [0..500] of Char;
  le: DWORD;

  NewKey: String;

  b: boolean;

begin
  b := false;
  CheckActive;
  GetActiveRecBuf(ActiveBuf);
  l := pTRecInfo(ActiveBuf + RecordSize).RecordRowID;
  if State = dsEdit then
  begin
    fOffset := FileSeek(DBFHandler, 0, 1);
    if RLock(l) then
      try
        FileSeek(DBFHandler, DBFHeader.data_offset + LongWord((l - 1) * FRecordSize), 0);
        RealRead := FileWrite(DBFHandler, ActiveBuf^, DBFHeader.rec_size);
        if RealRead = -1 then
        begin
          le := GetLastError();
          FormatMessage(
              FORMAT_MESSAGE_FROM_SYSTEM,
              nil,
              le,
              0, // Default language
              lpMsgBuf,
              500,
              nil
          );
          raise Exception.Create('TSmartDBF: ' + lpMsgBuf);
        end else begin
          Move(ActiveBuf^, (FBuffer + GetCurIndByRec(l) * FRecordSize)^, FRecordSize);
          for i := 0 to Indexes.Count - 1 do begin
            NewKey := Indexes[i].EvaluteKeyExpr;
            if NewKey <> Indexes[i].FOldEditKey then begin

              if  (Indexes.ActiveObject <> nil) and
                  (Indexes.ActiveObject = Indexes[i]) and
                  (Indexes[i].IsRanged) and
                  (not Indexes[i].InRange(NewKey)) then b := true;

              if  not (
                ( (FAccessMode.FLast and fmShareExclusive) = fmShareExclusive ) or
                ( (FAccessMode.FLast and fmShareDenyWrite) = fmShareDenyWrite ) or
                FFileLock
                  ) then begin
                if Indexes[i].FLock then
                  try
                    Indexes[i].StartUpdate(false);
                    Indexes[i].DeleteKey(Indexes[i].FOldEditKey, Indexes[i].FOldEditRec);
                    Indexes[i].AddKey(NewKey, l);
                  finally
                    Indexes[i].Flush;
                    Indexes[i].FUnLock;
                  end
                else
                  raise Exception.Create('TDBFDataSet.InternalPost: Can not Delete/add key to index file (FLock is false).');
              end else begin
                if Indexes[i].FLock then
                  try
                    Indexes[i].DeleteKey(Indexes[i].FOldEditKey, Indexes[i].FOldEditRec);
                    Indexes[i].AddKey(NewKey, l);
                  finally
                    Indexes[i].FUnLock;
                  end
                else
                  raise Exception.Create('TDBFDataSet.InternalPost: Can not Delete/add key to index file (FLock is false).');
              end;
            end;
          end;
          if not FFastPostRecord then begin
            if b then
              InternalFirst
            else
              GetBufferByRec(l);
          end;
        end;
      finally
        RUnLock(l);
        FileSeek(DBFHandler, fOffset, 0);
      end
    else
      raise Exception.Create('TSmartDBF.InternalPost: Can not lock DBF record.');
  end else begin
    InternalAddRecord(ActiveBuf, true);
  end;
end;

procedure TSmartDBF.InternalSetToRecord(Buffer: PChar);
var
  i, j: Longint;
begin
  i := pTRecInfo(Buffer + RecordSize).RecordRowID;
  for j := 0 to FBufCnt - 1 do begin
    if FBufDir = bdFromTop then
      if pLongint(pChar(FBufInd) + j * SizeOf(Longint))^ = i then begin
        FCurInd := j;
        FBOF := false;
        FEOF := false;
        Exit;
      end;
    if FBufDir = bdFromBottom then
      if pLongint(pChar(FBufInd) + (FRecordsPerBuf - j) * SizeOf(Longint))^ = i then begin
        FCurInd := FRecordsPerBuf - j;
        FBOF := false;
        FEOF := false;
        Exit;
      end;
  end;
  GetBufferByRec(i);
end;

function TSmartDBF.IsCursorOpen: Boolean;
var
  i: Integer;
begin
  Result := (DBFHandler > 0);
  if FIndexes <> nil then
    for i := 0 to FIndexes.Count - 1 do Result := Result and FIndexes.Items[i].IsOpen;
end;

procedure TSmartDBF.Loaded;
begin
  inherited Loaded();
  if FStreamedActive then Active := true;
end;

procedure TSmartDBF.SetActive(Value: boolean);
begin
  if csReading in ComponentState then
    FStreamedActive := Value
  else begin
    inherited SetActive(Value);
    if not IsCursorOpen then
      inherited SetActive(false);
  end;
end;

procedure TSmartDBF.SetBookmarkData(Buffer: PChar; Data: Pointer);
begin
  pTRecInfo(Buffer + RecordSize).RecordRowID := Longword(Data^);
end;

procedure TSmartDBF.SetBookmarkFlag(Buffer: PChar; Value: TBookmarkFlag);
begin
  pTRecInfo(Buffer + RecordSize).BookmarkFlag := Value;
end;

procedure TSmartDBF.SetFieldData(Field: TField; Buffer: Pointer);
var
  ss, ActiveBuf: pChar;
  qq: PFieldInfo;
  dd: double;
  sTS: TTimeStamp;
  Year, Month, Day: Word;
  dInt: Integer;
  dFloat: double;
  dBool: WordBool;
  q: String;
  p0, p1, p2, p3: byte;
  wE: boolean;
  w: char;
  i: Integer;
  SLen: WORD;
begin
  case Field.FieldKind of
  fkData:
    begin
      qq := PFieldInfo(FFieldInfo.Items[Field.FieldNo - 1]);
      GetActiveRecBuf(ActiveBuf);
      if (ActiveBuf <> nil) then begin
        if (Buffer <> nil) then begin
          ss := ActiveBuf + qq.FOff;
          case Field.DataType of
            ftDate:
              begin
                if qq.FType <> 'E' then begin
                  sTS.Date := pInteger(Buffer)^;
                  sTS.Time := 0;
                  dd := TimeStampToDateTime(sTS);
                  DecodeDate(dd, Year, Month, Day);
                  p0 := Year div 1000;
                  ss[0] := char( p0 + $30 );
                  p1 := (Year - p0 * 1000) div 100;
                  ss[1] := Char( p1 + $30 );
                  p2 := (Year - p0 * 1000 - p1 * 100) div 10;
                  ss[2] := Char( p2 + $30 );
                  p3 := (Year - p0 * 1000 - p1 * 100 - p2 * 10);
                  ss[3] := Char( p3 + $30 );
                  ss[4] := Char( (Month div 10) + $30 );
                  ss[5] := Char( (Month - (Month div 10) * 10 ) + $30 );
                  ss[6] := Char( (Day div 10) + $30 );
                  ss[7] := Char( (Day - (Day div 10) * 10) + $30 );
                  (*
                  dd := double(Buffer^);
                  sTS.Date := Trunc(dd/(3600.0*24*1000));
                  sTS.Time := Trunc(dd - sTS.Date*((3600.0*24*1000)));
                  dd := TimeStampToDateTime(sTS);
                  DecodeDate(dd, Year, Month, Day);
                  p0 := Year div 1000;
                  ss[0] := char( p0 + $30 );
                  p1 := (Year - p0 * 1000) div 100;
                  ss[1] := Char( p1 + $30 );
                  p2 := (Year - p0 * 1000 - p1 * 100) div 10;
                  ss[2] := Char( p2 + $30 );
                  p3 := (Year - p0 * 1000 - p1 * 100 - p2 * 10);
                  ss[3] := Char( p3 + $30 );
                  ss[4] := Char( (Month div 10) + $30 );
                  ss[5] := Char( (Month - (Month div 10) * 10 ) + $30 );
                  ss[6] := Char( (Day div 10) + $30 );
                  ss[7] := Char( (Day - (Day div 10) * 10) + $30 );
                  *)
                end else begin
                  case qq.FExType of
                    dbftDate: pInteger(ss)^ := pInteger(Buffer)^;
                    dbftDate_N:
                      begin
                        Byte(Pointer(ss)^) := 1;
                        pInteger(ss + 1)^ := pInteger(Buffer)^;
                      end;
                    dbftDate_NB: LongWord(Pointer(ss)^) := ( pLongWord(Buffer)^ or $80000000 );
                  end;
                end;
              end;
            ftTime:
              begin
                case qq.FExType of
                  dbftTime: pInteger(ss)^ := pInteger(Buffer)^;
                  dbftTime_N:
                    begin
                      Byte(Pointer(ss)^) := 1;
                      pInteger(ss + 1)^ := pInteger(Buffer)^;
                    end;
                  dbftTime_NB: LongWord(Pointer(ss)^) := ( pLongWord(Buffer)^ or $80000000 );
                end;
              end;
            ftBCD:
              begin
                pBcd(ss)^ := pBcd(Buffer)^;
                Byte(ss[1]) := ( Byte(ss[1]) or $40 );
              end;
            ftCurrency:
              begin
                case qq.FExType of
                  dbftCurrency: Currency(Pointer(ss)^) := pCurrency(Buffer)^;
                  dbftCurrency_N:
                    begin
                      Byte(Pointer(ss)^) := 1;
                      Currency(Pointer(ss + 1)^) := pCurrency(Buffer)^;
                    end;
                  dbftCurrency_NB:
                    begin
                      Currency(Pointer(ss)^) := pCurrency(Buffer)^;
                      Byte(Pointer(ss + 7)^) := ( pByte(ss + 7)^ or $80 );
                    end;
                end;
              end;
            ftWideString:
              begin
                //WideString(Pointer(ss + 1)^) := pWideString(Buffer)^;
              end;
            ftString:
              begin
                if qq.FType <> 'E' then begin
                  wE := false;
                  for i := 0 to qq.FLen - 1 do
                  begin
                    w := pChar(Buffer)[i];
                    if w = #0 then
                      wE := true;
                    if not wE then
                      ss[i] := w
                    else
                      ss[i] := ' ';
                  end;
                end else begin
                  case qq.FExType of
                    dbftString:
                      begin
                        SLen := StrLen(Buffer);
                        WORD(Pointer(ss)^) := SLen;
                        ss := ss + SizeOf(WORD);
                        Move(Buffer^, ss^, SLen);
                      end;
                    dbftString_N:
                      begin
                        Byte(Pointer(ss)^) := 1;
                        SLen := StrLen(Buffer);
                        ss := ss + 1;
                        WORD(Pointer(ss)^) := SLen;
                        ss := ss + SizeOf(WORD);
                        Move(Buffer^, ss^, SLen);
                      end;
                    dbftFixedChar:
                      begin
                        SLen := StrLen(Buffer);
                        Move(Buffer^, ss^, SLen);
                        ss[SLen] := #0;
                      end;
                  end;
                end;
              end;
            ftDateTime:
              begin
                case qq.FExType of
                  dbftD1:
                    begin
                      dd := double(Buffer^);
                      sTS.Date := Trunc(dd/(3600.0*24*1000));
                      sTS.Time := Trunc(dd - sTS.Date*((3600.0*24*1000)));
                      dd := TimeStampToDateTime(sTS);
                      Double(Pointer(ss)^) := Double(dd);
                    end;
                  dbftD1_NB:
                    begin
                      dd := double(Buffer^);
                      sTS.Date := Trunc(dd/(3600.0*24*1000));
                      sTS.Time := Trunc(dd - sTS.Date*((3600.0*24*1000)));
                      dd := TimeStampToDateTime(sTS);
                      Double(Pointer(ss)^) := Double(dd);
                      Byte(Pointer(ss + 7)^) := ( pByte(ss + 7)^ or $80 );
                    end;
                  dbftD2: Double(Pointer(ss)^) := pDouble(Buffer)^;
                  dbftD2_NB:
                    begin
                      Double(Pointer(ss)^) := pDouble(Buffer)^;
                      Byte(Pointer(ss + 7)^) := ( pByte(ss + 7)^ or $80 );
                    end;
                  dbftD3:
                    begin
                      dd := double(Buffer^);
                      sTS.Date := Trunc(dd/(3600.0*24*1000));
                      sTS.Time := Trunc(dd - sTS.Date*((3600.0*24*1000)));
                      dd := TimeStampToDateTime(sTS);
                      Real48(Pointer(ss)^) := Double(dd);
                    end;
                  dbftD3_NB:
                    begin
                      Real48(Pointer(ss)^) := pDouble(Buffer)^;
                      Byte(Pointer(ss + 5)^) := ( pByte(ss + 5)^ or $80 );
                    end;
                  dbftD1_N:
                    begin
                      dd := double(Buffer^);
                      sTS.Date := Trunc(dd/(3600.0*24*1000));
                      sTS.Time := Trunc(dd - sTS.Date*((3600.0*24*1000)));
                      dd := TimeStampToDateTime(sTS);
                      Byte(Pointer(ss)^) := 1;
                      Double(Pointer(ss + 1)^) := Double(dd);
                    end;
                  dbftD2_N:
                    begin
                      Byte(Pointer(ss)^) := 1;
                      Double(Pointer(ss + 1)^) := pDouble(Buffer)^;
                    end;
                  dbftD3_N:
                    begin
                      dd := double(Buffer^);
                      sTS.Date := Trunc(dd/(3600.0*24*1000));
                      sTS.Time := Trunc(dd - sTS.Date*((3600.0*24*1000)));
                      dd := TimeStampToDateTime(sTS);
                      Byte(Pointer(ss)^) := 1;
                      Real48(Pointer(ss + 1)^) := Double(dd);
                    end;
                end;
              end;
            ftSmallint:
              begin
                case qq.FExType of
                  dbftS1: Shortint(Pointer(ss)^) := pShortint(Buffer)^;
                  dbftS2: Smallint(Pointer(ss)^) := pSmallint(Buffer)^;
                  dbftS1_N:     //Shortint with NULL
                    begin
                      Byte(Pointer(ss)^) := 1;
                      Shortint(Pointer(ss + 1)^) := pShortint(Buffer)^;
                    end;
                  dbftS2_N:     //Smallint with NULL
                    begin
                      Byte(Pointer(ss)^) := 1;
                      Smallint(Pointer(ss + 1)^) := pSmallint(Buffer)^;
                    end;
                end;
              end;
            ftWord:
              begin
                case qq.FExType of
                  dbftU1: Byte(Pointer(ss)^) := pByte(Buffer)^;
                  dbftU2: Word(Pointer(ss)^) := pWord(Buffer)^;
                  dbftU1_N:
                    begin
                      Byte(Pointer(ss)^) := 1;
                      Byte(Pointer(ss + 1)^) := pByte(Buffer)^;
                    end;
                  dbftU2_N:
                    begin
                      Byte(Pointer(ss)^) := 1;
                      Word(Pointer(ss + 1)^) := pWord(Buffer)^;
                    end;
                  dbftU1_NB: Byte(Pointer(ss)^) := ( pByte(Buffer)^ or $80 );
                  dbftU2_NB: Word(Pointer(ss)^) := ( pWord(Buffer)^ or $8000 );
                end;
              end;
            ftInteger:
              begin
                if qq.FType <> 'E' then begin
                  dInt := Integer(Buffer^);
                  Str(dInt:qq.FLen, q);
                  Move(pChar(q)^, ss^, qq.FLen);
                end else begin
                  case qq.FExType of
                    dbftS4, dbftU4:       //Longint, Longword
                      begin
                        Integer(Pointer(ss)^) := pInteger(Buffer)^;
                      end;
                    dbftS4_N, dbftU4_N:     //Longint with NULL, Longword with NULL
                      begin
                        Byte(Pointer(ss)^) := 1;
                        Integer(Pointer(ss + 1)^) := pInteger(Buffer)^;
                      end;
                    dbftU4_NB: LongWord(Pointer(ss)^) := ( pLongWord(Buffer)^ or $80000000 );
                  end;
                end;
              end;
            ftLargeint:
              begin
                case qq.FExType of
                  dbftS8:   Int64(Pointer(ss)^) := pInt64(Buffer)^;
                  dbftS8_N:
                    begin
                      Byte(Pointer(ss)^) := 1;
                      Int64(Pointer(ss + 1)^) := pInt64(Buffer)^;
                    end;
                end;
              end;
            ftFloat:
              begin
                if qq.FType <> 'E' then begin
                  dFloat := Double(Buffer^);
                  Str(dFloat:qq.FLen:qq.FPrec, q);
                  Move(pChar(q)^, ss^, qq.FLen);
                end else begin
                  case qq.FExType of
                    dbftR4: Single(Pointer(ss)^) := pDouble(Buffer)^;
                    dbftR4_NB:
                      begin
                        Single(Pointer(ss)^) := pDouble(Buffer)^;
                        Byte(Pointer(ss + 3)^) := ( pByte(ss + 3)^ or $80 );
                      end;
                    dbftR6_NB:
                      begin
                        Real48(Pointer(ss)^) := pDouble(Buffer)^;
                        Byte(Pointer(ss + 5)^) := ( pByte(ss + 5)^ or $80 );
                      end;
                    dbftR8_NB:
                      begin
                        Double(Pointer(ss)^) := pDouble(Buffer)^;
                        Byte(Pointer(ss + 7)^) := ( pByte(ss + 7)^ or $80 );
                      end;
                    dbftR6: Real48(Pointer(ss)^) := pDouble(Buffer)^;
                    dbftR8: Double(Pointer(ss)^) := pDouble(Buffer)^;
                    dbftR10: Extended(Pointer(ss)^) := pExtended(Buffer)^;
                    dbftR4_N:
                      begin
                        Byte(Pointer(ss)^) := 1;
                        Single(Pointer(ss + 1)^) := pDouble(Buffer)^;
                      end;
                    dbftR6_N:
                      begin
                        Byte(Pointer(ss)^) := 1;
                        Real48(Pointer(ss + 1)^) := pDouble(Buffer)^;
                      end;
                    dbftR8_N:
                      begin
                        Byte(Pointer(ss)^) := 1;
                        Double(Pointer(ss + 1)^) := pDouble(Buffer)^;
                      end;
                    dbftR10_N:
                      begin
                        Byte(Pointer(ss)^) := 1;
                        Extended(Pointer(ss + 1)^) := pExtended(Buffer)^;
                      end;
                  end;
                end;
              end;
            ftBoolean:
              begin
                dBool := WordBool(Buffer^);
                if dBool then
                   ss[0] := 'T'
                else
                  ss[0] := 'F';
              end;
          end;
        end else begin
          ss := ActiveBuf + qq.FOff;
          if qq.FType <> 'E' then begin
            for i := 0 to qq.FLen - 1 do ss[i] := ' ';
          end else begin
            case qq.FExType of
              dbftS1_N,     //Shortint with NULL
              dbftU1_N,     //Byte  with NULL
              dbftS2_N,     //Smallint with NULL
              dbftU2_N,     //Word with NULL
              dbftS4_N,     //Longint with NULL
              dbftU4_N,     //Longword with NULL
              dbftS8_N,     //Int64 with NULL
              dbftR4_N,     //Single with NULL
              dbftR6_N,     //Real48 with NULL
              dbftR8_N,     //Double with NULL
              dbftR10_N,    //Extended with NULL
              dbftD1_N,     //TDateTime
              dbftD2_N,     //DataSet DateTime
              dbftD3_N,     //Real48 DateTime
              dbftString_N, //String
              dbftString,
              dbftCurrency_N,
              dbftBCD,
              dbftDate_N,
              dbftTime_N:
                begin
                  ss[0] := ' ';
                  ss[1] := ' ';
                end;
              dbftFixedChar: for i := 0 to qq.FLen do ss[i] := ' ';
              dbftU1_NB: Byte(Pointer(ss)^) := ( pByte(ss)^ and $7F );
              dbftU2_NB: Word(Pointer(ss)^) := ( pWord(ss)^ and $7FFF );
              dbftU4_NB: Longword(Pointer(ss)^) := ( pLongword(ss)^ and $7FFFFFFF );
              dbftR4_NB: Byte(Pointer(ss + 3)^) := ( pByte(ss + 3)^ and $7F );
              dbftR6_NB: Byte(Pointer(ss + 5)^) := ( pByte(ss + 5)^ and $7F );
              dbftR8_NB: Byte(Pointer(ss + 7)^) := ( pByte(ss + 7)^ and $7F );
              dbftCurrency_NB: Byte(Pointer(ss + 7)^) := ( pByte(ss + 7)^ and $7F );
              dbftD1_NB: Byte(Pointer(ss + 7)^) := ( pByte(ss + 7)^ and $7F );
              dbftD2_NB: Byte(Pointer(ss + 7)^) := ( pByte(ss + 7)^ and $7F );
              dbftD3_NB: Byte(Pointer(ss + 5)^) := ( pByte(ss + 5)^ and $7F );
              dbftDate_NB: Longword(Pointer(ss)^) := ( pLongword(ss)^ and $7FFFFFFF );
              dbftTime_NB: Longword(Pointer(ss)^) := ( pLongword(ss)^ and $7FFFFFFF );
            end;
          end;
        end;
      end;
    end;
  fkCalculated:
    begin
      GetActiveRecBuf(ActiveBuf);
      if (Buffer <> nil) and (ActiveBuf <> nil) then
      begin
        ss := ActiveBuf + FRecordSize + sizeof(TRecInfo) + Field.Offset;
        Move(Buffer^, ss^, Field.DataSize);
      end;
    end;
  end;
  if not (State in [dsCalcFields, dsFilter, dsNewValue]) then
    DataEvent(deFieldChange, Longint(Field));
end;

procedure TSmartDBF.SetFiltered(Value: Boolean);
begin
  if Active then
  begin
    CheckBrowseMode;
    if ((not Filtered) and Value) and (Filter <> '') then
      FFilterParser.SetExprParams(Filter, FilterOptions, [poExtSyntax], '');
    if Filtered <> Value then begin
      inherited SetFiltered(Value);
      First;
    end;
  end else
    inherited SetFiltered(Value);
end;

procedure TSmartDBF.SetIndexList(const Value: TIndexes);
begin
  FIndexes.Assign(Value);
end;

procedure TSmartDBF.SetRecNo(Value: Integer);
begin
  CheckBrowseMode;
  CursorPosChanged;
  DoBeforeScroll;
  GetBufferByRec(Value);
  Resync([]);
  DoAfterScroll;
end;

procedure TSmartDBF.SetSetDeleted(const Value: Boolean);
begin
  if Active then
  begin
    CheckBrowseMode;
    if FSetDeleted <> Value then
      FSetDeleted := Value;
    First;
  end else
    FSetDeleted := Value;
end;

function TSmartDBF.Translate(Src, Dest: PChar; ToOem: Boolean): Integer;
begin
  if FOEM then
  begin
    if not ToOem then begin
      if (Src <> nil) then
      begin
        if OemToChar(Src, Dest) then
          Result := StrLen(Dest)
        else
          Result := 0;
      end else
        Result := 0;
    end else begin
      if (Src <> nil) then
      begin
        if CharToOem(Src, Dest) then
          Result := StrLen(Dest)
        else
          Result := 0;
      end else
        Result := 0;
    end;
  end else
    Result := inherited Translate(Src, Dest, ToOem);
end;

procedure TSmartDBF.DefineProperties(Filer: TFiler);

  function WriteData: Boolean;
  begin
    if Filer.Ancestor <> nil then
      Result := not FIndexes.IsEqual(TDBFNTX(Filer.Ancestor).FIndexes) else
      Result := FIndexes.Count > 0;
  end;

begin
  inherited DefineProperties(Filer);
  Filer.DefineProperty('IndexData', ReadIndexData, WriteIndexData, WriteData);
end;

procedure TSmartDBF.ReadIndexData(Reader: TReader);
begin
  Reader.ReadValue;
  Reader.ReadCollection(Indexes);
end;

procedure TSmartDBF.WriteIndexData(Writer: TWriter);
begin
  Writer.WriteCollection(Indexes);
end;

function TSmartDBF.FirstByIndex(IndInd: Integer): TGetResult;
var
  i: Integer;
begin
  CheckBrowseMode;
  CursorPosChanged;
  DoBeforeScroll;
  Result := FIndexes[IndInd].GetFirstByIndex(i);
  if Result = grOK then begin
    GetBufferByRec(i);
    Resync([]);
  end;
  DoAfterScroll;
end;

function TSmartDBF.LastByIndex(IndInd: Integer): TGetResult;
var
  i: Integer;
begin
  CheckBrowseMode;
  CursorPosChanged;
  DoBeforeScroll;
  Result := FIndexes[IndInd].GetLastByIndex(i);
  if Result = grOK then begin
    GetBufferByRec(i);
    Resync([]);
  end;
  DoAfterScroll;
end;

function TSmartDBF.NextByIndex(IndInd: Integer): TGetResult;
var
  i: Integer;
begin
  CheckBrowseMode;
  CursorPosChanged;
  DoBeforeScroll;
  FIndexes[IndInd].SetToRecord;
  Result := FIndexes[IndInd].GetRecordByIndex(gmNext, i);
  if Result = grOK then begin
    GetBufferByRec(i);
    Resync([]);
  end;
  DoAfterScroll;
end;

function TSmartDBF.PriorByIndex(IndInd: Integer): TGetResult;
var
  i: Integer;
begin
  CheckBrowseMode;
  CursorPosChanged;
  DoBeforeScroll;
  FIndexes[IndInd].SetToRecord;
  Result := FIndexes[IndInd].GetRecordByIndex(gmPrior, i);
  if Result = grOK then begin
    GetBufferByRec(i);
    Resync([]);
  end;
  DoAfterScroll;
end;

function TSmartDBF.GetPrec(aField: TField): Integer;
begin
  Result := PFieldInfo(FFieldInfo.Items[aField.FieldNo - 1]).FPrec;
end;

function TSmartDBF.GetLen(aField: TField): Integer;
begin
  Result := PFieldInfo(FFieldInfo.Items[aField.FieldNo - 1]).FLen;
end;

function TSmartDBF.NextBuffer: Longint;
var
  i, RealRead: Integer;
  OldIndex, NextRec: Longint;
  OldKey: String;
  OldRec: Longint;
begin
  if FBufCnt > 0 then begin
    if FIndexes.ActiveObject = nil then begin
      if FBufDir = bdFromTop then
        NextRec := pLongint(pChar(FBufInd) + SizeOf(Longint)*(FBufCnt - 1))^
      else
        NextRec := pLongint(pChar(FBufInd) + SizeOf(Longint)*(FRecordsPerBuf - 1))^;
      FileSeek(DBFHandler, DBFHeader.data_offset + NextRec * FRecordSize, soFromBeginning);
      RealRead := FileRead(DBFHandler, FBuffer^, FRecordsPerBuf * FRecordSize);
      Result := RealRead div FRecordSize;
      if Result > 0 then begin
        FBufCnt := Result;
        FBufDir := bdFromTop;
        FCurInd := 0;
        for i := 0 to FBufCnt - 1 do pLongint(pChar(FBufInd) + SizeOf(Longint) * i)^ := NextRec + i + 1;
      end;
    end else begin
      //Next buffer by index
      OldIndex := FCurInd;
      OldKey := FIndexes.ActiveObject.CurrentKey;
      OldRec := FIndexes.ActiveObject.CurrentRec;
      if FBufDir = bdFromTop then
        FCurInd := FBufCnt - 1
      else
        FCurInd := FRecordsPerBuf - 1;
      FKeyCalk := true;
      try
        FIndexes.ActiveObject.SetToRecord(pLongint(pChar(FBufInd) + SizeOf(Longint)*FCurInd)^);
      finally
        FKeyCalk := false;
      end;
//      Result := 0;
      if FIndexes.ActiveObject.FLock then
        try
          Result := FIndexes.ActiveObject.NextBuffer(DBFHandler, FBuffer, FRecordsPerBuf, FRecordSize, FBufInd, DBFHeader.data_offset);
        finally
          FIndexes.ActiveObject.FUnLock;
        end
      else
        raise Exception.Create('TDBFDataSet: Can not read from index file (FLock is false).');
      if Result > 0 then begin
        FBufCnt := Result;
        FBufDir := bdFromTop;
        FCurInd := 0;
      end else begin
        FCurInd := OldIndex;
        FIndexes.ActiveObject.SetToRecord(OldKey, OldRec);
      end;
    end;
  end else Result := 0;
end;

function TSmartDBF.PriorBuffer: Longint;
var
  j: Longint;
  i: Integer;
  OldIndex, NextRec: Longint;
  OldKey: String;
  OldRec: Longint;
begin
  if FBufCnt > 0 then begin
    if FIndexes.ActiveObject = nil then begin
      if FBufDir = bdFromTop then
        NextRec := FBufInd^
      else
        NextRec := pLongint(pChar(FBufInd) + (FRecordsPerBuf - FBufCnt) * SizeOf(Longint))^;
      j := NextRec - FRecordsPerBuf;
      if j < 1 then j := 1;
      Result := NextRec - j;
      if Result > 0 then begin
        FBufCnt := Result;
        FileSeek(DBFHandler, DBFHeader.data_offset + ((j - 1) * FRecordSize), soFromBeginning);
        FileRead(DBFHandler, (FBuffer + (FRecordsPerBuf - FBufCnt) * FRecordSize)^, FBufCnt * FRecordSize);
        FBufDir := bdFromBottom;
        FCurInd := FRecordsPerBuf - 1;
        for i := 0 to FBufCnt - 1 do pLongint(pChar(FBufInd) + (FRecordsPerBuf - i - 1) * SizeOf(LongInt))^ := NextRec - i - 1;
      end;
    end else begin
      //Prior buffer by index
      OldIndex := FCurInd;
      OldKey := FIndexes.ActiveObject.CurrentKey;
      OldRec := FIndexes.ActiveObject.CurrentRec;
      if FBufDir = bdFromTop then
        FCurInd := 0
      else
        FCurInd := FRecordsPerBuf - FBufCnt;
      FKeyCalk := true;
      try
        FIndexes.ActiveObject.SetToRecord(pLongint(pChar(FBufInd) + SizeOf(Longint)*FCurInd)^);
      finally
        FKeyCalk := false;
      end;
//      Result := 0;
      if FIndexes.ActiveObject.FLock then
        try
          Result := FIndexes.ActiveObject.PriorBuffer(DBFHandler, FBuffer, FRecordsPerBuf, FRecordSize, FBufInd, DBFHeader.data_offset);
        finally
          FIndexes.ActiveObject.FUnLock;
        end
      else
        raise Exception.Create('TDBFDataSet: Can not read from index file (FLock is false).');
      if Result > 0 then begin
        FBufCnt := Result;
        FBufDir := bdFromBottom;
        FCurInd := FRecordsPerBuf - 1;
        FIndexes.ActiveObject.SetToRecord;
      end else begin
        FCurInd := OldIndex;
        FIndexes.ActiveObject.SetToRecord(OldKey, OldRec);
      end;
    end;
  end else Result := 0;
end;

procedure TSmartDBF.GetBufferByRec(Rec: Integer);
var
  i, RealRead: Integer;
  NewRec: Longint;
  Result: Longint;
begin
  if FIndexes.ActiveObject = nil then begin
    NewRec := Rec - ( FRecordsPerBuf div 2);
    if NewRec < 1 then NewRec := 1;
    FileSeek(DBFHandler, DBFHeader.data_offset + (NewRec - 1) * FRecordSize, soFromBeginning);
    RealRead := FileRead(DBFHandler, FBuffer^, FRecordsPerBuf * FRecordSize);
    FBufCnt := RealRead div FRecordSize;
    if FBufCnt = 0 then begin
      FBOF := true;
      FEOF := true;
    end else begin
      FBOF := false;
      FEOF := false;
      FCurInd := 0;
    end;
    FBufDir := bdFromTop;
    for i := 0 to FBufCnt - 1 do begin
      pLongint(pChar(FBufInd) + i * SizeOf(Longint))^ := NewRec + i;
      if pLongint(pChar(FBufInd) + SizeOf(Longint)*i)^ = Rec then
        FCurInd := i;
    end;
  end else begin
    if Rec < 1 then Rec := 1;
    FileSeek(DBFHandler, DBFHeader.data_offset + (Rec - 1) * FRecordSize, soFromBeginning);
    RealRead := FileRead(DBFHandler, FBuffer^, FRecordSize);
    if RealRead = FRecordSize then begin
      FBufInd^ := Rec;
      FCurInd := 0;
      FBufDir := bdFromTop;
      FKeyCalk := true;
      try
        if not FIndexes.ActiveObject.SetToRecord(Rec) then begin
          FCurInd := -1;
          FBufDir := bdFromTop;
          FBufCnt := 0;
          FBOF := true;
          FEOF := true;
          Exit;
        end else begin
          FBOF := false;
          FEOF := false;
        end;
      finally
        FKeyCalk := false;
      end;
      if FIndexes.ActiveObject.CurrentRec <> DWORD(Rec) then begin
        Rec := FIndexes.ActiveObject.CurrentRec;
        FileSeek(DBFHandler, DBFHeader.data_offset + (Rec - 1) * FRecordSize, soFromBeginning);
        FileRead(DBFHandler, FBuffer^, FRecordSize);
        FBufInd^ := Rec;
        FCurInd := 0;
      end;
//      Result := 0;
      if FIndexes.ActiveObject.FLock then
        try
          Result := FIndexes.ActiveObject.NextBuffer(DBFHandler, FBuffer + FRecordSize, FRecordsPerBuf - 1, FRecordSize, pLongint(pChar(FBufInd) + SizeOf(Longint)), DBFHeader.data_offset);
        finally
          FIndexes.ActiveObject.FUnLock;
        end
      else
        raise Exception.Create('TDBFDataSet: Can not read from index file (FLock is false).');
      FBufCnt := Result + 1
    end else begin
      FCurInd := -1;
      FBufDir := bdFromTop;
      FBufCnt := 0;
    end;
  end;
end;

function TSmartDBF.GetRecBuf: pChar;
begin
  if ( 0 <= FCurInd ) and ( FCurInd <= FRecordsPerBuf ) then
    Result := FBuffer + FCurInd * FRecordSize
  else
    Result := nil;
end;

procedure TSmartDBF.NextIndexBuf;
begin
  FBOF := false;
  Inc(FCurInd);
  if FBufDir = bdFromTop then begin
    if FCurInd >= FBufCnt then if NextBuffer = 0 then begin
      FCurInd := FBufCnt;
      FEOF := true;
    end;
  end else begin
    if FCurInd >= FRecordsPerBuf then if NextBuffer = 0 then begin
      FCurInd := FRecordsPerBuf;
      FEOF := true;
    end;
  end;
end;

procedure TSmartDBF.PriorIndexBuf;
begin
  FEOF := false;
  Dec(FCurInd);
  if FBufDir = bdFromTop then begin
    if FCurInd < 0 then if PriorBuffer = 0 then begin
      //FCurInd := 0;
      FBOF := true;
    end;
  end else begin
    if FCurInd < FRecordsPerBuf - FBufCnt then  if PriorBuffer = 0 then begin
      //FCurInd := FRecordsPerBuf - FBufCnt;
      FBOF := true;
    end;
  end;
end;

function TSmartDBF.GetRecNoBuf: Longint;
begin
  Result := pLongint(pChar(FBufInd) + (FCurInd)*SizeOf(Longint))^;
end;

function TSmartDBF.GetDataSource: TDataSource;
begin
  Result := FDataLink.DataSource;
end;

procedure TSmartDBF.SetDataSource(const Value: TDataSource);
begin
  FDataLink.DataSource := Value;
  SetRngInt;
end;

procedure TSmartDBF.SetRngInt;
begin
  if FDataLink.DataSource <> nil then begin
    if FMasterFields <> '' then begin
      if FRange then ClearRange;
      FRange := true;
      ListMasterFields.Clear;
      if FDataLink.DataSource.DataSet <> nil then begin
        if FDataLink.DataSource.DataSet.Active then begin
          FDataLink.DataSource.DataSet.GetFieldList(ListMasterFields, FMasterFields);
          SetRange(FMasterFields, GetMasterFields);
        end;
      end;
    end;
  end;
end;

function TSmartDBF.FileWrite(Handle: Integer; const Buffer;
  Count: LongWord): Integer;
var
  i: Integer;
  l: boolean;
  Ok: boolean;
begin
  i := 0;
  l := (( (FAccessMode.FLast and fmShareExclusive) = fmShareExclusive ) or ( (FAccessMode.FLast and fmShareDenyWrite) = fmShareDenyWrite ));
  repeat
    Result := SysUtils.FileWrite(Handle, Buffer, Count);
    Ok := (Result <> -1);
    if not Ok then begin
      if l then
        Ok := true
      else begin
        Wait(0.001, false);
        Inc(i);
        if i = FWaitBusyRes then Ok := true;
      end;
    end;
  until Ok;
end;

function TSmartDBF.FileRead(Handle: Integer; var Buffer;
  Count: LongWord): Integer;
var
  i: Integer;
  l: boolean;
  Ok: boolean;
begin
  i := 0;
  l := (( (FAccessMode.FLast and fmShareExclusive) = fmShareExclusive ) or ( (FAccessMode.FLast and fmShareDenyWrite) = fmShareDenyWrite ));
  repeat
    Result := SysUtils.FileRead(Handle, Buffer, Count);
    Ok := (Result <> -1);
    if not Ok then begin
      if l then
        Ok := true
      else begin
        Wait(0.001, false);
        Inc(i);
        if i = FWaitBusyRes then Ok := true;
      end;
    end;
  until Ok;
end;

function TSmartDBF.RLock: Boolean;
begin
  Result := RLock(RecNo);
end;

function TSmartDBF.RLock(nRec: Integer): Boolean;
var
  i, k: Integer;
  l: boolean;
begin
  if FFileLock then
    Result := true
  else begin
    k := FLockRecords.IndexOf(Pointer(nRec));
    if k = -1 then begin
      l := ( ( (FAccessMode.FLast and fmShareExclusive) = fmShareExclusive ) or ( (FAccessMode.FLast and fmShareDenyWrite) = fmShareDenyWrite ) );
      if not l then begin
        i := 0;
        repeat
          Result := FileLock(DBFHandler, 1000000000 + nRec, 1);
          if not Result then begin
            Wait(0.001, false);
            Inc(i);
            if i = FWaitBusyRes then Exit;
          end else
            FLockRecords.Add(Pointer(nRec));
        until Result;
      end else
        Result := true;
    end else
      Result := true;
  end;
end;

function TSmartDBF.RUnLock: Boolean;
begin
  Result := RUnLock(RecNo);
end;

function TSmartDBF.RUnLock(nRec: Integer): Boolean;
var
  l: boolean;
  k: Integer;
begin
  l := ( ( (FAccessMode.FLast and fmShareExclusive) = fmShareExclusive ) or ( (FAccessMode.FLast and fmShareDenyWrite) = fmShareDenyWrite ) );
  if not l then begin
    Result := FileUnLock(DBFHandler, 1000000000 + nRec, 1);
    if Result then begin
      k := FLockRecords.IndexOf(Pointer(nRec));
      if k <> -1 then FLockRecords.Delete(k);
    end;
  end else
    Result := true;
end;

function TSmartDBF.LockHeader: boolean;
var
  i: Integer;
  l: boolean;
begin
  i := 0;
  l := ( ( (FAccessMode.FLast and fmShareExclusive) = fmShareExclusive ) or ( (FAccessMode.FLast and fmShareDenyWrite) = fmShareDenyWrite ) );
  repeat
    if not l then begin
      Result := FileLock(DBFHandler, 1000000000, 1);
      if not Result then begin
        Wait(0.001, false);
        Inc(i);
        if i = FWaitBusyRes then Exit;
      end;
    end else
      Result := true;
  until Result;
end;

function TSmartDBF.UnlockHeader: boolean;
var
  l: boolean;
begin
  l := ( ( (FAccessMode.FLast and fmShareExclusive) = fmShareExclusive ) or ( (FAccessMode.FLast and fmShareDenyWrite) = fmShareDenyWrite ) );
  if not l then
    Result := FileUnLock(DBFHandler, 1000000000, 1)
  else
    Result := true;
end;

function TSmartDBF.GetCurIndByRec(nRec: Integer): Integer;
var
  i: Integer;
begin
  Result := 0;
  if FBufDir = bdFromTop then begin
    for i := 0 to FBufCnt - 1 do
      if pLongint(pChar(FBufInd) + SizeOf(Longint)*i)^ = nRec then begin
        Result := i;
        Break;
      end;
  end else begin
    for i := 0 to FBufCnt - 1 do
      if pLongint(pChar(FBufInd) + (FRecordsPerBuf - i - 1) * SizeOf(LongInt))^ = nRec then begin
        Result := (FRecordsPerBuf - i - 1);
        Break;
      end;
  end;
end;

function TSmartDBF.FLock: Boolean;
var
  i: Integer;
  l: boolean;
begin
  Result := false;
  if FFileLock then begin
    Result := true;
    Exit;
  end else begin
    try
      i := 0;
      l := ( ( (FAccessMode.FLast and fmShareExclusive) = fmShareExclusive ) or ( (FAccessMode.FLast and fmShareDenyWrite) = fmShareDenyWrite ) );
      repeat
        if not l then begin
          Result := FileLock(DBFHandler, 1000000001, 1000000000);
          if not Result then begin
            Wait(0.001, false);
            Inc(i);
            if i = FWaitBusyRes then Exit;
          end;
        end else
          Result := true;
      until Result;
    finally
      FFileLock := Result;
    end;
    if FFileLock then
      for i := 0 to Indexes.Count - 1 do Indexes[i].StartUpdate;
  end;
end;

function TSmartDBF.UnLock: Boolean;
var
  l: boolean;
  i: Integer;
begin
  l := ( ( (FAccessMode.FLast and fmShareExclusive) = fmShareExclusive ) or ( (FAccessMode.FLast and fmShareDenyWrite) = fmShareDenyWrite ) );
  if not l then
    Result := FileUnLock(DBFHandler, 1000000001, 1000000000)
  else
    Result := true;
  FFileLock := not Result;
  for i := 0 to Indexes.Count - 1 do Indexes[i].Flush;
end;

procedure TSmartDBF.InternalEdit;
var
  i, l: Integer;
  ActiveBuf: pChar;
begin
  GetActiveRecBuf(ActiveBuf);
  l := pTRecInfo(ActiveBuf + RecordSize).RecordRowID;
  for i := 0 to Indexes.Count - 1 do begin
    Indexes[i].FOldEditKey := Indexes[i].EvaluteKeyExpr;
    Indexes[i].FOldEditRec := l;
  end;
end;

procedure TSmartDBF.InternalRefresh;
var
  ActiveBuf: pChar;
  Rec: Longint;
begin
  GetActiveRecBuf(ActiveBuf);
  if ActiveBuf <> nil then begin
    Rec := pTRecInfo(ActiveBuf + RecordSize).RecordRowID;
    GetBufferByRec(Rec);
  end;
end;

procedure TSmartDBF.SetOrder(nOrd: Integer);
begin
  if ( nOrd = 0 ) and ( FIndexes.ActiveObject <> nil ) then
    FIndexes.ActiveObject.Active := false;
  if (( nOrd - 1 ) >= 0 ) and ( ( nOrd - 1 ) < FIndexes.Count ) then
    FIndexes[nOrd - 1].Active := true;
end;

procedure TSmartDBF.SetOrder(sOrd: ShortString);
var
  i: Integer;
begin
  for i := 0 to FIndexes.Count - 1 do
    if UpperCase(TNTXIndex(FIndexes[i]).Order) = UpperCase(sOrd) then
      TNTXIndex(FIndexes[i]).Active := true;
end;

function TSmartDBF.GetCretateNow: Boolean;
begin
  Result := False;
end;

procedure TSmartDBF.SetCretateNow(const Value: Boolean);
begin
  if Value then begin
    CreateTable;
    if csDesigning in ComponentState then ShowMessage(Format('Table %s create successfully!', [DBFFileName]));
  end;
end;

function TSmartDBF.LocateRecord(const KeyFields: string; const KeyValues: Variant; Options: TLocateOptions): Integer;
var
  m, i, j, k, l, n, p, o: Integer;
  FFields: TList;

  procedure CntFld;
  var
    I: Integer;
  begin
    I := p;
    while (I <= Length(KeyFields)) and (KeyFields[I] <> ';') do Inc(I);
    Inc(o);
    if (I <= Length(KeyFields)) and (KeyFields[I] = ';') then Inc(I);
    p := I;
  end;

  function LocatePass: Integer;
  var
    RecPareBuf, i: Integer;
    ReadSize, RealRead, BufCnt: Integer;
    Ok: boolean;
    Rec: Integer;
  begin
    IndState := true;
    Result := 0;
    Rec := 0;
    Ok := false;
    try
      RecPareBuf := LOCATE_BUFFER_SIZE div Header.rec_size;
      if RecPareBuf >= 1 then begin
        ReadSize := RecPareBuf * Header.rec_size;
        SysUtils.FileSeek(Handle, Header.data_offset, 0);
        repeat
          RealRead := SysUtils.FileRead(Handle, FLocateBuffer^, ReadSize);
          BufCnt := RealRead div Header.rec_size;
          for i := 0 to BufCnt - 1 do begin
            IndRecBuf := FLocateBuffer + Header.rec_size * i;
            Inc(Rec);
            if CompareLocateField(FFields, KeyValues, Options) = 0 then begin
              Ok := true;
              Exit;
            end;
          end;
        until ( BufCnt <= 0 );
      end else raise Exception.Create('TSmartDBF: Record size too lage');
    finally
      IndState := false;
      IndRecBuf := nil;
      if Ok then Result := Rec;
    end;
  end;

begin
  m := 0;
  k := 0;
  p := 1;
  o := 0;
  while p <= Length(KeyFields) do CntFld;
  j := Indexes.Count - 1;
  for i := 0 to j do begin
    l := Indexes[i].SuiteFieldList(KeyFields, n);
    if l > m then begin
      m := l;
      k := i;
    end;
  end;
  if (m > 0) and (o = m) then
    Result := Indexes[k].SeekFields(KeyFields, KeyValues)
  else begin
    FFields := TList.Create;
    try
      GetFieldList(FFields, KeyFields);
      Result := LocatePass;
    finally
      FFields.Free;
    end;
  end;
end;

function TSmartDBF.Locate(const KeyFields: string;
  const KeyValues: Variant; Options: TLocateOptions): Boolean;
var
  Rec: Integer;
begin
  Result := false;
  Rec := LocateRecord(KeyFields, KeyValues, Options);
  if Rec <> 0 then begin
    RecNo := Rec;
    Result := true;
  end;
end;

function TSmartDBF.Lookup(const KeyFields: string;
  const KeyValues: Variant; const ResultFields: string): Variant;
var
  Rec: Integer;
begin
  Rec := LocateRecord(KeyFields, KeyValues, []);
  if Rec <> 0 then begin
    SysUtils.FileSeek(Handle, Header.data_offset + (Rec - 1) * Header.rec_size, 0);
    SysUtils.FileRead(Handle, FLocateBuffer^, Header.rec_size);
    IndRecBuf := FLocateBuffer;
    IndState := true;
    try
      Result := FieldValues[ResultFields];
    finally
      IndState := false;
      IndRecBuf := nil;
    end;
  end else
    Result := Unassigned;
end;

function TSmartDBF.GetMasterFields: Variant;
var
  i: Integer;
begin
  if ListMasterFields.Count > 0 then begin
    Result := VarArrayCreate([0, ListMasterFields.Count - 1], varVariant);
    for i := 0 to ListMasterFields.Count - 1 do
      Result[i] := TField(ListMasterFields[i]).AsVariant;
  end else
    Result := Null;
end;

procedure TSmartDBF.SetMasterFields(const Value: String);
begin
  if Value = '' then begin
    FMasterFields := Value;
    if FRange then ClearRange;
    FRange := false;
  end;
  if FMasterFields <> Value then begin
    FMasterFields := Value;
    if FRange then ClearRange;
    FRange := true;
    ListMasterFields.Clear;
    if DataSource <> nil then begin
      if DataSource.DataSet <> nil then begin
        if DataSource.DataSet.Active then begin
          DataSource.DataSet.GetFieldList(ListMasterFields, FMasterFields);
          SetRange(FMasterFields, GetMasterFields);
        end;
      end;
    end;
  end;
end;

procedure TSmartDBF.ClearRange;
begin
  //
end;

procedure TSmartDBF.SetRange(FieldList: String;
  FieldValues: array of const);
begin
  //
end;

procedure TSmartDBF.SetRange(FieldList: String; FieldValues: variant);
begin
  //
end;

procedure TSmartDBF.Reindex;
var
  i, j: Integer;
begin
  j := Indexes.Count - 1;
  for i := 0 to j do
    Indexes[i].Reindex;
end;

procedure TSmartDBF.SetDBFFieldDefs(const Value: TDBFFieldDefs);
begin
  FDBFFieldDefs.Assign(Value);
end;

function TSmartDBF.CreateBlobStream(Field: TField;
  Mode: TBlobStreamMode): TStream;
var
  ActiveBuf: pChar;
  qq: PFieldInfo;
  ss: pChar;
  ss1: array [0..10] of char;
  iCode, dInt: Integer;
  dbfBuf: array [0..511] of byte;
  eof: boolean;
  i: Integer;
  rr, rr1: Integer;
  LenLob: Integer;
begin
  qq := PFieldInfo(FFieldInfo.Items[Field.FieldNo - 1]);
  GetActiveRecBuf(ActiveBuf);
  if ActiveBuf <> nil then
  begin
    if Mode = bmWrite then begin
      Result := TDBTStream.CreateDBTStream(self, Field);
      TDBTStream(Result).FModified := true;
      Exit;
    end;
    ss := ActiveBuf + qq.FOff;
    if  ( qq.FType = 'M' ) or
        ( ( qq.FType = 'E' ) and
          ( qq.FExType in [dbftClob, dbftFmtMemo,
                            dbftBlob, dbftGraphic] )) then begin
      Move(ss^, ss1, 10);
      ss1[10] := #0;
      Val(ss1, dInt, iCode);
    end else begin
      if ( ( pLongword(ss)^ and $80000000 ) = $80000000 ) then begin
        dInt := ( pLongword(ss)^ and $7FFFFFFF );
        iCode := 0;
      end else begin
        dInt := 0;
        iCode := 1;
      end;
    end;
    if (iCode = 0) and (LobHandler > 0) then
    begin
      Result := TDBTStream.CreateDBTStream(self, Field);
      case qq.FType of
        'M':
          begin
            eof := false;
            FileSeek(LobHandler, 512 * dInt, 0);
            repeat
              rr := FileRead(LobHandler, dbfBuf, 512);
              rr1 := rr;
              for i := 0 to rr - 1 do begin
                if dbfBuf[i] = $1A then begin
                  eof := true;
                  rr1 := i;
                  break;
                end;
              end;
              Result.Write(dbfBuf, rr1);
            until eof;
            Result.Position := 0;
            if TDBTStream(Result).Memory <> nil then begin
              pChar(TDBTStream(Result).Memory)[TDBTStream(Result).Size] := #0;
              Translate(TDBTStream(Result).Memory, TDBTStream(Result).Memory, false);
            end;
            TDBTStream(Result).FModified := false;
          end;
        'E':
          begin
            case qq.FExType of
              dbftClob, dbftFmtMemo:
                begin
                  FileSeek(LobHandler, 512 * dInt, 0);
                  FileRead(LobHandler, LenLob, SizeOf(Integer));
                  Result.Size := LenLob;
                  FileRead(LobHandler, TDBTStream(Result).Memory^, LenLob);
                  Result.Position := 0;
                  if TDBTStream(Result).Memory <> nil then begin
                    pChar(TDBTStream(Result).Memory)[TDBTStream(Result).Size] := #0;
                    Translate(TDBTStream(Result).Memory, TDBTStream(Result).Memory, false);
                  end;
                  TDBTStream(Result).FModified := false;
                end;
              dbftBlob, dbftGraphic:
                begin
                  FileSeek(LobHandler, 512 * dInt, 0);
                  FileRead(LobHandler, LenLob, SizeOf(Integer));
                  Result.Size := LenLob;
                  FileRead(LobHandler, TDBTStream(Result).Memory^, LenLob);
                  Result.Position := 0;
                  TDBTStream(Result).FModified := false;
                end;
              dbftClob_NB, dbftFmtMemo_NB:
                begin
                  FileSeek(LobHandler, dInt, 0);
                  FileRead(LobHandler, LenLob, SizeOf(Integer));
                  Result.Size := LenLob;
                  FileRead(LobHandler, TDBTStream(Result).Memory^, LenLob);
                  Result.Position := 0;
                  if TDBTStream(Result).Memory <> nil then begin
                    pChar(TDBTStream(Result).Memory)[TDBTStream(Result).Size] := #0;
                    Translate(TDBTStream(Result).Memory, TDBTStream(Result).Memory, false);
                  end;
                  TDBTStream(Result).FModified := false;
                end;
              dbftBlob_NB, dbftGraphic_NB:
                begin
                  FileSeek(LobHandler, dInt, 0);
                  FileRead(LobHandler, LenLob, SizeOf(Integer));
                  Result.Size := LenLob;
                  FileRead(LobHandler, TDBTStream(Result).Memory^, LenLob);
                  Result.Position := 0;
                  TDBTStream(Result).FModified := false;
                end;
            else
              raise Exception.Create('TSmartDBF: Lob stream create error!');
            end;
          end;
      else
        raise Exception.Create('TSmartDBF: Lob stream create error!');
      end;
    end else
      Result := TDBTStream.CreateDBTStream(self, Field);
  end else
    Result := TDBTStream.CreateDBTStream(self, Field);
end;

procedure TSmartDBF.SaveToDBT(Source: TMemoryStream; Field: TField);
var
  ActiveBuf: pChar;
  qq: PFieldInfo;
  ss: pChar;
  ss1: array [0..10] of char;
  lEnd, dInt: Integer;
  LenLob: Integer;
begin
  qq := PFieldInfo(FFieldInfo.Items[Field.FieldNo - 1]);
  GetActiveRecBuf(ActiveBuf);
  if (ActiveBuf <> nil) and (LobHandler > 0) then
  begin
    ss := ActiveBuf + qq.FOff;
    if Source.Memory <> nil then begin
      lEnd := FileSeek(LobHandler, 0, 2);
      if  ( qq.FType = 'M' ) or
          ( ( qq.FType = 'E' ) and
            ( qq.FExType in [ dbftClob, dbftFmtMemo,
                              dbftBlob, dbftGraphic] )) then begin
        dInt := lEnd div 512;
        if (lEnd mod 512) > 0 then Inc(dInt);
        FileSeek(LobHandler, dInt * 512, 0);
        Str(dInt:10, ss1);
        Move(ss1, ss^, 10);
      end else begin
        dInt := lEnd;
        FileSeek(LobHandler, dInt, 0);
        LongWord(Pointer(ss)^) := ( LongWord(dInt) or $80000000 );
      end;
      case qq.FType of
        'M':
          begin
            pChar(Source.Memory)[Source.Size] := #0;
            Translate(Source.Memory, Source.Memory, true);
            FileWrite(LobHandler, Pointer(Source.Memory)^, Source.Size);
            ss1[0] := #$1A;
            FileWrite(LobHandler, ss1, 1);
          end;
        'E':
          begin
            case qq.FExType of
              dbftClob, dbftFmtMemo:
                begin
                  LenLob := Source.Size;
                  pChar(Source.Memory)[Source.Size] := #0;
                  Translate(Source.Memory, Source.Memory, true);
                  FileWrite(LobHandler, LenLob, SizeOf(Integer));
                  FileWrite(LobHandler, Pointer(Source.Memory)^, LenLob);
                end;
              dbftBlob, dbftGraphic:
                begin
                  LenLob := Source.Size;
                  FileWrite(LobHandler, LenLob, SizeOf(Integer));
                  FileWrite(LobHandler, Pointer(Source.Memory)^, LenLob);
                end;
              dbftClob_NB, dbftFmtMemo_NB:
                begin
                  LenLob := Source.Size;
                  pChar(Source.Memory)[Source.Size] := #0;
                  Translate(Source.Memory, Source.Memory, true);
                  FileWrite(LobHandler, LenLob, SizeOf(Integer));
                  FileWrite(LobHandler, Pointer(Source.Memory)^, LenLob);
                end;
              dbftBlob_NB, dbftGraphic_NB:
                begin
                  LenLob := Source.Size;
                  FileWrite(LobHandler, LenLob, SizeOf(Integer));
                  FileWrite(LobHandler, Pointer(Source.Memory)^, LenLob);
                end;
            else
              raise Exception.Create('TSmartDBF: Lob stream save error!');
            end;
          end;
      else
        raise Exception.Create('TSmartDBF: Lob stream save error!');
      end;
    end else begin
      if  ( qq.FType = 'M' ) or
          ( ( qq.FType = 'E' ) and
            ( qq.FExType in [ dbftClob, dbftFmtMemo,
                              dbftBlob, dbftGraphic] )) then begin
        FillChar(ss1, 10, ' ');
        Move(ss1, ss^, 10);
      end else
        Byte(Pointer(ss + 3)^) := ( pByte(ss + 3)^ and $7F );
    end;
  end;
end;

procedure TSmartDBF.BeginAddBuffered(RecInBuffer: Integer);
begin
  if not FAddBuffered then begin
    FAddBuffered := true;
    FAddBufferCount := RecInBuffer;
    FAddBuffer := MemMgr.oMem.GetMem(self, FAddBufferCount * FRecordSize);
    FAddBufferCurrent := -1; // 0 - (FAddBufferCount - 1)
  end;
end;

procedure TSmartDBF.EndAddBuffered;
begin
  if FAddBuffered then begin
    if FAddBufferCount > -1 then FlushAddBuffer;
    MemMgr.oMem.FreeMem(FAddBuffer);
    FAddBuffered := false;
    FAddBuffer := nil;
    FAddBufferCount := -1;
    FAddBufferCurrent := -1;
  end;
end;

procedure TSmartDBF.FlushAddBuffer;
var
  i, j, RealRead: Integer;
  lpMsgBuf: array [0..500] of Char;
  le: DWORD;
  NewKey: String;
  NewRec: LongInt;
  LockR, b: boolean;
begin
  if FAddBuffered then begin
    if FAddBufferCurrent > -1 then begin
      CheckActive;
      if LockHeader then begin
        try
          DBFHeader.last_rec := ( (FileSeek(DBFHandler, 0, 2) - DBFHeader.data_offset) div DBFHeader.rec_size );
          NewRec := DBFHeader.last_rec + 1;
          FileSeek(DBFHandler, 0, 2);
          FileSeek(DBFHandler, DBFHeader.data_offset + LongWord(DBFHeader.last_rec * FRecordSize), 0);
          RealRead := FileWrite(DBFHandler, FAddBuffer^, DBFHeader.rec_size * ( FAddBufferCurrent + 1 ) );
          if RealRead = -1 then begin
            le := GetLastError();
            FormatMessage(
              FORMAT_MESSAGE_FROM_SYSTEM,
              nil,
              le,
              0, // Default language
              lpMsgBuf,
              500,
              nil
              );
            raise Exception.Create('TSmartDBF: ' + lpMsgBuf);
          end else begin
            Inc(DBFHeader.last_rec, FAddBufferCurrent + 1);
            FileSeek(DBFHandler, 0, 0); //go to the begin
            RealRead := FileWrite(DBFHandler, DBFHeader, SizeOf(DBFHeader));
            if RealRead = -1 then begin
              le := GetLastError();
              FormatMessage(
                FORMAT_MESSAGE_FROM_SYSTEM,
                nil,
                le,
                0, // Default language
                lpMsgBuf,
                500,
                nil
                );
              raise Exception.Create('TSmartDBF: ' + lpMsgBuf);
            end else begin
              FIndState := true;
              try
                for i := 0 to Indexes.Count - 1 do begin
                  b :=  ( (FAccessMode.FLast and fmShareExclusive) = fmShareExclusive ) or
                        ( (FAccessMode.FLast and fmShareDenyWrite) = fmShareDenyWrite ) or
                        FFileLock;
                  LockR := Indexes[i].FLock;
                  if not b then Indexes[i].StartUpdate(false);
                  if LockR then
                    try
                      for j := 0 to FAddBufferCurrent do begin
                        FIndRecBuf := FAddBuffer + j * DBFHeader.rec_size;
                        NewKey := Indexes[i].EvaluteKeyExpr;
                        Indexes[i].AddKey(NewKey, NewRec + j);
                      end;
                    finally
                      if not b then Indexes[i].Flush;
                      Indexes[i].FUnLock;
                    end
                  else
                    raise Exception.Create('TSmartDBF.FlushAddBuffer: Can not add key to index file (FLock is false).');
                end;
              finally
                FIndRecBuf := nil;
                FIndState := false;
              end;
            end;
          end;
        finally
          UnLockHeader;
        end;
      end else
        raise Exception.Create('TSmartDBF.FlushAddBuffer: Can not lock DBF header.');
    end;
    FAddBufferCurrent := -1;
  end;
end;

{ TDBFNTX }

procedure TDBFNTX.ClearRange;
begin
  if Indexes.ActiveObject <> nil then
    TNTXIndex(Indexes.ActiveObject).NTXRange.Active := false;
end;

constructor TDBFNTX.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FIndexes := TIndexes.Create(self, TNTXIndex);
end;

destructor TDBFNTX.Destroy;
begin
  FIndexes.Destroy;
  FIndexes := nil;
  inherited Destroy;
end;

procedure TDBFNTX.SetRange(FieldList: String; FieldValues: array of const);
var
  m, i, j, k, l, n, p, o: Integer;

  procedure CntFld;
  var
    I: Integer;
  begin
    I := p;
    while (I <= Length(FieldList)) and (FieldList[I] <> ';') do Inc(I);
    Inc(o);
    if (I <= Length(FieldList)) and (FieldList[I] = ';') then Inc(I);
    p := I;
  end;

begin
  m := 0;
  k := 0;
  o := 0;
  p := 1;
  while p <= Length(FieldList) do CntFld;
  j := Indexes.Count - 1;
  for i := 0 to j do begin
    l := Indexes[i].SuiteFieldList(FieldList, n);
    if l > m then begin
      m := l;
      k := i;
    end;
  end;
  if (m > 0) and (o = m) then
    Indexes[k].SetRangeFields(FieldList, FieldValues)
  else
    raise Exception.Create('TSmartDBF: There is no suitable index for range!');
end;

procedure TDBFNTX.SetRange(FieldList: String; FieldValues: variant);
var
  m, i, j, k, l, n, p, o: Integer;

  procedure CntFld;
  var
    I: Integer;
  begin
    I := p;
    while (I <= Length(FieldList)) and (FieldList[I] <> ';') do Inc(I);
    Inc(o);
    if (I <= Length(FieldList)) and (FieldList[I] = ';') then Inc(I);
    p := I;
  end;

begin
  m := 0;
  k := 0;
  o := 0;
  p := 1;
  while p <= Length(FieldList) do CntFld;
  j := Indexes.Count - 1;
  for i := 0 to j do begin
    l := Indexes[i].SuiteFieldList(FieldList, n);
    if l > m then begin
      m := l;
      k := i;
    end;
  end;
  if (m > 0) and (o = m) then
    Indexes[k].SetRangeFields(FieldList, FieldValues)
  else
    raise Exception.Create('TSmartDBF: There is no suitable index for range!');
end;

{ TDBFCDX }

constructor TDBFCDX.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FIndexes := TIndexes.Create(self, TCDXIndex);
end;

destructor TDBFCDX.Destroy;
begin
  FIndexes.Destroy;
  inherited Destroy;
end;

{ TVKDataLink }

procedure TVKDataLink.DataEvent(Event: TDataEvent; Info: Integer);
begin
  inherited;
  if Event = deDataSetChange then begin
    if FDBFDataSet.FRange then begin
      if FDBFDataSet.ListMasterFields.Count = 0 then
        DataSet.GetFieldList(FDBFDataSet.ListMasterFields, FDBFDataSet.FMasterFields);
      FDBFDataSet.SetRange(FDBFDataSet.FMasterFields, FDBFDataSet.GetMasterFields);
    end else begin
      if bof then FDBFDataSet.First;
      if eof then FDBFDataSet.Last;
    end;
  end;
end;

procedure TVKDataLink.DataSetScrolled(Distance: Integer);
begin
  inherited;
  if FDBFDataSet.FRange then begin
    if FDBFDataSet.ListMasterFields.Count = 0 then
      DataSet.GetFieldList(FDBFDataSet.ListMasterFields, FDBFDataSet.FMasterFields);
    FDBFDataSet.SetRange(FDBFDataSet.FMasterFields, FDBFDataSet.GetMasterFields);
  end else
    FDBFDataSet.MoveBy(Distance);
end;

{ TDBFFieldDefs }

procedure TDBFFieldDefs.AssignValues(Value: TDBFFieldDefs);
var
  I: Integer;
  P: TDBFFieldDef;
begin
  for I := 0 to Value.Count - 1 do
  begin
    P := FindIndex(Value[I].Name);
    if P <> nil then
      P.Assign(Value[I]);
  end;
end;

constructor TDBFFieldDefs.Create(AOwner: TPersistent);
begin
  inherited Create(AOwner, TDBFFieldDef);
end;

function TDBFFieldDefs.FindIndex(const Value: string): TDBFFieldDef;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
  begin
    Result := TDBFFieldDef(inherited Items[I]);
    if AnsiCompareText(Result.Name, Value) = 0 then Exit;
  end;
  Result := nil;
end;

function TDBFFieldDefs.GetCollectionOwner: TPersistent;
begin
  Result := GetOwner;
end;

function TDBFFieldDefs.GetItem(Index: Integer): TDBFFieldDef;
begin
  Result := TDBFFieldDef(inherited Items[Index]);
end;

function TDBFFieldDefs.IsEqual(Value: TDBFFieldDefs): Boolean;
var
  I: Integer;
begin
  Result := (Count = Value.Count);
  if Result then
    for I := 0 to Count - 1 do
    begin
      Result := TDBFFieldDef(Items[I]).IsEqual(TDBFFieldDef(Value.Items[I]));
      if not Result then Break;
    end
end;

procedure TDBFFieldDefs.SetItem(Index: Integer; const Value: TDBFFieldDef);
begin
  inherited SetItem(Index, TCollectionItem(Value));
end;

{ TDBFFieldDef }

procedure TDBFFieldDef.AssignTo(Dest: TPersistent);
begin
  with Dest as TDBFFieldDef do begin
    Name := self.Name;
    field_type := self.field_type;
    extend_type := self.extend_type;
    len := self.len;
    dec := self.dec;
  end;
end;

constructor TDBFFieldDef.Create(Collection: TCollection);
begin
  inherited Create(Collection);
end;

destructor TDBFFieldDef.Destroy;
begin
  inherited Destroy;
end;

function TDBFFieldDef.GetDataSize: Word;
begin
  //C N D L M     E
  case FieldRec.field_type of
    'C', 'N': Result := len;
    'D': Result := 8;
    'L': Result := 1;
    'M': Result := 10;
    'E':
      case FieldRec.extend_type of
        dbftS1:     Result := 1;  //Shortint
        dbftU1:     Result := 1;  //Byte
        dbftS2:     Result := 2;  //Smallint
        dbftU2:     Result := 2;  //Word
        dbftS4:     Result := 4;  //Longint
        dbftU4:     Result := 4;  //Longword
        dbftS8:     Result := 8;  //Int64
        dbftR4:     Result := 4;  //Single
        dbftR6:     Result := 6;  //Real48
        dbftR8:     Result := 8;  //Double
        dbftR10:    Result := 10; //Extended
        dbftD1:     Result := 8;  //TDateTime
        dbftD2:     Result := 8;  //DataSet DateTime
        dbftD3:     Result := 6;  //Real48 DateTime
        dbftS1_N:   Result := 2;  //Shortint with NULL
        dbftU1_N:   Result := 2;  //Byte  with NULL
        dbftS2_N:   Result := 3;  //Smallint with NULL
        dbftU2_N:   Result := 3;  //Word with NULL
        dbftS4_N:   Result := 5;  //Longint with NULL
        dbftU4_N:   Result := 5;  //Longword with NULL
        dbftS8_N:   Result := 9;  //Int64 with NULL
        dbftR4_N:   Result := 5;  //Single with NULL
        dbftR6_N:   Result := 7;  //Real48 with NULL
        dbftR8_N:   Result := 9;  //Double with NULL
        dbftR10_N:  Result := 11; //Extended with NULL
        dbftD1_N:   Result := 9;  //TDateTime with NULL
        dbftD2_N:   Result := 9;  //DataSet DateTime with NULL
        dbftD3_N:   Result := 7;  //Real48 DateTime
        dbftClob:   Result := 10;
        dbftBlob:   Result := 10;
        dbftGraphic:  Result := 10;
        dbftFmtMemo:  Result := 10;
        dbftClob_NB:   Result := 4;
        dbftBlob_NB:   Result := 4;
        dbftGraphic_NB:  Result := 4;
        dbftFmtMemo_NB:  Result := 4;
        dbftString:   Result := len + 2;
        dbftString_N: Result := len + 3;
        dbftFixedChar: Result := len + 1;
        dbftWideString: Result := len * 2 + 4;
        dbftCurrency: Result := 8;
        dbftCurrency_N: Result := 9;
        dbftCurrency_NB: Result := 8;
        dbftBCD: Result := SizeOf(TBcd);
        dbftDate:   Result := 4;  //ftDate
        dbftDate_N:   Result := 5;  //ftDate with NULL byte
        dbftTime:   Result := 4;  //ftTime
        dbftTime_N:   Result := 5;  //ftTime with NULL byte
        dbftU1_NB: Result := 1;
        dbftU2_NB: Result := 2;
        dbftU4_NB: Result := 4;
        dbftR4_NB: Result := 4;
        dbftR6_NB: Result := 6;
        dbftR8_NB: Result := 8;
        dbftD1_NB: Result := 8;
        dbftD2_NB: Result := 8;
        dbftD3_NB: Result := 6;
        dbftDate_NB: Result := 4;
        dbftTime_NB: Result := 4;
      else
        raise Exception.Create('Extend_type incarect!');
      end;
  else
    raise Exception.Create('Field_type incarect!');
  end;
end;

function TDBFFieldDef.GetDisplayName: string;
begin
  Result := FieldRec.field_name;
end;

function TDBFFieldDef.GetField: FIELD_REC;
begin
  //C N D L M     E
  case FieldRec.field_type of
    'C':
      begin
        FieldRec.lendth.char_len := len;
      end;
    'N':
      begin
        FieldRec.lendth.num_len.len := Byte(len);
        FieldRec.lendth.num_len.dec := Byte(dec);
      end;
    'D':
      begin
        FieldRec.lendth.num_len.len := Byte(8);
        FieldRec.lendth.num_len.dec := Byte(0);
      end;
    'L':
      begin
        FieldRec.lendth.num_len.len := Byte(1);
        FieldRec.lendth.num_len.dec := Byte(0);
      end;
    'M':
      begin
        FieldRec.lendth.num_len.len := Byte(10);
        FieldRec.lendth.num_len.dec := Byte(0);
      end;
    'E':
      begin
        if FieldRec.extend_type in [  dbftString, dbftString_N,
                                      dbftFixedChar, dbftWideString] then begin
          FieldRec.lendth.num_len.len := Byte(len);
        end else if FieldRec.extend_type in [dbftBCD] then begin
          FieldRec.lendth.num_len.len := Byte(len);
          FieldRec.lendth.num_len.dec := Byte(dec);
        end else begin
          FieldRec.lendth.num_len.len := Byte(0);
          FieldRec.lendth.num_len.dec := Byte(0);
        end;
      end;
  else
    raise Exception.Create('Field_type incarect!');
  end;
  Result := FieldRec;
end;

function TDBFFieldDef.IsEqual(Value: TDBFFieldDef): Boolean;
begin
  Result := false;
  if Value.Name = Name then Result := true;
end;

procedure TDBFFieldDef.SetDisplayName(const Value: string);
var
  l: Integer;
begin
  l := Length(Value);
  if l > 10 then raise ERangeError.CreateFmt('Field length must be <= 10, but no %d!', [l]);
  FillChar(FieldRec.field_name, 11, 0);
  Move(pChar(Value)^, FieldRec.field_name, l);
end;

{ TDBTStream }

procedure TDBTStream.Clear;
begin
  inherited Clear;
  FModified := true;
end;

constructor TDBTStream.Create;
begin
  inherited Create;
  FModified := false;
end;

constructor TDBTStream.CreateDBTStream(dbf: TSmartDBF; field: TField);
begin
  inherited Create;
  FModified := false;
  FSmartDBF := dbf;
  FField := field;
end;

destructor TDBTStream.Destroy;
begin
  if FModified then
    SaveToDBT;
  inherited Destroy;
end;

procedure TDBTStream.SaveToDBT;
begin
  FSmartDBF.SaveToDBT(self, FField);
end;

procedure TDBTStream.SetSize(NewSize: Integer);
begin
  inherited SetSize(NewSize);
  FModified := true;
end;

function TDBTStream.Write(const Buffer; Count: Integer): Longint;
begin
  FModified := true;
  Result := inherited Write(Buffer, Count);
end;

end.
